/* access.c - p2pd client access control check functions */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: access.c,v 1.23 2010/06/30 08:37:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <errno.h>
#include <string.h>

#include "p2pd.h"
#include "peer.h"
#include "access.h"
#include "conn.h"
#include "subnet.h"

int peerlocal(struct connection *);
extern int localrec;
extern struct subnet scc_rec_subnet;
extern struct subnet scc_http_subnet;

/*
 * access control functions
 *
 * LHC
 *     The LHC uses the allow_recur() and allow_streamreq() functions
 *     to check whether to allow recursive and stream requests,
 *     respectively. Both of these will normally only be accepted from
 *     the local host.
 *
 * SCC
 *     The SCC never accepts stream requests, but allows recursive
 *     requests from clients at the local site or domain.
 *     XXX specified either via domain name or address range.
 *
 * Server
 *     Accepts neither stream requests or recursive requests.
 *
 *     The server and SCC might have use for more fine grained access
 *     control over connections, e.g. using the hosts_access(3) might
 *     be appropriate, but the LHC is expected to be run as the owner
 *     of the application, so no global configuration files should be
 *     required.
 *
 */


/* allow_recur - allow recursive requests from peer? */
int
allow_recur(struct connection *c)
{
	/* recursive requests should only come from localhost */
	if ((lhc_mode == 1) && (peerlocal(c) == 1))
		return 1;

	/* the scc serves clients in specified subnet */
	if (scc_mode == 1 && sccclient(c))
		return 1;

	return 0;
}


/* allow_streamreq - allow stream/http requests from peer? */
int
allow_streamreq(struct connection *c)
{
	/* stream requests should only come from localhost in lhc mode */
	if ((lhc_mode == 1) && (peerlocal(c) == 1))
		return 1;
	else if (scc_mode && peerinsubnet(&scc_http_subnet,
					  (struct sockaddr *)&c->ss_remote))
		return 1;
	else
		return 0;
}


/* peerlocal - return 1 if other end of connection is at the local host */
int
peerlocal(struct connection *c)
{
	struct in_addr sin_addr;

	if (c->ss_local_len == 0)
		ERRX("local end sockname not identified");

	/* local end must be loopback address */
	if (((struct sockaddr *)&c->ss_local)->sa_family == AF_INET) {
		sin_addr.s_addr = htonl(INADDR_LOOPBACK);
		if (bcmp(&((struct sockaddr_in *)&c->ss_local)->sin_addr.s_addr,
			 &sin_addr.s_addr, sizeof(sin_addr.s_addr)) == 0)
			return 1;
	} else if (((struct sockaddr *)&c->ss_local)->sa_family == AF_INET6) {
		if (bcmp(&((struct sockaddr_in6 *)&c->ss_local)->sin6_addr,
			 &in6addr_loopback, sizeof(in6addr_loopback)) == 0)
			return 1;
	} else
		ERRX("unknown address family");

	return 0;
}


/* sccclient - return true if other end of connection is an scc client */
int
sccclient(struct connection *c)
{
	if (peerinsubnet(&scc_rec_subnet,
			 (struct sockaddr *)&c->ss_remote) == 1)
		return 1;
	else
		return 0;
}
