/* dlagent.c - file download manager */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: dlagent.c,v 1.25 2010/11/03 10:22:31 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include <time.h>

#include "p2pd.h"
#include "aqueue.h"
#include "block.h"
#include "conn.h"
#include "cachefile.h"
#include "diskcache.h"
#include "dlagent.h"
#include "file.h"
#include "filemeta.h"
#include "igot.h"
#include "peerset.h"
#include "peer.h"
#include "peerlist.h"
#include "socket.h"
#include "stream.h"
#include "spp.h"
#include "timers.h"
#include "urient.h"

#ifdef notyet
/* how far ahead should we stay? */
#define prefetch(c) (((c)->age < 5 ? 0 : \
		     (((c)->age < 7 ? 2 : \
		      (((c)->age < 9 ? 4 : \
		       (((c)->age < 11 ? 6 : \
			(((c)->age < 13 ? 8 : \
			 (((c)->age < 15 ? 10 : 12))))))))))))

#define CONNECTION_RETRY_INTERVAL 3 /* seconds */

#define FILE_PEERSREQ_DELAY 30 /* minimum number of seconds between requests */
#define FILE_PEERSREQ_REFRESH 30 /* at least request this often */
#endif

struct dlagents dlagents = SLIST_HEAD_INITIALIZER(dlatents);

struct dlagent *dlagent_new(struct urient *, off_t, off_t);
struct dlagent *dlagent_split(struct dlagent *, off_t, off_t);
void dlagent_free(struct dlagent *);
void dlagent_init(struct dlagent *);
void dlagent_getrange(struct dlagent *);
off_t dlagent_curroff(struct dlagent *);
off_t node_reqlen(struct urient *, struct peer *);
off_t leftblock(struct urient *, off_t);
off_t dlagent_clientmaxoff(struct dlagent *);
off_t dlagent_clientminoff(struct dlagent *);
void handle_dlchain(struct dlagent *);


/* dlagent_search - return pointer to dlagent for offset (create if needed) */
struct dlagent *
dlagent_search(struct urient *ue, off_t start, off_t end)
{
	const char *f = "dlagent_search";
	struct dlagent *dl, *l_d;

	dbg("%s: %lld %lld", f, (long long)start, (long long)end);

	/* look for dlagent with range containing start */
	SLIST_FOREACH(l_d, &dlagents, n) {
		if ((l_d->ue == ue) &&
		    /* create new dlagent if distance too long */
		    (l_d->start <= start) && (l_d->end >= start)) {
			/* gap too large? */
			off_t curroff;
			off_t gap;

			curroff = dlagent_curroff(l_d);
			gap = (start - curroff) / ue->meta->blocksize;
/*			gap = (start - l_d->start) / ue->meta->blocksize;*/
			if (gap < 2/*XXX*/)
				/* XXX might need to reactivate */
				return l_d;
			else
				return dlagent_split(l_d, start, end);
		}
	}

	if ((dl = dlagent_new(ue, start, end)) == NULL)
		return NULL;

	dlagent_init(dl);
	dlagent_status();

	return dl;
}


/* dlagent_split - create new dlagent with responsibility for part of range */
struct dlagent *
dlagent_split(struct dlagent *dl, off_t start, off_t end)
{
	const char *f = "dlagent_split";
	struct dlagent *dl2;

/*	ASSERT(dl->curroff < start, "start offset too low");*/
	ASSERT(start > 0, "start offset too low");

	dbg("%s: splitting existing dlagent: endoff %lld -> %lld", f,
		(long long)dl->end, (long long)start - 1);
	/* reduce responsibility of existing dlagent */
	dl->end = start - 1;

	dl2 = dlagent_new(dl->ue, start, end);
	dlagent_init(dl2);
	dlagent_status();
	return dl2;
}


/* dlagent_new - create new dlagent for file/range */
struct dlagent *
dlagent_new(struct urient *ue, off_t start, off_t end)
{
	const char *f = "dlagent_new";
	struct dlagent *dl;

	dbg("%s: for %lld to %lld (%lld)", f, (long long)start, (long long)end,
	    (long long)ue->meta->size);
	/* XXX resource management, limit max number */

	ASSERT(ue->meta->size > end, "end too high");

	dl = xmalloc(sizeof(struct dlagent));
	bzero(dl, sizeof(struct dlagent));

	dl->ue = ue;
	dl->start = start;
	dl->end = end;
	dl->to = timeout_add(dlagent_timeout, NULL, dl);
	dl->readers = aqueue_init(NULL);
	dl->state = DLAGENT_STATE_INACTIVE;
	dl->currblock = off2block(ue, dl->start);

	SLIST_INSERT_HEAD(&dlagents, dl, n);
	ASSERT(!SLIST_EMPTY(&dlagents), "list empty");
	dlagent_status();

	urient_fileref(dl->ue);

	return dl;
}


/* dlagent_free - release dlagent resources */
void
dlagent_free(struct dlagent *dl)
{
	const char *f = "dlagent_free";

	dbg("%s", f);

	timeout_del(dl->to);
	urient_fileunref(dl->ue);
	aqueue_free(dl->readers);

	SLIST_REMOVE(&dlagents, dl, dlagent, n);
	free(dl);
}


/* dlagent_init - initialize new dlagent structure */
void
dlagent_init(struct dlagent *dl)
{
/*	const char *f = "dlagent_init";*/
	struct sppconn *sc;

	ASSERT(dl->state == DLAGENT_STATE_INACTIVE, "state already set");

#if 0
	ASSERT(data not present, "data already exists");
#endif

/*	dbg("%s", f);*/

	dl->state = DLAGENT_STATE_PEERS;
	sc = sppconn_new(CLIENT_CMD_PEERS);
	sppconn_setfile(sc, dl->ue);
	sc->dl = dl;
	sc->dlblock = off2block(dl->ue, dl->start);
	spp_parentreq(sc);
}


/* dlagent_addclient - add reader client */
void
dlagent_addclient(struct dlagent *dl, struct sppconn *sc)
{
	aqueue_enqueue(dl->readers, sc);
	dlagent_addref(dl);
}


/* dlagent_delclient - remove client from read chain */
void
dlagent_delclient(struct dlagent *dl, struct sppconn *sc)
{
	aqueue_delval(dl->readers, sc);
	sc->dl = NULL;
	if (sc->stopped) {
		dl->stopcnt--;
		sc->stopped = 0;
	}
	ASSERT(dl->stopcnt >= 0, "negative stopcnt");

	dlagent_unref(dl);
}


/* dlagent_getrange - request file data range */
void
dlagent_getrange(struct dlagent *dl)
{
	const char *f = "dlagent_getrange";
	struct connection *c;
	struct peerset *pset;
	struct sppconn *sc;
	off_t clientoff;
	off_t curroff;
	off_t minoff;
	off_t reqlen;
	int32_t gap;

	dbg("%s", f);

	if ((curroff = dlagent_curroff(dl)) == -1) {
		dbg("%s: download complete", f);
		dl->state = DLAGENT_STATE_COMPLETE;
		if (dl->refcnt == 0)
			dlagent_free(dl);
		return;
	}
	dl->currblock = off2block(dl->ue, curroff);

	/* always start download if no clients (possibly not yet added) */
	if ((clientoff = dlagent_clientmaxoff(dl)) != -1) {
		/* request might be block aligned even if client req is not */
/*		ASSERT(curroff >= clientoff, "invalid curroff/clientoff");*/
		gap = (curroff - clientoff) / dl->ue->meta->blocksize;
		if (gap < 0) {
			dbg("%s: negative gap, adjusting start offset", f);
			minoff = dlagent_clientminoff(dl);
			if (minoff != dl->start) {
				dl->start = minoff;
				dlagent_getrange(dl);
				return;
			} else /* XXX split dlagents */
				ERRX("negative gap");
		}
		if (gap > 2) {
			dl->state = DLAGENT_STATE_INACTIVE;
			dbg("%s: setting dlagent inactive, gap too large: %d",
			    f, gap);
			return;
		} else {
			if (dl->state == DLAGENT_STATE_INACTIVE)
				dbg("%s: gap too small, "
				    "reactivating dlagent: %d", f, gap);
		}
	}

	pset = peerlist_gen(dl->ue, off2block(dl->ue, curroff));
	if (pset == NULL) {
		dbg("%s: NULL peerlist, reverting state", f);
		dl->state = DLAGENT_STATE_PEERS;
		dlagent_timeout(dl); /*XXX should be peers or something*/
		return;
	}

	/* attempt to get data */
	dl->state = DLAGENT_STATE_GET;

	dbg("%s: peerlist: %d hosts, unsorted", f, pset->peercnt);
	peerlist_dump(pset);

	if (pset->peercnt > 1) {
		peerlist_ratesort(pset);
		dbg("%s: peerlist: %d hosts, sorted", f, pset->peercnt);
		peerlist_dump(pset);

		/* XXXpermutate nodelist, round robin */
		peerlist_bestshift(pset);
		dbg("%s: peerlist: %d hosts, permutated", f, pset->peercnt);
		peerlist_dump(pset);
	}

	reqlen = MIN(node_reqlen(dl->ue, pset_peer(pset, 0)) - 1,
		     leftblock(dl->ue, curroff));

	sc = sppconn_new(CLIENT_CMD_GET);
	sppconn_setfile(sc, dl->ue);
	sc->dl = dl;
	sc->dlblock = dl->currblock;
	sc->dlstart = curroff;
	sc->dlend = curroff + reqlen;
	dbg("%s: get %s block %d, first node %s", f, sc->ue->uri.uri,
	    sc->dlblock, pset_peer(pset, 0)->id);
	/* check if a cached conn exists (only check first enty in peerlist) */
	if ((c = conn_getcached(pset_peer(pset, 0), 0)) != NULL) {
		sppconn_setconn(sc, c);
		spp_cmd_get(sc, sc->dlstart, sc->dlend);
/*		spp_conndone(sc);*/
		pset_free(pset);
		return;
	}

	/* do direct connect if only one alternative, mconn otherwise */
	if (pset->peercnt == 1) {
		sppconn_setconn(sc, conn_conninit(pset_peer(pset, 0)));
		conn_connect(sc->conn);
	} else {
		sppconn_setconn(sc, conn_mconninit(pset));
		conn_mconn(sc->conn);
	}

	pset_free(pset);
}


/* dlagent_purgedblock - block lost, update dlagents with block in range */
void
dlagent_purgedblock(struct urient *ue, int32_t bnum)
{
	const char *f = "dlagent_purgeblock";
	struct dlagent *dl, *dl2;
	struct dlagent *l_d;
	struct sppconn *sc;
	off_t bstart, bend;
	off_t oldstart;
	int neednew;
	int i;

	dbg("%s: %s block %d", f, ue->uri.uri, bnum);

	/*
	 * blocks can be purged after igot, which occurs after a block
	 * has been completely downloaded; there should be no active
	 * download for a purged block. the dlagent is most likely
	 * to be at a point after the purged block, but can also
	 * not yet be past it (if the block was in the disk cache
	 * or if it was the last block downloaded).
	 */

	bstart = block2off(ue, bnum);
	bend = block2off(ue, bnum + 1) - 1;
	ASSERT(bstart >= 0, "negative range start");
	ASSERT(bend < ue->meta->size, "range end exceeds file size");

	dl = NULL;
	/* look for dlagent with range containing start XXX dup */
	SLIST_FOREACH(l_d, &dlagents, n) {
		if ((l_d->ue == ue) &&
		    (l_d->start <= bstart) && (l_d->end >= bstart)) {
			dl = l_d;
			break;
		}
	}
	if (dl == NULL) {
		dbg("%s: no relevant dlagent found", f);
		return; /* no relevant dlagent active */
	}

	if (dl->currblock < bnum) {
		/*
		 * nothing needs to be done, dlagent will fetch block
		 * when it is reached.
		 */
		dbg("%s: block not yet reached by dlagent purged", f);
		return;
	} else if (dl->currblock == bnum) {
		/* only need to restart dlagent if stopped (bnum last block) */
		dbg("%s: most recently downloaded block purged", f);
		if (dl->state == DLAGENT_STATE_INACTIVE)
			dlagent_getrange(dl);
		return;
	} /* else, dl->currblock > bnum */

	/*
	 * active clients might have been served invalid data, but
	 * this should only affect stream clients (unavoidable if
	 * latency is to be reduced). block contents might change
	 * midstream; will need to have some failure tolerance with
	 * regards to handling corrupt media streams.
	 */

	/* don't interrupt current dlagent, set start point to after bnum */
	ASSERT(dl->currblock > bnum, "internal error");
	oldstart = dl->start;

	dl->start = bend + 1;
	ASSERT(dl->start != dl->end, "empty dlagent");

	dbg("dlagent_purgeblock: adjusting dlagent start point to %lld",
	    (long long)dl->start);

	neednew = 0;
	/* create new dlagent if any readers are in or before block */
	for (i = 0; (size_t)i < ALIST_SIZE(dl->readers); i++) {
		sc = ALIST_PVAL(dl->readers, i);
		ASSERT(sc->conn->type == CONN_TYPE_INCOMING, "invalid type");
		ASSERT(sc->conn->stream != NULL, "no stream in client");

		if (sc->conn->stream->bnum <= bnum) {
			neednew = 1;
			break;
		}
	}
	if (neednew) {
		dbg("%s: moving client to new dlagent", f);
		dl2 = dlagent_new(ue, oldstart, bend);
		for (i = 0; (size_t)i < ALIST_SIZE(dl->readers); i++) {
			sc = ALIST_PVAL(dl->readers, i);
			if (sc->conn->stream->bnum <= bnum) {
				/* move to new dlagent */
				dlagent_delclient(dl, sc);
				dlagent_addclient(dl2, sc);
				sc->dl = dl2;
			}
		}
		dlagent_init(dl2);
	}
}


/* dlagent_cmdok - notify dlagent of successful command response */
void
dlagent_cmdok(struct dlagent *dl, struct sppconn *sc)
{
	switch(dl->state) {
	case DLAGENT_STATE_PEERS:
		UTOSPP(dl->ue)->lastpeers = time(NULL);

		/*XXX some checking would be prudent */
		dl->state = DLAGENT_STATE_GET;
		dlagent_getrange(dl);
		break;
	case DLAGENT_STATE_GET:
		ASSERT(sc != NULL, "NULL sc");
		stream_recvstart(STREAM_DEADLINE, sc->conn, sc->ue,
				 sc->dlstart, sc->dlend);
		break;
	default:
		ERRX("unexpected dlagent state");
	}
}


/* dlagent_sendok - full block sent to dlagent reader */
void
dlagent_sendok(struct sppconn *sc, int32_t bnum)
{
	const char *f = "dlagent_sendok";

	dbg("%s: successfully transmitted block %d", f, bnum);
	if (sc->dl->state == DLAGENT_STATE_INACTIVE)
		dlagent_getrange(sc->dl);
}


/* dlagent_cmdfail - notify dlagent of command failure */
void
dlagent_cmdfail(struct dlagent *dl, struct sppconn *sc)
{
	const char *f = "dlagent_cmdfail";

	dbg("%s", f);

	if (sc->conn->type == CONN_TYPE_INCOMING)
		return; /* ignore */

	if (dl->state == DLAGENT_STATE_PEERS) {
		timeout_secupd(dl->to, RETRY_DELAY_PEERS);
		dl->retries++;
	} else if (dl->state == DLAGENT_STATE_GET) {
		dl->retries++;
		if (dl->retries > RETRY_GET_MAX) {
			/*
			 * XXX call urifail; should have 'might have
			 * some data but won't retrieve any more
			 * before this error times out because server
			 * is down' (i.e., disable recursion for file)
			 */
			dbg("%s: too many retries, aborting", f);
			handle_dlchain(dl);
			/* want to terminate dlagent */
			dl->state = DLAGENT_STATE_MULTFAIL;
		} else {
			dbg("%s: retrying (retry %d)", f, dl->retries);
			timeout_secupd(dl->to, RETRY_DELAY_GET * dl->retries);
		}
	} else
		ERRX("unexpected state");
}


/* dlagent_readdata - notification of data being read for dlagent */
void
dlagent_readdata(struct dlagent *dl, struct sppconn *insc)
{
	struct sppconn *sc;
	int i;

	/* any stopped readers waiting for data? */
	if (dl->stopcnt == 0)
		return;

	for (i = 0; (size_t)i < ALIST_SIZE(dl->readers); i++) {
		sc = ALIST_PVAL(dl->readers, i);
		if (sc->stopped && stream_availbytes(sc->conn->stream)) {
			dbg("stopped: off %d in off: %d",
			    sc->conn->stream->blockoff,
			    insc->conn->stream->blockoff);
			stream_wake(sc->conn->stream);
			sc->stopped = 0;
			dl->stopcnt--;
		}
	}
}


/* dlagent_getcomp - notify dlagent of successful get data transfer */
void
dlagent_getcomp(struct dlagent *dl, struct sppconn *sc)
{
	const char *f = "dlagent_getcomp";

	/*XXX moved to stream.c */
/*	urient_gotblock(dl->ue, sc->dlblock);*/

	/* report upstream that we got block */
	igot_add(dl->ue, sc->dlblock, dl->ue->file->blocks[sc->dlblock]->sum,
		 sc->conn->peer);

	dbg("%s: get complete, caching connection", f);
	cache_connection(sc->conn);

	dlagent_getrange(dl);
}


/* dlagent_tooslow - notify dlagent of too slow data transfer */
void
dlagent_tooslow(struct dlagent *dl)
{
	const char *f = "dlagent_tooslow";

	dbg("%s", f);

	dlagent_getrange(dl);
/*	dlagent_cmdfail(dl, NULL);*/
}


/* dlagent_streamstop - notify dlagent of stream being stopped */
void
dlagent_streamstop(struct dlagent *dl, struct sppconn *sc)
{
	const char *f = "dlagent_streamstop";

	dbg("%s", f);

	ASSERT(sc->conn->type == CONN_TYPE_INCOMING, "invalid type");
	ASSERT(sc->currcmd == CLIENT_CMD_GET, "invalid cmd");
	ASSERT(sc->stopped == 0, "stream already stopped");

	dl->stopcnt++;
	sc->stopped = 1;
}


/* dlagent_timeout - notify dlagent of connection timeout */
void
dlagent_timeout(void *dat)
{
	const char *f = "dlagent_timeout";
	struct sppconn *sc;
	struct dlagent *dl;
	off_t curroff;

	dbg("%s", f);

	dl = dat;
	curroff = dlagent_curroff(dl);
	if (curroff == -1) {
		dbg("%s: download complete", f);
		dl->state = DLAGENT_STATE_COMPLETE;
		return;
	}
	if (dl->state == DLAGENT_STATE_PEERS) {
		sc = sppconn_new(CLIENT_CMD_PEERS);
		sppconn_setfile(sc, dl->ue);
		sc->dl = dl;
		sc->dlblock = off2block(dl->ue, curroff);
		spp_parentreq(sc);
	} else if (dl->state == DLAGENT_STATE_GET) {
		dlagent_getrange(dl);
	} else
		ERRX("unexected state");
}


/* dlagent_connend - notification of connection state ending */
void
dlagent_connend(struct sppconn *sc)
{
	ASSERT(sc->dl != NULL, "NULL dl");

	if (sc->conn->type == CONN_TYPE_INCOMING)
		dlagent_delclient(sc->dl, sc);
}


/* dlagent_addref - increase dlagent reference count */
void
dlagent_addref(struct dlagent *dl)
{
	dl->refcnt++;
}


/* dlagent_unref - decrease dlagent reference count */
void
dlagent_unref(struct dlagent *dl)
{
	const char *f = "dlagent_unref";

/*	dbg("%s", f);*/

	dl->refcnt--;
	ASSERT(dl->refcnt >= 0, "negative refcnt");

	/*
	 * XXX want dlagent to continue if it should complete a file,
	 *     even if there currently are no downloaders.
	 *     a proper check for whether the file should be downloaded
	 *     will probably have to be a little more complicated than
	 *     the one below.
	 */
	if ((dl->refcnt == 0) &&
	    ((dl->state == DLAGENT_STATE_COMPLETE) ||
	     (dl->state == DLAGENT_STATE_MULTFAIL))) {
		dbg("%s: dlagent unused, freeing", f);
		dlagent_free(dl);
	}
}


/* leftblock - bytes until end of block at offset */
off_t
leftblock(struct urient *ue, off_t offset)
{
/*	const char *f = "leftblock";*/
	off_t left;
	int bnum;

	ASSERT(ue->meta != NULL, "NULL meta");
	ASSERT(offset < ue->meta->size, "offset larger than filesize");

/*	dbg("%s: size %lld", f, ue->meta->size);*/

	bnum = off2block(ue, offset);
	if (bnum + 1 == ue->meta->blockcnt)
		left = ue->meta->size - (offset + 1);
	else {
		left = block2off(ue, bnum) + ue->meta->blocksize;
		left -= (offset + 1);
	}

	ASSERT(left >= 0, "negative offset");

	return left;
}


/* dlagent_curroff - return start point for request, -1 if finished */
off_t
dlagent_curroff(struct dlagent *dl)
{
/*	const char *f = "dlagent_curroff";*/
	struct file *file;
	off_t offset;
	int bnum;

/*	dbg("%s", f);*/

	/*
	 * if block not saved to disk (e.g., due to disk full), actlen
	 * might be equal to blocklen, but bit in blockmap will not be
	 * set
	 */
	file = dl->ue->file;
	for (bnum = off2block(dl->ue, dl->start);
	     bnum < dl->ue->cf->bm.size;
	     bnum++)
		if (blockmap_test(&dl->ue->cf->bm, bnum) == 0 &&
		    file->blocks[bnum]->actlen < file->blocks[bnum]->blocklen)
			break;

	/* nothing to download? */
	offset = block2off(dl->ue, bnum);
	if (offset > dl->end)
		return -1;
	offset += file->blocks[bnum]->actlen;
/*	dbg("%s: bnum: %d actlen: %d", f, bnum, file->blocks[bnum]->actlen);*/
/*	dbg("%s: off %lld end: %lld", f, offset, dl->end);*/
	if (offset > dl->end)
		return -1;
	else
		return offset;
}


/* node_reqlen - peer request length, initially short, longer if stable */
off_t
node_reqlen(struct urient *ue, struct peer *node)
{
	(void)node;

	return ue->meta->blocksize; /*XXX*/
}


/* dlagent_clientmaxoff - higest reader offset */
off_t
dlagent_clientmaxoff(struct dlagent *dl)
{
	struct sppconn *sc;
	off_t           curroff;
	off_t           maxoff;
	int             i;

	if (ALIST_SIZE(dl->readers) == 0)
		return -1;

	maxoff = 0;

	for (i = 0; (size_t)i < ALIST_SIZE(dl->readers); i++) {
		sc = ALIST_PVAL(dl->readers, i);
		ASSERT(sc->conn->type == CONN_TYPE_INCOMING, "invalid type");
		ASSERT(sc->conn->stream != NULL, "no stream in client");

		curroff = stream_curroff(sc->conn->stream);
		if (curroff > maxoff)
			maxoff = curroff;
	}

	return maxoff;
}


/* dlagent_clientminoff - lowest reader offset */
off_t
dlagent_clientminoff(struct dlagent *dl)
{
	struct sppconn *sc;
	off_t           curroff;
	off_t           minoff;
	int             i;

	minoff = dlagent_clientmaxoff(dl); /* start with highest value */

	for (i = 0; (size_t)i < ALIST_SIZE(dl->readers); i++) {
		sc = ALIST_PVAL(dl->readers, i);
		ASSERT(sc->conn->type == CONN_TYPE_INCOMING, "invalid type");
		ASSERT(sc->conn->stream != NULL, "no stream in client");

		curroff = stream_curroff(sc->conn->stream);
		if (curroff < minoff)
			minoff = curroff;
	}

	return minoff;
}


/* handle_dlchain - abort client connections in case of multiple failures */
void
handle_dlchain(struct dlagent *dl)
{
	const char *f = "handle_dlchain";
	struct sppconn *sc;
	int i;

	dbg("%s: ending waiting dlagent reader connections", f);
	/*XXX uses ALIST_foo macros */
	for (i = 0; (size_t)i < ALIST_SIZE(dl->readers); i++) {
		sc = ALIST_PVAL(dl->readers, i);
		ASSERT(sc->conn->type == CONN_TYPE_INCOMING, "invalid type");

		conn_changestate(sc->conn, CONNECTION_STATUS_ENDED);
		dlagent_delclient(dl, sc);
		sc->dl = NULL;
	}
}


/* dlagent_status - list all dlagents */
void
dlagent_status(void)
{
	struct dlagent *l_dl;

	if (SLIST_EMPTY(&dlagents))
		return;

	dbg("dlagents:");
	SLIST_FOREACH(l_dl, &dlagents, n) {
		dbg(" %s, range %lld-%lld (%s), refcnt: %d", l_dl->ue->uri.uri,
		    (long long)l_dl->start, (long long)l_dl->end,
		    dlagent_state2str(l_dl->state), l_dl->refcnt);
	}
}


/* dlagen_state2str - return string representation of dlagent state */
char *
dlagent_state2str(enum dlstate state)
{
	switch (state) {
	case DLAGENT_STATE_INACTIVE:
		return "DLAGENT_STATE_INACTIVE";

	case DLAGENT_STATE_PEERS:
		return "DLAGENT_STATE_PEERS";

	case DLAGENT_STATE_GET:
		return "DLAGENT_STATE_GET";

	case DLAGENT_STATE_COMPLETE:
		return "DLAGENT_STATE_COMPLETE";

	case DLAGENT_STATE_MULTFAIL:
		return "DLAGENT_STATE_MULTFAIL";
	}

	ERRX("no state matched");
	/*NOTREACHED*/
}
