/* p2pd.c - main spp request processing */

/*
 * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: p2pd.c,v 1.191 2010/10/06 18:51:33 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/stat.h>

#include <err.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "p2pd.h"
#include "block.h"
#include "file.h"
#include "filter.h"
#include "callback.h"
#include "conn.h"
#include "diskcache.h"
#include "memcache.h"
#include "mkdigest.h"
#include "igot.h"
#include "iostat.h"
#include "mconn.h"
#include "peer.h"
#include "proc.h"
#include "queue.h"
#include "recov.h"
#include "resolve.h"
#include "request.h"
#include "scc.h"
#include "serv.h"
#include "servfile.h"
#include "servop.h"
#include "socket.h"
#include "spp.h"
#include "ssl.h"
#include "stream.h"
#include "subnet.h"
#include "sumfile.h"
#include "urient.h"

/*
 * process requests, accept or deny commands
 */

char *lport = NULL;
char *servname = NULL; /* public name of server */
int lhc_mode; /* LHC mode; recursive requests accepted from localhost */
int scc_mode; /* SCC mode; recursive requests accepted from other hosts */
int serv_mode; /* server mode; authorative server for content */
int igot_sslonly = 0; /* require SSL connection for IGOT requests? */
int cleancache = 0;  /* clean cache upon exit */

char mysockname[NI_MAXHOST]; /*XXX local ip address */

struct subnet scc_rec_subnet;
struct subnet scc_http_subnet;

extern int debug;
extern int logstderr;
extern int igot_accept;
extern FILE *logfp;

void inthand(int);
void usage(void);


/* usage - print p2pd usage string and exit */
void
usage(void)
{
	errx(1, "usage: p2pd %s\n%s\n%s\n%s",
	     "[-sdDit] [-A host name] [-b block size] [-C diskcache size]",
	     "  [-c cachedir] [-L logfile] [-l mirror] [-m default digest]",
	     "  [-M memcache size] [-N subnet] [-n subnet]",
	     "  [-p publicdir] [-P portnum] [-S address:port] [-Z]");
}


/* main - it all starts here */
int
main(int argc, char *argv[])
{
	const char *f = "main";
	struct sigaction act;
	enum logtype logtype;
	unsigned int diskmax;
	unsigned int memmax;
	const char *errstr;
	uint32_t blocksize;
	struct stat sb;
	int standalone;
	char *cachedir;
	char *altport;
	char *logfile;
	char *pidport;
	char *pubdir;
	char *mdname;
	int mirrors;
	int filter;
	int ch;

	bzero(mysockname, sizeof(mysockname));
	bzero(&scc_rec_subnet, sizeof(struct subnet));
	bzero(&scc_http_subnet, sizeof(struct subnet));

	urient_preinit();
	peer_preinit(peer_opsinit, peer_opsfree);
	queue_preinit();
	filter_preinit();

	opterr = 0;

	blocksize = BLOCKSIZE_DEFAULT;
	standalone = 0;
	diskmax = 0;
	memmax = 0;
	mdname = NULL;
	altport = NULL;
	pubdir = NULL;
	cachedir = NULL;
	logfile = NULL;
	pidport = NULL;
	igot_accept = 0;
	filter = 0;
	mirrors = 0;

	/* make lhc mode the default */
	lhc_mode = 0;
	scc_mode = 0;
	serv_mode = 0;
	while ((ch = getopt(argc, argv, "A:b:C:c:dDg:iL:l:m:M:N:n:P:p:rsS:TVZ")) != -1) {
		switch (ch) {
		case 'A':
			filter = 1;
			if (filter_passhost(optarg) == -1)
				errx(1, "invalid filter specification: %s",
				     optarg);
			break;

		case 'b':
			blocksize = strtonum(optarg, 0, FILESIZE_MAX, &errstr);
			if (errstr || !block_validsize(blocksize))
				errx(1, "blocksize: %s", errstr);
			break;

		case 'D':
			standalone = 1;
			break;

		case 'd':
			debug = 1;
			break;

		case 'C': /* cache size */
			diskmax = strtonum(optarg, 1, 10000, &errstr);
			if (errstr)
				errx(1, "cache size: %s", errstr);
			break;

		case 'c':
			if (serv_mode)
				errx(1, "cannot use both -p and -c");
			lhc_mode = 1;
			if ((stat(optarg, &sb) == -1) ||
			    (sb.st_mode & S_IFDIR) == 0)
				errx(1, "invalid cachedir: %s: %s", optarg,
				     strerror(errno));
			cachedir = optarg;
			break;

		case 'g': /* store pid + bound port number in file */
			pidport = optarg;
			break;

		case 'i':
			/* enable p2p/client sharing mode */
			igot_accept = 1;
			break;

		case 'L': /* specify logfile */
			logfile = optarg;
			break;

		case 'l': /* add mirror */
			mirrors++;
			if (serv_parsemirror(optarg) == -1)
				errx(1, "invalid -l option");
			break;

		case 'm': /* set default message digest */
			mdname = optarg;
			break;

		case 'M': /* mem cache size */
			memmax = strtonum(optarg, 1, 1000, &errstr);
			if (errstr)
				errx(1, "memory cache size: %s", errstr);
			break;

		case 'N': /* http access */
			if (subnet_parse(&scc_http_subnet, optarg) == -1)
				errx(1, "invalid subnet specification: %s",
				    optarg);
			break;

		case 'n': /* recursion access */
			if (serv_mode)
				errx(1, "cannot use both -n and -p");
			/* enable SCC mode if -n argument is given */
			lhc_mode = 0;
			scc_mode = 1;
			if (subnet_parse(&scc_rec_subnet, optarg) == -1)
				errx(1, "invalid subnet specification: %s",
				    optarg);
			break;

		case 'P':
			altport = optarg;
			break;

		case 'p':
			if (lhc_mode || scc_mode)
				errx(1, "cannot use -p in lhc/scc mode");
			/* enable server mode if -p is given */
			serv_mode = 1;
			if ((stat(optarg, &sb) == -1) ||
			    (sb.st_mode & S_IFDIR) == 0)
				errx(1, "invalid pubdir: %s: %s", optarg,
				     strerror(errno));
			pubdir = optarg;
			break;

		case 'S':
			if (scc_parsearg(optarg) == -1)
				errx(1, "invalid address: %s", optarg);
			break;

		case 's':
			serversumfiles = 1;
			break;

		case 'T':
			http_servcompat = 1;
			break;

		case 'V':
#if HAVE_OPENSSL
			igot_sslonly = 1;
#else
			errx(1, "the -V option requires OpenSSL");
#endif /* HAVE_OPENSSL */
			break;

		case 'Z':
			cleancache = 1;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (!lhc_mode && !scc_mode && !serv_mode)
		errx(1, "need to specify one of -c and -p");

	if ((scc_rec_subnet.active || scc_http_subnet.active) && !scc_mode)
		errx(1, "-n and -N are only used in scc mode");

	if (lhc_mode && igot_accept)
		errx(1, "-i option only applies in scc or server mode");

	if (serv_mode && filter)
		errx(1, "filtering requested in server mode");

	if (mirrors && !serv_mode)
		errx(1, "-l option only valid in server mode");

	/* XXX might want option in lhc mode to forward http requests */
	if (http_servcompat && !serv_mode)
		errx(1, "-T option is only valid in server mode");

	/* loginit, no syslog output if debugging */
	if (logfile != NULL)
		logtype = log_file;
	else if (standalone)
		logtype = log_syslog;
	else
		logtype = log_stderr; /* log to stderr if not detaching */
	if (log_preinit(APPNAME_P2PD, logtype, logfile) == -1)
		serr("log_preinit");

	/*
	 * check option sanity
	 */
	if (serv_mode && cleancache) {
		sinfo("%s: error: -Z cannot be used in server mode", f);
		usage();
	}

	if (scc_mode && serv_mode) {
		sinfo("%s: error: both SCC and server mode cannot be "
		      "enabled simultanously", f);
		usage();
	}

	if (serv_mode) {
		if (pubdir == NULL) {
			sinfo("%s: error: pubdir not specified.", f);
			usage();
		}
	} else {
		ASSERT(lhc_mode || scc_mode, "invalid mode");

		if (cachedir == NULL) {
			sinfo("%s: error: cachedir not specified.", f);
			usage();
		}
	}

	/* digest only makes sense on server */
	if (!serv_mode && mdname != NULL) {
		sinfo("%s: error: -m option can only be used by server", f);
		usage();
	}

	/*
	 * start setup
	 */

	memcache_preinit(memmax);
	callback_preinit();
	/* both http and spp handled by spp code */
	callback_add(PROTNAME_SPP,
		     (void (*)(void *))accept_request,
		     (void *(*)(void *))sppfile_new);
	if (lhc_mode ||
	    (scc_mode && scc_http_subnet.active) ||
	    (serv_mode && http_servcompat))
		callback_add(PROTNAME_HTTP,
			     (void (*)(void *))accept_request,
			     NULL); /* no caching of http files? */

	digest_preinit(mdname);
#if HAVE_OPENSSL
	if (ssl_preinit() == -1) {
		if (igot_sslonly)
			serr("ssl init failed");
	}
#endif /* HAVE_OPENSSL */

	if (altport != NULL) {
		if (strcmp(altport, "_any_") == 0)
			/* magic value '_any_', let kernel decide */
			lport = bind_port(NULL);
		else
			lport = bind_port(altport);
	} else {
		if ((lport = bind_port(SPP_SERV)) == NULL)
			/* try once more with numeric port number */
			lport = bind_port(SPP_PORT);
	}
	if (lport == NULL)
		errx(1, "unable to bind any sockets");

	connstatus();

	/* run as standalone daemon? */
	if (standalone)
		if (daemon(0, 0) == -1)
			err(1, "daemon");

	if (pidport != NULL) {
		FILE *fp;

		if ((fp = fopen(pidport, "w")) == NULL)
			serr("fopen: %s: %m", pidport);
		if (fprintf(fp, "%d\n%s\n", getpid(), lport) == -1)
			serr("write to pid file failed");
		fclose(fp);
	}

	/* ignore SIGPIPE */
	bzero(&act, sizeof(act));
	act.sa_handler = SIG_IGN;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	if (sigaction(SIGPIPE, &act, NULL) == -1)
		serr("sigaction");

	/* exit handler */
	bzero(&act, sizeof(act));
	act.sa_handler = inthand;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	if (sigaction(SIGTERM, &act, NULL) == -1)
		serr("sigaction");

	/* XXX during debug mode only? */
	bzero(&act, sizeof(act));
	act.sa_handler = inthand;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	if (sigaction(SIGINT, &act, NULL) == -1)
		serr("sigaction");

	bzero(&ios, sizeof(struct iostats));

	if (lhc_mode || scc_mode) {
		dbg("%s: calling proc_init", f);
		proc_preinit();
		/*XXX should have error return value if fork fails */
		proc_init(PROC_RESOLVE, 1, 2, 0, PROC_TYPE_GROW, resolve_serv);
		proc_init(PROC_MCONN, 1, 2, 0, PROC_TYPE_GROW, mconn_serv);
	}

	if (serv_mode)
		if (servfile_preinit(blocksize, pubdir) == -1)
			serr("server startup failed");

	if (diskmax != 0 && cachedir == NULL)
		dbg("%s: disk cache size specified, but not -c", f);

	/*XXX parse cache before binding socket? */
	if (cachedir != NULL) {
		diskcache_preinit(diskmax, cachedir);
		diskcache_loadidx();

		/* let server know that we are back (XXX flag to disable?) */
		dbrefresh_all();
	}

	/* XXX might try to regenerate cache in case of lhc_mode */
	if (scc_mode || serv_mode)
		recovery_check();

	/* create status file immediately after running recovery check */
	server_alive();

	/* XXX read leechscores (from db or similar) */

	if (lhc_mode)
		sinfo("%s: starting normal operation (LHC)", f);
	else if (scc_mode)
		sinfo("%s: starting normal operation (SCC)", f);
	else if (serv_mode)
		sinfo("%s: starting normal operation (server)", f);
	else
		ERRX("no mode specified");

	if (parent_scc)
		dbg("%s: using proxy: %s:%s", f, parent_name, parent_port);

	if (scc_mode || serv_mode) {
		if (igot_accept)
			sinfo("%s: p2p mode enabled, adding clients to peerlist", f);
		else
			sinfo("%s: p2p mode disabled, peerlist is fixed", f);
	}
	if (http_servcompat)
		sinfo("%s: http compatibility mode enabled", f);

	if (!serv_mode)
		sppfile_revconninit();

	sinfo("%s: using blocksize: %d", f, blocksize);

	apptimers_preinit();

	serverop();

	sinfo("%s: ending normal operation", f);

	p2pd_exit(0);
	/* NOTREACHED */
}


/* inthand - signal handler for SIGTERM/INT */
void
inthand(int sig)
{
	if (sig == SIGTERM || sig == SIGINT)
		pollend();
}


/* p2pd_exit - called upon server shutdown */
void
p2pd_exit(int status)
{
	if (logfp != NULL)
		fflush(logfp);

	if (cleancache)
		diskcache_clean();
	else if (lhc_mode  || scc_mode)
		diskcache_saveidx();

	proc_shutdown(PROC_RESOLVE);
	proc_shutdown(PROC_MCONN);
	safe_shutdown();
	exit(status);
}
