/* parse_igots.c - IGOT data parsing function */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: parse_igots.c,v 1.15 2010/06/30 08:37:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>

#include "p2pd.h"
#include "igot.h"


/* igots_parse - parse client IGOT entries */
int
igots_parse(struct igot *igots, int size, char *buf)
{
	const char *f = "igots_parse";
	char *bnum, *digest, *addr, *port, *credit;
	const char *errstr;
	size_t len;
	char *line;
	char *p;
	char *e;
	int i;

	ASSERT(size > 0, "invalid array size");

/*	dbg("%s: %s", f, buf);*/

	i = 0;
	line = buf;
	for (;;) {
		/* end of data if no terminating newline */
		if ((e = strstr(line, "\r\n")) == NULL)
			break;
		*e = '\0';

		if (i >= size) {
			dbg("%s: too many IGOT entries, ignoring", f);
			break;
		}

		bnum = line;

		if ((p = strchr(line, ' ')) == NULL) {
			dbg("%s: invalid IGOT entry", f);
			return -1;
		}
		*p = '\0';
		digest = p + 1;

		addr = NULL;
		port = NULL;
		credit = NULL;
		if ((p = strchr(digest, ' ')) != NULL) {
			*p = '\0';
			addr = p + 1;

			if ((p = strchr(addr, ' ')) != NULL) {
				*p = '\0';
				port = p + 1;

				if ((p = strchr(port, ' ')) != NULL) {
					*p = '\0';
					credit = p + 1;
				}
			}
		}

		/* bnum */
		igots[i].bnum = (int32_t)strtonum(bnum, 0, FILEBLOCKS_MAX, &errstr);
		if (errstr) {
			dbg("%s: block number parse failure: %s", f, errstr);
			return -1;
		}

		/* digest */
		len = sizeof(igots[i].digest);
		if (strlcpy(igots[i].digest, digest, len) >= len) {
			dbg("%s: IGOT digest too long", f);
			return -1;
		}

		/* addr */
		if (addr != NULL) {
			len = sizeof(igots[i].addr);
			if (strlcpy(igots[i].addr, addr, len) >= len) {
				dbg("%s: IGOT peer address too long", f);
				return -1;
			}
		}

		/* port */
		if (port != NULL) {
			len = sizeof(igots[i].port);
			if (strlcpy(igots[i].port, port, len) >= len) {
				dbg("%s: IGOT peer port too long", f);
				return -1;
			}
		}

		/* credits */
		if (credit != NULL) {
			igots[i].cred = strtonum(credit, 0, CREDITS_MAX, &errstr);
			if (errstr) {
				dbg("%s: credits parse failure: %s", f, errstr);
				return -1;
			}
		}

		line = e + 2;
		i++;
	}

	if (i == 0) {
		dbg("%s: no IGOT entries found", f);
		return -1;
	} else
		return i;
}
