/* parse_req.c - SPP request parsing */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: parse_req.c,v 1.55 2010/10/06 08:17:24 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>

#include "p2pd.h"
#include "conn.h"
#include "spp.h"
#include "request.h"
#include "parse_req.h"
#include "uri.h"


/* request_parse - parse spp/http request */
char *
request_parse(struct req *req, char *buf)
{
	const char *f = "request_parse";
	char fulluri[URI_MAXLEN];
	char *p;

/*	dbg("%s: : %s", f, buf);*/

	if ((p = parse_reqcmd(req, buf)) == NULL)
		return NULL;

	for (;;) {
		/* empty line (end of request)? */
		if (strncmp(NEWLINE, p, strlen(NEWLINE)) == 0) {
			p += strlen(NEWLINE);
			break;
		}

		if ((p = parse_reqopt(req, p)) == NULL)
			return NULL;
	}

	/*
	 * Populate struct uri, basically following HTTP1.1 RFC2616.
	 *
	 * The Host header should be ignored in case of an absolute
	 * URI. Use the request URI if possible, otherwise attempt to
	 * construct an URI based on the Host header.
	 */
	if (strstr(req->requri, "://") != NULL) {
		if (uri_parse(&req->uri, req->requri) == -1)
			return NULL;

		/* sanity checking of request URI */
		if (strncasecmp(SCHEME_HTTP, req->uri.scheme,
				strlen(SCHEME_HTTP)) == 0) {
			if (req->http == 0) {
				dbg("%s: HTTP URI in non-HTTP request", f);
				return NULL;
			}

			/* replace scheme with spp */
			strlcpy(req->uri.scheme, SCHEME_SPP,
				sizeof(req->uri.scheme));
			/* port number missing? */
			if (req->uri.port[0] == '\0')
				strlcpy(req->uri.port, HTTP_PORT,
					sizeof(req->uri.port));
			/* rebuild normalized uri */
			if (uri_joinparts(&req->uri) == -1)
				return NULL;
		} else if (strncasecmp(SCHEME_SPP, req->uri.scheme,
				       strlen(SCHEME_SPP)) == 0) {
			if (req->http == 1) {
				dbg("%s: SPP URI in HTTP request", f);
				return NULL;
			}
			/* port number missing? */
			if (req->uri.port[0] == '\0')
				strlcpy(req->uri.port, SPP_PORT,
					sizeof(req->uri.port));
			/* rebuild normalized uri */
			if (uri_joinparts(&req->uri) == -1)
				return NULL;
		} else {
			dbg("%s: unsupported scheme '%s'", f, req->uri.scheme);
			return NULL;
		}
	} else {
		/*
		 * The request is not a complete URI, attempt to
		 * create a full URI based on known values.
		 */

		/* scheme, always set spp */
		strlcpy(fulluri, SCHEME_SPP, sizeof(fulluri));
		strlcat(fulluri, "://", sizeof(fulluri));

		/* host */
		if (strlen(req->reqhost) == 0) {
			dbg("%s: incomplete request, missing host", f);
			return NULL;
		}

		/* XXX might require encoding */
		strlcat(fulluri, req->reqhost, sizeof(fulluri));

		/* port */
		if (strlen(req->reqport) > 0) {
			strlcat(fulluri, ":", sizeof(fulluri));
			strlcat(fulluri, req->reqport, sizeof(fulluri));
		}

		/* path */
		if (req->requri[0] != '/')
			strlcat(fulluri, "/", sizeof(fulluri));
		strlcat(fulluri, req->requri, sizeof(fulluri));

/*		dbg("%s: constructing uri: %s", f, fulluri);*/

		if (uri_parse(&req->uri, fulluri) == -1)
			return NULL;
	}

	/*
	 * HTTP clients require return code 206 to indicate support
	 * for partial requests (SPP requests are by their nature
	 * partial and always return 200).
	 */
	if ((req->http == 1) && (req->offset != 0))
		req->flags |= REQ_FLAG_PROXYRES;

	/* unexpected content-length header? */
	if ((req->size > 0) && (req->cmd != CLIENT_CMD_IGOT)) {
		dbg("%s: content-length header only valid with IGOT", f);
		return NULL;
	}

	/* unexpected subseq-blocks? */
	if ((req->bcnt > 0) && (req->cmd != CLIENT_CMD_REVCONN)) {
		dbg("%s: subsequent-blocks header only valid with REVCONN", f);
		return NULL;
	}

	return p;
}


/* parse_reqcmd - parse request command line and return pointer to next line */
char *
parse_reqcmd(struct req *req, char *line)
{
	const char *f = "parse_reqcmd";
	size_t maxlen;
	char *uri;
	char *ver;
	char *e;

	/* all commands should end in a '\r\n' pair */
	if ((e = strstr(line, "\r\n")) == NULL)
		ERRX("line not newline terminated");
	*e = '\0';

/*	dbg("%s: %s", f, line);*/

	if (strncmp(COMMAND_CRED, line, strlen(COMMAND_CRED)) == 0) {
		req->cmd = CLIENT_CMD_CRED;
		uri = line + strlen(COMMAND_CRED);
	} else if (strncmp(COMMAND_GET, line, strlen(COMMAND_GET)) == 0) {
		req->cmd = CLIENT_CMD_GET;
		uri = line + strlen(COMMAND_GET);
	} else if (strncmp(COMMAND_HEAD, line, strlen(COMMAND_HEAD)) == 0) {
		req->cmd = CLIENT_CMD_HEAD;
		uri = line + strlen(COMMAND_HEAD);
	} else if (strncmp(COMMAND_IGOT, line, strlen(COMMAND_IGOT)) == 0) {
		req->cmd = CLIENT_CMD_IGOT;
		uri = line + strlen(COMMAND_CRED);
	} else if (strncmp(COMMAND_PEERS, line, strlen(COMMAND_PEERS)) == 0) {
		req->cmd = CLIENT_CMD_PEERS;
		uri = line + strlen(COMMAND_PEERS);
	} else if (strncmp(COMMAND_PING, line, strlen(COMMAND_PING)) == 0) {
		req->cmd = CLIENT_CMD_PING;
		uri = line + strlen(COMMAND_PING);
	} else if (strncmp(COMMAND_REVCONN, line, strlen(COMMAND_REVCONN)) == 0) {
		req->cmd = CLIENT_CMD_REVCONN;
		uri = line + strlen(COMMAND_REVCONN);
	} else if (strncmp(COMMAND_STARTTLS, line, strlen(COMMAND_STARTTLS)) == 0) {
		req->cmd = CLIENT_CMD_STARTTLS;
		uri = line + strlen(COMMAND_STARTTLS);
	} else {
		dbg("%s: invalid command in request", f);
		return NULL;
	}

	/* should be only a single separating space (included in COMMAND_FOO) */
	if (*uri == ' ') {
		dbg("%s: invalid request: %s", f, line);
		return NULL;
	}

	/* SPP/HTTP version should be after second space  */
	if ((ver = strchr(uri, ' ')) == NULL) {
		dbg("%s: invalid request: %s", f, line);
		return NULL;
	}

	*ver = '\0'; /* end of request URI at this point */
	ver++;

	maxlen = sizeof(req->requri);
	if (strlcpy(req->requri, uri, maxlen) >= maxlen) {
		dbg("%s: uri too long", f);
		return NULL;
	}

	if (strncmp(VALUE_VERSION_SPP, ver, strlen(VALUE_VERSION_SPP)) != 0) {
		if ((strncmp(VALUE_VERSION_HTTP10, ver,
			     strlen(VALUE_VERSION_HTTP10)) == 0) ||
		    (strncmp(VALUE_VERSION_HTTP11, ver,
			     strlen(VALUE_VERSION_HTTP11)) == 0)){
			req->http = 1;
			dbg("%s: http request detected", f);
		} else {
			dbg("%s: invalid request type: %s", f, ver);
			return NULL;
		}
	}

	return e + 2; /* next line */
}


/* parse_reqopt - parse request option fields */
char *
parse_reqopt(struct req *req, char *line)
{
	const char *f = "parse_reqopt";
	const char *errstr;
	char *p, *q, *e;
	size_t maxlen;

	/* options should end in a single "\r\n" pair */
	if ((e = strstr(line, "\r\n")) == NULL)
		ERRX("line not newline terminated");

	*e = '\0';

/*	dbg("%s: %s", f, line);*/

	if (strncmp(OPTION_RANGE, line, strlen(OPTION_RANGE)) == 0) {
		p = line + strlen(OPTION_RANGE);

		/* XXX blocks=%d-%d? */
		if (strncmp(PARAM_BYTERANGE, p, strlen(PARAM_BYTERANGE)) == 0)
			p += strlen(PARAM_BYTERANGE);
		else {
			dbg("%s: unsupported range unit: %s", f, p);
			return NULL;
		}

		/* look for interval char */
		if ((q = strchr(p, '-')) != NULL) {
			*q = '\0';
			q++;
		}

		/* parse first value */
		req->offset = strtonum(p, 0, FILESIZE_MAX, &errstr);
		if (errstr) {
			dbg("%s: unable to parse range header: %s", f, errstr);
			return NULL;
		}

		/* parse second value, if it exists */
		if ((q != NULL) && (strlen(q) > 0)) {
			req->offend = strtonum(q, 0, FILESIZE_MAX, &errstr);
			if (errstr) {
				dbg("%s: range value: %s", f, errstr);
				return NULL;
			}
			req->withend = 1;
		}

		if ((req->offend > 0) && (req->offset > req->offend)) {
			dbg("%s: invalid range: %lld > %lld", f,
			    (long long)req->offset, (long long)req->offend);
			return NULL;
		}
	} else if (strncmp(OPTION_RTT, line, strlen(OPTION_RTT)) == 0) {
		p = line + strlen(OPTION_RTT);

		req->rtt = strtonum(p, 0, UINT16_MAX, &errstr);
		if (errstr)
			dbg("%s: ignoring rtt value: %s", f, errstr);
	} else if (strncmp(OPTION_HOST, line, strlen(OPTION_HOST)) == 0) {
		p = line + strlen(OPTION_HOST);

		if ((q = strchr(p, ':')) != NULL)
			*q = '\0';

		maxlen = sizeof(req->reqhost);
		if (strlcpy(req->reqhost, p, maxlen) >= maxlen) {
			dbg("%s: hostname too long", f);
			return NULL;
		}

		/* port number? */
		if (q != NULL) {
			q++;
			maxlen = sizeof(req->reqport);
			if (strlcpy(req->reqport, q, maxlen) >= maxlen) {
				dbg("%s: port number too long", f);
				return NULL;
			}
		}
	} else if (strncmp(OPTION_AGENT, line, strlen(OPTION_AGENT)) == 0) {
		p = line + strlen(OPTION_AGENT);
#if 0
		if (strstr(line, "vlc media player") != NULL)
			req->flags |= REQ_FLAG_FOO;
#endif
		dbg("%s: user agent: %s", f, p);
	} else if (strncmp(OPTION_BINDPORT, line,
			   strlen(OPTION_BINDPORT)) == 0) {
		p = line + strlen(OPTION_BINDPORT);

		maxlen = sizeof(req->bindport);
		if (strlcpy(req->bindport, p, maxlen) >= maxlen) {
			dbg("%s: port number too long", f);
			return NULL;
		}
		(void)strtonum(p, 1, 65536, &errstr);
		if (errstr) {
			dbg("%s: bindport: %s", f, errstr);
			return NULL;
		}
	} else if (strncmp(OPTION_CONTLEN, line, strlen(OPTION_CONTLEN)) == 0) {
		p = line + strlen(OPTION_CONTLEN);

		req->size = strtonum(p, 0, FILESIZE_MAX, &errstr);
		if (errstr) {
			dbg("%s: content length: %s", f, errstr);
			return NULL;
		}
	} else if (strncmp(OPTION_NOUNCE, line, strlen(OPTION_NOUNCE)) == 0) {
		p = line + strlen(OPTION_NOUNCE);

		req->have_nounce = 1;
		req->nounce = strtonum(p, 0, SPP_NOUNCE_MAX, &errstr);
		if (errstr) {
			dbg("%s: nounce: %s", f, errstr);
			return NULL;
		}
	} else if (strncmp(OPTION_SUBSEQ, line, strlen(OPTION_SUBSEQ)) == 0) {
		p = line + strlen(OPTION_SUBSEQ);

		req->bcnt = strtonum(p, 0, FILEBLOCKS_MAX, &errstr);
		if (errstr) {
			dbg("%s: subsequent-blocks: %s", f, errstr);
			return NULL;
		}
	} else {
		if (req->http == 1) {
			dbg("%s: ignoring HTTP option: %s", f, line);
		} else {
			dbg("%s: invalid option: %s", f, line);
			return NULL;
		}
	}

	return e + 2;
}
