/* parse_resp.c - spp response parsing */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: parse_resp.c,v 1.25 2010/06/30 08:37:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>

#include "p2pd.h"
#include "spp.h"
#include "parse_resp.h"


/* response_parse - parse response XXX similar to request_parse() */
char *
response_parse(struct rep *rep, char *buf)
{
/*	const char *f = "response_parse";*/
	char *p;

/*	dbg("%s: %s", f, buf);*/

	if ((p = parse_respcode(rep, buf)) == NULL)
		return NULL;

	for (;;) {
		/* empty line (end of request)? */
		if (strncmp(NEWLINE, p, strlen(NEWLINE)) == 0) {
			p += strlen(NEWLINE);
			break;
		}

		if ((p = parse_respopt(rep, p)) == NULL)
			return NULL;
	}

	return p;
}


/* parse_respcode - parse response status message */
char *
parse_respcode(struct rep *rep, char *line)
{
	const char *f = "parse_respcode";
	const char *errstr;
	char *code, *msg;
	char *e;

	/*XXX data might not be null terminated (data after get) */
	/* each response line should end in a "\r\n" pair */
	if ((e = strstr(line, "\r\n")) == NULL)
		ERRX("line not newline terminated");
	*e = '\0';

/*	dbg("%s: %s", f, line);*/

	if (strncmp(VALUE_VERSION_SPP, line, strlen(VALUE_VERSION_SPP)) == 0) {
		code = line + strlen(VALUE_VERSION_SPP);
		if (*code == ' ')
			code++; /* skip space */
		else {
			dbg("%s: invalid response format: %s", f, line);
			return NULL;
		}

		if ((msg = strchr(code, ' ')) != NULL)
			*msg = '\0';

		rep->code = strtonum(code, 100, 599, &errstr);
		if (errstr) {
			dbg("%s: unable to parse response code: %s", f, errstr);
			return NULL;
		}

		/* XXX preserve response message? */
	} else {
		dbg("%s: invalid response format: %s", f, line);
		return NULL;
	}

	return e + 2;
}


/* parse_respopt - parse response option keyword */
char *
parse_respopt(struct rep *rep, char *line)
{
	const char *f = "parse_respopt";
	const char *errstr;
	char *num;
	char *e;

	/* each response line should end in a "\r\n" pair */
	if ((e = strstr(line, "\r\n")) == NULL)
		ERRX("line not newline terminated");
	*e = '\0';

/*	dbg("%s: %s", f, line); */

	if (strncmp(OPTION_BLOCKSIZE, line, strlen(OPTION_BLOCKSIZE)) == 0) {
		num = line + strlen(OPTION_BLOCKSIZE);

		rep->blocksize = strtonum(num, 0, FILESIZE_MAX, &errstr);
		if (errstr) {
			dbg("%s: blocksize: %s", f, errstr);
			return NULL;
		}
	} else if (strncmp(OPTION_CONTLEN, line, strlen(OPTION_CONTLEN)) == 0) {
		num = line + strlen(OPTION_CONTLEN);

		rep->size = strtonum(num, 0, FILESIZE_MAX, &errstr);
		if (errstr) {
			dbg("%s: content length: %s", f, errstr);
			return NULL;
		}

	} else if (strncmp(OPTION_DBREFRESH, line,
			   strlen(OPTION_DBREFRESH)) == 0) {
		/* OPTION_DBREFRESH includes only legal parameter */
		dbg("%s: got database refresh request", f);
		rep->dbrefresh = 1;
	} else if (strncmp(OPTION_NOUNCE, line, strlen(OPTION_NOUNCE)) == 0) {
		num = line + strlen(OPTION_NOUNCE);

		rep->have_nounce = 1;
		rep->nounce = strtonum(num, 0, SPP_NOUNCE_MAX, &errstr);
		if (errstr) {
			dbg("%s: nounce: %s", f, errstr);
			return NULL;
		}
	} else if (strncmp(OPTION_SUBSEQ, line, strlen(OPTION_SUBSEQ)) == 0) {
		num = line + strlen(OPTION_SUBSEQ);
		/*XXX only valid for some commands */
		rep->bcnt = strtonum(num, 0, FILEBLOCKS_MAX, &errstr);
		if (errstr) {
			dbg("%s: subsequent-blocks: %s", f, errstr);
			return NULL;
		}
	} else {
/*		dbg("response data: %s 0x%02x %02x", line, line[0], line[1]);*/
		dbg("%s: invalid response header: %s", f, line);
		return NULL;
	}

	return e + 2;
}
