/* peerlist.c - functions for generation of rate ordered peerlists */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: peerlist.c,v 1.12 2010/06/30 11:19:58 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <math.h>
#include <string.h>

#include "p2pd.h"
#include "filemeta.h"
#include "ip.h"
#include "urient.h"
#include "peer.h"
#include "peerlist.h"
#include "peerset.h"
#include "tcpmodel.h"

/*
 *  XXX want functions for working on ranks, e.g. rankcount,
 *  ranklength, rank_shift (shift out all values to next rank)
 *
 *  would be useful in ratesort function; first rank by score,
 *  then subsort each rank by rate.
 */

/*
 * XXX
 * -return all peers matching uri/block number
 * -return all peers matching uri
 * -return all uris matching peer
 * (-return all peers with uri and N next blocks
 */


/* peerlist_gen - return peerlist with peers having given block */
struct peerset *
peerlist_gen(struct urient *ue, int bnum)
{
	struct peerset *pset;
	struct fm_peer *pf;
	struct alist *al;
	int i;

/*	dbg("%s: (%p, %d)", f, uri, bnum);*/

	pset = pset_create();

	al = ue->meta->al_peers;
	for (i = 0; (size_t)i < ALIST_SIZE(al); i++) {
		pf = ALIST_PVAL(al, i);
		/* XXX bnum == -1 for all peers */
		if ((bnum >= 0) && bit_test(pf->bm.map, bnum) == 0)
			continue; /* no match */

		/* entry matches all criteria, add to peerlist */
		pset_addpeer(pset, pf->peer);
	}

	/* return pset if any matches found */
	if (pset->peercnt == 0) {
		pset_free(pset);
		return NULL;
	} else
		return pset;
}


/* peerlist_allnew - returns true if all peers in peerset lack rate/rtt info */
int
peerlist_allnew(struct peerset *pset)
{
	struct peer *peer;
	int i;

	/* return false if set is empty */
	if (pset->peercnt == 0)
		return 0;

	for (i = 0; i < pset->peercnt; i++) {
		peer = pset_peer(pset, i);
#if 0 /*XXX server will have rtt */
		if ((peerpar_getfloat2(peer, PEERPAR_LINK_RTT, NULL) != -1) ||
		    (peerpar_getfloat2(peer, PEERPAR_LINK_RATE, NULL) != -1))
			break;
#else
		if (peerpar_getfloat2(peer, PEERPAR_LINK_RATE, NULL) != -1)
			break;
#endif
	}

	if (i == pset->peercnt)
		return 1;
	else
		return 0;
}


/* peerlist_firstsort - sort based on server rank and shared ip prefix */
void
peerlist_firstsort(struct peerset *pset)
{
	const char *f = "peerlist_firstsort";

	dbg("%s", f);

	/* rank based on server score values (best peer has lowest sscore) */
	pset_sort(pset, PEERPAR_SPP_SSCORE, 0);
	pset_setintrank(pset, PEERPAR_SPP_SSCORE, PEERPAR_TMP_INT1);

	/* rank based on ip distance (best peer has lowest ipdist) */
	pset_applyint(pset, PEERPAR_LINK_IPDIST, mysockname, peer_ipdist);
	pset_sort(pset, PEERPAR_LINK_IPDIST, 0);
	pset_setintrank(pset, PEERPAR_LINK_IPDIST, PEERPAR_TMP_INT2);

	/* add rank values and sort (best peer should have lowest sum) */
	pset_fieldop(pset, PEERPAR_TMP_INT1, PEERPAR_FOP_ADD,
		     PEERPAR_TMP_INT2, PEERPAR_TMP_INT3);
	pset_sort(pset, PEERPAR_TMP_INT3, 0);

/*	peerlist_dump(pset);*/
}


/* peerlist_ratesort - sort peers, first based on score, then estimated rate */
void
peerlist_ratesort(struct peerset *pset)
{
	const char *f = "peerlist_ratesort";
	int origcnt;
	int ival;

	dbg("%s", f);

	if (pset->peercnt == 0)
		return;

	/* store orignal size */
	origcnt = pset->peercnt;

	/* XXX could do next op with intrank + subsort */
	/* sort peers based on peerscore, then trim nodes with negative score */
	pset_applyint(pset, PEERPAR_TMP_INT1, NULL, peerscore);
	ival = 0;
	pset_evalsofttrim(pset, PEERPAR_TMP_INT1, PEERPAR_CMPOP_LT, &ival);

	/* revert to full size now if no non-negative nodes */
	if (pset->peercnt == 0)
		pset_softtrim(pset, origcnt);

	/* use firstsort if no latency/rate information exists */
	if (peerlist_allnew(pset)) {
		peerlist_firstsort(pset);
		pset_softtrim(pset, origcnt);
		return;
	}
	/* rate estimate based sorting of ok peers */
	if (pset->peercnt > 1)
		pset_applyfloat(pset, PEERPAR_SPP_BESTEST, NULL, bestest_rate);
	pset_sort(pset, PEERPAR_SPP_BESTEST, 1);

	/* revert size of working set */
	pset_softtrim(pset, origcnt);

	/* XXX perform peerlist_bestshift here, or leave to caller? */

/*	peerlist_dump(pset);*/
}


/* peerlist_servsort - sort based on estimated suitability for client */
void
peerlist_servsort(struct peerset *pset, struct peer *client)
{
	const char *f = "peerlist_servsort";
	float peerrtt;

	if (pset->peercnt == 0)
		return;

	/* XXX can use ipdist here too */

#if 0
	pset_dump(pset, PEERPAR_PEER_LASTACTIVE);
	dbg("%s: peercnt: %d", f, pset->peercnt);
#endif
	peerrtt = peerpar_getfloat(client, PEERPAR_LINK_RTT);
	dbg("%s: peerrtt: %f", f, peerrtt);
	if (peerrtt > 0) {
		/* sort based on estimated RTT distance */
		pset_applyfloat(pset, PEERPAR_TMP_FLOAT1, &peerrtt,
				rttdist_peer);
		pset_sort(pset, PEERPAR_TMP_FLOAT1, 0);
		dbg("%s: pltrrdist sort:", f);
		pset_dump(pset, PEERPAR_TMP_FLOAT1);
	}
}


/* peer_ipdist - return ip distance between peer and reference ip */
int
peer_ipdist(struct peer *peer, int u, void *refip)
{
	(void)u;

	/* ip_prefixbits returns shared bits, invert to get distance */
	return 32 - ip_prefixbits(peer->name, (char *)refip);
}


/* bestest_rate - function for estimating TCP transfer rate for peer */
float
bestest_rate(struct peer *peer, int u, void *u2)
{
	(void)u;
	(void)u2;

	return peer_rate(peer);
}


/* peerlist_bestshift - lastreq shift of fastest peers in sorted peerlist */
void
peerlist_bestshift(struct peerset *pset)
{
	float okrate;
	int origcnt;
	int ival;

	if (pset->peercnt <= 1)
		return;

	/* store set size */
	origcnt = pset->peercnt;

	/* reduce working set to nodes with similar speed as fastest node */
	okrate = peerpar_getfloat(pset_peer(pset, 0), PEERPAR_SPP_BESTEST);
	okrate /= peerdiv(pset_peer(pset, 0));
	pset_evalsofttrim(pset, PEERPAR_SPP_BESTEST, PEERPAR_CMPOP_LT, &okrate);

	/* reduce set to nodes with positive peeerscore */
	/*XXX PEER_PREVOK currently 0, need to remove negative peers */
	pset_applyint(pset, PEERPAR_TMP_INT1, NULL, peerscore);
	pset_sort(pset, PEERPAR_TMP_INT1, 1);
	ival = 0;
	pset_evalsofttrim(pset, PEERPAR_TMP_INT1, PEERPAR_CMPOP_LT, &ival);

	/* sort based on time of last request */
	pset_sort(pset, PEERPAR_SPP_LASTREQ, 0);

	/* revert softtrim, to allow connect to be attempted to all nodes */
	pset_softtrim(pset, origcnt);
}


/* peerscore - peerset function for peer sorting */
int
peerscore(struct peer *p, int u, void *u2)
{
	int s;

	(void)u;
	(void)u2;

	s = peerpar_getint(p, PEERPAR_SPP_LASTRES) * PEER_PREVWEIGHT -
	    peerpar_getint(p, PEERPAR_PEER_ACTIVECONN) * PEER_ACTIVECONNWEIGHT;

	return s;
}


/* rttdist_peer - calculate distance between peer rtt value and rttp */
float
rttdist_peer(struct peer *peer, int u, void *rttp)
{
	const char *f = "rttdist_peer";
	float refrtt, rtt;

	(void)u;

	refrtt = *(float *)rttp;

	if (peerpar_getfloat2(peer, PEERPAR_LINK_RTT, &rtt) != -1)
		rtt = fabs(rtt - refrtt);
	else
		rtt = PEER_RTT_UNKN;

	dbg("%s: rtt dist: %s: %f", f, peer->id, rtt);

	return rtt;
}


/* peerlist_dump - show peerlist order, with relevant values */
void
peerlist_dump(struct peerset *pset)
{
	struct peer *peer;
	int i;

	for (i = 0; i < pset->peercnt; i++) {
		peer = pset_peer(pset, i);
		dbg(" %d %s rtt: %f bw: %f est: %f ippre: %d sscore: %d",
		    i, peer->id, peerpar_getfloat(peer, PEERPAR_LINK_RTT),
		    peerpar_getfloat(peer, PEERPAR_LINK_RATE),
		    peerpar_getfloat(peer, PEERPAR_SPP_BESTEST),
		    peerpar_getint(peer, PEERPAR_LINK_IPDIST),
		    peerpar_getint(peer, PEERPAR_SPP_SSCORE));
	}
}
