/* recov.c - post-crash server recovery functions */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: recov.c,v 1.15 2010/06/30 08:38:00 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/stat.h>

#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "p2pd.h"
#include "recov.h"

/*
 * Server recovery
 *
 *  -improper server shutdown causes the server to enter a
 *   recovery mode where it asks clients to resend information
 *   about blocks stored locally.
 */

#define P2PD_RUNNING ".p2pd_running" /* existence indicates likely crash */
#define RECOVERY_INTERVAL (60 * 60 * 2) /* attempt recovery if crash recent */
#define RECOVERY_DURATION (60 * 5) /* duration of recovery mode */

int recovery = 0; /* if positive, seconds until recovery mode is finished */
int recov_disable = 0;


/* server_alive - refresh function called regularely during normal operation */
void
server_alive(void)
{
	const char *f = "server_alive";
	int fd;

	if (recov_disable)
		return;

	/* indicate liveliness */
	if ((fd = open(P2PD_RUNNING, O_WRONLY|O_TRUNC|O_CREAT, S_IRWXU)) == -1) {
		swarn("%s: creation of liveliness file: %m", f);
		swarn("%s: no recovery possible in case of failure", f);
		recov_disable = 1;
		return;
	}

	if (close(fd) == -1)
		serr("close: %m");
}


/* safe_shutdown - called during controlled server shutdown */
void
safe_shutdown(void)
{
	const char *f = "safe_shutdown";

	if (recov_disable)
		return;

	if (unlink(P2PD_RUNNING) == -1)
		dbg("%s: unable to remove liveliness file: %m", f);
}


/* recovery_check - called during startup to check for recent server failure */
void
recovery_check(void)
{
	const char *f = "recovery_check";
	struct stat sb;
	time_t now, tdiff;

	if (stat(P2PD_RUNNING, &sb) == 0) {
		now = time(NULL);
		tdiff = now - sb.st_mtime;
		if ((tdiff >= 0) && (tdiff <= RECOVERY_INTERVAL)) {
			swarn("%s: recent server failure detected"
			      "entering recovery mode", f);
			recovery = RECOVERY_DURATION;
		} else
			sinfo("%s: non-recent server failure detected"
			      "ignoring", f);
	}
}
