/* request.c - main spp request handling */

/*
 * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: request.c,v 1.550 2010/07/31 09:48:09 karls Exp $ */

/* XXX rename to spp_req */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "p2pd.h"
#include "conn.h"
#include "access.h"
#include "filemeta.h"
#include "filter.h"
#include "parse_req.h"
#include "peer.h"
#include "spp.h"
#include "request.h"
#include "serv.h"
#include "subnet.h"
#include "tcpmodel.h"
#include "urient.h"

/*
 *
 * We support this subset of the HTTP standard:
 *
 * URI:spp://spp.example.org/file.avi
 *
 * GET file.avi SPP/1.0
 * [Range: offset]
 * Host: spp.example.org[:port]
 * \n
 *
 * Range is optional
 *
 * TCP is used, STOP command not needed
 * Closing of socket stops transfer [? or closes connection]
 *
 * XXX port
 */

/*
 * define stream request as something without end offset or
 * with a request length longer than the file block size.
 */
#define streamreq(r, ue) ((r)->withend == 0 || \
			  (r)->offend + 1 - (r)->offset > (ue)->meta->blocksize)

int errno_spp = 0;

int handle_directreq(struct urient *, struct sppconn *, struct req *req);
int needrecursion(struct urient *, struct req *);


/* accept_request - new (spp) request received on connection */
void
accept_request(struct connection *c)
{
	const char *f = "accept_request";
	struct sppconn *sc;
	struct sppfile *ss;
	struct urient *ue;
	struct speer *sp;
	struct req *req;
	int err;
	char *p;

	ASSERT(c->uldat == NULL, "uldat set");

	req = xmalloc(sizeof(struct req));
	bzero(req, sizeof(struct req));
	/*XXX as long as connection has req pointer, unify handle_foo funcs */

	/*
	 * XXX allow conn layer to refuse connection if not recognizable as
	 * spp request (no valid uri obtainable)
	 */
	if ((p = request_parse(req, c->ibuf)) == NULL) {
		dbg("%s: request parsing failed", f);
		c->ibuflen = 0;
		conn_writerep(c, "SPP/1.0 400 Request parsing error\r\n\r\n");
		conn_changestate(c, CONNECTION_STATUS_ENDED);
		return;
	}
	conn_ibufsync(c, p);

	/* XXX ip+BOUND PORT based peer identification, requires request parsing */
	if (c->peer == NULL) {
		if (req->http)
			sp = speer_add(c->peername, "http");
		else {
			if (strlen(req->bindport) == 0)
				ERRX("no identity for peer");
			sp = speer_add(c->peername, req->bindport);
		}
		if (sp == NULL)
			ERRX("XXX unable to add peer");
		c->peer = SPTOP(sp);
		/*
		 * XXX network level values would be separate from peers;
		 * having separate information for network and peers,
		 * multiple peers can be on the same ip/subnet,
		 * aggregating some information would make sense.
		 * would allow these values to be set upon connection being
		 * made.
		 */
		peerpar_adjint(c->peer, PEERPAR_PEER_TOTCONN, 1);
		peerpar_setint(c->peer, PEERPAR_LINK_MSS, c->mss); /*XXX*/
	}

#if 0
	/*
	 * XXX setting status to something other than readreq makes
	 * netcat work, since netcat closes the connection after
	 * writing the request. the result is essentially that we do
	 * not detect this.
	 */
	conn_changestate(c, CONNECTION_STATUS_PROTNEG);
#else
	conn_changestate(c, CONNECTION_STATUS_READREQ);
#endif

	sc = sppconn_new(req->cmd);
	sppconn_setconn(sc, c);

	if (req->http) {
		sc->httpreq = 1;
		dbg("%s: treating request as HTTP request", f);
	}

	/* XXX if (scc_mode && !permitted_uri(uri)) deny */

	ss = sppfile_clientreq(&req->uri);
	if (ss == NULL)
		ue = NULL;
	else
		ue = SPPTOU(ss);

	if (!serv_mode && (ue == NULL || needrecursion(ue, req))) {
		/* permitted uri? */
		if (filter_checkuri(&req->uri) == -1) {
			swarn("%s: filter denied request from %s for %s", f,
			      c->peer->id, req->uri.uri);
			spp_cmd_rep(sc, 403);
			return;
		}
		/* recursive requests permitted for client? */
		if (!allow_recur(c)) {
			swarn("%s: denying recursive request from %s", f,
			      c->peer->id);
			spp_cmd_rep(sc, 403);
			return;
		} else
			dbg("%s: allowing recursive request from %s", f,
			    c->peer->id);
	}

	if (ue == NULL) {
		if (serv_mode)
			ue = sppfile_servinit(&req->uri);
		else if (scc_mode || lhc_mode)
			/* start retrieving meta data */
			ue = sppfile_headinit1(&req->uri);
		else
			ERRX("invalid mode");

		/* still NULL? */
		if (ue == NULL) {
			err = 404;
			dbg("%s: request failed (%d)", f, err);
			spp_cmd_rep(sc, err);
			return;
		}
	}
	ASSERT(ue != NULL, "NULL ue");

	ss = UTOSPP(ue); /*XXX*/
	/* XXX might want to do temporary failure for first head on scc */
	if (ss->state == SPP_STATE_ERR) {
		dbg("%s: request failed (%d)", f, ss->spp_err);
		spp_cmd_rep(sc, ss->spp_err);
		return;
	}

	sppconn_setfile(sc, ue);

	/* p2pd can currently be in only one mode at a time */
	if (serv_mode) {
		if (sc->httpreq) {
			/* http compatibility mode? */
			if (http_servcompat && (req->cmd == CLIENT_CMD_HEAD ||
			    req->cmd == CLIENT_CMD_GET)) {
				sc->currcmd = req->cmd;
				serv_httpreply(ue, sc);
			} else {
				swarn("%s: denying invalid http request", f);
				spp_cmd_rep(sc, 404);
			}
		} else if (handle_directreq(ue, sc, req) == -1)
			dbg("%s: request failure", f);
	} else {
		/* cache request */
		if (UTOSPP(ue)->headactive) {
			/* connection needs to wait for result of head */
			sc->req = req; /* free later */
			sppfile_waithead(UTOSPP(ue), sc);
			return;
		}

		handle_cachereq(ue, sc, req);
	}
	free(req);

/*	connstatus();*/
}


/* handle_directreq - direct request handling */
int
handle_directreq(struct urient *ue, struct sppconn *sc, struct req *req)
{
	const char *f = "handle_directreq";
	struct connection *c;
	int rtt;

	ASSERT(serv_mode == 1, "direct request, but not server");

	c = sc->conn;

	sc->dlstart = req->offset;
	sc->dlend = req->offend;

	/* req is freed if idle_connection is called */
	rtt = req->rtt;

	switch (req->cmd) {
	case CLIENT_CMD_CRED:
		handle_cred(sc, req);
		break;
	case CLIENT_CMD_GET:
		if (streamreq(req, ue)) {
			swarn("%s: direct stream requests denied", f);
			spp_cmd_rep(sc, 403);
			return -1;
		}

		if (handle_get(sc, req) == -1)
			return -1;
		break;
	case CLIENT_CMD_HEAD:
		handle_head(sc, req);
		break;
	case CLIENT_CMD_IGOT:
		if (handle_igot(sc, req) == -1)
			return -1;
		break;
	case CLIENT_CMD_PEERS:
		handle_peers(sc, req);
		break;
	case CLIENT_CMD_PING:
		handle_ping(sc, req);
		break;
	case CLIENT_CMD_REVCONN:
		dbg("%s: revconn to server, ignoring", f);
		spp_cmd_rep(sc, 468); /* Not Interested */
		return -1;
		/* NOTREACHED */
	case CLIENT_CMD_STARTTLS:
		handle_starttls(sc);
		break;
	default:
		swarn("%s: invalid server command %d", f, req->cmd);
		spp_cmd_rep(sc, 403);
		return -1;
	}

	if (rtt != 0)
		peer_addrtt(c->peer, ((float)rtt) / 10);

	return 0;
}


/* handle_cachereq - handling of requests for cached files */
int
handle_cachereq(struct urient *ue, struct sppconn *sc, struct req *req)
{
	const char *f = "handle_cachereq";
	struct connection *c;

	ASSERT(lhc_mode || scc_mode, "cache request, but not cache");
	ASSERT(sc->ue != NULL, "NULL ue");

/*	dbg("%s", f);*/

	c = sc->conn;

	/* note: dlend might be adjusted in spp_cmd_get */
	sc->dlstart = req->offset;
	sc->dlend = req->offend;

	switch (req->cmd) {
	case CLIENT_CMD_GET:
		if (streamreq(req, ue)) {
			dbg("%s: stream type request from %s", f, c->peer->id);
			if (!allow_streamreq(c)) {
				swarn("%s: stream requests not allowed from %s",
				    f, c->peer->id);
				spp_cmd_rep(sc, 403);
				return -1;
			}

			sc->streamreq = 1;
			c->noreuse = 1; /* stream req, cannot be reused */
		} else if (scc_mode && sccclient(c)) {
			dbg("%s: scc client request from %s", f, c->peer->id);
			sc->sccclient = 1;
			PTOSP(c->peer)->sccclient = 1;
		}

		if (handle_get(sc, req) == -1)
			return -1;

		return 0;
		/* NOTREACHED */
	case CLIENT_CMD_PING:
		handle_ping(sc, req);
		return 0;
		/* NOTREACHED */
	case CLIENT_CMD_REVCONN:
		/* scc: only for external clients */
		if (lhc_mode || (scc_mode && !sccclient(c)))
			handle_revconn(sc, req);
		else {
			dbg("%s: revconn from internal client", f);
			spp_cmd_rep(sc, 468); /* Not Interested */
			return -1;
		}
		return 0;
		/* NOTREACHED */
	case CLIENT_CMD_HEAD:
		/* HEAD might also be sent to lhc */
		handle_head(sc, req);
		return 0;
	default:
		if (!scc_mode) {
			dbg("%s: invalid command submitted to cache (%s)", f,
			    spp_cmd2str(req->cmd));
			spp_cmd_rep(sc, 403);
			return -1;
		}
		break;
	}

	/* additional scc mode command handling */
	ASSERT((scc_mode && !lhc_mode), "unexpected mode");

	switch (req->cmd) {
	case CLIENT_CMD_GET: /* FALLTHROUGH */
	case CLIENT_CMD_PING:/* FALLTHROUGH */
	case CLIENT_CMD_REVCONN: /* FALLTHROUGH */
	case CLIENT_CMD_HEAD:
		ERRX("command should already be handled");
		/* NOTREACHED */
	case CLIENT_CMD_CRED:
		handle_cred(sc, req);
		break;
	case CLIENT_CMD_IGOT:
		if (handle_igot(sc, req) == -1)
			return -1;
		break;
	case CLIENT_CMD_PEERS:
		handle_peers(sc, req);
		break;
	case CLIENT_CMD_STARTTLS:
		slog(LOG_ERR, "starttls");
		handle_starttls(sc);
		break;
	default:
		dbg("%s: invalid server command %s", f, spp_cmd2str(req->cmd));
		spp_cmd_rep(sc, 403);
		return -1;
	}

	if (req->rtt != 0)
		peer_addrtt(c->peer, ((float)req->rtt) / 10);

	return 0;
}


/* needrecursion - returns 1 if recursion required to fulfill request */
int
needrecursion(struct urient *ue, struct req *req)
{
	int32_t bnum;
	int needreq;

	/* recursion needed if only uri known */
	if (ue->state <= URIENT_STATE_URI)
		return 1;

	needreq = 1;
	switch (req->cmd) {
	case CLIENT_CMD_CRED:    /* FALLTHROUGH */
	case CLIENT_CMD_PEERS:   /* FALLTHROUGH */
	case CLIENT_CMD_PING:    /* FALLTHROUGH */
	case CLIENT_CMD_STARTTLS:/* FALLTHROUGH */
	case CLIENT_CMD_REVCONN: /* FALLTHROUGH */
		/* no recursion needed for either command */
		needreq = 0;
		break;
	case CLIENT_CMD_GET:
		bnum = off2block(ue, req->offset);
		if (ue->state == URIENT_STATE_COMPLETE)
			needreq = 0;
		else if (ue->state == URIENT_STATE_INCOMPLETE) {
			/* recursion needed if data not cached */
			if (ue->cf == NULL)
				needreq = 1;
			else {
				/* does not respond to invalidity */
				if (bnum < ue->meta->blockcnt &&
				    urient_haveblock(ue, bnum))
					needreq = 0;
				else
					needreq = 1;
			}

#ifdef notyet
			/*XXX could check if data is found in mem */
			if (needreq && ue->file != NULL) {
			}
#endif
		} else
			needreq = 1;
		break;
	case CLIENT_CMD_HEAD:
		/* recursion needed if meta data not known */
		if (ue->state == URIENT_STATE_URI)
			needreq = 1;
		else
			needreq = 0;
		break;
	case CLIENT_CMD_IGOT:
		/* no recursion needed, currently. ignored if sum known */
		needreq = 0;
		break;
	}

	return needreq;
}
