/* serv.c - server related functions like parsing and http compatibility */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: serv.c,v 1.5 2010/07/31 09:48:09 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <err.h>
#include <netdb.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "p2pd.h"
#include "ip.h"
#include "serv.h"
#include "spp.h"

/*
 * to aid integration with browsers and media players that do not have
 * native support for spp, the server can be set up to answer http
 * requests for files that exist on the server. rather than returning
 * the media file, it returns a playlist with the uri and the mime
 * type 'application/x-spp-playlist'. by adding a handler for this
 * mime type browsers can play files without having spp support.
 */

int http_servcompat = 0;

/* XXX won't work with ipv6 addresses, use separate arguments instead? */


/* serv_parsemirror - parse server mirror specification (adr:port) */
int
serv_parsemirror(char *adr)
{
	const char *errstr;
	char *host;
	char *port;
	char *p;

	host = adr;
	if ((p = strchr(host, ':')) == NULL) {
		port = SPP_PORT;
	} else {
		if (p == host)
			return -1; /* nothing before ':' */
		if (strlen(p) == 1)
			return -1; /* nothing after ':' */
		*p = '\0';
		p++; /* point at data beyond ':' */
		port = p;
	}

	if (!ip_isnumeric(host))
		errx(1, "invalid address: %s", host);
	(void)strtonum(port, 1, 65535, &errstr);
	if (errstr)
		errx(1, "port number: %s: %s", errstr, port);

	if (filemeta_addserv(host, port) == -1)
		return -1;

	return 0;
}


/* serv_httpreply - return dummy file with special mime type */
void
serv_httpreply(struct urient *ue, struct sppconn *sc)
{
	char playlist[1024];

	ASSERT(sc->conn->obuflen == 0, "buffered data");

	/* build playlist */
	snprintf(playlist, sizeof(playlist),
		 "[playlist]\n"
		 "numberofentries=2\n"
		 "File1=spp://%s:%s/%s\n"
		 "Title1=spp url, for supporting clients\n"
		 "Length=%lld\n"
		 "File2=http://%s:%s/%s\n"
		 "Title2=http fallback url\n"
		 "Length2=%lld\n",
		 ue->uri.host, ue->uri.port, ue->uri.path,
		 (long long)ue->meta->size,
		 ue->uri.host, ue->uri.port, ue->uri.path,
		 (long long)ue->meta->size);

	snprintf(sc->conn->obuf, sizeof(sc->conn->obuf),
		 "HTTP/1.0 200 OK\r\n"
		 "Content-Length: %lld\r\n"
		 "Connection: close\r\n"
		 "Content-Type: application/x-spp-playlist\r\n"
		 "\r\n", (long long)strlen(playlist));
	/* add playlist data */
	if (sc->currcmd == CLIENT_CMD_GET) {
		if (strlcat(sc->conn->obuf, playlist, sizeof(sc->conn->obuf)) >=
		    sizeof(sc->conn->obuf))
			ERRX("reply too long");
	}

	conn_changestate(sc->conn, CONNECTION_STATUS_WRITEREP);
	sc->sppcode = 200;
	sc->conn->obuflen = strlen(sc->conn->obuf);
}
