/* speer.c - spp dependent peer data structures */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: speer.c,v 1.7 2010/06/26 01:09:01 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>

#include "p2pd.h"
#include "peer.h"
#include "spp.h"
#include "timers.h"
#include "urient.h"


/* speer_add - add new peer, adding speer structure */
struct speer *
speer_add(char *name, char *port)
{
	const char *f = "speer_add";
	struct peer *peer;
	struct speer *sp;

	if ((peer = peer_add(name, port)) == NULL)
		return NULL;

	/* speer struct already added? */
	if (PTOSP(peer) != NULL)
		return PTOSP(peer);

	dbg("%s: %s:%s", f, name, port);

	sp = xmalloc(sizeof(struct speer));
	bzero(sp, sizeof(*sp));

	SPTOP(sp) = peer;
/*	PTOSP(peer) = sp;*/ /*XXX problems with gcc 4.4.1 */
	peer->uldat[PROTO_SPP] = sp;

	return sp;
}


/* speer_free - release speer resources */
void
speer_free(struct speer *sp)
{
	if (sp->to_newpeer != NULL)
		speer_gettinginfo(sp);
	free(sp);
}


/* speer_wantinfo - bw/latency info on peer wanted for urient */
void
speer_wantinfo(struct urient *ue, struct speer *sp)
{
	int delay;

	ASSERT(!serv_mode, "speer_wantinfo called on server");

	/*
	 * XXX currently used to setup initial state, in order to ensure
	 *     all peers are attempted connected (if possible in prioritized
	 *     order).
	 *
	 *     would be possible to do regular monitoring of peers.
	 *     rather than deleting the timer, the speer_gettinginfo function
	 *     could be used to postpone any ping operation.
	 *     the time between each peer being contacted could be gradually
	 *     increased if the node is never used for data transfers.
	 */

	if (peerpar_gettime(SPTOP(sp), PEERPAR_PEER_LASTCONNECT) != 0)
		/* connect attempt already made, ignore */
		return;

	/* already set? */
	if (sp->to_newpeer != NULL)
		return; /* XXX could update timeout value if different */

	/* add connect timer for all new peers */
	sp->to_newpeer = timeout_add(newpeer_ping, NULL, SPTOP(sp));

	/*
	 * use server rank (to prioritize good candidates), or
	 * randomize if server rank is not available (to prevent lots
	 * of peers being contacted at once).
	 */
	if ((delay = peerpar_getint(SPTOP(sp), PEERPAR_SPP_SSCORE)) == 0)
		delay = SPP_NEWPEER_TIMEOUT + rand() % SPP_NEWPEER_TIMEOUT;
	timeout_secupd(sp->to_newpeer, delay);

	/*
	 * ping is currently per-file, add pointer to current file.
	 * could theoretically have the same peer occurring in multiple
	 * urients. purpose of operation is to obtain latency (and
	 * to a lesser extent, updated block info). for latency the
	 * urient does not matter, but it might be possible to make
	 * an argument for changing the urient if this function is called
	 * before timeout triggers (XXX but won't bother with that for now)
	 */
	sp->ue_newpeer = ue;
	sppfile_addref(UTOSPP(ue));
}


/* speer_gettinginfo - clear timeout and reduce sppfile refcount */
void
speer_gettinginfo(struct speer *sp)
{
	ASSERT(sp->to_newpeer != NULL, "newpeer timeout not set");

	/*
	 * remove newpeer timer.
	 * will be called on connect attempt, even if not ping command is
	 * used; purpose is not primarily to use ping, but to obtain info
	 * if no connect is done. if get or other command is used, this is
	 * sufficient and we don't care about ping.
	 */

#if 0 /*XXX never disable timer to test ping */
	timeout_del(sp->to_newpeer);
	sp->to_newpeer = NULL;
	sppfile_unref(UTOSPP(sp->ue_newpeer));
	sp->ue_newpeer = NULL;
#endif
}


#if 0
/* speer_status - return zero if peer ok, status code otherwise */
int
speer_status(struct speer *sp)
{
	if (SPTOP(sp)->
			peerpar_adjint(c->peer, PEERPAR_PEER_LASTRES,
				       PEER_MULTFAIL);

}
#endif
