/* spp.c - top level functions for spp */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: spp.c,v 1.44 2010/10/11 22:16:58 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>

#include "p2pd.h"
#include "conn.h"
#include "dlagent.h"
#include "iostat.h"
#include "request.h"
#include "spp.h"
#include "parse_resp.h"
#include "protio.h"
#include "peer.h"
#include "peerset.h"
#include "stream.h"
#include "tcpmodel.h"
#include "scc.h"
#include "serv.h"
#include "ssl.h"
#include "tv.h"
#include "urient.h"

void connevent(struct connection *, enum connevent, int);
void streamevent(struct sppconn *, enum streamevent, int32_t, ssize_t);
void resevent(struct connection *, enum resolveevent, int, struct peerset *);
void mconnevent(struct connection *, enum mconnevent, int, int,
		struct connection *);
void process_resp(struct sppconn *);
void process_respdat(struct sppconn *);
void spp_sslcomp(struct sppconn *);
void spp_resolvedone(struct sppconn *, struct peerset *);
void spp_resolveerr(struct sppconn *);
void spp_conndone(struct sppconn *);
void spp_mconndone(struct sppconn *, int, struct connection *);
void spp_mconnerr(struct sppconn *, int, int);
void process_reqdat(struct sppconn *);
void iodone_writerep(struct sppconn *);
void spp_failedreq(struct sppconn *);
void spp_streamstart(struct sppconn *);
void spp_streamend(struct sppconn *);
void spp_connend(struct sppconn *);
void sppconn_free(struct sppconn *);
void spp_wrotedata(struct sppconn *, ssize_t);
void spp_readdata(struct sppconn *, ssize_t);
void spp_streamslow(struct sppconn *);
void spp_streamstop(struct sppconn *);
void spp_streamfail(struct sppconn *);
void spp_streamcomp(struct sppconn *);
void spp_sentblock(struct sppconn *, int32_t);


/* XXX rename functions to be on format spp_foo */
struct protops proto_spp = {
	connevent,
	(void (*)(void *, enum streamevent, int32_t, ssize_t))streamevent,
	resevent,
	mconnevent
};

int extstreams = 0; /* external client streams */
int instreams = 0;  /* incoming streams */
int outstreams = 0; /* outgoing streams */


/* connevent - connection event handler */
void
connevent(struct connection *c, enum connevent event, int err)
{
	(void)err;

	dbg("connevent: %s", conn_connev2str(event));

	switch (event) {
	case CONNEVENT_CONNECT_OK:
		spp_conndone(c->uldat);
		break;
	case CONNEVENT_CONNECT_FAIL:
		/*XXX*/
		spp_failedreq(c->uldat);
		break;
	case CONNEVENT_CONNECTION_END:
		spp_connend(c->uldat);
		break;
	case CONNEVENT_REQUEST_FAIL:
		spp_failedreq(c->uldat);
		break;
	case CONNEVENT_GOT_REPLY:
		process_resp(c->uldat);
		break;
	case CONNEVENT_GOT_REPLYDAT:
		process_respdat(c->uldat);
		break;
	case CONNEVENT_GOT_REQUEST:
		accept_request(c);
		break;
	case CONNEVENT_GOT_REQUESTDAT:
		process_reqdat(c->uldat);
		break;
	case CONNEVENT_WROTE_REPLY:
		iodone_writerep(c->uldat);
		break;
	case CONNEVENT_SSLCONNECT_OK:
		spp_sslcomp(c->uldat);
		break;
	case CONNEVENT_SSLCONNECT_FAIL:
		/*XXX*/
		break;
	case CONNEVENT_SSLCONNECTION_END:
		/*XXX*/
		break;
	case CONNEVENT_FREEULDAT:
		sppconn_free(c->uldat);
		break;
	}
}


/* streamevent - stream event handler */
void
streamevent(struct sppconn *sc, enum streamevent event, int32_t bnum,
	    ssize_t iosize)
{
	switch (event) {
	case STREAMEVENT_STREAM_START:
		spp_streamstart(sc);
		break;
	case STREAMEVENT_STREAM_TOOSLOW:
		spp_streamslow(sc);
		break;
	case STREAMEVENT_STREAM_STOPPED:
		spp_streamstop(sc);
		break;
	case STREAMEVENT_STREAM_FAIL:
		spp_streamfail(sc);
		break;
	case STREAMEVENT_STREAM_COMPLETE:
		spp_streamcomp(sc);
		break;
	case STREAMEVENT_STREAM_END:
		spp_streamend(sc);
		break;
	case STREAMEVENT_STREAM_WROTEDATA:
		spp_wrotedata(sc, iosize);
		break;
	case STREAMEVENT_STREAM_READDATA:
		spp_readdata(sc, iosize);
		break;
	case STREAMEVENT_STREAM_SENTBLOCK:
		spp_sentblock(sc, bnum);
		break;
	}
}


/* resevent - resolver event handler */
void
resevent(struct connection *c, enum resolveevent event, int err,
	 struct peerset *resdata)
{
	(void)err;

	switch (event) {
	case RESEVENT_RESOLVE_OK:
		spp_resolvedone(c->uldat, resdata);
		break;
	case RESEVENT_RESOLVE_FAIL:
		spp_resolveerr(c->uldat);
		break;
	case RESEVENT_PROC_FAIL:
		/*XXX*/
		ERRX("resolver process failed");
		break;
	}
}


/* mconnevent - mconn event handler */
void
mconnevent(struct connection *c, enum mconnevent event, int idx, int err,
	   struct connection *peer)
{
	switch (event) {
	case MCONNEVENT_CONNECT_OK:
		spp_mconndone(c->uldat, idx, peer);
		break;
	case MCONNEVENT_CONNECT_FAIL:
		spp_mconnerr(c->uldat, idx, err);
		break;
	case MCONNEVENT_CONNECT_ALLFAIL:
		spp_failedreq(c->uldat);
		break;
	case MCONNEVENT_PROC_FAIL:
		/*XXX*/
		ERRX("mconn process failed");
		break;
	}
}


/* process_resp - process server/peer response */
void
process_resp(struct sppconn *sc)
{
	const char *f = "process_resp";
	struct connection *c;
	struct rep rep;
	float dur;
	char *p;

	c = sc->conn;

	/* command response time */
	dur = tv_dur2ms(&c->reqstart);
	peer_addrtt(c->peer, dur);
	dbg("%s: %s request response time to %s: %0.2f ms", f,
	    spp_cmd2str(sc->currcmd), c->peer->id, dur);

	bzero(&rep, sizeof(struct rep));
	if ((p = response_parse(&rep, c->ibuf)) == NULL) {
		dbg("%s: response parsing failed, ending connection", f);
		c->ibuflen = 0;
		if (sc->dl != NULL)
			dlagent_cmdfail(sc->dl, sc);
		else
			conn_changestate(c, CONNECTION_STATUS_ENDED);
		return;
	}
	conn_ibufsync(c, p);

	switch (sc->currcmd) {
	/* server commands, having gotten this far means the server is ok */
	case CLIENT_CMD_CRED:  /* FALLTHROUGH */
	case CLIENT_CMD_HEAD:  /* FALLTHROUGH */
	case CLIENT_CMD_IGOT:  /* FALLTHROUGH */
	case CLIENT_CMD_PEERS: /* FALLTHROUGH */
	case CLIENT_CMD_STARTTLS:
		/* aid peer selection */
		peerpar_setint(c->peer, PEERPAR_SPP_LASTRES, PEER_PREVOK);
		/* ignore refresh if in head? (not retrieved yet?)XXX*/
		if (sc->currcmd != CLIENT_CMD_HEAD)
			/* refresh request? */
			if (rep.dbrefresh)
				dbrefresh(sc->ue);
		break;

		/*
		 * client commands
		 *
		 * XXX as long as these commands are per file, it
		 * might not make sense to mark the nodes as failed
		 * because a peer can have multiple files.
		 *
		 * a related problem is that a node that has deleted a
		 * block can still be registered in the server as
		 * having the block. deleting the peer would only lead
		 * to it being added back the next time the server is
		 * contacted.
		 *
		 * information about peer usability would have to be
		 * maintained per file.
		 *
		 * synchronization of information between clients and
		 * servers is a general problem, especially in case of
		 * client block deletion. changing the way information
		 * is shared would require a change in the way
		 * checksums are verified/distributed.
		 *
		 * there is also the separation of node information on
		 * the connectivity level (whether the file is
		 * reachable or not), and on the (spp) protocol
		 * level. in the second case, peer information is
		 * per-file and not necessarily synchronized at the
		 * different nodes in the system.
		 *
		 * a solution might be to have (per-file) peer
		 * orderings on the dlagent level, meaning that
		 * commands such as failed ping/get requests cause
		 * peers to get a unusable flag or something similar,
		 * which only affects that dlagent sorting, which is
		 * time limited. this allows the use of information
		 * from the peers themselves, without affecting other
		 * files (could be relevant for other dlagents in same
		 * file though).
		 */

	case CLIENT_CMD_GET:
		if (rep.code == 200)
			peerpar_setint(c->peer, PEERPAR_SPP_LASTRES,
				       PEER_PREVOK);
		/* XXX do nothing for now in case of failure */
		break;

	case CLIENT_CMD_PING:
		if (rep.code == 200)
			peerpar_setint(c->peer, PEERPAR_SPP_LASTRES,
				       PEER_PREVOK);
		/* XXX do nothing for now in case of failure */
		break;

	case CLIENT_CMD_REVCONN:
		if (rep.code == 200)
			peerpar_setint(c->peer, PEERPAR_SPP_LASTRES,
				       PEER_PREVOK);
		else
			/* XXX separate error indicator for revconn;
			   does it make sense setting multfail? */
			peerpar_setint(c->peer, PEERPAR_SPP_LASTRES,
				       PEER_MULTFAIL);
		break;
	default:
		dbg("%s: invalid command: %s", f, spp_cmd2str(sc->currcmd));
		ERRX("unknown command");
	}

	switch (sc->currcmd) {
	case CLIENT_CMD_CRED:
		handleresp_cred(&rep, sc);
		break;

	case CLIENT_CMD_GET:
		handleresp_get(&rep, sc);
		break;

	case CLIENT_CMD_HEAD:
		handleresp_head(&rep, sc);
		break;

	case CLIENT_CMD_IGOT:
		handleresp_igot(&rep, sc);
		break;

	case CLIENT_CMD_PEERS:
		handleresp_peers(&rep, sc);
		break;

	case CLIENT_CMD_PING:
		/*XXX want callback called after command finished */
#if 0
		ue = sc->ue; /* deleted during cache_connection */
#endif
		handleresp_ping(&rep, sc);
		/* currently doing post-revconn pings? */
		if (PTOSP(c->peer)->revconn &&
		    PTOSP(c->peer)->revconn < PEER_REVCONN_PINGCNT) {
			PTOSP(c->peer)->revconn++;
			/*spp_cmd_ping(ue, c->peer);*/
			spp_cmd_ping(sc);
		} else
			cache_connection(sc->conn);
		break;

	case CLIENT_CMD_REVCONN:
		handleresp_revconn(&rep, sc);
		break;

	case CLIENT_CMD_STARTTLS:
		handleresp_starttls(&rep, sc);
		break;

	default:
		dbg("%s: invalid command: %s", f, spp_cmd2str(sc->currcmd));
		ERRX("unknown command");
	}
}


/* process_respdat - process additional data included with server response */
void
process_respdat(struct sppconn *sc)
{
	const char *f = "process_respdat";

	switch (sc->currcmd) {
	case CLIENT_CMD_CRED:
		parse_hostcred(sc);
		cache_connection(sc->conn);
		break;
	case CLIENT_CMD_PEERS:
		parse_peerlist(sc);
		break;
	case CLIENT_CMD_IGOT:
		parse_igoterrs(sc);
		break;
	default:
		dbg("%s: invalid command: %s", f, spp_cmd2str(sc->currcmd));
		ERRX("unknown command");
	}
}


/* spp_sslcomp - ssl connnection established */
void
spp_sslcomp(struct sppconn *sc)
{
	const char *f = "spp_sslcomp";

	/*XXX could be more generic */
	ASSERT(sc->currcmd == CLIENT_CMD_STARTTLS, "unexpect ssl command");
	PTOSP(sc->sslinitpeer)->sslinitactive = 0;
	dbg("%s: sslinit complete for %s", f, sc->sslinitpeer->id);
	sc->sslinitpeer = NULL;

	/* now do igot */
	sc->currcmd = CLIENT_CMD_IGOT;
	spp_cmd_igot(sc);
}


/* spp_resolvedone - usable dns resolution data available (new or cached) */
void
spp_resolvedone(struct sppconn *sc, struct peerset *dnspset)
{
	struct sppconn *peersc;
	struct peer *peer;
	int i;

	for (i = 0; i < dnspset->peercnt; i++) {
		peer = pset_peer(dnspset, i);
		if (speer_add(peer->name, peer->port) == NULL)
			ERRX("speer_add failed");
	}

	/* retry request (resolve only done for requests to parent) */
	peersc = sppconn_new(sc->currcmd);
	sppconn_setfile(peersc, sc->ue);
	if (sc->dl != NULL) {
		peersc->dl = sc->dl;
		peersc->dlblock = sc->dlblock;
		peersc->dlstart = sc->dlstart;
		peersc->dlend = sc->dlend;
	}
	spp_parentreq(peersc);
}


/* spp_resolveerr - dns lookup failed */
void
spp_resolveerr(struct sppconn *sc)
{
	const char *f = "spp_resolveerr";

	dbg("%s", f);

	ASSERT(sc->currcmd == CLIENT_CMD_HEAD, "unexpected command");
	sppfile_headfail(UTOSPP(sc->ue), sc, 404); /* not found */
}


/* spp_conndone - connection established */
void
spp_conndone(struct sppconn *sc)
{
	const char *f = "spp_conndone";

	dbg("%s", f);

	switch (sc->currcmd) {
	case CLIENT_CMD_CRED:
		spp_cmd_cred(sc);
		break;
	case CLIENT_CMD_GET:
		spp_cmd_get(sc, sc->dlstart, sc->dlend);
		break;
	case CLIENT_CMD_HEAD:
		if (spp_cmd_head(sc) == -1)
			sppfile_headfail(UTOSPP(sc->ue), NULL, errno_spp);
		break;
	case CLIENT_CMD_IGOT:
#if HAVE_OPENSSL
		if (use_ssl)
			ERRX("igot without ssl");
#endif /* HAVE_OPENSSL */
		spp_cmd_igot(sc);
		break;
	case CLIENT_CMD_PEERS:
		ASSERT(sc->dl != NULL, "NULL dl and peers");
		spp_cmd_peers(sc, sc->dlblock);
		break;
	case CLIENT_CMD_PING:
		spp_cmd_ping(sc);
		break;
	case CLIENT_CMD_REVCONN:
		spp_cmd_revconn(sc, sc->conn->peer);
		break;
	case CLIENT_CMD_STARTTLS:
		spp_cmd_starttls(sc);
		break;
	default:
		ERRX("unsupp cmd");
	}
}


/* spp_mconndone - connection established via mconn */
void
spp_mconndone(struct sppconn *sc, int idx, struct connection *peerconn)
{
	struct sppconn *peersc;
	struct peer *peer;

	/* XXX need common place for doing this */
	peer = pset_peer(sc->conn->mconnpset, idx);
	if (PTOSP(peer)->to_newpeer != NULL)
		speer_gettinginfo(PTOSP(peer));

	/* not first connected peer? */
	if (sc->conn->mconnrcpt > 0) {
		peersc = sppconn_new(CLIENT_CMD_PING);
		sppconn_setfile(peersc, sc->ue);
		sppconn_setconn(peersc, peerconn);
		spp_cmd_ping(peersc);
		return;
	}

	peersc = sppconn_new(sc->currcmd);
	sppconn_setfile(peersc, sc->ue);
	sppconn_setconn(peersc, peerconn);

	if (sc->dl != NULL) {
		peersc->dl = sc->dl;
		peersc->dlblock = sc->dlblock;
		peersc->dlstart = sc->dlstart;
		peersc->dlend = sc->dlend;
	}
	spp_conndone(peersc);
}


/* spp_mconnerr - connection establishment via mconn failed */
void
spp_mconnerr(struct sppconn *sc, int idx, int err)
{
	struct peer *peer;

	(void)err;

	peer = pset_peer(sc->conn->mconnpset, idx);
	if (PTOSP(peer)->to_newpeer != NULL)
		speer_gettinginfo(PTOSP(peer));

	/* XXX if all connections fail, need to call failedreq */
}


/* process_reqdat - process additional data supplied with request */
void
process_reqdat(struct sppconn *sc)
{
	const char *f = "process_reqdat";
	struct connection *c;

	c = sc->conn;

	switch (sc->currcmd) {
	case CLIENT_CMD_IGOT:
		if (sc->ue->state <= URIENT_STATE_META) {
			/* don't know anything about file yet, ignore */
			/*XXX could call dbrefresh for this file later */
			c->ibuflen = 0;
			spp_cmd_rep(sc, 200);
			return;
		}
		if (urient_fileref(sc->ue) == -1) {
			dbg("%s: urient_fileref failure", f);
			spp_cmd_rep(sc, 403); /*XXX errcode? */
		} else {
			parse_igots(sc);
			urient_fileunref(sc->ue);
		}
		break;
	default:
		dbg("%s: invalid command: %s", f, spp_cmd2str(sc->currcmd));
		ERRX("unknown command");
	}
}


/* iodone_writerep - operations performed after writing reply */
void
iodone_writerep(struct sppconn *sc)
{
	const char *f = "iodone_writerep";
	struct connection *c;

	c = sc->conn;

	dbg("%s", f);

	/* update status */
	dbg("%s: status: %s cmd: %s", f, conn_state2str(c->status),
	    spp_cmd2str(sc->currcmd));

	/* XXX some return values might always result in closed connection */
	if (sc->sppcode == 404 || sc->sppcode == 403) {
		conn_changestate(c, CONNECTION_STATUS_ENDED);
		return;
	}

	/* XXX currcmd should be set as a result of conn_changestate/idle/cache */
	switch (sc->currcmd) {
	case CLIENT_CMD_HEAD:
		/*
		 * lynx appears to hang if connection not closed after head
		 * XXX check http rfc for possible reason
		 */
		if (sc->httpreq)
			conn_changestate(c, CONNECTION_STATUS_ENDED);
		else if ((sc->sppcode == 200) || (sc->sppcode == 504))
			/*conn_changestate(c, CONNECTION_STATUS_IDLE);*/
			idle_connection(c);
		else
			/* no more requests are likely if HEAD fails */
			conn_changestate(c, CONNECTION_STATUS_ENDED);
#if 0
		sc->currcmd = 0;
#endif
		break;

	case CLIENT_CMD_GET:
		if (serv_mode && sc->httpreq) {
			/* XXX */
			ASSERT(http_servcompat != 0, "http get on server");
			conn_changestate(c, CONNECTION_STATUS_ENDED);
		} else if (sc->sppcode == 200 || sc->sppcode == 206) {
			sppget_postrep(sc);

#if 0
			if (sc->streamreq || sc->sccclient)
				conn_changestate(c, CONNECTION_STATUS_FOREVERSTREAM);
			else
				conn_changestate(c, CONNECTION_STATUS_DEADLINESTREAM);
#endif
		} else {
			dbg("%s: ending failed get request", f);
			conn_changestate(c, CONNECTION_STATUS_ENDED);
		}
/*		sc->currcmd = 0;*/ /*XXX?*/
		break;

	/* always idle after IGOT; will likely receive new connections */
	case CLIENT_CMD_IGOT:  /* FALLTHROUGH */
	case CLIENT_CMD_CRED:  /* FALLTHROUGH */
	case CLIENT_CMD_PEERS: /* FALLTHROUGH */
	case CLIENT_CMD_PING:
		/* idle after completing io */
		/*conn_changestate(c, CONNECTION_STATUS_IDLE);*/
		idle_connection(c);
#if 0
		sc->currcmd = 0;
#endif
		break;

	case CLIENT_CMD_STARTTLS:
		if (use_ssl) {
#if HAVE_OPENSSL
			/*XXX need to set connstate to something (SSL state?)*/
			conn_changestate(c, CONNECTION_STATUS_READREQ);
			ssl_starttls_serv(c);
#else
			ERRX("no ssl support");
#endif /* HAVE_OPENSSL */
		} else {
			/* can receive starttls even if not supported; idle */
			idle_connection(c);
		}
		break;

	case CLIENT_CMD_REVCONN:
		/* revconn, cache, don't idle */
		if (sc->sppcode == 200) {
			ASSERT(sc->conn->type == CONN_TYPE_INCOMING,
			       "not incoming connection");
#if 0
			ue = sc->ue; /* save copy before cache_connection */
			/* XXX conn_revconn - change INCOMING->OUTGOING */
			cache_connection(c);
#endif
			conn_revconn(c);

			/* determine peer rtt XXX should do at least twice */
			dbg("%s: got revconn, sending ping", f);
			PTOSP(c->peer)->revconn++;
			/*XXX set currcmd directly */
			sc->currcmd = CLIENT_CMD_PING;
			spp_cmd_ping(sc);
			/*XXX do in conn_revconn?*/
			changecstatus(c, CONNECTION_STATUS_WRITEREQ);
			xgettimeofday(&c->reqstart, NULL);
		} else
			conn_changestate(c, CONNECTION_STATUS_ENDED);
		break;
	}

	/* XXX set currcmd to 0? */
}


/* spp_failedreq - notify upper layer that request failed */
void
spp_failedreq(struct sppconn *sc)
{
	const char *f = "spp_failedreq";

	dbg("%s", f);

	if (sc == NULL) { /* XXX caller should not call if NULL */
		dbg("%s: zero state, ignoring", f);
		return;
	}

	if (sc->conn->type == CONN_TYPE_INCOMING)
		return; /* ignore */

	if (sc->ue != NULL)
		sppfile_connfail(UTOSPP(sc->ue), sc);
	if (sc->dl != NULL)
		dlagent_cmdfail(sc->dl, sc);
}


/* spp_connend - notification of connection state ending (idle/ended) */
void
spp_connend(struct sppconn *sc)
{
	const char *f = "spp_connend";

	dbg("%s", f);

	if (sc == NULL) { /* XXX caller should not call if NULL */
		dbg("%s: zero state, ignoring", f);
		ERRX("spp_connend without data");
		return;
	}

	if (sc->dl != NULL)
		dlagent_connend(sc);

	if (sc->conn->type == CONN_TYPE_OUTGOING &&
	    sc->currcmd == CLIENT_CMD_IGOT)
		igot_restore(sc->ue); /* restore in-transit IGOTs */

	if (sc->headwait) {
		dbg("%s: deleting connection from waiting queue", f);
		aqueue_delval(UTOSPP(sc->ue)->waiting, sc);
		sc->headwait = 0;
	}

	if (sc->ue != NULL) {
		sppfile_unref(UTOSPP(sc->ue));
		sc->ue = NULL;
	}

	/* XXX free earlier? */
	if (sc->req != NULL) {
		free(sc->req);
		sc->req = NULL;
	}
}


/* spp_streamstart - notification of stream starting */
void
spp_streamstart(struct sppconn *sc)
{
	const char *f = "spp_streamstart";

	if (spp_extclient(sc)) {
		extstreams++;
		dbg("%s: external streams: %d", f, extstreams);
	}
	if (sc->conn->stream->type == STREAM_FROM_MEM)
		outstreams++;
	else
		instreams++;
	dbg("%s: instreams: %d outstreams: %d", f, instreams, outstreams);
}


/* spp_streamend - notification of (outgoing) stream over connection ending */
void
spp_streamend(struct sppconn *sc)
{
	const char *f = "spp_streamend";

	ASSERT(sc->ue != NULL, "NULL ue");

	if (spp_extclient(sc)) {
		ASSERT(extstreams > 0 , "negative extstreams");
		extstreams--;
		dbg("%s: external streams: %d", f, extstreams);
	}
	if (sc->conn->stream->type == STREAM_FROM_MEM) {
		ASSERT(outstreams > 0 , "negative outstreams");
		outstreams--;
	} else {
		ASSERT(instreams > 0 , "negative instreams");
		instreams--;
	}
	dbg("%s: instreams: %d outstreams: %d", f, instreams, outstreams);
}


/* spp_streamslow - stream transfer too slow */
void
spp_streamslow(struct sppconn *sc)
{
	/*XXX not in server mode */
	if ((sc->dl != NULL) && (sc->dl->state == DLAGENT_STATE_GET))
		dlagent_tooslow(sc->dl);
}


/* spp_streamstop - notification that stream is stopped (data unavailable) */
void
spp_streamstop(struct sppconn *sc)
{
	if (sc->dl != NULL)
		dlagent_streamstop(sc->dl, sc);
}


/* spp_streamfail - notification that stream failure */
void
spp_streamfail(struct sppconn *sc)
{
	if (sc->dl != NULL)
		dlagent_cmdfail(sc->dl, sc);
}


/* XXX have one failure function, reason parameter? */
/* spp_streamcomp - spp transfer completed */
void
spp_streamcomp(struct sppconn *sc)
{
	/* XXX checksum ok? mark source as failed if not */
	peerpar_adjint(sc->conn->peer, PEERPAR_SPP_OKBLOCKS, 1);
	peerpar_setint(sc->conn->peer, PEERPAR_SPP_LASTRES, PEER_PREVOK);

	/*XXX not in server mode */
	if ((sc->dl != NULL) && (sc->dl->state == DLAGENT_STATE_GET))
		dlagent_getcomp(sc->dl, sc);
}


/* spp_parent - return pointer to parent node peer structure */
struct peer *
spp_parent(struct uri *uri)
{
	struct speer *sp;

	/* XXX should call speer_add(), returning result */
	/* use peer_add, node will only be added if it does not exist */
	if (parent_scc)
		sp = speer_add(parent_name, parent_port);
	else
		sp = speer_add(uri->host, uri->port);

	if (sp == NULL)
		return NULL;
	else
		return SPTOP(sp);
}


/* spp_parentreq - send specified request to parent node */
void
spp_parentreq(struct sppconn *sc)
{
	const char *f = "spp_parentreq";
	struct connection *c;
	struct peer *parent;
	struct peer *peer;

	parent = UTOSPP(sc->ue)->parent;

	/* dns resolution needed? */
	if (!peer_validaddr(parent)) {
		sppconn_setconn(sc, conn_resinit(parent));
		conn_resolve(sc->conn);
		return;
	}

	if ((c = conn_getcached(parent, 0)) != NULL) {
		dbg("%s: doing request with cached connection", f);
		sppconn_setconn(sc, c);
		spp_conndone(sc);
		return;
	}

	/* do direct connect if only one alternative, mconn otherwise */
	if (parent->addr != NULL) {
		/* not symbolic */
		sppconn_setconn(sc, conn_conninit(parent));
		conn_connect(sc->conn);
	} else if (parent->dns->peercnt == 1) {
		peer = pset_peer(parent->dns, 0);
		sppconn_setconn(sc, conn_conninit(peer));
		conn_connect(sc->conn);
	} else {
		sppconn_setconn(sc, conn_mconninit(parent->dns));
		conn_mconn(sc->conn);
	}
}


/* spp_peerreq - send specified request to peer */
void
spp_peerreq(struct sppconn *sc, struct peer *peer)
{
	const char *f = "spp_peerreq";
	struct connection *c;

	ASSERT(peer->addr != NULL, "peer missing address info");

	if ((c = conn_getcached(peer, 0)) != NULL) {
		dbg("%s: doing request with cached connection", f);
		sppconn_setconn(sc, c);
		spp_conndone(sc);
		return;
	}

	sppconn_setconn(sc, conn_conninit(peer));
	conn_connect(sc->conn);
}


/* sppconn_new - return pointer to new spp connection structure */
struct sppconn *
sppconn_new(enum sppcmd cmd)
{
	struct sppconn *sc;

	ASSERT(cmd != 0, "cmd not set");

	sc = xmalloc(sizeof(struct sppconn));
	bzero(sc, sizeof(struct sppconn));

	sc->currcmd = cmd;

	return sc;
}


/* sppconn_setconn - specify underlying connection */
void
sppconn_setconn(struct sppconn *sc, struct connection *c)
{
	ASSERT(sc->conn == NULL, "connection already set");

	sc->conn = c;
	sc->conn->uldat = sc;
/*	ASSERT(sc->conn->protops == NULL, "protops already set for conn");*/
	sc->conn->protops = &proto_spp;
}


/* sppconn_setfile - specify ue/file */
void
sppconn_setfile(struct sppconn *sc, struct urient *ue)
{
	ASSERT(sc->ue == NULL, "ue already set");

	sc->ue = ue;
	sppfile_addref(UTOSPP(ue));
}


/* sppconn_free - release sppconn resources */
void
sppconn_free(struct sppconn *sc)
{
	free(sc);
}


/* spp_wrotedata - notification of data written on connection */
void
spp_wrotedata(struct sppconn *sc, ssize_t wres)
{
	const char *f = "spp_wrotedata";

	/* ignore at server, and for stream request clients */
	/*XXX only for scc clients in scc_mode */

	if (!serv_mode && !scc_mode && sc->streamreq == 0) {
		PTOSP(sc->conn->peer)->uncredited += wres;
		dbg("%s: uncredited by %s %lld", f, sc->conn->peer->id,
		    (long long)PTOSP(sc->conn->peer)->uncredited);
	}

	if (sc->streamreq)
		ios.net_wdat_s += wres;
}


/* spp_readdata - notification of data read on connection */
void
spp_readdata(struct sppconn *sc, ssize_t rres)
{
	PTOSP(sc->conn->peer)->unreported += rres;

	if (sc->dl != NULL)
		dlagent_readdata(sc->dl, sc);
}


/* spp_sentblock - block transmission ended */
void
spp_sentblock(struct sppconn *sc, int32_t bnum)
{
	const char *f = "spp_sentblock";

	ASSERT(sc->currcmd != 0, "currcmd not set");

	if (sc->dl != NULL)
		dlagent_sendok(sc, bnum);
	else
		dbg("%s: transfer without dl", f);
}


/* spp_extclient - true if request from external client (not scc/lhc client) */
int
spp_extclient(struct sppconn *sc)
{
	/*XXX not 100% correct maybe, should have single proxyclient value */
	if (sc->streamreq == 0 && sc->cachereq == 0 && sc->sccclient == 0)
		return 1;
	else
		return 0;
}


/* spp_cmd2str - return string representation of spp cmd */
char *
spp_cmd2str(enum sppcmd cmd)
{
	switch (cmd) {
	case CLIENT_CMD_CRED:
		return "CRED";
	case CLIENT_CMD_GET:
		return "GET";
	case CLIENT_CMD_HEAD:
		return "HEAD";
	case CLIENT_CMD_IGOT:
		return "IGOT";
	case CLIENT_CMD_PEERS:
		return "PEERS";
	case CLIENT_CMD_PING:
		return "PING";
	case CLIENT_CMD_STARTTLS:
		return "STARTTLS";
	case CLIENT_CMD_REVCONN:
		return "REVCONN";
	}
	ERRX("no command matched");
	/*NOTREACHED*/
}
