/* spp_cred.c - spp cred command handling */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: spp_cred.c,v 1.55 2010/07/31 09:48:09 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "p2pd.h"
#include "block.h"
#include "dlagent.h"
#include "cred.h"
#include "peer.h"
#include "socket.h"
#include "spp.h"
#include "urient.h"

#define MAX_CRED 50


/* spp_cmd_cred - send cred command to parent node */
void
spp_cmd_cred(struct sppconn *sc)
{
	const char *f = "spp_cmd_cred";
	struct connection *c;
	char str[100];
	char *enc;
	float rtt;

	c = sc->conn;

	dbg("%s", f);

	ASSERT(c->obuflen == 0, "buffered data");

	strlcpy(c->obuf, COMMAND_CRED, sizeof(c->obuf));
	if ((enc = uri_encode(sc->ue->uri.path)) == NULL)
		return;
	strlcat(c->obuf, enc, sizeof(c->obuf));
	strlcat(c->obuf, " ", sizeof(c->obuf));
	strlcat(c->obuf, VALUE_VERSION_SPP, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	strlcat(c->obuf, OPTION_HOST, sizeof(c->obuf));
	if ((enc = uri_encode(sc->ue->uri.host)) == NULL)
		return;
	strlcat(c->obuf, enc, sizeof(c->obuf));
	strlcat(c->obuf, ":", sizeof(c->obuf));
	strlcat(c->obuf, sc->ue->uri.port, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	strlcat(c->obuf, OPTION_BINDPORT, sizeof(c->obuf));
	strlcat(c->obuf, lport, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	/* return RTT in ms/10 */
	if (peerpar_getfloat2(c->peer, PEERPAR_LINK_RTT, &rtt) != -1) {
		strlcat(c->obuf, OPTION_RTT, sizeof(c->obuf));
		snprintf(str, sizeof(str), "%d", (int)(rtt * 10));
		strlcat(c->obuf, str, sizeof(c->obuf));
		strlcat(c->obuf, "\r\n", sizeof(c->obuf));
	}

	if (strlcat(c->obuf, "\r\n", sizeof(c->obuf)) >= sizeof(c->obuf))
		ERRX("request too long");

/*	dbg("%s: %s", f, c->obuf);*/

	c->obuflen = strlen(c->obuf);

	UTOSPP(sc->ue)->lastcredcheck = time(NULL);
}


/* handleresp_cred - handle response to cred request */
void
handleresp_cred(struct rep *rep, struct sppconn *sc)
{
	const char *f = "handleresp_cred";

	ASSERT(sc->currcmd == CLIENT_CMD_CRED, "noncred command");

	if (rep->code != 200) {
		dbg("%s: CRED failed: %d", f, rep->code);
		/* received a proper reply, reuse possible  */
		cache_connection(sc->conn);
		return;
	}

	/*XXX handle more generally? */
	if (rep->size == 0) {
		dbg("%s: error: successful CRED with no content", f);
		errno_spp = 403; /*XXX resp code? */
		conn_changestate(sc->conn, CONNECTION_STATUS_ENDED);
		return;
	}

	/* prepare to read CRED data */
	sc->conn->ibufexplen = rep->size;
	conn_changestate(sc->conn, CONNECTION_STATUS_READREPDAT);
}


/* handle_cred - cred request handling (server/scc) */
void
handle_cred(struct sppconn *sc, struct req *req)
{
	const char *f = "handle_cred";
	struct credline hlist[MAX_CRED];
	char redir[MAX_REQREPLEN];
	struct connection *c;
	unsigned int i, j;
	char str[100];

	(void)req;

	c = sc->conn;

	dbg("%s: CRED request from %s", f, c->peer->id);

	ASSERT(c->obuflen == 0, "buffered data");

	/* XXX when to return error? */
	strlcpy(c->obuf, "SPP/1.0 200 OK\r\n", sizeof(c->obuf));

	ASSERT(PTOSP(c->peer)->pcredlen <= MAX_CRED, "MAX_CRED");

	/* format is a newline separated "<ip> [port] cnt [L]" list */
	for (i = 0; i < PTOSP(c->peer)->pcredlen; i++) {
		strlcpy(hlist[i].addr, PTOSP(c->peer)->pcred[i].rcpt->name, sizeof(hlist[i].addr));
		strlcpy(hlist[i].port, PTOSP(c->peer)->pcred[i].rcpt->port, sizeof(hlist[i].port));
		hlist[i].cred = PTOSP(c->peer)->pcred[i].totlen;
		if (PTOSP(PTOSP(c->peer)->pcred[i].rcpt)->leechscore >
		    PEER_LEECHSCORE_MAXBLOCKS * sc->ue->meta->blocksize)
			hlist[i].leecher = 1;
		else
			hlist[i].leecher = 0;
	}
	PTOSP(c->peer)->pcredlen = 0;

	bzero(redir, sizeof(redir));
	for (j = 0; j < i; j++) {
		dbg(" %s:%s (%lld)", hlist[j].addr, hlist[j].port,
		     (long long)hlist[j].cred);
		strlcat(redir, hlist[j].addr, sizeof(redir));
		/* XXX only send if different from SPP_PORT */
		strlcat(redir, " ", sizeof(redir));
		strlcat(redir, hlist[j].port, sizeof(redir));
		strlcat(redir, " ", sizeof(redir));
		snprintf(str, sizeof(str), "%lld", (long long)hlist[j].cred);
		strlcat(redir, str, sizeof(redir));
		if (hlist[j].leecher == 1) {
			strlcat(redir, " L", sizeof(redir));
			dbg("%s: %s:%s marked as leecher", f, hlist[j].addr,
			    hlist[j].port);
		}
		if (strlcat(redir, "\r\n", sizeof(redir)) >= sizeof(redir))
			ERRX("request too long");
		dbg("%s: redir: %s", f, redir);
	}

	/*XXX return error if content length zero? */

	strlcat(c->obuf, "Content-Length: ", sizeof(c->obuf));
	snprintf(str, sizeof(str), "%d", (int)strlen(redir));
	strlcat(c->obuf, str, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));
	if (strlcat(c->obuf, redir, sizeof(c->obuf)) >= sizeof(c->obuf))
		ERRX("request too long");

	c->obuflen = strlen(c->obuf);
	conn_changestate(c, CONNECTION_STATUS_WRITEREP);
}


/* parse_hostcred - parse list of peers which have credited us for uploading */
int
parse_hostcred(struct sppconn *sc)
{
	const char *f = "parse_hostcred";
	struct credline hlist[MAX_CRED];
	struct connection *c;
	struct speer *sp;
	off_t cred;
	int credcnt;
	int i;

	c = sc->conn;
/*	dbg("%s: %s", f, c->obuf); */

	if ((credcnt = cred_parse(&hlist[0], MAX_CRED, c->ibuf)) == -1) {
		c->ibuflen = 0;
		return -1;
	}
	c->ibuflen = 0;

	ASSERT(credcnt <= MAX_CRED, "credcnt exceeds MAX_CRED");

	for (i = 0; i < credcnt; i++) {
		if ((sp = speer_add(hlist[i].addr, hlist[i].port)) == NULL) {
			dbg("%s: speer_add failure for %s:%s", f, hlist[i].addr,
			    hlist[i].port);
			continue;
		}
		peerpar_setint(SPTOP(sp), PEERPAR_SPP_LEECHER, hlist[i].leecher);
		cred = hlist[i].cred;

		/* XXX int overflow? */
		if (cred > sp->uncredited) {
			dbg("%s: too many bytes credited (%lld > %lld)", f,
			     (long long)cred, (long long)sp->uncredited);
			/* XXX ignore? */
			continue;
		}
		sp->uncredited -= cred;
		dbg("%s: credited for %lld bytes by %s, "
		    "%lld uncredited bytes remaining", f, (long long)cred,
		    SPTOP(sp)->id, (long long)sp->uncredited);
	}

	UTOSPP(sc->ue)->lastcredres = time(NULL);

	return 0;
}
