/* spp_head.c - HEAD spp command handling */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: spp_head.c,v 1.53 2010/06/29 15:48:06 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include <time.h>
#include <unistd.h>

#include "p2pd.h"
#include "block.h"
#include "filemeta.h"
#include "peer.h"
#include "request.h"
#include "spp.h"
#include "urient.h"


/* spp_cmd_head - construct head request */
int
spp_cmd_head(struct sppconn *sc)
{
	const char *f = "spp_cmd_head";
	struct connection *c;
	char *enc;

	ASSERT(sc->ue->master == 0, "head by master")

	dbg("%s", f);

	c = sc->conn;
	ASSERT(c->obuflen == 0, "buffered data");

	strlcpy(c->obuf, COMMAND_HEAD, sizeof(c->obuf));
	if ((enc = uri_encode(sc->ue->uri.path)) == NULL)
		return -1;
	strlcat(c->obuf, enc, sizeof(c->obuf));
	strlcat(c->obuf, " ", sizeof(c->obuf));
	strlcat(c->obuf, VALUE_VERSION_SPP, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	strlcat(c->obuf, OPTION_HOST, sizeof(c->obuf));
	if ((enc = uri_encode(sc->ue->uri.host)) == NULL)
		return -1;
	strlcat(c->obuf, enc, sizeof(c->obuf));
	strlcat(c->obuf, ":", sizeof(c->obuf));
	strlcat(c->obuf, sc->ue->uri.port, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	strlcat(c->obuf, OPTION_BINDPORT, sizeof(c->obuf));
	strlcat(c->obuf, lport, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	if (strlcat(c->obuf, "\r\n", sizeof(c->obuf)) >= sizeof(c->obuf))
		ERRX("request too long");

	c->obuflen = strlen(c->obuf);

	dbg("%s: %s", f, c->obuf);

	return 0;
}


/* handleresp_head - handle head reply from parent node */
void
handleresp_head(struct rep *rep, struct sppconn *sc)
{
	const char *f = "handleresp_head";
	struct connection *c;
	struct urient *ue;
	size_t maxlen;

	c = sc->conn;
	ASSERT(sc->currcmd == CLIENT_CMD_HEAD, "incorrect command");

	dbg("%s", f);

	/*XXX should have whoami command */
	if (strlen(mysockname) == 0) {
		maxlen = sizeof(mysockname);
		if (strlcpy(mysockname, c->sockname, maxlen) >= maxlen)
			ERRX("mysocklen too long");
	}

	ue = sc->ue;
	if (rep->code != 200) {
		sppfile_headfail(UTOSPP(ue), sc, rep->code);
		return;
	}

	if (!block_validsize(rep->blocksize)) {
		swarn("%s: %s: invalid blocksize: %d", f, sc->ue->uri.uri,
		      rep->blocksize);
		sppfile_headfail(UTOSPP(ue), sc, rep->code);
		return;
	}

	if (urient_cacheinit(ue, rep->size, rep->blocksize,
			     DIGEST_DEFAULT) == -1) {
		sppfile_headfail(UTOSPP(ue), sc, 502); /* Bad Gateway */
		return;
	}

	sppfile_headok(UTOSPP(ue), sc);
	/* XXX sppfile_headok() caches connection */
}


/* handle_head - handle head request */
void
handle_head(struct sppconn *sc, struct req *req)
{
	const char *f = "handle_head";
#ifdef TESTFAIL
	static int cnt;
#endif /* TESTFAIL */
	struct rep rep;

	(void)req;

	dbg("%s: HEAD request from %s", f, sc->conn->peer->id);

	bzero(&rep, sizeof(rep));

#ifdef TESTFAIL
	/* ensure that headchain code gets tested */
	if (!lhc_mode && cnt++ < 2)
		rep.code = 504;
	else
#endif /* TESTFAIL */
		rep.code = 200;

	rep.size = sc->ue->meta->size;
	if (sc->httpreq == 0)
		rep.blocksize = sc->ue->meta->blocksize;

	spp_cmd_rep2(sc, &rep);
}
