/* spp_get.c - spp IGOT command handling */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: spp_igot.c,v 1.83 2010/10/30 09:25:03 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "p2pd.h"
#include "alist.h"
#include "block.h"
#include "blocksum.h"
#include "dlagent.h"
#include "file.h"
#include "filemeta.h"
#include "igot.h"
#include "peer.h"
#include "request.h"
#include "socket.h"
#include "ssl.h"
#include "spp.h"
#include "timers.h"
#include "urient.h"

/* send outstanding igots after flushtime */
#define SPP_IGOT_FLUSHTIME 5

#define REQ_IGOT_MAX 50 /* max number of entries in IGOT command */

extern int igot_sslonly;
int igot_accept = 0; /* ignore igot commands by default */

void igot_endtransit(struct sppconn *);


/*XXX should return status */
/* spp_cmd_get - construct igot request */
void
spp_cmd_igot(struct sppconn *sc)
{
	const char *f = "spp_cmd_igot";
	struct connection *c;
	struct peer *parent;
	struct igot *l_i;
	char igotdat[4096];/*XXX*/
	char str[1024];
	int igotcnt;
	float rtt;
	char *enc;
	int i;

	ASSERT(sc != NULL, "NULL sc");
	ASSERT(sc->ue != NULL, "NULL ue");

	dbg("%s", f);

	c = sc->conn;
#if HAVE_OPENSSL
	if (use_ssl && c->peer->sslactive == 0)
		ERRX("peer not sslactive");
#endif /* HAVE_OPENSSL */

	parent = UTOSPP(sc->ue)->parent;

	ASSERT(ALIST_SIZE(PTOSP(parent)->al_igots) > 0,
	       "no igots in spp_cmd_igot");

	ASSERT(c->obuflen == 0, "buffered data");

	strlcpy(c->obuf, COMMAND_IGOT, sizeof(c->obuf));
	if ((enc = uri_encode(sc->ue->uri.path)) == NULL)
		return;
	strlcat(c->obuf, enc, sizeof(c->obuf));
	strlcat(c->obuf, " ", sizeof(c->obuf));
	strlcat(c->obuf, VALUE_VERSION_SPP, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	strlcat(c->obuf, OPTION_HOST, sizeof(c->obuf));
	if ((enc = uri_encode(sc->ue->uri.host)) == NULL)
		return;
	strlcat(c->obuf, enc, sizeof(c->obuf));
	strlcat(c->obuf, ":", sizeof(c->obuf));
	strlcat(c->obuf, sc->ue->uri.port, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	strlcat(c->obuf, OPTION_BINDPORT, sizeof(c->obuf));
	strlcat(c->obuf, lport, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	/* return RTT in ms/10 */
	if (peerpar_getfloat2(c->peer, PEERPAR_LINK_RTT, &rtt) != -1) {
		strlcat(c->obuf, OPTION_RTT, sizeof(c->obuf));
		snprintf(str, sizeof(str), "%d", (int)(rtt * 10));
		strlcat(c->obuf, str, sizeof(c->obuf));
		strlcat(c->obuf, "\r\n", sizeof(c->obuf));
	}

	bzero(igotdat, sizeof(igotdat));
	/* format: bnum digest [peer credit] */
	igotcnt = 0;
	/*
	 * do one urient at at time, in the submitted order.
	 * igots for ue are added, skipping any igots for other urients.
	 */
	for (i = 0; (size_t)i < ALIST_SIZE(PTOSP(parent)->al_igots); i++) {
		l_i = ALIST_PVAL(PTOSP(parent)->al_igots, i);
		/* skip igots for other files */
		if (l_i->ue != sc->ue)
			continue;

		/* stop once igots for next file is found */
		ASSERT(l_i->sent == 0, "duplicate igot in transit");

		if (igotcnt >= REQ_IGOT_MAX)
			break;

		dbg("%s: %s bnum %d", f, l_i->ue->uri.uri, l_i->bnum);

		snprintf(str, sizeof(str), "%d", l_i->bnum);
		strlcat(igotdat, str, sizeof(igotdat));
		strlcat(igotdat, " ", sizeof(igotdat));
		strlcat(igotdat, l_i->digest, sizeof(igotdat));
		if (l_i->peer != NULL) {
			strlcat(igotdat, " ", sizeof(igotdat));
			strlcat(igotdat, l_i->peer->name, sizeof(igotdat));
			strlcat(igotdat, " ", sizeof(igotdat));
			strlcat(igotdat, l_i->peer->port, sizeof(igotdat));

			if (PTOSP(l_i->peer)->unreported > 0)
				dbg("%s: still uncredited: %lld", f,
				     (long long)PTOSP(l_i->peer)->unreported);

			/* XXX could drop credit if same as length of range */
			/* XXX per block accounting? */
			strlcat(igotdat, " ", sizeof(igotdat));
			snprintf(str, sizeof(str), "%lld",
				 (long long)PTOSP(l_i->peer)->unreported);
			strlcat(igotdat, str, sizeof(igotdat));
			/* maintain value in igot until sent or aborted */
			l_i->cred = PTOSP(l_i->peer)->unreported;
			PTOSP(l_i->peer)->unreported = 0;
		}
		l_i->sent = 1;
		if (strlcat(igotdat, "\r\n", sizeof(igotdat)) >= sizeof(igotdat))
			ERRX("too much igot data");
		igotcnt++;
	}

	dbg("%s: sending %d postponed IGOT commands", f, igotcnt);

	ASSERT(igotcnt != 0, "no igots transmitted");

	strlcat(c->obuf, "Content-Length: ", sizeof(c->obuf));
	snprintf(str, sizeof(str), "%d", (int)strlen(igotdat));
	strlcat(c->obuf, str, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));
	if (strlcat(c->obuf, igotdat, sizeof(c->obuf)) >= sizeof(c->obuf))
		ERRX("too much igot command data");

	c->obuflen = strlen(c->obuf);

	PTOSP(parent)->igotactive = 1;

/*	dbg("%s: %s %s", f, ue->uri.uri, c->obuf);*/

/*	conn_changestate(c, CONNECTION_STATUS_SSLCACHED);*/ /*XXX*/
}


/* handleresp_igot - handle response to igot command */
void
handleresp_igot(struct rep *rep, struct sppconn *sc)
{
	const char *f = "handleresp_igot";
	struct connection *c;
	struct peer *parent;

	ASSERT(sc->currcmd == CLIENT_CMD_IGOT, "incorrect command");

	c = sc->conn;
	parent = UTOSPP(sc->ue)->parent;

	dbg("%s: respcode: %d", f, rep->code);

	if (rep->code != 200) {
		ASSERT(rep->code == 409, "unexpected error");

		dbg("%s: igot failures, reading errors", f);
		sc->sppcode = rep->code;
		sc->conn->ibufexplen = rep->size;
		conn_changestate(sc->conn, CONNECTION_STATUS_READREPDAT);
		return;
	}

	/* start revconn timeout */
	if (UTOSPP(sc->ue)->sentigot == 0) {
		UTOSPP(sc->ue)->sentigot = 1;
		dbg("%s: enabling revconn timer for %s", f, sc->ue->uri.uri);
		timeout_secupd(UTOSPP(sc->ue)->to_revconn,
			       FILE_PROACTIVEREQ_DELAY);
	}

	/* remove submitted igots from queue */
	igot_endtransit(sc);

	PTOSP(parent)->igotactive = 0;

	cache_connection(c);

	/* send one more request if still outstanding data */
	if (ALIST_SIZE(PTOSP(parent)->al_igots) > 0)
		igot_flushnode(parent);
}


/* parse_igoterrs - read list of blocks with bad checksums */
void
parse_igoterrs(struct sppconn *sc)
{
	const char *f = "parse_igoterrs";
	struct peer *parent;
	const char *errstr;
	struct igot *l_i;
	int32_t bnum;
	char *line;
	char *e;
	int i;

	parent = UTOSPP(sc->ue)->parent;

	PTOSP(parent)->igotactive = 0;

#define LINEEND " BAD\r\n"

	line = sc->conn->ibuf;
	for (;;) {
		/* end of data if no terminating newline */
		if ((e = strstr(line, LINEEND)) == NULL)
			break;
		*e = '\0';

		bnum = strtonum(line, 0, sc->ue->meta->blockcnt - 1, &errstr);
		if (errstr) {
			swarn("%s: bnum parse error: %s", f, errstr);
			conn_changestate(sc->conn, CONNECTION_STATUS_ENDED);
			return;
		}

		/*
		 * need to remove block from disk and memory,
		 *  should stop all ongoing transfers of block from memory.
		 */
		dbg("%s: invalid digest: %s (block %d)", f, sc->ue->uri.uri,
		    bnum);
		urient_lostblock(sc->ue, bnum);
		urient_purgeblock(sc->ue, bnum);
		dlagent_purgedblock(sc->ue, bnum);

		/* remove from igot queue */
		for (i = ALIST_SIZE(PTOSP(parent)->al_igots) - 1; i >= 0; i--) {
			l_i = ALIST_PVAL(PTOSP(parent)->al_igots, i);
			if (l_i->ue != sc->ue || l_i->bnum != bnum)
				continue;
			ASSERT(l_i->sent, "error on unsent igot");

			if (l_i->peer != NULL) {
				/* restore unreported value */
				PTOSP(l_i->peer)->unreported += l_i->cred;
				l_i->cred = 0;

				/*
				 * this peer either sent bad data
				 * (or it got corrupted somewhere).
				 * increase badblocks counter.
				 */
				peerpar_adjint(l_i->peer,
					       PEERPAR_SPP_BADBLOCKS, 1);
			}
			alist_delidx(PTOSP(parent)->al_igots, i);
		}

		line += strlen(LINEEND);
	}
	sc->conn->ibuflen = 0;

#if 0 /*XXX should surely resend the rest? */
	/* remove submitted igots from queue */
	igot_endtransit(sc);
#endif
	igot_restore(sc->ue);

	cache_connection(sc->conn);

	/* resend if still any igots remaining */
	if (ALIST_SIZE(PTOSP(parent)->al_igots) > 0)
		igot_flushnode(parent);
}


/* igot_add - add information on downloaded block for reporting to server */
void
igot_add(struct urient *ue, int32_t bnum, char *digest, struct peer *source)
{
	const char *f = "igot_add";
	struct peer *parent;
	struct igot *igot;
	struct igot *l_i;
	int i;

	/* send to parent */
	parent = UTOSPP(ue)->parent;

	/* first igot for parent node? */
	if (PTOSP(parent)->al_igots == NULL) {
		/* init igot structures */
		dbg("%s: preparing for igots to %s", f, parent->id);

		PTOSP(parent)->to_igot =
			timeout_add((void (*)(void *))igot_flushnode, NULL,
				    parent);
		PTOSP(parent)->al_igots = alist_init(NULL, NULL, 1,
						   ALIST_SIZE_GROW,
						   ALIST_TYPE_UNSORTED_PTR);
		ASSERT(PTOSP(parent)->al_igots != NULL, "alist_init failure");
	}

	/*
	 * not actually optimal, but walk through and skip duplicates
	 * (can occur as a result of dbrefresh, if igot already exists)
	 * (ignoring duplicates is an alternative; either waste cpu
	 *  here or bandwidth on transmission).
	 */
	for (i = 0; (size_t)i < ALIST_SIZE(PTOSP(parent)->al_igots); i++) {
		l_i = ALIST_PVAL(PTOSP(parent)->al_igots, i);
		if (l_i->ue == ue && l_i->bnum == bnum) {
			dbg("%s: duplicate igot (%s, %d)", f, ue->uri.uri, bnum);
			/* XXX should rarely occur, ERR for now */
			ERRX("duplicate igot");
			return;
		}
	}

	/*
	 * refcount is increased for each igot. having it increased only for
	 * unique files in queue would be more elegant but this way is simpler.
	 */
	sppfile_addref(UTOSPP(ue));

	/*
	 * maintain timer to ensure igots are sent
	 * (can also be sent if igot queue reaches REQ_IGOT_MAX limit (below)
	 */
	if (ALIST_SIZE(PTOSP(parent)->al_igots) == 0)
		timeout_secupd(PTOSP(parent)->to_igot, SPP_IGOT_FLUSHTIME);

	igot = xmalloc(sizeof(struct igot));
	bzero(igot, sizeof(struct igot));

	igot->ue = ue;
	igot->bnum = bnum;
	if (strlcpy(igot->digest, digest, sizeof(igot->digest)) >= sizeof(igot->digest)) {
		ERRX("digest too long");
	}
	igot->peer = source;

	/* use value as key (unsorted array, key not used) */
	if (alist_add(PTOSP(parent)->al_igots, igot, igot) == NULL)
		ERRX("alist_add failed"); /* should not occur for GROW array */

	/* initiate flush direcly when request limit is reached */
	if (ALIST_SIZE(PTOSP(parent)->al_igots) == REQ_IGOT_MAX)
		igot_flushnode(parent);
}


/* igot_endtransit - remove in-transit igots, command ended */
void
igot_endtransit(struct sppconn *sc)
{
	const char *f = "igot_endtransit";
	struct peer *parent;
	struct igot *l_i;
	int i;

	ASSERT(sc->currcmd == CLIENT_CMD_IGOT, "incorrect command");

	parent = UTOSPP(sc->ue)->parent;

	for (i = ALIST_SIZE(PTOSP(parent)->al_igots) - 1; i >= 0; i--) {
		l_i = ALIST_PVAL(PTOSP(parent)->al_igots, i);
		if (l_i->sent == 1) {
			dbg("%s: reported %s block %d", f, l_i->ue->uri.uri,
				l_i->bnum);
			/* decrease ref count for each igot*/
			sppfile_unref(UTOSPP(sc->ue));

			alist_delidx(PTOSP(parent)->al_igots, i);
			free(l_i);
		}
	}
}


/* igot_flushnode - transmit outstanding igots to peer */
void
igot_flushnode(struct peer *peer)
{
	const char *f = "igot_flushnode";
#if HAVE_OPENSSL
	struct connection *c;
#endif /* HAVE_OPENSSL */
	struct sppconn *sc;
	struct urient  *ue;

	ASSERT(ALIST_SIZE(PTOSP(peer)->al_igots) != 0, "flush on empty queue");

	/* abort if already active */
	if (PTOSP(peer)->igotactive) {
		dbg("%s: igot already active for %s, skipping", f, peer->id);
		return;
	}
	if (PTOSP(peer)->sslinitactive) {
		dbg("%s: sslinit already active for %s, skipping", f, peer->id);
		return;
	}

	/* disable node igot timer until finished flushing */
	timeout_off(PTOSP(peer)->to_igot);

	/*
	 * igots for multiple files can be outstanding, start with
	 * first file.
	 */
	ue = ((struct igot *)ALIST_PVAL(PTOSP(peer)->al_igots, 0))->ue;
	ASSERT(ue != NULL, "NULL ue");
	if (use_ssl) {
#if HAVE_OPENSSL
		if (peer_validaddr(peer) &&
		    (c = conn_getcached(peer, 1)) != NULL) {
			sc = sppconn_new(CLIENT_CMD_IGOT);
			sppconn_setfile(sc, ue);
			sppconn_setconn(sc, c);
			spp_cmd_igot(sc);
		} else if (PTOSP(peer)->sslinitactive == 0) {
			dbg("no ssl connection active, issuing starttls");
			sc = sppconn_new(CLIENT_CMD_STARTTLS);
			sppconn_setfile(sc, ue);
			spp_parentreq(sc);
			PTOSP(peer)->sslinitactive = 1;
			sc->sslinitpeer = peer;
		} else
			ERRX("ssl init failure");
#else
		ERRX("no ssl support");
#endif /* HAVE_OPENSSL */
	} else {
		sc = sppconn_new(CLIENT_CMD_IGOT);
		sppconn_setfile(sc, ue);
		spp_parentreq(sc);
		/*
		 * XXX avoid igot_flushnode() being called multiple times.
		 *     should have generic way of checking whether
		 *     a specific command is being underway,
		 *     or something like dlagent to keep state.
		 */
		PTOSP(peer)->igotactive = 1;
	}
}


/* dbrefresh - send igots for all blocks kept in cache */
void
dbrefresh(struct urient *ue)
{
	const char *f = "dbrefresh";
	int i;

	/*
	 * (should not be possible for non-scc/server nodes to
	 * cause this function to be called).
	 */

	ASSERT(ue != NULL, "NULL ue");
	ASSERT(ue->cf != NULL, "dbrefresh on noncache file");

	dbg("%s: %s", f, ue->uri.uri);

	/* ignore too frequent dbrefresh requests */
	if (UTOSPP(ue)->lastrefresh + SPP_DBREFRESH_INTERVAL > time(NULL)) {
		dbg("%s: recently did dbrefresh, ignoring", f);
		return;
	}

	sppfile_addref(UTOSPP(ue));
	if (urient_fileref(ue) == -1) {
		dbg("%s: unable to open file, no dbrefresh for %s", f,
		    ue->uri.uri);
		return;
	}

	for (i = 0; i < ue->meta->blockcnt; i++)
		if (urient_haveblock(ue, i)) {
			ASSERT(ue->file->blocks[i]->sum != NULL,
			       "have block but no sum");
			/*
			 * skip source here, only included as part of
			 * crediting uploads, which is only done once
			 * (dbrefresh would typically result in
			 * already sent igots being sent again).
			 */
			/*
			 * dbrefresh might be called for file which
			 * still has outstanding igots for same file,
			 * igot_add need to handle dupliate igot
			 * entries.
			 */
			igot_add(ue, i, ue->file->blocks[i]->sum, NULL);
		}

	urient_fileunref(ue);
	UTOSPP(ue)->lastrefresh = time(NULL);
	sppfile_unref(UTOSPP(ue));
}


/*XXX test how current igot handling handles big cache */
/*XXX if cache big, should all files be reported? */
/* dbrefresh_all - send updates for all (cached) files */
void
dbrefresh_all(void)
{
	struct urient *ue;
	int i;

	for (i = 0; (size_t)i < ALIST_SIZE(al_urient); i++) {
		ue = ALIST_PVAL(al_urient, i);
		if (ue->cf == NULL)
			continue; /* not created on disk */
		dbrefresh(ue);
	}
}


/* igot_restore - restore in-transit IGOTs */
void
igot_restore(struct urient *ue)
{
	const char *f = "igot_restore";
	struct peer *parent;
	struct igot *l_i;
	int i;

	dbg("%s: ending igot, restoring outstanding igots", f);

	parent = UTOSPP(ue)->parent;

	for (i = 0; (size_t)i < ALIST_SIZE(PTOSP(parent)->al_igots); i++) {
		l_i = ALIST_PVAL(PTOSP(parent)->al_igots, i);
		if (l_i->ue != ue)
			continue;

		if (l_i->sent) {
			dbg("%s: igot for block %d restored", f, l_i->bnum);

			/* restore unreported value */
			if (l_i->peer != NULL)
				PTOSP(l_i->peer)->unreported += l_i->cred;
			l_i->cred = 0;
			l_i->sent = 0;
		}
	}
}


/* handle_igot - process igot request (server/scc) */
int
handle_igot(struct sppconn *sc, struct req *req)
{
	const char *f = "handle_igot";
	struct connection *c;

	dbg("%s", f);

	c = sc->conn;

#if HAVE_OPENSSL
	if (igot_sslonly && (c->ssl_serv == 0)) {
		dbg("%s: IGOT on non-ssl connection from %s, denying", f,
		    c->peer->id);
		spp_cmd_rep(sc, 469); /* XXX Not Auhenticated */
		return -1;
	}
#endif /* HAVE_OPENSSL */

	/* prepare to read IGOT data */
	c->ibufexplen = req->size;
	conn_changestate(c, CONNECTION_STATUS_READREQDAT);

	return 0;
}


/* parse_igots - process igot data from client */
int
parse_igots(struct sppconn *sc)
{
	const char *f = "parse_igots";
	struct igot pigots[REQ_IGOT_MAX];
	struct peer *uploader;
	struct connection *c;
	struct igot *igot;
	struct file *file;
#ifdef TESTFAIL
	static int testfail;
#endif /* TESTFAIL */
	char failstr[4096]; /*XXX*/
	struct speer *sp;
	unsigned int i;
	int firstcred;
	char str[10];
	int32_t bnum;
	int igotcnt;
	int nofail;
	int j;

	c = sc->conn;
	file = sc->ue->file;

	dbg("igot: %s", c->ibuf);
	bzero(&pigots, sizeof(pigots));
	if ((igotcnt = igots_parse(&pigots[0], REQ_IGOT_MAX, c->ibuf)) == -1) {
		spp_cmd_rep(sc, 403); /*XXX errcode? */
		return -1;
	}
	c->ibuflen = 0;

	/*interpret */
	if (igotcnt == 0) {
		dbg("%s: empty igot", f);
		spp_cmd_rep(sc, 403);
		return -1;
	}
	dbg("%s: %d igots from %s", f, igotcnt, c->peer->id);

	nofail = 1;
	failstr[0] = '\0';
	for (j = 0; j < igotcnt; j++) {
		igot = &pigots[j];
		bnum = igot->bnum;

		if (strlen(igot->addr) > 0)
			dbg("%s: %s got %s block %d from %s:%s", f,
			    c->peer->id, sc->ue->uri.uri, bnum, igot->addr,
			    igot->port);
		else
			dbg("%s: %s got %s block %d", f, c->peer->id,
			    sc->ue->uri.uri, bnum);

		/*XXX check that offend is sane */
		if (bnum >= sc->ue->meta->blockcnt) {
			swarn("%s: IGOT invalid block number %d", f, bnum);
			spp_cmd_rep(sc, 403);
			return -1;
		}

		if (strlen(igot->addr) > 0) {
			if ((sp = speer_add(igot->addr, igot->port)) == NULL)
				ERRX("speer_add failure");
			igot->peer = SPTOP(sp);
		}

		/*XXX lots of sanity checking */

		if (igot->cred > sc->ue->meta->size) {
			swarn("%s: large credit value: %lld > %lld", f,
			      (long long)igot->cred,
			      (long long)sc->ue->meta->size);
#if 0 /* XXX could be larger if some data downloaded multiple times */
			spp_cmd_rep(sc, 403);
			return -1;
#endif
		}

		/*XXX block_init*/

		if (file->blocks[bnum]->sum == NULL) {
			dbg("%s: IGOT for block %d with unknown digest, "
			    "calculating", f, bnum);
			if (urient_haveblock(sc->ue, bnum))
				block_gendigest(sc->ue, bnum);
			else
				/*
				 * XXX how to handle this?
				 * scc does (no longer?) have block
				 * located at scc client
				 * -resending igot request is probably
				 *  simplest solution (doing igot to
				 *  parent node with value received by
				 *  client).
				 * -should build batch request with all
				 *  blocks in received request not
				 *  found locally, and build
				 *  response based on that.
				 *  (will need 'do not know' response)
				 * -problem is that parent not will think
				 *  we have the blocks -> will need to
				 *  to decline requests or do priority
				 *  retrieval
				 *
				 * ...
				 *
				 * -a progmatic alternative might be to
				 *  use this as reason to go from
				 *  using igot as a 'malicious user'
				 *  preventive measure and simply
				 *  return checksums for a requested
				 *  range in the head request.
				 * -will mean that multiple head requests
				 *  might need to be issued
				 * -having initial request be large
				 *  would place larger load on server
				 *  from clients that never retrieve more
				 *  than a few blocks
				 * -might think of problem in relation
				 *  to live streaming and file download
				 * -would need classifier that specified
				 *  what kind of anti-leecher techniques
				 *  which could be used.
				 * -current block for live streaming?
				 * -type: random|linear|linear-range-limited
				 *     (download|vod   |live stream)
				 *
				 * -alternative, keep the current
				 *  status solution, simply say
				 *  digests are ok and ignore igot
				 *  messages. if for popular file, it
				 *  will probably be requested by
				 *  different client, in which case it
				 *  can be downloaded then. the
				 *  clients that report the blocks
				 *  before the scc knows what is
				 *  correct will be ignored (i.e.,
				 *  reduction in efficiency), but this
				 *  might not matter in practice.
				 *
				 * -situation in itself should be relatively
				 *  rare, only if new scc node is inserted,
				 *  or if the scc node crashes hard, etc.
				 *  in effect situations where all checksum
				 *  data is lost
				 * -some related questions, how long should
				 *  checksum data for files not in cache be
				 *  retained (if at all).
				 * -should do simulations to evaluate the
				 *  benefits of clients having to
				 *  verify checksums, due to some of
				 *  the problematic situations it
				 *  creates. if clients know the checksums
				 *  in advance, there is at least some cost
				 *  associated with attempting to spam the
				 *  network with invalid clients (that
				 *  report that they have block but
				 *  only send garbage/altered content.
				 *  one malicious user with many clients
				 *  will only need to download a
				 *  single block once, while many malicious
				 *  users will each need to download at least
				 *  one block once. since multiple clients can
				 *  run on the same machine, this does in
				 *  practice not limit this to users having
				 *  many addresses.
				 *
				 * -one alternative to having checksums
				 *  returned in response to the head command,
				 *  would be to have it returned in the response
				 *  to the get command, which typically
				 *  only allows requests for single blocks.
				 *  this would effectively spread out
				 *  the load of checksum transmission over the
				 *  whole session (and be practical for
				 *  live streaming).
				 * -would rather need to have it returned
				 *  in peers request, cannot trust info
				 *  from other peers, perhaps using head is
				 *  simplest after all.
				 *
				 * -strategies usable to protect against
				 *  malicious users.
				 *
				 *  -by endusers:
				 *
				 *   -downloading block with invalid checksum
				 *    from peer; down prioritize that peer
				 *    (and all peers on that ip-address)
				 *    -consequence; 1) each ip with a
				 *     malicious user would only get
				 *     one chance to upload a block
				 *     to a client
				 *    -consequence; 2) if all blocks invalid,
				 *     the malicious user would never
				 *     be reported as having uploaded to other
				 *     peers. unless the user employed another
				 *     node to report uploads, could also
				 *     ignore crediting between clients
				 *     on the same ip (assume that this
				 *     has no practical value). sameipfilter
				 *    -consequence; 3) if only some blocks
				 *     are invalid, the malicious node would
				 *     be credited, but practical effect
				 *     might not be too dissimilar from
				 *     network transmission errors.
				 *     (what if peer only transmits some
				 *      invalid data, and another peer
				 *      supplies the rest, how to know which
				 *      one was bad?)
				 *     -one way to address this might be
				 *      to reduce priority of nodes where
				 *      connetion is too slow/aborted or similar
				 *
				 *   -conc: the more discreet a client is,
				 *      the less effective it will be.
				 *      some natural lossage, due to network
				 *      errors, peer memory corruption, etc.
				 *      system will need to have some tolerance
				 *      to these problems. the goal would be
				 *      make it difficult for a malicious
				 *      peer to introduce more errors in a
				 *      system than would normally be there,
				 *      without being detected.
				 *
				 *  -by scc/server
				 *
				 *   -a malicious user that only sends bad
				 *    data would never be credited by normal
				 *    users. having mix of never credited
				 *    and credited clients would be a step
				 *    towards adressing this (all clients
				 *    start with never being credited).
				 *
				 *   -a malicious user that only sometimes
				 *    sends bad data be be credited as
				 *    a good user, no good way to detect
				 *    this in server.
				 *
				 *   -only including one peer per-ip address
				 *    would prevent any users from creating
				 *    many clients without having access to
				 *    more than one ip.
				 *
				 *   -conc: ensuring that peerlists include
				 *    as much diversity as possible is
				 *    should probably be the goal of the scc,
				 *    to give each client an as good as possible
				 *    starting point for taking care of itself.
				 *
				 * -the most efficient approach by a malicious
				 *  node might be to only send part of a block,
				 *  which would make detection difficult and
				 *  cause problems with little cost.
				 *
				 *  -could address this by having two download
				 *   modes:
				 *    -normal, for when everybody
				 *     appear to be nice, and do
				 *     nothing bad, in which case a
				 *     client will do partial
				 *     downloads from peers.
				 *
				 *    -careful, after a bad block checksum
				 *     has been received, will result in
				 *     peers only sending partial blocks
				 *     being marked as suspicious (PARTIALPEER)
				 *     and data discarded if the entire block
				 *     connot be downloaded from a single peer.
				 *
				 *   -the result of this would be that
				 *    it would be possible for a
				 *    malicious user to reduce the
				 *    performance of the network by
				 *    potentially causing
				 *    retransmissions of data, which
				 *    again would lead to increased
				 *    latency. the alternative would on
				 *    the other hand be to make it easy for
				 *    a malicious user to disrupt the network
				 *    to a much larger degree.
				 */
				continue; /* simply ignore for now */
		}

		/*XXX fail once while testing */
		if (
#ifdef TESTFAIL
			testfail++ == 0 ||
#endif /* TESTFAIL */

			strncmp(file->blocks[bnum]->sum, igot->digest, sizeof(file->blocks[bnum]->sum)) != 0) {
			nofail = 0;
			dbg("%s: IGOT digest mismatch: %s != %s", f,
			    igot->digest, file->blocks[bnum]->sum);

			snprintf(str, sizeof(str), "%d", bnum);
			strlcat(failstr, str, sizeof(c->obuf));
			strlcat(failstr, " BAD\r\n", sizeof(c->obuf));
			continue;
		}
		/* valid digest */

		/* don't update peer data if ignoring igots */
		if (igot_accept == 0)
			continue;

		/* XXX do reverse connect to verify? */

		filemeta_peerupd(sc->ue->meta, c->peer, bnum, 0, BLOCKINFO_IGOT);

		/* XXX make sure we don't add ourself */
		if (igot->peer == NULL) {
			dbg("%s: no peer", f);
			continue;
		}
		if ((strcmp(c->sockname, igot->peer->name) == 0) &&
		    (strcmp(c->sockport, igot->peer->port) == 0)) {
/*			dbg("%s: we are credited for block %d, ignoring", f,
			    bnum);*/
			continue;
		}

		uploader = igot->peer;
		filemeta_peerupd(sc->ue->meta, uploader, bnum, 0, BLOCKINFO_CRED);
		if (igot->cred == 0) {
			dbg("%s: IGOT: no peer credits", f);
			continue;
		}

		firstcred = 1;
		for (i = 0; i < PTOSP(uploader)->pcredlen; i++) {
			if (PTOSP(uploader)->pcred[i].rcpt == c->peer) {
				firstcred = 0;
				break;
			}
		}
		if (firstcred) {
			dbg("%s: firstcred in slot %d (%s:%s)", f, i,
			    uploader->name, uploader->port);
			PTOSP(uploader)->pcred[i].totlen = 0;
			PTOSP(uploader)->pcred[i].rcpt = c->peer;
			PTOSP(uploader)->pcredlen++;
			dbg("%s: pcredlen -> %d", f, PTOSP(uploader)->pcredlen);
			if (PTOSP(uploader)->pcredlen > PEER_CREDLEN)
				ERRX("PEER_CREDLEN");
		}

		PTOSP(uploader)->pcred[i].totlen += igot->cred;
		/*XXX check for overflow */
		PTOSP(c->peer)->leechscore -= igot->cred;
		/* XXX only increase if peer neither a leecher nor a nobody */
		PTOSP(uploader)->leechscore += igot->cred;
		dbg("%s: leechscore sender (%s) %lld", f,
		     c->peer->id, (long long)PTOSP(c->peer)->leechscore);
		dbg("%s: leechscore receiver (%s) %lld", f,
		    uploader->id, (long long)PTOSP(uploader)->leechscore);
	}

/*	dbg("%s: replying to IGOT", f);*/
	if (nofail)
		spp_cmd_rep(sc, 200);
	else {
		/*XXX which code? */
		strlcpy(c->obuf, "SPP/1.0 409 Bad Digest\r\n", sizeof(c->obuf));

		strlcat(c->obuf, "Content-Length: ", sizeof(c->obuf));
		snprintf(str, sizeof(str), "%d", (int)strlen(failstr));
		strlcat(c->obuf, str, sizeof(c->obuf));
		strlcat(c->obuf, "\r\n", sizeof(c->obuf));
		strlcat(c->obuf, "\r\n", sizeof(c->obuf));
		if (strlcat(c->obuf, failstr, sizeof(c->obuf)) >= sizeof(c->obuf))
			ERRX("request too long");

		c->obuflen = strlen(c->obuf);
		conn_changestate(c, CONNECTION_STATUS_WRITEREP);
	}

	return 0;
}
