/* spp_head.c - spp PING command handling */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: spp_ping.c,v 1.51 2010/06/30 08:38:00 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "p2pd.h"
#include "block.h"
#include "filemeta.h"
#include "peer.h"
#include "peerlist.h"
#include "peerset.h"
#include "request.h"
#include "socket.h"
#include "spp.h"
#include "urient.h"

/*
 *XXX need to include a secret value in ping requests, which must be
 *    included in reply, to protect against peers that attempt to
 *    create a lower than actual rtt by sending replies to multiple
 *    ping requests (e.g., after revconn).
 */


/* spp_cmd_ping - generate ping request to peer */
void
spp_cmd_ping(struct sppconn *sc)
{
	const char *f = "spp_cmd_ping";
	struct connection *c;
	char str[100];
	char *enc;

	c = sc->conn;

	ASSERT(c->obuflen == 0, "buffered data");
	ASSERT(c->ibuflen == 0, "buffered data");

	strlcpy(c->obuf, COMMAND_PING, sizeof(c->obuf));
	if ((enc = uri_encode(sc->ue->uri.path)) == NULL)
		return;
	strlcat(c->obuf, enc, sizeof(c->obuf));
	strlcat(c->obuf, " ", sizeof(c->obuf));
	strlcat(c->obuf, VALUE_VERSION_SPP, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	strlcat(c->obuf, OPTION_HOST, sizeof(c->obuf));
	if ((enc = uri_encode(sc->ue->uri.host)) == NULL)
		return;
	strlcat(c->obuf, enc, sizeof(c->obuf));
	strlcat(c->obuf, ":", sizeof(c->obuf));
	strlcat(c->obuf, sc->ue->uri.port, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	strlcat(c->obuf, OPTION_BINDPORT, sizeof(c->obuf));
	strlcat(c->obuf, lport, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	sc->nounce = rand() % SPP_NOUNCE_MAX; /*XXX proper random value*/
	strlcat(c->obuf, OPTION_NOUNCE, sizeof(c->obuf));
	snprintf(str, sizeof(str), "%d", sc->nounce);
	strlcat(c->obuf, str, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	if (strlcat(c->obuf, "\r\n", sizeof(c->obuf)) >= sizeof(c->obuf))
		ERRX("request too long");

	c->obuflen = strlen(c->obuf);

	dbg("%s: to %s", f, c->peer->id);

	peerpar_settime(c->peer, PEERPAR_SPP_LASTPING);
}


/* handleresp_ping - handle ping response */
void
handleresp_ping(struct rep *rep, struct sppconn *sc)
{
	const char *f = "handleresp_ping";
	struct connection *c;

	c = sc->conn;
	if (rep->code != 200) { /* XXX should remove peer */
		dbg("%s: ping unsuccessful, reply code: %d", f, rep->code);
		return;
	}

	if (!rep->have_nounce || rep->nounce != sc->nounce) {
		dbg("%s: correct nounce missing in ping, ignoring", f);
		return;
	}

	/* ignore meta info if lacking meta data XXX ping called to early? */
	if (sc->ue->state >= URIENT_STATE_META) {
		/*
		 * the total number of blocks at peer is returned in bcnt,
		 * filemeta_peerupd() only wants additional blocks, so
		 * reduce by one of set.
		 */
		if (rep->bcnt > 0)
			rep->bcnt--;
		filemeta_peerupd(sc->ue->meta, c->peer, 0, rep->bcnt, BLOCKINFO_PING);
	}

/*	cache_connection(c);*/
}


/* newpeer_ping - locate peer suitable for ping request */
void
newpeer_ping(void *data)
{
	const char *f = "newpeer_ping";
	struct sppconn *sc;
	struct peer    *peer;

	peer = data;

	dbg("%s: proactive rtt estimation to %s", f, peer->id);

	sc = sppconn_new(CLIENT_CMD_PING);
	sppconn_setfile(sc, PTOSP(peer)->ue_newpeer);
	spp_peerreq(sc, peer);
}


/* handle_ping - process ping request */
void
handle_ping(struct sppconn *sc, struct req *req)
{
	const char *f = "handle_ping";
	struct connection *c;
	unsigned int cnt;
	struct rep rep;

	c = sc->conn;
	ASSERT(sc->ue != NULL, "NULL ue");

	dbg("%s: pinged by %s", f, c->peer->id);

	if (!req->have_nounce)
		dbg("%s: ping without nounce from %s", f, c->peer->id);

	/* how many blocks do we have, starting with block 0? */
	/* XXX scc, server like value reduction? */
	cnt = urient_subseq(sc->ue, 0);
	cnt++; /* block zero not counted by urient_subseq */

	bzero(&rep, sizeof(rep));
	rep.code = 200;
	rep.nounce = req->nounce;
	rep.bcnt = cnt;
	spp_cmd_rep2(sc, &rep);
}
