/* spp_revconn.c - spp revconn command handling */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: spp_revconn.c,v 1.45 2010/07/31 09:48:09 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "p2pd.h"
#include "conn.h"
#include "filemeta.h"
#include "peer.h"
#include "peerlist.h"
#include "peerset.h"
#include "request.h"
#include "socket.h"
#include "spp.h"
#include "urient.h"

/*
 * XXX
 *  -run per-file revconn timer while file active (rationale,
 *   desirable to have priority for files that are downloaded by
 *   scc/lhc client)
 *
 *  -run global revconn timer that triggers if no requests have been
 *   received at all within given time.  (rationale, lhc/scc behind
 *   firewall, only able to initiate connections. if no active
 *   clients, no file will be active)
 *
 * --
 *  -two possible approaches:
 *   -handle each file independently
 *   -choose file where revconn most likely to succeed
 *
 *  -possible problem with first approach:
 *   -big cache, lots of revconn activity (many files) shortly after startup
 *
 * -likely best approach (combined):
 *  -do revconn for each active file
 *  -have global revconn timer, choose likely best file if timeout triggers
 *
 */


/* spp_cmd_revconn - send REVCONN request to peer */
void
spp_cmd_revconn(struct sppconn *sc, struct peer *peer)
{
	const char *f = "spp_cmd_revconn";
	struct connection *c;
	char str[1024];
	char *enc;
	int bcnt;

	c = sc->conn;

	ASSERT(c->obuflen == 0, "buffered data");
	ASSERT(c->ibuflen == 0, "buffered data");

	strlcpy(c->obuf, COMMAND_REVCONN, sizeof(c->obuf));
	if ((enc = uri_encode(sc->ue->uri.path)) == NULL)
		return;
	strlcat(c->obuf, enc, sizeof(c->obuf));
	strlcat(c->obuf, " ", sizeof(c->obuf));
	strlcat(c->obuf, VALUE_VERSION_SPP, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	strlcat(c->obuf, OPTION_HOST, sizeof(c->obuf));
	if ((enc = uri_encode(sc->ue->uri.host)) == NULL)
		return;
	strlcat(c->obuf, enc, sizeof(c->obuf));
	strlcat(c->obuf, ":", sizeof(c->obuf));
	strlcat(c->obuf, sc->ue->uri.port, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	strlcat(c->obuf, OPTION_BINDPORT, sizeof(c->obuf));
	strlcat(c->obuf, lport, sizeof(c->obuf));
	strlcat(c->obuf, "\r\n", sizeof(c->obuf));

	bcnt = urient_subseq(sc->ue, 0);
	if (bcnt != 0) {
		strlcat(c->obuf, OPTION_SUBSEQ, sizeof(c->obuf));
		snprintf(str, sizeof(str), "%d", bcnt);
		strlcat(c->obuf, str, sizeof(c->obuf));
		strlcat(c->obuf, "\r\n", sizeof(c->obuf));
	}

	if (strlcat(c->obuf, "\r\n", sizeof(c->obuf)) >= sizeof(c->obuf))
		ERRX("request too long");

	c->obuflen = strlen(c->obuf);

	dbg("%s: to %s", f, sc->conn->peer->id);

	peerpar_settime(peer, PEERPAR_SPP_LASTREVCONN);
	UTOSPP(sc->ue)->lastrevconn = time(NULL);
}


/* handleresp_revconn - process response to revconn command */
void
handleresp_revconn(struct rep *rep, struct sppconn *sc)
{
	const char *f = "handleresp_revconn";

	if (rep->code != 200) { /* XXX not interested, remove from peerlist? */
		dbg("%s: revconn failed, peer not interested, failed %d", f,
		    rep->code);
		conn_changestate(sc->conn, CONNECTION_STATUS_ENDED);
	} else {
		conn_revconn(sc->conn); /* reverse connection */
		/* revconn -> idle, not cache */
		idle_connection(sc->conn);
	}
}


/* revconn_file - attempt to get request from peers (e.g. if firewalled) */
void
revconn_file(void *dat)
{
	const char *f = "revconn_file";
	struct peerset *pset;
	struct sppconn *sc;
	struct urient *ue;
	struct peer *peer;
	int32_t maxblock;
	time_t rint;
	int conncnt;
	time_t now;
	int ival;
	int i;

	ue = dat;
	ASSERT(ue->cf != NULL, "only revconn for files in cache");

	dbg("%s", f);

/*	revconn_timerupd(ue);*/

	now = time(NULL);

#if 0
	/* XXX we assume that everybody start with block 0 */
	if (spp_cmd_peers(file, 0, blockend(0)) == NULL)
		dbg("%s: unable to retrive peerlist for first block for"
		    "revconn candidates", f);
#endif

	/* how many blocks do we have from start */
	/*XXX int value, overflow */
	maxblock = urient_blockcount(ue, 0);
	if (maxblock == 0) {
		/* ASSERT(maxblock > 0,
			  "revconn on file with no blocks cached");*/
		swarn("%s: revconn, but missing at least first block", f);
		return;
	}
	/*
	 * XXX reduce slightly to give some slack for slow updates?
	 *     (to avoid peers that already have the same block that the
	 *      node doing spp_revconn does)
	 */
	if (maxblock < FILE_PROACTIVEREQ_MINBEHIND) {
		dbg("%s: too few blocks, aborting revconn", f);
		return;
	}
	maxblock -= FILE_PROACTIVEREQ_MINBEHIND;

	/* linear playback; obtain list for first block */
	if ((pset = peerlist_gen(ue, 0)) == NULL) {
		dbg("%s: got no peerlist, aborting revconn", f);
		return;
	}

	/* exclude nodes not suitable for revconn, then sort by rate */
	for (i = pset->peercnt; i > 0; i--) {
		peer = pset_peer(pset, i - 1);
		if (filemeta_peerseq(ue->meta, 0, peer) >= maxblock)
			pset_delpeer(pset, peer);
	}
	if (pset->peercnt == 0) {
		dbg("%s: nothing to offer, aborting revconn", f);
		pset_free(pset);
		return;
	}

	/*
	 *XXX should have 'not interested in revconn message'
	 *    1: for use by server
	 *    2: for use by clients that no longer download file
	 *
	 *    having a 'file does not exist' message would solve
	 *    the second scenario (by deleting filemeta data for that peer),
	 *    but would not work if the server issued (would interfer with
	 *    downloading from server).
	 */

	/* trim peers to which we have a connection */
	ival = 0;
	pset_evalsofttrim(pset, PEERPAR_PEER_TOTCONN, PEERPAR_CMPOP_GT, &ival);
	/* trim peers to which an error has occured */
	ival = PEER_PREVOK;
	pset_evalsofttrim(pset, PEERPAR_SPP_LASTRES, PEERPAR_CMPOP_NE, &ival);
	/* trim peers which have been tried recently */
	rint = now - PEER_REVCONN_INTERVAL;
	pset_evalsofttrim(pset, PEERPAR_SPP_LASTREVCONN, PEERPAR_CMPOP_GT,
			  &rint);
	if (pset->peercnt == 0) {
		dbg("%s: no suitable peers, aborting revconn", f);
		pset_free(pset);
		return;
	}

	/* sort based on score and estimated rate */
	peerlist_ratesort(pset);

	conncnt = 0;
	for (i = 0; i < pset->peercnt; i++) {
		peer = pset_peer(pset, i);

		/*
		 * XXXskip server? should have message which causes it
		 * to never be retried more than once
		 */

		dbg("%s: found candidate revconn: %s", f, peer->id);
		sc = sppconn_new(CLIENT_CMD_REVCONN);
		sppconn_setfile(sc, ue);
		spp_peerreq(sc, peer);

		if (conncnt++ > 2)
			break;
	}
	pset_free(pset);
}


/* handle_revconn - process reverse connection request (offered by peer) */
void
handle_revconn(struct sppconn *sc, struct req *req)
{
	const char *f = "handle_revconn";
	struct connection *c;

	c = sc->conn;

	dbg("%s: got revconn from %s", f, c->peer->id);

	filemeta_peerupd(sc->ue->meta, c->peer, 0, req->bcnt,
			 BLOCKINFO_REVCONN);
	spp_cmd_rep(sc, 200);

	peerpar_setint(c->peer, PEERPAR_SPP_LASTRES, PEER_PREVOK);
}
