/* sppfile.c - state management for spp files */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: sppfile.c,v 1.10 2010/10/11 17:44:38 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>

#include "p2pd.h"
#include "file.h"
#include "filemeta.h"
#include "proc.h"
#include "socket.h"
#include "spp.h"
#include "timers.h"
#include "urient.h"


/* states:
 *
 * -attempting to get meta data
 * -error
 * -downloading data (incomplete)
 * -complete
 *
 * -doing revconn|not doing revonn (if have blocks)
 * -want credupdate (occurs during handle_get)
 * -want to update igot (after cacheinit|dbrefresh|download)
 *
 * -creation:
 * -no meta data
 * -error
 * -incomplete
 * -complete
 *
 * -no meta data|error
 * -no revconn
 *
 * -incomplete|complete
 * -(maybe) do revconn
 *
 * ----
 * per-file:
 * -cred
 * -igot?
 *
 * igot timer per server (send all igots (might be for muliple files)
 * when igot_flush() called (e.g., in dbrefresh_all(), after all entries
 * added), or in timeout
 *
 * global spp actions:
 * -revconn (on limited number of files)?
 *  try to determine file|peer most likely to be useful
 *  of all files, which file has peer most likely to be interested
 *  (would be most recently arrived peer, use file requested by that peer)
 */

extern int totheadactive;


/* XXX rename functions to be on format spp_foo */
struct ueops ueops_spp = {
	sppfile_free,
	(void (*)(void *))sppfile_cacheinit,
};


/*XXX err state handled without sppfile? */

/* sppfile_new - initialize new sppfile */
struct sppfile *
sppfile_new(struct urient *ue)
{
	const char *f = "sppfile_new";
	struct sppfile *ss;

	ss = xmalloc(sizeof(struct sppfile));
	bzero(ss, sizeof(*ss));

	SPPTOU(ss) = ue;
	ss->waiting = aqueue_init(NULL);

	ss->to = timeout_add((void (*)(void *))sppfile_timeout, NULL, ss);

	ss->parent = spp_parent(&ue->uri);
	ASSERT(ss->parent != NULL, "NULL parent node");
	dbg("%s: parent: %s", f, ss->parent->id);

	/* state changed to active later if referenced */
	ss->state = SPP_STATE_UNUSED;
	sppfile_timersync(ss);

	ue->ueops = &ueops_spp;

	return ss;
}


/* sppfile_free - release sppfile resources */
void
sppfile_free(void *dat)
{
	struct sppfile *ss;

	ss = dat;

	aqueue_free(ss->waiting);
	timeout_del(ss->to);
	if (ss->to_revconn != NULL)
		timeout_del(ss->to_revconn);

	if (ss->to_head != NULL)
		timeout_del(ss->to_head);

	free(dat);
}


/* sppfile_addref - increase sppfile reference count */
void
sppfile_addref(struct sppfile *ss)
{
	const char *f = "sppfile_addref";

	ASSERT(ss->state != SPP_STATE_ERR, "err addref");

	ss->refcnt++; /*XXX overflow */

	dbg("%s: -> %d", f, ss->refcnt);

	if (ss->state == SPP_STATE_UNUSED) {
		ss->state = SPP_STATE_ACTIVE;
		sppfile_timersync(ss);
		/*
		 * reenable revconn if at least one igot has been sent.
		 * reset timer to give system some time before igot initiated.
		 */
		if (ss->sentigot) {
			dbg("reenabling revconn timeout");
			timeout_secupd(ss->to_revconn, FILE_PROACTIVEREQ_DELAY);
		}
	}
}


/* sppfile_unref - decrease sppfile reference count */
void
sppfile_unref(struct sppfile *ss)
{
	const char *f = "sppfile_unref";

	ASSERT(ss->refcnt > 0, "invalid refcnt");
	ASSERT(ss->state != SPP_STATE_UNUSED, "unused sppfile with ref");
	ss->refcnt--;

	dbg("%s: -> %d", f, ss->refcnt);

	if (ss->refcnt == 0) {
		switch (ss->state) {
		case SPP_STATE_ACTIVE:
			ss->state = SPP_STATE_UNUSED;
			sppfile_timersync(ss);
			break;
		case SPP_STATE_ERR:
			/* might wish to retry at once, delete if unused */
			urient_del(SPPTOU(ss));
			break;
		case SPP_STATE_UNUSED:
			ERRX("UNUSED sppfile had references");
		}
	}
}


/* sppfile_clientreq - client request for uri, return corresponding sppfile */
struct sppfile *
sppfile_clientreq(struct uri *uri)
{
	struct urient *ue;

	if ((ue = urient_find(uri)) == NULL)
		return NULL;

	/* new request, update urient lastreq and any timers */
	urient_gotreq(ue);
	sppfile_timersync(UTOSPP(ue));

	return UTOSPP(ue);
}


/* sppfile_servinit - add urient for file, if it exists */
struct urient *
sppfile_servinit(struct uri *uri)
{
	struct urient *ue;

	if ((ue = urient_servinit(uri)) == NULL)
		return NULL;

/*	UTOSPP(ue) = sppfile_new(ue);*/ /*XXX gcc 4.4.1 */
	ue->uldat = sppfile_new(ue);

	return ue;
}


/* sppfile_headinit1 - start head for new file */
struct urient *
sppfile_headinit1(struct uri *uri)
{
	const char *f = "sppfile_headinit";
	struct urient *ue;

	dbg("%s", f);

	/* setup urient for new request */
	if ((ue = urient_add(uri)) == NULL) {
		dbg("%s: unable to add uri, aborting headinit", f);
		return NULL;
	}
/*	ue->state = URIENT_STATE_URI;*/

#if 0
	/*
	 * want to limit effect of dos attempts. dns resolution is a
	 * blocking operation (when threads are not used).  the number
	 * of simultaneous resolution attempts is limited by the number
	 * of processes that are available for doing this.  submitting
	 * a large number of requests (to an scc) with uris containing
	 * servers that take a (sufficiently) long time to resolve
	 * should not result in an increasing amount of resource usage
	 * by the server, which could happen if requests are allowed
	 * to pile up.
	 *
	 * with a fixed number of resolution processes, and a fixed
	 * size queue for sppconns that want to resolve a peer, the
	 * queue will fill up if many requests that need dns
	 * resolution are received. rather than increasing the size of
	 * the queue, we at this point start refusing requests with
	 * requests that need dns resolution (with a temporary error).
	 *
	 * this will limit the scope of any such dos attack to simply
	 * leading to many connections requests being accepted (and
	 * relatively quickly refused).  other clients are most likely
	 * to experience this by difficulty in connecting (listen
	 * queue filling up), but apart from this, no lasting effect
	 * should result and clients that get through will still be
	 * able to make requests for files where the server is already
	 * resolved.
	 *
	 * an lhc/scc with a scc parent node will not be affected by
	 * this due to the server not being contacted directly.
	 */
	/*
	 * XXX use grow alist for queue internally to avoid the need
	 * to handle the queue being full for internal parts of code
	 * (e.g., if dns resolution value expires and needs to be
	 * resolved again during spp_cred or similar connect operation.
	 * simply return error to client if queued entries is too high.
	 */
	if (!ip_isnumeric(parent->name) && resolve_isqueued() < 4)
		return temporary_error;
	if (peer_resolve(spp_parent(ue)) == 0)
		ue->state = URIENT_STATE_RESOLVED;
#endif

/*	UTOSPP(ue) = sppfile_new(ue);*/ /*XXX gcc 4.4.1 */
	ue->uldat = sppfile_new(ue);
	if (!peer_validaddr(UTOSPP(ue)->parent) &&
	    proc_queuesize(PROC_RESOLVE) > SPP_HEAD_MAXRESQUEUE) {
		dbg("%s: refusing request, too many resolve queue entries", f);
		sppfile_seterr(UTOSPP(ue), 503); /* temporary err */
		return ue;
	}

	if (totheadactive > SPP_HEAD_OUTGOING_MAX) {
		dbg("%s: too many ongoing head requests, refusing", f);
		sppfile_seterr(UTOSPP(ue), 503); /* temporary err */
		return ue;
	}

	sppfile_headinit(UTOSPP(ue));

	return ue;
}


/*XXX take filename of cache file as argument? */
/* sppfile_cacheinit - add urient for file cached on disk */
void
sppfile_cacheinit(struct sppfile *ss)
{
	const char *f = "sppfile_cacheinit";

	dbg("%s: %s", f, SPPTOU(ss)->uri.uri);

/*	ue->state = URIENT_STATE_UNUSED;*/
/*	UTOSPP(ue)->state = SPP_STATE_UNUSED;*/
	/*
	 * XXX when to disable to_revconn timeout?
	 * only relevant for cached files, and should run for as long
	 * as file is cached? (but might wish to limit number of files
	 * for which this is done if it is large)
	 */
	ss->to_revconn = timeout_add(revconn_file, NULL, SPPTOU(ss));
}


/* sppfile_seterr - cache an spp error value */
void
sppfile_seterr(struct sppfile *ss, int err)
{
	const char *f = "sppfile_seterr";

	slog(LOG_DEBUG, "%s: %s (%d)", f, SPPTOU(ss)->uri.uri, err);

	ss->spp_err = err;
	ss->state = SPP_STATE_ERR;
	sppfile_timersync(ss);
}


/* sppfile_timersync - refresh state dependent timers */
void
sppfile_timersync(struct sppfile *ss)
{
	int timeout;


	/* alt 1: check if underlying ue has ERR state, then increase
	 *  timeout value (i.e., no ERR state for ss)
	 *
	 *  alt 2: (probably better) urient state only reflects
	 *  information, and no urient ERR state, instead set if to
	 *  URI, which correctly reflects the info that it stores (but
	 *  an uri for which there is no data for some reason (e.g.,
	 *  server does not exist)). the err state is in the sppfile
	 *  object, which might also have a text version of the error
	 *  message.  timeout set based on sppfile state only,
	 *  including err state.
	 */

	timeout = 0;
	switch (ss->state) {
	case SPP_STATE_ACTIVE:
		timeout = 0; /* no timeout */
		break;
	case SPP_STATE_ERR:
		timeout = 0; /* no timeout, placed on expiry queue */
		break;
	case SPP_STATE_UNUSED:
		timeout = SPP_TIMEOUT_UNUSED;
		break;
	}
	if (timeout > 0)
		timeout_secupd(ss->to, timeout);
	else
		timeout_off(ss->to);
}


/* sppfile_timeout - handle spp timeout */
void
sppfile_timeout(struct sppfile *ss)
{
	const char *f = "sppfile_timeout";

	switch (ss->state) {
	case SPP_STATE_ACTIVE: /* FALLTHROUGH */
	case SPP_STATE_ERR:
		ERRX("unexpected timeout for sppfile state");

	case SPP_STATE_UNUSED:
		ASSERT(ss->refcnt == 0, "nonzero refcnt on unused sppfile");
		if (SPPTOU(ss)->file != NULL) {
			file_close(SPPTOU(ss));

			/* restart timeout, additional cleanup next time */
			sppfile_timersync(ss);
		} else {
			if (SPPTOU(ss)->master) {
				/*
				 * remove sppfile/urient on master
				 */
				ASSERT(SPPTOU(ss)->master,
				       "uri del on non-master");
				urient_del(SPPTOU(ss));
			} else {
				/*
				 * disable revconn on cache
				 * (but keep file in memory)
				 */
				/* timeout only set after get request */
				if (ss->to_revconn != NULL) {
					dbg("%s: disabling revconn timeout", f);
					timeout_off(ss->to_revconn);
				}
			}
		}
		break;
	}
}


/* sppfile_connfail - failed request on connection, do cleanup */
void
sppfile_connfail(struct sppfile *ss, struct sppconn *sc)
{
	const char *f = "sppfile_connfail";

	ASSERT(sc->currcmd != 0, "no command set");

	dbg("%s", f);

	/*XXX should have generic retry state (try 3 times then fail) */

	switch (sc->currcmd) {
	case CLIENT_CMD_HEAD:
		if (sc->conn->type == CONN_TYPE_OUTGOING ||
		    sc->conn->type == CONN_TYPE_RESOLVE ||
		    sc->conn->type == CONN_TYPE_MCONN) {
			ASSERT(ss->headactive, "not doing head");
			sppfile_headfail(ss, NULL, 504); /* gateway timeout?XXX */
		} else if (sc->conn->type == CONN_TYPE_INCOMING) {
			ASSERT(ss->headactive, "not doing head");
			dbg("%s: deleting conn from head wait queue", f);
			aqueue_delval(ss->waiting, sc);
		} else
			ERRX("invalid sppconn type")
		break;
	case CLIENT_CMD_STARTTLS:
		if (sc->conn->type == CONN_TYPE_OUTGOING) {
			PTOSP(sc->sslinitpeer)->sslinitactive = 0;
			dbg("%s: sslinit failure for %s", f,
			    sc->sslinitpeer->id);
			sc->sslinitpeer = NULL;
			/*XXX retry? */
		}

	default:
		dbg("%s: unhandled command: %s", f,
		    spp_cmd2str(sc->currcmd));
		break;
	}
}
