/* sppfile_head.c - head command related state handling for spp files */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: sppfile_head.c,v 1.8 2010/06/26 01:09:02 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include "p2pd.h"
#include "peer.h"
#include "peerset.h"
#include "spp.h"
#include "request.h"
#include "socket.h"
#include "timers.h"
#include "urient.h"

/*
 * counter for number of active outgoing head requests.
 * should be limited to prevent overloading.
 */
int totheadactive = 0;


/* sppfile_headinit - start head for sppfile */
void
sppfile_headinit(struct sppfile *ss)
{
	const char *f = "sppfile_headinit";

	ASSERT(SPPTOU(ss)->state == URIENT_STATE_URI, "invalid urient state");

	dbg("%s: %s", f, SPPTOU(ss)->uri.uri);

	totheadactive++;

	/* add and start timer */
	ss->headactive = 1;
	ss->to_head = timeout_add((void (*)(void *))sppfile_tryhead, NULL, ss);

	sppfile_addref(ss);
	sppfile_tryhead(ss);
}


/* sppfile_waithead - add connection waiting for head result */
void
sppfile_waithead(struct sppfile *ss, struct sppconn *sc)
{
	const char *f = "sppfile_waithead";

	ASSERT(ss->headactive, "not headactive");

	dbg("%s: adding connection to head wait queue", f);

	ASSERT(sc->conn->type == CONN_TYPE_INCOMING, "not incoming");

/*	sc->uestate = UESTATE_CONN_HEADWAIT;*/
	aqueue_enqueue(ss->waiting, sc);
	sc->headwait = 1;
}


/* sppfile_headfail - handle failure of head command */
void
sppfile_headfail(struct sppfile *ss, struct sppconn *sc, int code)
{
	const char *f = "sppfile_headfail";

	ASSERT(code != 200, "return code is not failure");
/*	ASSERT(sc->conn->type == CONN_TYPE_OUTGOING, "invalid sppconn");*/

	dbg("%s: %s %p", f, SPPTOU(ss)->uri.uri, sc);

	/* retry in case of timeout/non-fatal error */
	if ((code != 403) && (code != 404) &&
	    (ss->head_attempt < RETRIES_MAX_HEAD)) {
		dbg("%s: head failed %d, retrying later", f, errno_spp);
		/* setup new delayed head request */
		/*XXX headactive not set if eg. resolve headinit1 queue full */
		ASSERT(ss->headactive, "not headactive");
		timeout_secupd(ss->to_head, RETRY_DELAY_HEAD);
		ss->head_attempt++;

		if (sc != NULL)
			cache_connection(sc->conn);
	} else {
/*			ERRX("headfail");*/
		sppfile_seterr(ss, code);
		if (sc != NULL) {
#if 0
			sppfile_unref(ss);
			sc->ue = NULL;
#endif
			changecstatus(sc->conn, CONNECTION_STATUS_ENDED);
		}
		handle_headchain(ss);

		ss->headactive = 0;

		ASSERT(totheadactive >= 1, "negative headactive");
		totheadactive--;

		timeout_del(ss->to_head); /* no longer needed */
		ss->to_head = NULL;
	}
}


/* sppfile_headok - head returned ok, return reply to waiting connections */
void
sppfile_headok(struct sppfile *ss, struct sppconn *sc)
{
	const char *f = "sppfile_headok";

	/* should not occur on master */
	ASSERT(SPPTOU(ss)->master == 0, "sppfile_headok on master");

	dbg("%s: %s", f, SPPTOU(ss)->uri.uri);

	/* head completed, disable timer etc. */
	ss->headactive = 0;

	ASSERT(totheadactive >= 1, "negative headactive");
	totheadactive--;

	timeout_del(ss->to_head); /* no longer needed */
	ss->to_head = NULL;

	handle_headchain(ss);

	sppfile_unref(UTOSPP(sc->ue));
	sc->ue = NULL;

	cache_connection(sc->conn);
}


/* handle_headchain - process waiting head requests */
void
handle_headchain(struct sppfile *ss)
{
	const char *f = "sppfile_headchain";
	struct sppconn *nc;

	dbg("%s: processing delayed HEAD requests", f);
	while ((nc = aqueue_dequeue(ss->waiting)) != NULL) {
/*		ASSERT(nc->uestate == UESTATE_CONN_HEADWAIT, "invalid state");*/
#if 0 /* for scc, can be any command */
		/* head would typically be initiated as a result of head/get */
		ASSERT(nc->currcmd == CLIENT_CMD_HEAD ||
		       nc->currcmd == CLIENT_CMD_GET, "invalid command");
#endif
		ASSERT(nc->conn->type == CONN_TYPE_INCOMING, "not incoming");
		ASSERT(nc->req != NULL, "null req");
		ASSERT(nc->headwait, "headwait not set");

/*		nc->ue = NULL;*/
		if (ss->state == SPP_STATE_ERR) {
			dbg("%s: sending error notification", f);
			spp_cmd_rep(nc, ss->spp_err);
		} else {
			dbg("%s: processing cachereq", f);
/*			nc->ue = SPPTOU(ss);*/
			if (handle_cachereq(SPPTOU(ss), nc, nc->req) == -1)
				dbg("cachereq failed for delayed head client");
		}
		free(nc->req);
		nc->req = NULL;
		nc->headwait = 0;
	}

/*	ss->head_attempt = 0;*/ /* keep this value */
}


/* sppfile_tryhead - try to initiate spp head (for use with timeout) */
void
sppfile_tryhead(struct sppfile *ss)
{
	struct sppconn *sc;

	ASSERT(ss->headactive, "head timeout but not headactive");

	sc = sppconn_new(CLIENT_CMD_HEAD);
	sppconn_setfile(sc, SPPTOU(ss));
	spp_parentreq(sc);
}
