/* sppfile_revconn.c - revconn command related state handling for spp files */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: sppfile_revconn.c,v 1.7 2010/06/29 15:14:38 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <time.h>

#include "p2pd.h"
#include "alist.h"
#include "spp.h"
#include "request.h"
#include "socket.h"
#include "timers.h"
#include "urient.h"

/* start revconn for any file if no requests received within this time */
#define SPP_NOREQS_TIMEOUT 10

#define SPP_REVCONN_INTERVAL 10 /* time between revconn attempts */

/* have timer that will trigger if no requests received */
struct timeout *to_revconnany = NULL;


/* sppfile_revconninit - setup global revconn timer */
void
sppfile_revconninit(void)
{
	to_revconnany = timeout_add(revconn_anyfile, NULL, NULL);
}


/* revconn_timerupd - enable/postpone initiation of reverse connections */
void
revconn_timerupd(void)
{
	const char *f = "sppfile_timerupd";

	if (to_revconnany == NULL) /* not used on server */
		return;

	dbg("%s: postponing revconn", f);
	timeout_secupd(to_revconnany, SPP_NOREQS_TIMEOUT);
}


/* revconn_anyfile - try to find a file to do revconn on */
void
revconn_anyfile(void *data)
{
	const char *f = "sppfile_anyfile";
	struct urient *recent_ue;
	struct urient *ue;
	int i;

	(void)data;

	dbg("%s", f);

	/*
	 * current approach uses lastreq time to sort urients.
	 * other parameters that might be considered include
	 * the number of blocks stored locally.XXX
	 */
	recent_ue = NULL;
	for (i = 0; (size_t)i < ALIST_SIZE(al_urient); i++) {
		ue = ALIST_PVAL(al_urient, i);
		if (ue->cf == NULL ||
		    !UTOSPP(ue)->sentigot ||
		    UTOSPP(ue)->lastrevconn + SPP_REVCONN_INTERVAL > time(NULL))
			continue;

		if (recent_ue == NULL) {
			recent_ue = ue;
			continue;
		}
		if (ue->lastreq > recent_ue->lastreq)
			recent_ue = ue;
	}

	if (recent_ue == NULL)
		dbg("%s: no files usable for revconn", f);
	else
		/* attempt revconn for most recent file */
		revconn_file(recent_ue);

	/* reset timer */
	timeout_secupd(to_revconnany, SPP_NOREQS_TIMEOUT);
}
