/* subnet.c - subnet parsing and classification functions */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: subnet.c,v 1.8 2010/06/30 08:38:00 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string.h>

#include "p2pd.h"
#include "subnet.h"


/* parse 10.0.0.0/8 type address specifications (XXX only ipv4 for now) */
int
subnet_parse(struct subnet *subnet, char *str)
{
	const char *f = "subnet_parse";
	const char *errstr;
	char netstr[strlen("255.255.255.255/32") + 1];
	uint32_t mask;
	char *p;
	int n;

	if (strlcpy(netstr, str, sizeof(netstr)) >= sizeof(netstr)) {
		swarn("%s: subnet specification too long", f);
		return -1;
	}

	/* XXX allow comma separated list of values? */
	if ((p = strchr(netstr, '/')) == NULL) {
		swarn("%s: invalid address specification", f);
		return -1;
	}
	*p = '\0';
	p++; /* width */
	subnet->width = strtonum(p, 0, 32, &errstr);
	if (errstr) {
		swarn("%s: invalid address specification", f);
		return -1;
	}

	if (inet_pton(AF_INET, netstr, &subnet->v4subnet) == 0) {
		swarn("%s: invalid ipv4 address", f);
		return -1;
	}

	/* convert to host format; address is not being transmitted */
	subnet->v4subnet.s_addr = ntohl(subnet->v4subnet.s_addr);

	/* verify that all subnet bits zero */
	for (n = subnet->width; n < 32; n++) {
		mask = (uint32_t)0x80000000 >> n;
		if (mask & subnet->v4subnet.s_addr) {
			swarn("%s: invalid subnet specification", f);
			return -1;
		}
	}

	subnet->active = 1;

	return 0;
}

/* returns true if address is part of scc client subnet */
int
peerinsubnet(struct subnet *subnet, struct sockaddr *sockaddr)
{
	const char *f = "peerinsubnet";
	struct in_addr peeraddr;
	uint32_t mask;
	int n;

	if (subnet->active == 0)
		return 0;

	/* XXX ipv6 */
	if (sockaddr->sa_family != AF_INET) {
		swarn("%s: unsupported address format", f);
		return 0;
	}

	peeraddr.s_addr =
		ntohl(((struct sockaddr_in *)sockaddr)->sin_addr.s_addr);

	/* verify that all network bits match */
	for (n = 0; n < subnet->width; n++) {
		mask = (uint32_t)0x80000000 >> n;
		if ((mask & subnet->v4subnet.s_addr) !=
		    (mask & peeraddr.s_addr)) {
			swarn("%s: address not in subnet", f);
			return 0;
		}
	}

	return 1;
}
