/* alist.c - array based list management functions */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: alist.c,v 1.23 2010/06/25 22:02:25 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <errno.h>
#include <search.h>
#include <string.h>
#include <stddef.h>
#include <time.h>

#include "log.h"
#include "alist.h"

#include "alist_int.h"

/*XXX have resize callback function called before resize operation, to
 *     be used for removing entries that are no longer needed,
 *     possibly leading to the resize operation not being needed?
 */

/* XXX have alist_delidxrange function (deletes range in one op) */

/* global used to by alist_datacmp for value comparison */
int (*datacmpfunc)(const void *, const void *) = NULL;


/* alist_pcmp - sorting function for pointer key values */
int
alist_pcmp(const void *a, const void *b)
{
	struct alent const *pa, *pb;

	pa = a;
	pb = b;
	if (pa->key < pb->key)
		return -1;
	else if (pa->key > pb->key)
		return 1;
	else
		return 0;
}


/* alist_strcmp - sorting function for string key values */
int
alist_strcmp(const void *a, const void *b)
{
	struct alent const *pa, *pb;

	pa = a;
	pb = b;
	return strcmp(pa->key, pb->key);
}


/* alist_datacmp - sorting function for data */
int
alist_datacmp(const void *a, const void *b)
{
	struct alent const *pa, *pb;

	ASSERT(datacmpfunc != NULL, "NULL comparison function");

	pa = a;
	pb = b;
	return datacmpfunc(pa->data, pb->data);
}


/* alist_init - initialize and return pointer to array list structure */
struct alist *
alist_init(void (*printfunc)(const void *), void (*freefunc)(void *),
	   size_t size, enum alsizing sizing, enum altype type)
{
	struct alist *al;

	/* function should never return NULL (callers currently expect this) */

	al = xmalloc(sizeof(struct alist));
	bzero(al, sizeof(struct alist));
	al->print = printfunc;
	al->free = freefunc;
	al->type = type;

	al->arr = xcalloc(size, sizeof(struct alist));
	al->size = 0;
	al->arrsize = size;
	al->sizing = sizing;

	switch (al->type) {
	case ALIST_TYPE_PTR:
		al->sort = alist_pcmp;
		break;
	case ALIST_TYPE_STR:
		al->sort = alist_strcmp;
		break;
	case ALIST_TYPE_UNSORTED_STR:
		al->sort = NULL;
		break;
	case ALIST_TYPE_UNSORTED_PTR:
		al->sort = NULL;
		break;
	default:
		ERRX("invalid alist type");
	}

	return al;
}


/* alist_add - add new entry with pointer key, if not already existing */
void *
alist_add(struct alist *al, void *key, void *data)
{
	const char *f = "alist_add";
	struct alent *newp;
	size_t newsize;
	time_t lrumin;
	size_t len;
	int lruoff;
	int i;

#if 0
	dbg("%s", f);
	al->print(data);
#endif

	if (alist_find(al, key) != NULL) {
		dbg("%s: entry already exists", f);
		return data;
	}

	/* array full? */
	if (al->size == al->arrsize) {
		switch (al->sizing) {
		case ALIST_SIZE_FIXED:
			swarn("%s: array full", f);
			return NULL;
		case ALIST_SIZE_GROW:
			newsize = al->arrsize > 0 ? al->arrsize *= 2 : 1;
/*			dbg("%s: new size: %zd", f, newsize);*/
			newp = xrealloc(al->arr, newsize * sizeof(struct alist));
			al->arr = newp;
			al->arrsize = newsize;
			break;
		case ALIST_SIZE_LRU:
			if (al->size == 0) {
				swarn("%s: zero size array", f);
				return NULL;
			}
			/* all entries should be older than this */
			lrumin = time(NULL);
			lruoff = 0;
			for (i = 0; (size_t)i < al->size; i++)
				if (al->arr[i].lru < lrumin) {
					lrumin = al->arr[i].lru;
					lruoff = i;
				}
			alist_delidx(al, lruoff);
			break;
		case ALIST_SIZE_FIFO:
			/* pointless, but allow */
			if (al->size == 0) {
				swarn("%s: zero size fifo", f);
				return NULL;
			}
			/* delete first entry */
			alist_delidx(al, 0);
			break;
		default:
			ERRX("invalid sizing type");
		}
	}

	/* add entry */
	len = sizeof(struct alent);
	bzero(&al->arr[al->size], len);
	al->arr[al->size].data = data;
	if (al->sizing == ALIST_SIZE_LRU)
		al->arr[al->size].lru = time(NULL);

	switch (al->type) {
	case ALIST_TYPE_PTR:
		al->arr[al->size].key = key;
		break;
	case ALIST_TYPE_STR:
		/* make a copy of key in string case */
		al->arr[al->size].key = xstrdup(key);
		break;
	case ALIST_TYPE_UNSORTED_STR:
		/* make a copy of key in string case */
		al->arr[al->size].key = xstrdup(key);
		break;
	case ALIST_TYPE_UNSORTED_PTR:
		al->arr[al->size].key = key;
		break;
	default:
		ERRX("invalid alist type");
	}

	al->size++;
	if (al->sort != NULL)
		qsort(al->arr, al->size, len, al->sort);
	return data;
}


/* alist_delidx - remove entry at specified array index from array */
void
alist_delidx(struct alist *al, size_t idx)
{
	size_t entlen;

	ASSERT(idx < al->size, "invalid index");

	if (al->free != NULL)
		al->free(al->arr[idx].data);
	/* a copy is made of the key in the string case, free now */
	if ((al->type == ALIST_TYPE_STR) ||
	    (al->type == ALIST_TYPE_UNSORTED_STR))
		free(al->arr[idx].key);

	/*XXX overflow */
	entlen = (al->size - (idx + 1)) * sizeof(struct alent);
	bcopy(&al->arr[idx + 1], &al->arr[idx], entlen);
	bzero(&al->arr[al->size - 1], sizeof(struct alent));
	al->size--;

	/* XXX shrink array? */
}


/* alist_del - remove entry from array */
void
alist_del(struct alist *al, void *key)
{
	ssize_t idx;

	if ((idx = alist_findidx(al, key, al->size)) == -1)
		ERRX("alist_del: entry not found");

	alist_delidx(al, idx);
}


/* alist_findidx - return entry matching key, or -1 if not found */
/*XXX size allows specification of end point in array for matching,
      more general would be to have a start/end range specification */
ssize_t
alist_findidx(struct alist *al, void *key, size_t size)
{
	struct alent *match;
	struct alent ae;
	size_t nel;
	size_t len;

	ASSERT(size <= al->size, "invalid size");

	if (al->size == 0)
		return -1;

	/* fill dummy structure for search */
	bzero(&ae, sizeof(ae));
	ae.key = key;
	len = sizeof(struct alent);
	switch (al->type) {
	case ALIST_TYPE_PTR:
		match = bsearch(&ae, al->arr, size, len, alist_pcmp);
		break;
	case ALIST_TYPE_STR:
		match = bsearch(&ae, al->arr, size, len, alist_strcmp);
		break;
	case ALIST_TYPE_UNSORTED_STR:
		nel = size;
		match = lfind(&ae, al->arr, &nel, len, alist_strcmp);
		break;
	case ALIST_TYPE_UNSORTED_PTR:
		nel = size;
		match = lfind(&ae, al->arr, &nel, len, alist_pcmp);
		break;
	default:
		ERRX("invalid alist type");
	}

	if (match == NULL)
		return -1;
	else {
		if (al->type == ALIST_TYPE_PTR) {
			ASSERT(match->key == key, "bad lookup");
		} else {
			ASSERT(strcmp(match->key, key) == 0, "bad lookup");
		}

		/* update usage time */
		if (al->sizing == ALIST_SIZE_LRU)
			match->lru = time(NULL);

		return (ptrdiff_t)(match - al->arr);
	}
}


/* alist_find - return pointer to data matching key, or NULL if not found */
void *
alist_find(struct alist *al, void *key)
{
/*	const char *f = "alist_find";*/
	int idx;

	if ((idx = alist_findidx(al, key, al->size)) == -1) {
/*		dbg("%s: entry not found", f);*/
		return NULL;
	} else {
/*		dbg("%s: found entry at idx %d", f, idx);*/
		return al->arr[idx].data;
	}
}


/* alist_rfind - return pointer to data matching key, if found in range */
void *
alist_rfind(struct alist *al, void *key, size_t end)
{
/*	const char *f = "alist_rfind";*/
	int idx;

	if ((idx = alist_findidx(al, key, end)) == -1) {
/*		dbg("%s: entry not found", f);*/
		return NULL;
	} else {
/*		dbg("%s: found entry at idx %d", f, idx);*/
		return al->arr[idx].data;
	}
}


/* alist_trim - reduce alist size, deleting entries as necessary */
/* XXX alist_resize would be more general */
void
alist_trim(struct alist *al, size_t newsize)
{
	ASSERT(newsize <= al->size, "newsize larger than size");

	while (al->size > newsize)
		alist_delidx(al, al->size - 1);
}


/* alist_shift - shift values in range in given direction */
void
alist_shift(struct alist *al, size_t start, size_t end, int dir)
{
	struct alent ae;
	size_t len;

	ASSERT(start < al->size && end < al->size, "invalid range");
	ASSERT(start <= end, "invalid range");
	ASSERT(dir == ALIST_SHIFT_LEFT || dir == ALIST_SHIFT_RIGHT, "bad dir");

	if (al->size < 2)
		return;

	len = (end - start) * sizeof(struct alent);
	if (dir == ALIST_SHIFT_LEFT) {
		ae = al->arr[start];
		bcopy(&al->arr[start + 1], &al->arr[start], len);
		al->arr[end] = ae;
	} else {
		ae = al->arr[end];
		bcopy(&al->arr[start], &al->arr[start + 1], len);
		al->arr[start] = ae;
	}
}


/* alist_mysort - sort specified subset of array with given function */
void
alist_mysort(struct alist *al, size_t end,
	     int (*cmpfunc)(const void *, const void *))
{
	size_t len;

	ASSERT(end <= al->size, "invalid end specification");

	datacmpfunc = cmpfunc;
	len = sizeof(struct alent);
	qsort(al->arr, end, len, alist_datacmp);
	datacmpfunc = NULL;
}


/* alist_free - release all alist resources */
void
alist_free(struct alist *al)
{
	int i;

	for (i = 0; (size_t)i < al->size; i++) {
		if (al->free != NULL)
			al->free(al->arr[i].data);
		if (al->type == ALIST_TYPE_STR ||
		    al->type == ALIST_TYPE_UNSORTED_STR)
			free(al->arr[i].key);
	}
	free(al->arr);
	free(al);
}


/* alist_dump - apply alist print function to all elements in alist */
void
alist_dump(struct alist *al)
{
	int i;

	ASSERT(al->print != NULL, "alist_dump called with NULL printfunc");

	for (i = 0; (size_t)i < al->size; i++)
		al->print(al->arr[i].data);
}
