/* aqueue.c - alist based queue management functions */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: aqueue.c,v 1.8 2010/06/25 22:02:25 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include "log.h"
#include "alist.h"
#include "aqueue.h"

/*
 * simple wrapper around alist for queue management
 *
 * -entries in queues are assumed to not require memory management
 *  (i.e., they are referenced from elsewhere too, but listed in
 *   queue to be ordered in some way)
 */


/* aqueue_init - return alist for queue use */
aqueue_t *
aqueue_init(void (*printfunc)(const void *))
{
	struct alist *al;

	/* start at size zero to use as little memory as possible */
	al = alist_init(printfunc, NULL, 0, ALIST_SIZE_GROW,
			ALIST_TYPE_UNSORTED_PTR);

	return (aqueue_t *)al;
}


/* aqueue_free - release queue resources */
void
aqueue_free(aqueue_t *aq)
{
	alist_free(aq);
}


/* aqueue_enqueue - place entry at the end of the queue */
void
aqueue_enqueue(aqueue_t *aq, void *val)
{
	void *key;

	/* use value as key */
	key = val;

	ASSERT(alist_find(aq, key) == NULL, "queue entry already exists");
	if (alist_add(aq, key, val) == NULL)
		ERRX("alist_add failed"); /* should not occur for GROW array */
}


/* aqueue_dequeue - remove entry from beginning of queue */
void *
aqueue_dequeue(aqueue_t *aq)
{
	void *val;

	if (ALIST_SIZE(aq) == 0)
		return NULL;

	/* beginning of queue is in array index 0, obtain and delete */
	val = ALIST_PVAL(aq, 0);
	alist_delidx(aq, 0);

	return val;
}


/* aqueue_peek - return first value in queue (without removing), if any */
void *
aqueue_peek(aqueue_t *aq)
{
	void *val;

	if (ALIST_SIZE(aq) == 0)
		return NULL;

	/* beginning of queue is in array index 0, return only */
	val = ALIST_PVAL(aq, 0);

	return val;
}


/* aqueue_delval - remove all occurences of val in queue, O(N) */
void
aqueue_delval(aqueue_t *aq, void *val)
{
	alist_del(aq, val);
}


/* aqueue_isempty - return true if queue is empty */
int
aqueue_isempty(aqueue_t *aq)
{
	if (ALIST_SIZE(aq) == 0)
		return 1;
	else
		return 0;
}

/* XXX both aqueue_size and aqueue_isempty needed? */

/* aqueue_size - return size of queue */
int
aqueue_size(aqueue_t *aq)
{
	return ALIST_SIZE(aq);
}


/* aqueue_dump - print queue entries */
void
aqueue_dump(aqueue_t *aq)
{
	alist_dump(aq);
}
