/* block.c - block structure management functions */

/*
 * Copyright (c) 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: block.c,v 1.28 2010/06/29 21:17:37 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include <time.h>

#include "log.h"
#include "block.h"
#include "file.h"
#include "filemeta.h"
#include "memcache.h"
#include "sumfile.h"
#include "urient.h"


/* block_init - block initialization */
struct block *
block_init(struct urient *ue, int32_t bnum)
{
/*	const char *f = "block_init";*/
	struct block *block;

	ASSERT(bnum < ue->meta->blockcnt, "invalid bnum");

/*	dbg("%s: block %d", bnum, f);*/

	block = xmalloc(sizeof(struct block));
	bzero(block, sizeof(struct block));

	/*
	 * in case of sendfile(), memory might not be needed, obtained
	 * via memcache_get() otherwise.
	 */
	block->blockdat = NULL;

	if (ue->meta->blockcnt - 1 == bnum)
		block->blocklen = MIN(ue->meta->size, ue->meta->size -
				      bnum * ue->meta->blocksize);
	else
		block->blocklen = ue->meta->blocksize;
	ASSERT(block->blocklen != 0, "invalid blocklen");

	return block;
}


/*XXX remove? */
/* block_access - io operation accessed block */
void
block_access(struct block *block)
{
	block->lastio = time(NULL);
}


/* block_ref - increment active block users */
void
block_ref(struct urient *ue, int32_t bnum, enum blockref_t type)
{
	struct block *block;

	/*XXX postpone calling block_init until this point? */

	block = ue->file->blocks[bnum];
	if (block->blockdat == NULL &&
	    block->readercnt == 0 && block->writercnt == 0) {
		/* data retrievable from cache? */
		if (memcache_lookup(ue, bnum) == -1)
			/* no, allocate */
			block->blockdat = xmalloc(block->blocklen);
	}

	if (type == BLOCKREF_READER)
		block->readercnt++;
	else if (type == BLOCKREF_WRITER) {
		block->writercnt++;
		ASSERT(block->writercnt <= 1, "multiple block writers");
	} else
		ERRX("unexpected blockref type");

	block_access(block);
}


/* block_unref - one less user of block */
void
block_unref(struct urient *ue, int32_t bnum, enum blockref_t type)
{
	struct block *block;

	block = ue->file->blocks[bnum];

	if (type == BLOCKREF_READER) {
		ASSERT(block->readercnt > 0, "readercnt too low");
		block->readercnt--;
	} else if (type == BLOCKREF_WRITER) {
		ASSERT(block->writercnt > 0, "writercnt too low");
		block->writercnt--;
	} else
		ERRX("unexpected blockref type");

	if (block->readercnt == 0 && block->writercnt == 0)
		/* cache data */
		memcache_store(ue, bnum);
}


/* block_free - free block resources */
void
block_free(struct block *block)
{
	ASSERT(block->readercnt == 0, "attempt to free active block");
	ASSERT(block->writercnt == 0, "attempt to free active block");

	/* blockdat should be gone or in cache */
	ASSERT(block->blockdat == NULL, "blockdat still set");

	if (block->sum != NULL)
		free(block->sum);
	free(block);
}


/* block_validsize - return true if blocksize is valid, false otherwise */
int
block_validsize(uint32_t blocksize)
{
	const char *f = "block_validsize";

	/* place some limits on block sizes */
	if (blocksize % 1024 != 0 ||
	    blocksize < BLOCKSIZE_MIN || blocksize > BLOCKSIZE_MAX) {
		swarn("%s: invalid block size %d", f, blocksize);
		return 0;
	}

	return 1;
}
