/* blockmap.c - bit manipulation functions for block map */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: blockmap.c,v 1.8 2010/06/25 22:02:26 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include "log.h"
#include "blockmap.h"


/* blockmap_init - initialize blockmap to size */
void
blockmap_init(struct blockmap *bm, int32_t size)
{
	ASSERT(size > 0, "invalid size");

	if ((bm->map = bit_alloc(size)) == NULL)
		serr("bit_alloc");
	bm->size = size;
	bit_nclear(bm->map, 0, bm->size - 1);
}


/* blockmap_set - bit_set wrapper with range test */
void
blockmap_set(struct blockmap *bm, int32_t bnum)
{
	ASSERT(bnum < bm->size, "bnum too large");

	bit_set(bm->map, bnum);
}


/* blockmap_clear - bit_clear wrapper with range test */
void
blockmap_clear(struct blockmap *bm, int32_t bnum)
{
	ASSERT(bnum < bm->size, "bnum too large");

	bit_clear(bm->map, bnum);
}


/* blockmap_nclear - bit_nclear wrapper with range test */
void
blockmap_nclear(struct blockmap *bm, int32_t start, int32_t stop)
{
	int32_t i;

	ASSERT(start < bm->size, "start too large");
	ASSERT(stop < bm->size, "stop too large");

	/* clear from start to (and including) stop */
	for (i = start; i <= stop; i++)
		blockmap_clear(bm, i);
}


/* blockmap_test - bit_test wrapper with range test */
int
blockmap_test(const struct blockmap *bm, int32_t bnum)
{
	ASSERT(bnum < bm->size, "bnum too large");

	return bit_test(bm->map, bnum);
}


/* blockmap_free - release allocated map memory in blockmap */
void
blockmap_free(struct blockmap *bm)
{
	free(bm->map);
}


/* blockmap_blockcount - return number of continous blocks from bnum */
int32_t
blockmap_blockcount(struct blockmap *bm, int32_t bnum)
{
	int32_t i;
	int cnt;

	ASSERT(bnum < bm->size, "bnum too large");

	cnt = 0;
	for (i = bnum; i < bm->size; i++) {
		if (bit_test(bm->map, i))
			cnt++;
		else
			break;
	}

	return cnt;
}


/* blockmap_subseq - return number of set blocks after bnum */
int32_t
blockmap_subseq(struct blockmap *bm, int32_t bnum)
{
	if (bnum + 1 == bm->size)
		return 0;
	else
		/* note: do not care if bnum is set */
		return blockmap_blockcount(bm, bnum + 1);
}


/* blockmap_iscomp - return true if all blocks set */
int
blockmap_iscomp(struct blockmap *bm)
{
	if (blockmap_blockcount(bm, 0) == bm->size)
		return 1;
	else
		return 0;
}
