/* blocksum.c - block digest creation functions */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: blocksum.c,v 1.20 2010/10/11 09:07:47 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>

#include "log.h"
#include "block.h"
#include "blocksum.h"
#include "file.h"
#include "filemeta.h"
#include "diskio.h"
#include "sumfile.h"
#include "urient.h"


/* block_gendigest - read full block, calculating digest if missing */
void
block_gendigest(struct urient *ue, int32_t bnum)
{
	const char *f = "block_gendigest";
	struct block *block;
	size_t rlen;

	ASSERT(ue->file != NULL, "NULL file");
	ASSERT(urient_haveblock(ue, bnum), "block not found on disk");

	block = ue->file->blocks[bnum];

/*	dbg("%s: %d", f, bnum);*/

	block_ref(ue, bnum, BLOCKREF_READER);

	rlen = block->blocklen - block->actlen;
	if (rlen > 0) {
		dbg("%s: reading block %d from disk", f, bnum);
		disk_blockread(ue, block, bnum, rlen);
	}

	/* new digest? */
	if (block->sum == NULL)
		digest_blocksum(ue, bnum);

	block_unref(ue, bnum, BLOCKREF_READER);
}


/* digest_blocksum - generate and fill sum field of complete block */
void
digest_blocksum(struct urient *ue, int32_t bnum)
{
	const char *f = "digest_blocksum";
	struct block *block;
	struct digest *md;
	char *sum;

/*	dbg("%s", f);*/

	block = ue->file->blocks[bnum];

	md = digest_init(ue->meta->digest);
	digest_update(md, block->blockdat, block->blocklen);
	sum = digest_end(md);
	digest_free(md);

	if (block->sum == NULL) {
		block->sum = sum;
		dbg("%s: storing new checksum %s, block %d", f, ue->uri.uri,
		    bnum);
		sumfile_write(ue->file->suf, bnum, sum);
	} else {
		if (strcmp(sum, block->sum) != 0) {
			swarn("%s: checksum mismatch %s, block %d: %s != %s",
			      f, ue->uri.uri, bnum, sum, block->sum);
			ERRX("checksum mismatch");
		} else
			dbg("%s: %s, block %d checksum already known", f,
			    ue->uri.uri, bnum);
		/* not needed in block, free */
		free(sum);
	}
}
