/* cachefile.c - cachefile operations */

/*
 * Copyright (c) 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: cachefile.c,v 1.11 2010/06/30 08:37:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/stat.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "log.h"
#include "blockmap.h"
#include "filemeta.h"
#include "cachefile.h"
#include "urient.h"


/* cachefile_new - return new cache file structure (file not created) */
struct cachefile *
cachefile_new(struct filemeta *fm)
{
	struct cachefile *cf;

/*	dbg("%s", f);*/

	cf = xmalloc(sizeof(struct cachefile));
	bzero(cf, sizeof(struct cachefile));

	blockmap_init(&cf->bm, fm->blockcnt);

	return cf;
}


/* cachefile_free - release cachefile resources (does not delete file) */
void
cachefile_free(struct cachefile *cf)
{
	blockmap_free(&cf->bm);
	free(cf);
}


/* cachefile_namegen - generate cachefile name based on urient (static buf) */
char *
cachefile_namegen(struct urient *ue)
{
	static char name[FILENAME_MAX];
	char tmp[FILENAME_MAX];
	size_t maxlen;
	char c;
	int i;

	/*
	 * create filename with timestamp and a randomized component.
	 * common prefix added to all files; no need to worry about
	 * leading '.' characters in filenames. slashes are removed below.
	 */
	maxlen = sizeof(name);
	strlcpy(name, CACHEFILE_PREFIX, maxlen);
	snprintf(tmp, sizeof(tmp), "%ld-%ld-", (long)time(NULL), random() % 10000);
	strlcat(name, tmp, maxlen);
	/*
	 * the uri path might be NUL, so include hostname as well
	 * to differentiate between files on different hosts.
	 *
	 * ignore truncation; file should still be unique
	 *
	 * copy values first, the remove unwanted chars
	 */
	strlcat(name, ue->meta->uri->host, maxlen);
	strlcat(name, "-", maxlen);
	strlcat(name, ue->meta->uri->path, maxlen);

	/* prettify (will remove any slashes) */
	for (i = 0; (size_t)i < strlen(name); i++) {
		c = name[i];
		if (isalnum(c) || (c == '.') || (c == '_'))
			continue;
		else
			name[i] = '-';
	}

	return name;
}


/*XXX take filemeta as argument? */
/* cachefile_create - create file of given size */
int
cachefile_create(char *path, off_t size)
{
	const char *f = "cachefile_create";
	char nul;
	int fd;

	/* attempt to create new file (exclusively) */
	if ((fd = open(path, O_RDWR|O_CREAT|O_EXCL, S_IRWXU)) == -1) {
		swarn("%s: open: %s: %m", f, path);
		return -1;
	}
	/* set size (by writing a single char at file end) XXXbetter way? */
	if (lseek(fd, size - 1, SEEK_SET) == -1) {
		swarn("%s: lseek: %m", f);
		return -1;
	}
	nul = '\0';
	if (write(fd, &nul, 1) != 1) {
		swarn("%s: write: %m", f);
		return -1;
	}
	close(fd);

	return 0;
}
