/* conn.c - connection management functions */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: conn.c,v 1.154 2010/10/31 10:20:02 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <errno.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "log.h"
#include "file.h"
#include "config.h"
#include "conn.h"
#include "ip.h"
#include "peer.h"
#include "peerset.h"
#include "protio.h"
#include "queue.h"
#include "servop.h"
#include "socket.h"
#include "ssl.h"
#include "stream.h"
#include "tcpmodel.h"
#include "timers.h"

struct connections connections = SLIST_HEAD_INITIALIZER(connections);

void free_connection(struct connection *);
void conntimersync(struct connection *);
int peertimeosec(struct peer *);
void peer_addconn(struct peer *);
void peer_delconn(struct peer *);


/* conn_changestate - upper layer version of changecstate() */
void
conn_changestate(struct connection *c, enum connstatus newstate)
{
	const char *f = "conn_changestate";

        if ((c->status == CONNECTION_STATUS_WRITEREP) &&
            ((newstate == CONNECTION_STATUS_IDLE) ||
             (newstate == CONNECTION_STATUS_FOREVERSTREAM) ||
             (newstate == CONNECTION_STATUS_DEADLINESTREAM) ||
             (newstate == CONNECTION_STATUS_ENDED) ||
             (newstate == CONNECTION_STATUS_CACHED))) {
		if (c->ioendstat != 0) {
/*			dbg("%s: curr: %d afterio: %d new: %d", f, c->status,
			    c->ioendstat, newstate); */
			dbg("%s: many status changes", f);
		}
		c->ioendstat = newstate;
	} else
		changecstatus(c, newstate);
}


/* new_connection - allocate new connection structure */
struct connection *
new_connection(struct peer *peer, int fd)
{
	const char *f = "new_connection";
	struct connection *c;

	if (peer != NULL)
		dbg("%s: %s", f, peer->id);

	c = xmalloc(sizeof(struct connection));
	bzero(c, sizeof(struct connection));

	c->fd = fd;
	c->timeout = timeout_add((void (*)(void*))conntimeout, NULL, c);
	c->pollidx = -1;
	c->stream = NULL;
	c->peer = peer;
	/* wanted state must be set by caller */
	c->status = CONNECTION_STATUS_UNCONNECTED;
	if (peer != NULL)
		/* increment now, even if connection not established */
		peerpar_adjint(peer, PEERPAR_PEER_TOTCONN, 1);
	else
		dbg("%s: connection without peer", f);

	SLIST_INSERT_HEAD(&connections, c, n);
	pollchange++;

/*	connstatus();*/

	return c;
}


/* cache_connection - cache connections that might be reused soon */
void
cache_connection(struct connection *c)
{
	const char *f = "cache_connection";

	ASSERT(c->status != CONNECTION_STATUS_CACHED, "caching cached conn");
	ASSERT(c->noreuse == 0, "stream");
	ASSERT(c->peer != NULL, "NULL peer");
	ASSERT(c->type == CONN_TYPE_OUTGOING, "caching invalid conn type");

	if (c->status == CONNECTION_STATUS_ENDED) {
		swarn("%s: not caching ended connection to %s", f, c->peer->id);
		return;
	}
#if HAVE_OPENSSL
	/* cache_connectin used for both ssl and non-ssl connections */
	if (c->ssl_active) {
		sslconn_cache(c);
		return;
	}
#endif /* HAVE_OPENSSL */

	/* cache connection */
	if (c->peer->cached != NULL) {
		/* old connection is more likely to be dead */
		dbg("%s: closing old cached connection to %s", f, c->peer->id);
		changecstatus(c->peer->cached, CONNECTION_STATUS_ENDED);
	}

	/* clear connection state */
	conn_endstate(c);

	c->peer->cached = c;
	changecstatus(c, CONNECTION_STATUS_CACHED);
	dbg("%s: cached conn: %s", f, c->peer->id);

/*	connstatus();*/
}


/* conn_havecached - return true if cached connection to peer exists */
int
conn_havecached(struct peer *peer, int sslconn)
{
	struct peer *respeer;
	int i;

	/* not symbolic hostname? */
	if (peer->dns == NULL) {
		ASSERT(ip_isnumeric(peer->name), "symbolic, missing dns");

		if (sslconn && peer->sslcached != NULL)
			return 1;
		else if (sslconn == 0 && peer->cached != NULL)
			return 1;
		else
			return 0;  /* no cached connection */
	} else {
		for (i = 0; i < peer->dns->peercnt; i++) {
			respeer = pset_peer(peer->dns, i);
			if (sslconn && respeer->sslcached != NULL)
				return 1;
			else if (sslconn == 0 && respeer->cached != NULL)
				return 1;
		}

		/* no cached connection to any host in peerlist */
		return 0;
	}
}


/* conn_getcached - return cached connection to peer, if any exists */
struct connection *
conn_getcached(struct peer *peer, int sslconn)
{
	struct connection *c;
	struct peer *respeer;
	int i;

	c = NULL;

	/* not symbolic hostname? */
	if (peer->dns == NULL) {
		ASSERT(ip_isnumeric(peer->name), "symbolic, missing dns");

		if (sslconn && peer->sslcached != NULL) {
			c = peer->sslcached;
			peer->sslcached = NULL;
		} else if (!sslconn && peer->cached != NULL) {
			c = peer->cached;
			peer->cached = NULL;
		} /* else, no cached connection */
	} else {
		for (i = 0; i < peer->dns->peercnt; i++) {
			respeer = pset_peer(peer->dns, i);
			if (sslconn && respeer->sslcached != NULL) {
				c = respeer->sslcached;
				respeer->sslcached = NULL;
			} else if (!sslconn && respeer->cached != NULL) {
				c = respeer->cached;
				respeer->cached = NULL;
			} /* else, no cached connection */
		}
	}

	if (c == NULL)
		return NULL;

	changecstatus(c, CONNECTION_STATUS_WRITEREQ);
	xgettimeofday(&c->reqstart, NULL);

	return c;
}


/* idle_connection - mark connection (from remote node) as inactive */
void
idle_connection(struct connection *c)
{
	const char *f = "idle_connection";

	ASSERT(c->peer != NULL, "NULL peer");
	ASSERT(c->type == CONN_TYPE_INCOMING, "idling invalid conn type");

#if 0
	if (c->peer->XXXidlecnt > PEER_IDLE_MAX) {
		dbg("%s: too many idle connections to %s, ending", f,
		    c->peer->id);
		changecstatus(c, CONNECTION_STATUS_ENDED);
		return;
	}
#endif

	if (c->noreuse) {
		dbg("%s: noreuse connection never idle, ending", f);
		changecstatus(c, CONNECTION_STATUS_ENDED);
		return;
	}

	/* clear connection state */
	conn_endstate(c);
	changecstatus(c, CONNECTION_STATUS_IDLE);
	dbg("%s: connection marked idle - %s", f, c->peer->id);

	if (c->ibuflen > 0) {
		dbg("%s: data remaining on idle connection: %ld bytes", f,
		     (long)c->ibuflen);
		/* new request might already have been read */
		queue_add(QUEUE_PROTIO, c);
	}
	ASSERT(c->obuflen == 0, "output data on idle connection");
}


/* end_connection - remove and free connection data */
void
end_connection(struct connection *c)
{
	const char *f = "end_connection";

	ASSERT(c->proc == NULL, "ending connection with proc pointer");

	/* clear state */
	conn_endstate(c);

	if (c->fd != -1) {
		xclose(c->fd);
		c->fd = -1;
	}

#if HAVE_OPENSSL
	if (c->ssl != NULL) {
		/* kill if not active (i.e., connecting/accepting) */
		dbg("%s: killing ssl conn to %s", f, c->peer->id);
		ssl_kill(c);
	}
#endif /* HAVE_OPENSSL */

	/* use source port if no peer data exists */
	if (c->type == CONN_TYPE_RESOLVE)
		dbg("%s: resolve connection", f);
	else if (c->type == CONN_TYPE_MCONN)
		dbg("%s: mconn connection", f);
	else if (c->peer == NULL)
		dbg("%s: %s:(%s)", f, c->peername, c->peerport);
	else
		dbg("%s: %s", f, c->peer->id);

	if (c->peer != NULL)
		peerpar_adjint(c->peer, PEERPAR_PEER_TOTCONN, -1);

	/* remove from peer if cached */
	if (c->peer != NULL) {
		if (c->peer->cached == c)
			c->peer->cached = NULL;
		if (c->peer->sslcached == c) {
			dbg("%s: removing cached ssl connection: %s", f,
			    c->peer->id);
			c->peer->sslcached = NULL;
		}
	}

	free_connection(c);

/*	connstatus();*/
}


/* free_connection - release connection resources */
void
free_connection(struct connection *c)
{
	SLIST_REMOVE(&connections, c, connection, n);
	pollchange++;

	ASSERT(c->timeout != NULL, "NULL timeout");
	timeout_del(c->timeout);
	free(c);
}


/* conn_revconn - swap incoming/outgoing connection status */
void
conn_revconn(struct connection *c)
{
	if (c->type == CONN_TYPE_INCOMING)
		c->type = CONN_TYPE_OUTGOING;
	else if (c->type == CONN_TYPE_OUTGOING)
		c->type = CONN_TYPE_INCOMING;
	else
		ERRX("invalid type");
}


/* conn_endstate - reset all state information for connection */
void
conn_endstate(struct connection *c)
{
	if (c->stream != NULL) {
		stream_end(c->stream);
		c->stream = NULL;
	}

	/* notify upper layer of connection end */
	if (c->uldat != NULL) {
		c->protops->connevent(c, CONNEVENT_CONNECTION_END, 0);
		c->protops->connevent(c, CONNEVENT_FREEULDAT, 0);
/*		c->protops->connend(c->uldat);*/
/*		c->protops->freeuldat(c->uldat);*/
		c->uldat = NULL;
		/*XXX should really also reset protops, but skip for now */
	}

	if (c->mconnpset != NULL)
		pset_free(c->mconnpset);
}


/* conntimeout - handle state specific connection timeout */
void
conntimeout(struct connection *c)
{
	const char *f = "conntimeout";

	dbg("%s: expired timer for %s", f, conn_state2str(c->status));

	switch(c->status) {
	case CONNECTION_STATUS_CACHED:
		dbg("%s: %s: connection cached too long, ending", f,
		    c->peer->id);
		changecstatus(c, CONNECTION_STATUS_ENDED);
		break;

	case CONNECTION_STATUS_DEADLINESTREAM:
		stream_tooslow(c);
		break;

	case CONNECTION_STATUS_CONNECTING:
		peerpar_adjint(c->peer, PEERPAR_PEER_CONNTIMEO, 1);
		failedconn(c->peer);
		/* timeout is soft error, retry */
		if (c->retries++ < CONNECTION_MAX_RETRIES) {
			dbg("%s: %s: connect timeout, retrying (%d)", f,
			    c->peer->id, c->retries);
			xclose(c->fd);
			conn_connect(c);
		} else {
			dbg("%s: %s: connect timeout, giving up", f,
			     c->peer->id);
			c->protops->connevent(c, CONNEVENT_CONNECT_FAIL,
					      ETIMEDOUT);
/*			c->protops->connerr(c->uldat, ETIMEDOUT);*/
			changecstatus(c, CONNECTION_STATUS_ENDED);
		}
		break;

	case CONNECTION_STATUS_IDLE:
		dbg("%s: %s connection idle too long, ending", f, c->peer->id);
#if 0
		/* XXX simpler to have revconn counter and get counter */
		if (c->revconn && (c->age == 0)) {
			dbg("%s: profitless revconn attempt, not retrying", f);
			peerpar_adjint(c->peer, PEERPAR_PEER_LASTRES,
				       PEER_MULTFAIL);
		}
#endif
#if HAVE_OPENSSL
		if (c->ssl != NULL)
			ssl_close(c);
		else
#endif /* HAVE_OPENSSL */
			changecstatus(c, CONNECTION_STATUS_ENDED);
		break;

#if HAVE_OPENSSL
	case CONNECTION_STATUS_SSLCACHED:
		dbg("%s: ssl connection to %s cached too long, ending", f,
		     c->peer->id);
		ASSERT(c->ssl != NULL, "NULL ssl");
		ssl_close(c);
		break;
#endif /* HAVE_OPENSSL */

	case CONNECTION_STATUS_READREQ:   /* FALLTHROUGH */
	case CONNECTION_STATUS_READREQDAT:/* FALLTHROUGH */
	case CONNECTION_STATUS_READREP:   /* FALLTHROUGH */
	case CONNECTION_STATUS_READREPDAT:/* FALLTHROUGH */
	case CONNECTION_STATUS_WRITEREP:  /* FALLTHROUGH */
	case CONNECTION_STATUS_WRITEREQ:
		dbg("%s: reqrep timeout, ending connection to %s:(%s)", f,
		     c->peername, c->peerport);
		/* event less interesting if peer is connecting to us */
		if (c->type != CONN_TYPE_INCOMING) {
			if (c->peer != NULL)
				failedconn(c->peer);
			c->protops->connevent(c, CONNEVENT_REQUEST_FAIL,
					      ETIMEDOUT);
/*			c->protops->failedreq(c->uldat);*/
		}
		/*
		 * use of changecstatus() will cause ioendstat to be
		 * set if current state is writerep
		 */
		changecstatus(c, CONNECTION_STATUS_ENDED);
		break;

	default:
		dbg("%s: %s", f, conn_state2str(c->status));
		ERRX("unexpected status");
	}
}


/* failedconn - update failed connections */
void
failedconn(struct peer *peer)
{
	int val;

	peerpar_adjint(peer, PEERPAR_PEER_FAILEDCONN, 1);
	val = peerpar_getint(peer, PEERPAR_PEER_FAILEDCONN);

	/* XXX SPP values */
	if (val >= PEER_SEQFAIL_MAX)
		peerpar_setint(peer, PEERPAR_SPP_LASTRES, PEER_MULTFAIL);
	else
		peerpar_setint(peer, PEERPAR_SPP_LASTRES, PEER_PREVFAIL);
}


/* conn_conninit - start connection to host/port, nonblocking */
struct connection *
conn_conninit(struct peer *peer)
{
	struct connection *c;

	ASSERT(peer->addr != NULL, "no address info in peer");

	c = new_connection(peer, -1);
	c->type = CONN_TYPE_OUTGOING;

	return c;
}


/* conntimersync - set state dependent connection timeout */
void
conntimersync(struct connection *c)
{
	const char *f = "conntimersync";
	struct timeval to;
	int update;
	float dur;

	bzero(&to, sizeof(to));

	update = 0;
	switch (c->status) {
	case CONNECTION_STATUS_UNCONNECTED:/* FALLTHROUGH */
	case CONNECTION_STATUS_ACCEPTING:  /* FALLTHROUGH */
	case CONNECTION_STATUS_RESOLVING:  /* FALLTHROUGH */ /*XXX*/
	case CONNECTION_STATUS_RESOLVEWAIT:/* FALLTRHOUGH */
	case CONNECTION_STATUS_MCONN:      /* FALLTHROUGH */
	case CONNECTION_STATUS_MCONNWAIT:
	case CONNECTION_STATUS_ENDED:
		/* no timeout */
		break;

	case CONNECTION_STATUS_FOREVERSTREAM:
		dbg("%s: stream to %s, no transfer timeout", f, c->peer->id);
		break;

	case CONNECTION_STATUS_DEADLINESTREAM:
		dur = peer_xfrdeadline(c->peer, c->stream->endoff + 1 - c->stream->startoff);

		/* convert to timeval */
		to.tv_sec = (int)dur;
		dur -= to.tv_sec;
		to.tv_usec = (long)(dur * 1000 * 1000);
		/*XXX sanity check value */
		dbg("%s: transfer deadline: %ld.%06ld s", f, (long)to.tv_sec,
		    (long)to.tv_usec);

		update = 1;
		break;

	case CONNECTION_STATUS_IDLE:
		to.tv_sec = CONNECTION_IDLE_TIMEOUT;
		update = 1;
		break;

	case CONNECTION_STATUS_CACHED:
		to.tv_sec = CONNECTION_CACHED_TIMEOUT;
		update = 1;
		break;

	case CONNECTION_STATUS_READREQ:   /* FALLTHROUGH */
	case CONNECTION_STATUS_READREQDAT:/* FALLTHROUGH */
	case CONNECTION_STATUS_READREP:   /* FALLTHROUGH */
	case CONNECTION_STATUS_READREPDAT:/* FALLTHROUGH */
	case CONNECTION_STATUS_WRITEREP:  /* FALLTHROUGH */
	case CONNECTION_STATUS_WRITEREQ:
		to.tv_sec = CONNECTION_REQREP_TIMEOUT;
		update = 1;
		break;

	case CONNECTION_STATUS_SSLCACHED:
		to.tv_sec = CONNECTION_CACHED_TIMEOUT;
		update = 1;
		break;

	case CONNECTION_STATUS_CONNECTING:
		to.tv_sec = peertimeosec(c->peer);
		update = 1;
		break;
	}

	if (update) {
/*		dbg("%s: enabling timeout, duration %ld.%06ld", f,
		    to.tv_sec, to.tv_usec);*/
		timeout_upd(c->timeout, &to);
	} else {
/*		dbg("%s: disabling timeout %s", f,
		     conn_state2sdtr(c->status));*/
		timeout_off(c->timeout);
	}
}


/* peertimeosec - return a timeout value for connections to the given peer */
int
peertimeosec(struct peer *peer)
{
	int conntimeo;

	conntimeo = peerpar_getint(peer, PEERPAR_PEER_CONNTIMEO) + 1;
	/* increase timeout interval for each failure, up to max limit */
	if (conntimeo > CONNECTION_REMOTE_MAXFAIL)
		conntimeo = CONNECTION_REMOTE_MAXFAIL;
	return conntimeo * CONNECT_TIMEOUT_SEC;
}


/* peer_addconn - increase peer connection count */
void
peer_addconn(struct peer *peer)
{
	peerpar_adjint(peer, PEERPAR_PEER_ACTIVECONN, 1);
}


/* peer_delconn - decrease peer connection count */
void
peer_delconn(struct peer *peer)
{
	ASSERT(peerpar_getint(peer, PEERPAR_PEER_ACTIVECONN) > 0,
	       "activeconn < 0");
	peerpar_adjint(peer, PEERPAR_PEER_ACTIVECONN, -1);
}


/* changecstatus - change connection status/state */
void
changecstatus(struct connection *c, enum connstatus newstatus)
{
	const char *f = "changecstatus";

/*	dbg("%s: (%d) curr: %s afterio: %s new: %s", f, c->pollidx,
	    conn_state2str(c->status), conn_state2str(c->ioendstat),
	    conn_state2str(newstatus));*/
	if (c->status == CONNECTION_STATUS_RESOLVEWAIT &&
	    newstatus == CONNECTION_STATUS_ENDED)
		ERRX("ending resolvewait connection");
#if 0
	if (c->ibuflen != 0)
		dbg("%s: ibuflen nonzero: %s", f, c->ibuf);
#endif
#if 0
	if (c->obuflen != 0)
		dbg("%s: obuflen nonzero: %s", c->obuf);
#endif
	if (c->ioendstat != 0) {
		dbg("%s: clearing ioendstat (%s), curr: %s new %s", f,
		    conn_state2str(c->ioendstat), conn_state2str(c->status),
		    conn_state2str(newstatus));
		c->ioendstat = 0;
	}
	if (c->status == newstatus)
		/* dbg("%s: connection state the same (%d)", f,
		   conn_state2str(c->status))*/;
	else {
		if (c->status == CONNECTION_STATUS_ENDED) {
			swarn("%s: reuse of ended connection", f);
			ERRX("reuse");
		}
		if (c->status == CONNECTION_STATUS_FOREVERSTREAM ||
		    c->status == CONNECTION_STATUS_DEADLINESTREAM)
			peer_delconn(c->peer);
		if (newstatus == CONNECTION_STATUS_FOREVERSTREAM ||
		    newstatus == CONNECTION_STATUS_DEADLINESTREAM) {
			/* data already available? */
			if (c->ibuflen > 0)
				queue_add(QUEUE_DATAIO, c);
		}
		if ((newstatus == CONNECTION_STATUS_READREQDAT) ||
		    (newstatus == CONNECTION_STATUS_READREPDAT)) {
			ASSERT(c->ibufexplen != 0, "no explen");
			ASSERT(c->ibufexplen < (int)sizeof(c->ibuf),
			       "XXX reply too long");
			/* data already available? */
			if (ibuf_explen(c) == 0) {
				dbg("%s: IODONE", f);
				queue_add(QUEUE_IODONE, c);
			}
		}
		if (newstatus == CONNECTION_STATUS_FOREVERSTREAM ||
		    newstatus == CONNECTION_STATUS_DEADLINESTREAM)
			peer_addconn(c->peer);
		if (newstatus == CONNECTION_STATUS_ENDED)
			queue_add(QUEUE_CONNEND, c);
		c->status = newstatus;
		conntimersync(c);
#if 0
		if (c->status == CONNECTION_STATUS_IDLE)
			queue_add(QUEUE_CONNEND, c);
#endif
	}
	pollchange++;
}


/* ioendcstatus - set connection io completion status */
void
ioendcstatus(struct connection *c)
{
	c->status = c->ioendstat;
	c->ioendstat = 0;
	conntimersync(c);
	if (c->status == CONNECTION_STATUS_FOREVERSTREAM ||
	    c->status == CONNECTION_STATUS_DEADLINESTREAM)
		peer_addconn(c->peer);
	if (c->status == CONNECTION_STATUS_ENDED)
		queue_add(QUEUE_CONNEND, c);
	if (c->status == CONNECTION_STATUS_IDLE)
		idle_connection(c);
	pollchange++;
}


/* conn_state2str - return text representation of connection state */
char *
conn_state2str(enum connstatus status)
{
	switch (status) {
	case CONNECTION_STATUS_UNCONNECTED:
		return "CONNECTION_STATUS_UNCONNECTED";
	case CONNECTION_STATUS_ACCEPTING:
		return "CONNECTION_STATUS_ACCEPTING";
	case CONNECTION_STATUS_CACHED:
		return "CONNECTION_STATUS_CACHED";
	case CONNECTION_STATUS_CONNECTING:
		return "CONNECTION_STATUS_CONNECTING";
	case CONNECTION_STATUS_DEADLINESTREAM:
		return "CONNECTION_STATUS_DEADLINESTREAM";
	case CONNECTION_STATUS_ENDED:
		return "CONNECTION_STATUS_ENDED";
	case CONNECTION_STATUS_FOREVERSTREAM:
		return "CONNECTION_STATUS_FOREVERSTREAM";
	case CONNECTION_STATUS_IDLE:
		return "CONNECTION_STATUS_IDLE";
	case CONNECTION_STATUS_MCONN:
		return "CONNECTION_STATUS_MCONN";
	case CONNECTION_STATUS_MCONNWAIT:
		return "CONNECTION_STATUS_MCONNWAIT";
	case CONNECTION_STATUS_READREP:
		return "CONNECTION_STATUS_READREP";
	case CONNECTION_STATUS_READREPDAT:
		return "CONNECTION_STATUS_READREPDAT";
	case CONNECTION_STATUS_READREQ:
		return "CONNECTION_STATUS_READREQ";
	case CONNECTION_STATUS_READREQDAT:
		return "CONNECTION_STATUS_READREQDAT";
	case CONNECTION_STATUS_RESOLVEWAIT:
		return "CONNECTION_STATUS_RESOLVEWAIT";
	case CONNECTION_STATUS_RESOLVING:
		return "CONNECTION_STATUS_RESOLVING";
	case CONNECTION_STATUS_SSLCACHED:
		return "CONNECTION_STATUS_SSLCACHED";
	case CONNECTION_STATUS_WRITEREP:
		return "CONNECTION_STATUS_WRITEREP";
	case CONNECTION_STATUS_WRITEREQ:
		return "CONNECTION_STATUS_WRITEREQ";
	}
	ERRX("no state matched");
	/*NOTREACHED*/
}


/* conn_connev2str - return text representation of connection event */
char *
conn_connev2str(enum connevent event)
{
	switch (event) {
	case CONNEVENT_CONNECT_OK:
		return "CONNEVENT_CONNECT_OK";
	case CONNEVENT_CONNECT_FAIL:
		return "CONNEVENT_CONNECT_FAIL";
	case CONNEVENT_CONNECTION_END:
		return "CONNEVENT_CONNECTION_END";
	case CONNEVENT_REQUEST_FAIL:
		return "CONNEVENT_REQUEST_FAIL";
	case CONNEVENT_GOT_REPLY:
		return "CONNEVENT_GOT_REPLY";
	case CONNEVENT_GOT_REPLYDAT:
		return "CONNEVENT_GOT_REPLYDAT";
	case CONNEVENT_GOT_REQUEST:
		return "CONNEVENT_GOT_REQUEST";
	case CONNEVENT_GOT_REQUESTDAT:
		return "CONNEVENT_GOT_REQUESTDAT";
	case CONNEVENT_WROTE_REPLY:
		return "CONNEVENT_WROTE_REPLY";
	case CONNEVENT_SSLCONNECT_OK:
		return "CONNEVENT_SSLCONNECT_OK";
	case CONNEVENT_SSLCONNECT_FAIL:
		return "CONNEVENT_SSLCONNECT_FAIL";
	case CONNEVENT_SSLCONNECTION_END:
		return "CONNEVENT_SSLCONNECTION_END";
	case CONNEVENT_FREEULDAT:
		return "CONNEVENT_FREEULDAT";
	}
	ERRX("no state matched");
	/*NOTREACHED*/
}
