/* conn.h - connection related defined and structures */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: conn.h,v 1.92 2010/08/02 22:24:56 karls Exp $ */

#ifndef _CONN_H_
#define _CONN_H_

#include <sys/types.h>
#include <sys/socket.h>

#if HAVE_QUEUE_MACROS
#include <sys/queue.h>
#else
#include "bsd_queue.h"
#endif /* HAVE_QUEUE_MACROS */

#include <netdb.h>
#include <syslog.h>

#if HAVE_OPENSSL
#include "ssl.h"
#endif /* HAVE_OPENSSL */

#define CONNECT_TIMEOUT_SEC 1
#define CONNECT_TIMEOUT_MSEC 0
#define CONNECT_RETRIES 3

#define SOCKET_BUFSIZE (256 * 1024) /* XXX fancy window size calculation? */

#define CONNECTION_REMOTE_MAXFAIL 3

#define TIMEOUT_SHORT /*XXX*/
#ifdef TIMEOUT_SHORT
#define CONNECTION_REQREP_TIMEOUT 10 /* seconds */
#else
#define CONNECTION_REQREP_TIMEOUT 120 /* seconds */
#endif /* TIMEOUT_SHORT */
#define CONNECTION_IDLE_TIMEOUT 120
#define CONNECTION_CACHED_TIMEOUT 30

#define PEER_IDLE_MAX 2
#define PEER_SEQFAIL_MAX 3 /* avoid hosts which are likely down */

#define CONNECTION_MAX_RETRIES 3

enum connevent {
	CONNEVENT_CONNECT_OK = 1,
	CONNEVENT_CONNECT_FAIL,
	CONNEVENT_CONNECTION_END,
	CONNEVENT_REQUEST_FAIL,
	CONNEVENT_GOT_REPLY,
	CONNEVENT_GOT_REPLYDAT,
	CONNEVENT_GOT_REQUEST, /*XXX handled by callback.c */
	CONNEVENT_GOT_REQUESTDAT,
	CONNEVENT_WROTE_REPLY,
	CONNEVENT_SSLCONNECT_OK,
	CONNEVENT_SSLCONNECT_FAIL,
	CONNEVENT_SSLCONNECTION_END,
	CONNEVENT_FREEULDAT, /* XXX merge into CONNECTION_END? */
};

enum resolveevent {
	RESEVENT_RESOLVE_OK = 1,
	RESEVENT_RESOLVE_FAIL,
	RESEVENT_PROC_FAIL,
};

enum mconnevent {
	MCONNEVENT_CONNECT_OK = 1, /* might be returned multiple times */
	MCONNEVENT_CONNECT_FAIL,
	MCONNEVENT_CONNECT_ALLFAIL,
	MCONNEVENT_PROC_FAIL,
};

enum connstatus {
	CONNECTION_STATUS_UNCONNECTED = 1,
	CONNECTION_STATUS_ACCEPTING,
	CONNECTION_STATUS_CACHED,
	CONNECTION_STATUS_CONNECTING,
	CONNECTION_STATUS_DEADLINESTREAM,
	CONNECTION_STATUS_ENDED,
	CONNECTION_STATUS_FOREVERSTREAM,
	CONNECTION_STATUS_IDLE,
	CONNECTION_STATUS_MCONN,
	CONNECTION_STATUS_MCONNWAIT,
	CONNECTION_STATUS_READREP,
	CONNECTION_STATUS_READREPDAT,
	CONNECTION_STATUS_READREQ,
	CONNECTION_STATUS_READREQDAT,
	CONNECTION_STATUS_RESOLVEWAIT,
	CONNECTION_STATUS_RESOLVING,
	CONNECTION_STATUS_SSLCACHED,
	CONNECTION_STATUS_WRITEREP,
	CONNECTION_STATUS_WRITEREQ,
};

enum conntype {
	CONN_TYPE_INCOMING = 1,
	CONN_TYPE_OUTGOING,
	CONN_TYPE_RESOLVE,
	CONN_TYPE_MCONN,
};

/* XXX if positive, peers with higher speed can have lower score (e.g. zero) */
/* XXX better place to define this? */
#define PEER_PREVOK 0 /* or unknownXXX */
#define PEER_PREVBUSY -1
#define PEER_PREVFAIL -2
#define PEER_MULTFAIL -3000

struct connection {
	int fd;
	enum conntype type;
	void *uldat; /* upper layer data */
	struct timeout *timeout;
	enum connstatus status;
	int pollevent; /* POLLIN/POLLOUT */
	uint16_t ioendstat; /* status code when protocol I/O finished */
	unsigned int age; /* number of blocks transferred */
	int noreuse; /* set if connection cannot be reused (idled/cached) */
	struct timeval connstart; /* start time for connection attempt */
	struct timeval reqstart;  /* start time for request (after connect) */
	struct peer *peer;
	int pollidx;
	int retries; /* connect retries (for non-fatal errors only) */
	struct stream *stream;
	struct proc *dnsproc; /* placeholder for proc when resolving */
	struct proc *proc; /* placeholder for proc when using mconn/resolver */
	struct peerset *mconnpset; /* mconn targets */
	int mconnrcpt; /* number of entries received */
	int mconnok; /* number of successfully connected entries */
	struct sockaddr_storage ss_local; /* ip address of our end */
	socklen_t ss_local_len; /* length of ip address of our end XXXneeded?*/
	char sockname[NI_MAXHOST]; /* ip address of our end, text */
	char sockport[NI_MAXSERV]; /* port number on our end, text */
	struct sockaddr_storage ss_remote; /* ip address of our end */
	socklen_t ss_remote_len; /* length of ip address of our end XXXneeded?*/
	char peername[NI_MAXHOST]; /* ip address of our end, text */
	char peerport[NI_MAXSERV]; /* port number on our end, text */
	int mss; /* TCP_MAXSEG value for connection */
	char ibuf[MAX_REQREPLEN]; /* requests/responses (read) */
	int ibuflen; /* length of data in buffer */
	int ibufexplen; /* expected buffer length */
	char obuf[MAX_REQREPLEN]; /* requests/responses (write) */
	int obuflen; /* length of data in buffer */
	int obufexplen; /* expected buffer length */
	int obufoff; /* offset for write operations */
	struct protops *protops;
#if HAVE_OPENSSL
	SSL *ssl;
	BIO *sbio;
	int ssl_serv;
	int ssl_active; /* 1 if connection data is encrypted */
	enum sslcmd ssl_cmd; /* current ssl command */
	enum sslcmdstate ssl_cmdstate; /* want read/write */
	struct timeval sslstart; /* conn start time */
#else
	void *ssl;
#endif /* HAVE_OPENSSL */
	SLIST_ENTRY(connection) n;
};

SLIST_HEAD(connections, connection);
extern struct connections connections;

/* conn.c */
void conn_changestate(struct connection *, enum connstatus);
struct connection *new_connection(struct peer *, int);
void cache_connection(struct connection *);
int conn_havecached(struct peer *, int);
struct connection *conn_getcached(struct peer *, int);
void idle_connection(struct connection *);
void end_connection(struct connection *);
void conn_revconn(struct connection *);
void conn_endstate(struct connection *);
void conntimeout(struct connection *);
void failedconn(struct peer *);
struct connection *conn_conninit(struct peer *);
void changecstatus(struct connection *, enum connstatus);
void ioendcstatus(struct connection *);
char *conn_state2str(enum connstatus);
char *conn_connev2str(enum connevent);

/* connio.c */
ssize_t ibufread(struct connection *, void *, size_t);
int read_conn(struct connection *);
int write_obuf(struct connection *);
void conn_ibufsync(struct connection *, char *);
int ibuf_nlnlterm(struct connection *);
int ibuf_explen(struct connection *);
void conn_writerep(struct connection *, char *);

/* conn_mconn */
struct connection *conn_mconninit(struct peerset *);
void conn_mconn(struct connection *);
void conn_gotmconnproc(struct connection *, struct proc *);
void conn_getmconnrep(struct connection *);
void conn_mconnfail(struct connection *);
char *conn_mconnev2str(enum mconnevent);

/* conn_resolve */
struct connection *conn_resinit(struct peer *);
void conn_resolve(struct connection *);
void conn_gotdnsproc(struct connection *, struct proc *);
void conn_getdnsrep(struct connection *);
void conn_dnsfail(struct connection *);
char *conn_resolveev2str(enum resolveevent);

/* conn_socket */
char *bind_port(char *);
struct connection *accept_connection(int);
void conn_connect(struct connection *);
int conncomp(struct connection *);
int conn_getsockname(struct connection *);
int conn_getpeername(struct connection *);
int conn_getmss(struct connection *);

/* debug */
void connstatus(void);

#endif /* _CONN_H_ */
