/* conn_mconn.c - connection interface to mconn (parallel connections) */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: conn_mconn.c,v 1.7 2010/08/02 22:24:56 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <errno.h>

#include "log.h"
#include "conn.h"
#include "peer.h"
#include "peerset.h"
#include "proc.h"
#include "protio.h"
#include "pset_addr.h"
#include "tcpmodel.h"
#include "mconn.h"


/* conn_mconninit - prepare for mconnect */
struct connection *
conn_mconninit(struct peerset *pset)
{
	const char *f = "conn_mconninit";
	struct connection *c;

	ASSERT(pset->peercnt <= MCONNMAXCONN, "pset too large");
	ASSERT(pset->peercnt > 0, "empty pset");

	dbg("%s: %d entries", f, pset->peercnt);

	c = new_connection(NULL, -1);
	c->mconnpset = pset_cpy(pset); /* create duplicate of argument */
	c->type = CONN_TYPE_MCONN;

	return c;
}


/* conn_mconn - initiate parallell mconnect to peerset */
void
conn_mconn(struct connection *c)
{
	if (proc_enabled[PROC_MCONN]) {
		changecstatus(c, CONNECTION_STATUS_MCONNWAIT);
		proc_enqueue(PROC_MCONN, c,
			     (void (*)(void *, struct proc *))conn_gotmconnproc,
			     (void (*)(void *))conn_mconnfail);
		return;
	}

	/* mconn process down */
	ERRX("mconn process down");

	/*XXX could connect to first entry/try sequentially */
}


/* conn_gotmconnproc - callback called upon being assigned mconn process */
void
conn_gotmconnproc(struct connection *c, struct proc *proc)
{
	ASSERT(c->status == CONNECTION_STATUS_MCONNWAIT, "invalid status");

	if (mconn_sendreq(proc->fd, pset_pset2addr(c->mconnpset)) == -1) {
		proc_kill(proc);
		proc_returnproc(proc);
		conn_mconnfail(c);
		return;
	}
	/* setup connection to wait for reply from mconn process */
	ASSERT(c->fd == -1, "connection descriptor already set");
	c->fd = proc->fd;
	ASSERT(c->proc == NULL, "proc already set");
	c->proc = proc;
	changecstatus(c, CONNECTION_STATUS_MCONN);
}


/* conn_getmconnrep - called upon mconn descriptor becoming readable */
void
conn_getmconnrep(struct connection *c)
{
	const char *f = "conn_getmconnrep";
	struct connection *peerconn;
	struct peer *peer;
	float conntime;
	int   idx, err;
	int   peerfd;

	ASSERT(c->proc != NULL, "NULL proc");
	ASSERT(c->mconnrcpt < c->mconnpset->peercnt,
	       "all descriptors already received");

	dbg("%s", f);

	peerfd = mconn_readrep(c->fd, &idx, &err, &conntime);
	if (peerfd == -1)
		/* mconn process communication */
		conn_mconnfail(c);
	else {
		ASSERT(idx <= c->mconnpset->peercnt, "invalid idx value");
		peer = pset_peer(c->mconnpset, idx);
		/* did we get a descriptor? */
		/*XXX zero is a valid descriptor */
		if (peerfd != 0) {
			c->mconnok++;
			dbg("%s: got connection: %s conntime: %f", f, peer->id,
			    conntime);
			peer_addrtt(peer, conntime);
			peerconn = new_connection(peer, peerfd);
			peerconn->type = CONN_TYPE_OUTGOING;
			peerconn->protops = c->protops;
			if (conncomp(peerconn) == 0)
				c->protops->mconnevent(c, MCONNEVENT_CONNECT_OK,
						       idx, 0, peerconn);
/*				c->protops->mconndone(c->uldat, idx, peerconn);*/
		} else {
			errno = err;
			dbg("%s: connect failure: %s: %m", f, peer->id);
			failedconn(peer);
			c->protops->mconnevent(c, MCONNEVENT_CONNECT_FAIL, idx,
					       err, NULL);
/*			c->protops->mconnerr(c->uldat, idx, err);*/
		}
	}

	c->mconnrcpt++;
	if (c->mconnrcpt == c->mconnpset->peercnt) {
		proc_returnproc(c->proc);
		c->proc = NULL;
		c->fd = -1;
		pset_free(c->mconnpset);
		c->mconnpset = NULL;
		/* report as request failure if no connections succeeded */
		if (c->mconnok == 0)
			/*XXXconnerr? */
			c->protops->mconnevent(c, MCONNEVENT_CONNECT_ALLFAIL,
					       0, 0, NULL);
/*			c->protops->failedreq(c->uldat);*/
		changecstatus(c, CONNECTION_STATUS_ENDED);
	}
}


/* conn_mconnfail - handle failure of mconn process */
void
conn_mconnfail(struct connection *c)
{
	c->protops->mconnevent(c, MCONNEVENT_PROC_FAIL, 0, 0, NULL);
}


/* conn_mconnev2str - return text representation of mconn event */
char *
conn_mconnev2str(enum mconnevent event)
{
	switch (event) {
	case MCONNEVENT_CONNECT_OK:
		return "MCONNEVENT_CONNECT_OK";
	case MCONNEVENT_CONNECT_FAIL:
		return "MCONNEVENT_CONNECT_FAIL";
	case MCONNEVENT_CONNECT_ALLFAIL:
		return "MCONNEVENT_CONNECT_ALLFAIL";
	case MCONNEVENT_PROC_FAIL:
		return "MCONNEVENT_PROC_FAIL";
	}
	ERRX("no state matched");
	/*NOTREACHED*/
}
