/* conn_resolve.c - non-blocking dns resolution via proc managed resolver */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: conn_resolve.c,v 1.9 2010/08/02 22:24:56 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>

#include <netdb.h>
#include <string.h>

#include "ip.h"
#include "log.h"
#include "config.h"
#include "conn.h"
#include "peer.h"
#include "pset_addr.h"
#include "proc.h"
#include "protio.h"
#include "resolve.h"
#include "socket.h"

/*
 * functions for doing non-blocking dns resolution via connection interface.
 * the caller needs to create a new connection via conn_resinit(peer),
 * followed by conn_resolve(), to resolve the address of peer.
 * after the resolve_done() callback has been called, it will be possible
 * to connect to the peer using conn_conninit() and conn_connect().
 */


/* conn_resinit - return connection for dns resolution */
/*
 * XXX connection never used to connect remotely, but having separate struct
 *     would make things more compliated.
 */
struct connection *
conn_resinit(struct peer *peer)
{
	struct connection *c;

	c = new_connection(peer, -1);
	c->type = CONN_TYPE_RESOLVE;

	return c;
}


/* conn_resolve - perform (or initiate) resolution of peer address */
void
conn_resolve(struct connection *c)
{
	const char *f = "conn_resolve";
	struct addrinfo hints, *res;
	int n;

	/* can connect proceed directly? */
	if (peer_validaddr(c->peer)) {
		c->protops->resevent(c, RESEVENT_RESOLVE_OK, 0, c->peer->dns);
/*		c->protops->resolve_done(c->uldat, c->peer->dns);*/
		return;
	}

	dbg("%s: %s", f, c->peer->id);

	/*
	 * no, time to do/redo dns lookup. if resolver is up,
	 * enqueue dns request, if not, fallback to blocking lookup.
	 *
	 * (any old data is freed when results are received).
	 */
	if (proc_enabled[PROC_RESOLVE]) {
		changecstatus(c, CONNECTION_STATUS_RESOLVEWAIT);
		proc_enqueue(PROC_RESOLVE, c,
			     (void (*)(void *, struct proc *))conn_gotdnsproc,
			     (void (*)(void *))conn_dnsfail);
		return;
	}
	/* die to detect resolver errors for now */
	ERRX("resolver is down");
	swarn("%s: doing fallback to blocking dns lookup", f);

	/* if the resolver is down, do blocking dns resolution) */
	bzero(&hints, sizeof(struct addrinfo));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	if ((n = getaddrinfo(c->peer->name, c->peer->port, &hints, &res)) != 0) {
		swarn("%s: getaddrinfo failure for %s %s", f, c->peer->id,
		      gai_strerror(n));
		peer_setdns(c->peer, NULL, n);
		failedconn(c->peer); /* XXX failedresolv? */
		c->protops->resevent(c, RESEVENT_RESOLVE_FAIL, n, NULL);
/*		c->protops->resolve_err(c->uldat);*/
		changecstatus(c, CONNECTION_STATUS_ENDED);
	} else {
		peer_setdns(c->peer, pset_addr2pset(res), 0);
		c->protops->resevent(c, RESEVENT_RESOLVE_OK, 0, c->peer->dns);
/*		c->protops->resolve_done(c->uldat, c->peer->dns);*/
	}
}


/* conn_gotdnssproc - callback called upon being assigned resolver process */
void
conn_gotdnsproc(struct connection *c, struct proc *proc)
{
	ASSERT(c->status == CONNECTION_STATUS_RESOLVEWAIT, "invalid status");

	/* if we were queued, somebody else might have already resolved name */
	if (peer_validaddr(c->peer)) {
		proc_returnproc(proc); /* no longer needed */
		c->protops->resevent(c, RESEVENT_RESOLVE_OK, 0, c->peer->dns);
/*		c->protops->resolve_done(c->uldat, c->peer->dns);*/
	} else {
		ASSERT(!ip_isnumeric(c->peer->name), "resolving numeric addr");
		if (resolve_sendreq(proc->fd, c->peer->name, c->peer->port) == -1) {
			proc_kill(proc);
			proc_returnproc(proc);
			conn_dnsfail(c); /* dnsfail might reenqueue request */
			return;
		}
		/* setup connection to wait for reply from resolver */
		ASSERT(c->fd == -1, "connection descriptor already set");
		c->fd = proc->fd;
		ASSERT(c->dnsproc == NULL, "dnsproc already set");
		c->dnsproc = proc;
		changecstatus(c, CONNECTION_STATUS_RESOLVING);
	}
}


/* conn_getdnsrep - called upon resolve descriptor becoming readable */
void
conn_getdnsrep(struct connection *c)
{
	const char *f = "conn_getdnsrep";
	struct addrinfo *res;
	int code;
	int ret;

	ASSERT(c->dnsproc != NULL, "NULL proc");

	dbg("%s", f);

	ret = resolve_readrep(c->fd, &res, &code);
	proc_returnproc(c->dnsproc);
	c->dnsproc = NULL;
	c->fd = -1;

	if (ret == -1 || code == EAI_AGAIN)
		/* resolver communication/dns lookup failed */
		conn_dnsfail(c);
	else {
		if (code == 0) {
			peer_setdns(c->peer, pset_addr2pset(res), 0);
			c->protops->resevent(c, RESEVENT_RESOLVE_OK, 0,
					     c->peer->dns);
/*			c->protops->resolve_done(c->uldat, c->peer->dns);*/
		} else {
			peer_setdns(c->peer, NULL, code);
			c->protops->resevent(c, RESEVENT_RESOLVE_FAIL, code, NULL);
/*			c->protops->resolve_err(c->uldat);*/
		}
	}
	changecstatus(c, CONNECTION_STATUS_ENDED);
}


/* conn_dnsfail - handle failure of dns resolution process */
void
conn_dnsfail(struct connection *c)
{
	if (c->peer->resolveretries++ >= RESOLVE_MAXRETRIES) {
		failedconn(c->peer); /* XXX failedresolv? */
		c->protops->resevent(c, RESEVENT_PROC_FAIL, EAI_FAIL, NULL);
/*		c->protops->failedreq(c->uldat);*/
		changecstatus(c, CONNECTION_STATUS_ENDED);
	} else
		/* retry (will be placed at end of queue) */
		conn_resolve(c);
}


#if 0 /* notyet, killing increases child mortality rate XXX should it? */
/* conn_dnstimeout - handle timeout of resolver request */
void
conn_dnstimeout(struct connection *c)
{
	ASSERT(c->status == CONNECTION_STATUS_RESOLVING, "invalid status");
	/* should not be necessary with timeout for RESOLVEWAIT */

	/* resolver process appears to be hanging, kill it */
	proc_kill(c->dnsproc);

	conn_dnsfail(c);
}
#endif


/* conn_resolveev2str - return text representation of resolve event */
char *
conn_resolveev2str(enum resolveevent event)
{
	switch (event) {
	case RESEVENT_RESOLVE_OK:
		return "RESEVENT_RESOLVE_OK";
	case RESEVENT_RESOLVE_FAIL:
		return "RESEVENT_RESOLVE_FAIL";
	case RESEVENT_PROC_FAIL:
		return "RESEVENT_PROC_FAIL";
	}
	ERRX("no state matched");
	/*NOTREACHED*/
}
