/* conn_socket.c - sockets api related function for connection interface */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: conn_socket.c,v 1.22 2010/10/11 22:01:55 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "log.h"
#include "conn.h"
#include "peer.h"
#include "peerset.h"
#include "protio.h"
#include "servop.h"
#include "socket.h"
#include "tcpmodel.h"
#include "tv.h"

#define LISTEN_BACKLOG 50


/* bind_port - bind server sockets */
char *
bind_port(char *altport)
{
	const char *f = "bind_port";
	struct addrinfo hints, *res, *res0;
	struct connection *c;
	int optval;
	int nsock;
	int error;
	int bound;
	int s;

	bzero(&hints, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	hints.ai_protocol = IPPROTO_TCP;

	res0 = NULL;
	if (altport == NULL)
		error = getaddrinfo(NULL, "0", &hints, &res0);
	else
		error = getaddrinfo(NULL, altport, &hints, &res0);
	if (error)
		return NULL;

	nsock = 0;
	for (res = res0; res; res = res->ai_next)
		nsock++;
	ASSERT(nsock < POLL_MAX, "too many bound ports");

	bound = 0;
	c = NULL;
	for (res = res0; res; res = res->ai_next) {
		/* XXX skip ipv6 for now */
		if (res->ai_family == PF_INET6)
			continue;

		if ((s = socket(res->ai_family, res->ai_socktype,
				res->ai_protocol)) == -1) {
			swarn("%s: socket: %m", f);
			continue;
		}

#ifdef IPV6_V6ONLY
		optval = 1;
		if (res->ai_family == PF_INET6)
			if (setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY, &optval,
				       sizeof(optval)) == -1)
				swarn("%s: setsockopt: %m", f);
#endif /* IPV6_V6ONLY */

		optval = 1;
		if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &optval,
			       sizeof(optval)) == -1) {
			swarn("%s: setsockopt: %m", f);
			xclose(s);
			continue;
		}

		socketprep(s);

		if (bind(s, res->ai_addr, res->ai_addrlen) == -1) {
			swarn("%s: bind: %m", f);
			xclose(s);
			continue;
		}

		if (listen(s, LISTEN_BACKLOG) == -1) {
			swarn("%s: listen: %m", f);
			xclose(s);
			continue;
		}
		c = new_connection(NULL, s);
		changecstatus(c, CONNECTION_STATUS_ACCEPTING);

		bound++;

		if (conn_getsockname(c) == -1)
			serr("getsockname: %m");
		dbg("%s: bound to port %s", f, c->sockport);
	}
	freeaddrinfo(res0);

	/*XXX only handle binding of one port correctly */
	if (c == NULL)
		return NULL;
	else
		return c->sockport;
}


/* accept_connection - handle new connection */
struct connection *
accept_connection(int acceptfd)
{
	const char *f = "accept_connection";
	struct sockaddr_storage ss;
	struct connection *c;
	socklen_t salen;
	int s;

	/* new connection */
	salen = sizeof(ss);
	/* XXX check peer before accept() */
	if ((s = accept(acceptfd, (struct sockaddr *)&ss, &salen)) == -1) {
		swarn("accept: %m");
		return NULL;
	}

	/*
	 * linux does not let connection inherit nonblocking flag;
	 * call socketprep also after accept().
	 */
	if (socketprep(s) == -1)
		return NULL;

	c = new_connection(NULL, s);
	c->type = CONN_TYPE_INCOMING;

	if (conn_getmss(c) == -1 ||
	    conn_getsockname(c) == -1 ||
	    conn_getpeername(c) == -1) {
		changecstatus(c, CONNECTION_STATUS_ENDED);
		return NULL;
	}

	/*
	 * note: the peerport is different from the port the peer receives
	 *       connections to (being peer->port).
	 */
	dbg("%s: got connection from %s:%s", f, c->peername, c->peerport);

	changecstatus(c, CONNECTION_STATUS_READREQ);

	return c;
}


/* conn_connect - attempt nonblocking connect */
void
conn_connect(struct connection *c)
{
	const char *f = "conn_connect";
	struct addrinfo *res;
	float conntime;
	int s;

	ASSERT(peer_validaddr(c->peer), "no address information for peer");

	res = c->peer->addr;
	if ((s = socket(res->ai_family, res->ai_socktype,
			res->ai_protocol)) == -1) {
		swarn("%s: %s: socket: %m", f, c->peer->id);
		c->protops->connevent(c, CONNEVENT_CONNECT_FAIL, errno);
/*		c->protops->connerr(c->uldat, errno);*/
		return;
	}
	socketprep(s);
	c->fd = s;

	peerpar_settime(c->peer, PEERPAR_PEER_LASTCONNECT);
	xgettimeofday(&c->connstart, NULL);
	dbg("%s: connect start: %s", f, c->peer->id);
	if (connect(c->fd, res->ai_addr, res->ai_addrlen) == 0) {
		conntime = tv_dur2ms(&c->connstart);
		peer_addrtt(c->peer, conntime);
		dbg("%s: to %s in %0.2f s, immediate", f, c->peer->id,
		    conntime);
		if (conncomp(c) == 0)
			c->protops->connevent(c, CONNEVENT_CONNECT_OK, 0);
/*			c->protops->conndone(c->uldat);*/
		return;
	}
	if (errno == EINPROGRESS) {
		dbg("%s: connect in progress: %s", f, c->peer->id);
		changecstatus(c, CONNECTION_STATUS_CONNECTING);
	} else {
		swarn("%s: connect: %s: %m", f, c->peer->id);
		c->protops->connevent(c, CONNEVENT_CONNECT_FAIL, errno);
/*		c->protops->connerr(c->uldat, errno);*/
		changecstatus(c, CONNECTION_STATUS_ENDED);
	}
}


/* conncomp - complete connection establishment attempt */
int
conncomp(struct connection *c)
{
	const char *f = "conncomp";

	/* successful connect, reset failedconn count */
	peerpar_setint(c->peer, PEERPAR_PEER_FAILEDCONN, 0);

	if (conn_getmss(c) == -1 ||
	    conn_getsockname(c) == -1 ||
	    conn_getpeername(c) == -1) {
		changecstatus(c, CONNECTION_STATUS_ENDED);
		return -1;
	}

	dbg("%s: new connection completed: %s:%s (%s)", f, c->peername,
	    c->peerport, c->sockname);

	changecstatus(c, CONNECTION_STATUS_WRITEREQ);
	xgettimeofday(&c->reqstart, NULL);

	return 0;
}


/* conn_getsockname - obtain address/name of local end of connection */
int
conn_getsockname(struct connection *c)
{
	const char *f = "conn_getsockname";
	int n;

	/*
	 * XXX get ip address of local end of connection, used when we
	 * refer to ourself, or need to recognize ourself in data
	 * received from the other end of the connection.
	 */
	c->ss_local_len = sizeof(c->ss_local);
	if (getsockname(c->fd, (struct sockaddr *)&c->ss_local,
			&c->ss_local_len) == -1) {
		swarn("%s: getsockname: %m", f);
		return -1;
	}
	if ((n = getnameinfo((struct sockaddr *)&c->ss_local, c->ss_local_len,
			     c->sockname, sizeof(c->sockname),
			     c->sockport, sizeof(c->sockport),
			     NI_NUMERICHOST|NI_NUMERICSERV)) != 0) {
		swarn("%s: getnameinfo: %s", f, gai_strerror(n));
		ERRX("getsockname getnameinfo failure");
	}

	return 0;
}


/* conn_getpeername - obtain address/name of remote end of connection */
int
conn_getpeername(struct connection *c)
{
	const char *f = "conn_getpeername";
	int n;

	ASSERT(strlen(c->peername) == 0, "peername already set");

	/* get name of client */
	c->ss_remote_len = sizeof(c->ss_remote);
	if (getpeername(c->fd, (struct sockaddr *)&c->ss_remote,
			&c->ss_remote_len) == -1) {
		swarn("%s: getpeername: %m", f);
		return -1;
	}
	if ((n = getnameinfo((struct sockaddr *)&c->ss_remote, c->ss_remote_len,
			c->peername, sizeof(c->peername),
			c->peerport, sizeof(c->peerport),
			NI_NUMERICHOST|NI_NUMERICSERV)) != 0) {
		swarn("%s: getnameinfo: %s", f, gai_strerror(n));
		ERRX("getpeername getnameinfo failure");
	}

	return 0;
}


/* conn_getmss - set max segment size value for connection */
int
conn_getmss(struct connection *c)
{
	const char *f = "conn_getmss";
	socklen_t optlen;

	optlen = sizeof(c->mss);
	if (getsockopt(c->fd, IPPROTO_TCP, TCP_MAXSEG, &c->mss, &optlen) == -1) {
		swarn("%s: getsockopt: %m", f);
		return -1;
	}
	/* mss is per link, but set in peer as, 'most recent mss value' */
	if (c->peer != NULL)
		peerpar_setint(c->peer, PEERPAR_LINK_MSS, c->mss);

	return 0;
}
