/* connio.c - i/o buffering functions, primarily for protocol text (non-data) */

/*
 * Copyright (c) 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: connio.c,v 1.17 2010/06/30 08:37:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <unistd.h>

#include <errno.h>
#include <string.h>

#include "log.h"
#include "conn.h"
#include "iostat.h"
#include "uri.h"
#include "ssl.h"


/* ibufread - call read_conn, then copy requested bytelen from ibuf */
ssize_t
ibufread(struct connection *c, void *buf, size_t nbytes)
{
	int len;

	/* read new data */
	if (read_conn(c) == -1)
		return -1;

	/* enough data available? */
	if (c->ibuflen < (int)nbytes)
		len = c->ibuflen;
	else
		len = nbytes;

	/* copy data */
	bcopy(c->ibuf, buf, len);

	/* full buffer copied? */
	if (len == c->ibuflen)
		c->ibuflen = 0;
	else
		conn_ibufsync(c, c->ibuf + len);

	return len;
}


/* read_conn - fill connection ibuf, returns 0 on success, -1 on error */
int
read_conn(struct connection *c)
{
	const char *f = "read_conn";
	size_t left;
	ssize_t rres;

	/* calculate length value to supply to read() */
	left = sizeof(c->ibuf) - c->ibuflen - 1; /* remaining buffer space */
	ASSERT(left > 0, "zero read length");

#if HAVE_OPENSSL
	if (c->ssl != NULL) {
/*		ASSERT(c->ssl_cmd == SSLCMD_NONE, "ssl command set");*/
		ASSERT(c->status != CONNECTION_STATUS_DEADLINESTREAM &&
		       c->status != CONNECTION_STATUS_FOREVERSTREAM,
		       "ssl data io");
		rres = ssl_read(c, c->ibuf + c->ibuflen, left);
	} else
#endif /* HAVE_OPENSSL */
		rres = read(c->fd, c->ibuf + c->ibuflen, left);

	if (rres > 0) {
		if (c->status == CONNECTION_STATUS_DEADLINESTREAM ||
		    c->status == CONNECTION_STATUS_FOREVERSTREAM)
			ios.net_rdat += rres;
		else
			ios.net_roh += rres;
		c->ibuflen += rres;
		ASSERT(c->ibuflen < (int)sizeof(c->ibuf), "too much data");
		c->ibuf[c->ibuflen] = '\0';

/*		dbg("%s: read %ld bytes (%d/%d)", f, (long long)rres,
		    c->ibuflen, c->ibufexplen);*/

		return 0;
	}

	/* nothing was read, handle error */
	if (rres == -1) {
		if (errno == EAGAIN) {
			if (c->ssl == NULL)
				dbg("%s: EAGAIN", f);
			return 0; /* not yet finished, but not errror */
		}

		swarn("read_conn: %m");
	} else if (rres == 0)
		if (c->status != CONNECTION_STATUS_IDLE)
			dbg("read_conn: EOF (%d unread)", c->ibufexplen);

	c->ibuflen = 0; /*XXX? */
	changecstatus(c, CONNECTION_STATUS_ENDED);

	return -1;
}


/* write_obuf - write nonwritten data in connection obuf */
int
write_obuf(struct connection *c)
{
	const char *f = "write_obuf";
	ssize_t wres;
	size_t left;

	ASSERT(c->obuflen != 0, "nothing to write");

/*	dbg("%s: fd %d", f, c->fd);*/

	left = (size_t)(c->obuflen - c->obufoff);

	/*
	 * XXX note: this approach will likely lead to reply and start
	 * of data being in different packets. alternative would be
	 * to have obuf handled in stream_send().
	 */

#if HAVE_OPENSSL
	if (c->ssl != NULL) {
		ASSERT(c->ssl_cmd == SSLCMD_NONE, "ssl command set");
		ASSERT(c->status != CONNECTION_STATUS_DEADLINESTREAM &&
		       c->status != CONNECTION_STATUS_FOREVERSTREAM,
		       "ssl data io");
		wres = ssl_write(c, c->obuf + c->obufoff, left);
	} else
#endif /* HAVE_OPENSSL */
		wres = write(c->fd, c->obuf + c->obufoff, left);

	if (wres > 0) {
		c->obufoff += wres;
		ios.net_woh += wres;
/*		dbg("%s: %ld/%d", (long long)wres, c->obuflen);*/

		if (c->obufoff != c->obuflen) {
			errno = EAGAIN;
			dbg( "%s: %m", f);
			return -1; /* not yet finished */
		}

/*		dbg("%s: finished");*/
		/* data in obuf written, finished */
		c->obufoff = 0;
		c->obuflen = 0;
		return 0;
	}

	if (wres == -1) {
		if (errno == EAGAIN)
			return -1; /* not yet finished */

		swarn("%s: write_obuf: %m", f);
	} else
		ERRX("unexpected write return value");

	c->obuflen = 0;
	changecstatus(c, CONNECTION_STATUS_ENDED);

	return -1;
}


/* conn_ibufsyn - move data at offset 'newstart' to the start of ibuf buffer */
void
conn_ibufsync(struct connection *c, char *newstart)
{
	const char *f = "conn_ibufsync";
	size_t len;

	/* move any additional data to start of buffer */
	if (newstart < c->ibuf + c->ibuflen) {
		len = c->ibuflen - (newstart - c->ibuf);
		dbg("%s: data remaining: %ld bytes after text", f, (long)len);
		bcopy(newstart, c->ibuf, len);
		c->ibuflen = len;
		c->ibuf[c->ibuflen] = '\0';
/*		dbg("%s: rem: %s", f, c->ibuf);*/
/*		dbg("%s: rem: %x %x %x %x %x", f, *p, *(p + 1), *(p + 2),
		    *(p + 3), *(p + 4));*/
	} else
		c->ibuflen = 0;
}


/* ibuf_nlnlterm - check whether an empty line has been read */
int
ibuf_nlnlterm(struct connection *c)
{
	/* multiple commands might have been read previously */
	if (c->ibuflen > 0) {
		ASSERT((size_t)c->ibuflen < sizeof(c->ibuf),
		       "ibuflen too long");
		ASSERT(c->ibuf[c->ibuflen] == '\0', "not NUL terminated");
		/* request already complete? */
		if (strstr(c->ibuf, "\r\n\r\n") != NULL) {
			return 0;
		}
	}

	return -1; /* not yet finished */
}


/* ibuf_explen - check whether expected data length is available */
int
ibuf_explen(struct connection *c)
{
	ASSERT(c->ibufexplen > 0, "nonpositive ibufexplen");
	ASSERT(c->ibufexplen <= (int)sizeof(c->ibuf), "ibufexplen too long");
	ASSERT(c->ibuf[c->ibuflen] == '\0', "data not NUL terminated");

	if (c->ibufexplen <= c->ibuflen) {
		c->ibufexplen = 0;
		return 0; /* finished */
	}

	return -1;
}


/* conn_writerep - write message to peer */
void
conn_writerep(struct connection *c, char *msg)
{
	ASSERT(c->obuflen == 0, "unwritten data");

	if (strlcpy(c->obuf, msg, sizeof(c->obuf)) >= sizeof(c->obuf))
		ERRX("message too long for output buffer");
	c->obuflen = strlen(c->obuf);
}
