/* debug.c - debug information on connections */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: debug.c,v 1.36 2010/06/25 22:02:26 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>

#include <stdio.h>
#include <string.h>

#include "log.h"
#include "block.h"
#include "file.h"
#include "filemeta.h"
#include "conn.h"
#include "peer.h"
#include "stream.h"
#include "urient.h"

extern int active_streams; /*XXX*/

struct peer *activesource(struct connection *);


/* connstatus - show status of all connections */
void
connstatus(void)
{
	char status[URI_MAXLEN * 2], tmp[100]; /*XXXlen*/
	struct connection *l_c;
	struct peer *peer;
	size_t statlen;
	int i;

	dbg("active streams: %d", active_streams);

	i = 0;
	statlen = sizeof(status);
	SLIST_FOREACH(l_c, &connections, n) {
		snprintf(tmp, sizeof(tmp), " %d ", i);
		strlcpy(status, tmp, statlen);

		strlcat(status, conn_state2str(l_c->status), statlen);
		strlcat(status, " ", statlen);

		if (l_c->noreuse)
			strlcat(status, "S ", statlen);

		snprintf(tmp, sizeof(tmp), " (%d)", l_c->pollidx);
		strlcat(status, tmp, statlen);
		strlcat(status, " ", statlen);

		snprintf(tmp, sizeof(tmp), " i/obuflen: %d/%d", l_c->ibuflen,
			 l_c->obuflen);
		strlcat(status, tmp, statlen);

		snprintf(tmp, sizeof(tmp), " (age: %d)", l_c->age);
		strlcat(status, tmp, statlen);
		strlcat(status, " ", statlen);

		if (l_c->peer != NULL) {
			strlcat(status, l_c->peer->name, statlen);
			strlcat(status, ":", statlen);
			strlcat(status, l_c->peer->port, statlen);
		}

		if (l_c->stream != NULL) {
			strlcat(status, " -", statlen);
			strlcat(status, stream_type2str(l_c->stream->type),
				statlen);

			snprintf(tmp, sizeof(tmp), " (%d, %d) writercnt: %d ",
				 l_c->stream->bnum, l_c->stream->blockoff,
				 l_c->stream->ue->file->blocks[l_c->stream->bnum]->writercnt);
			strlcat(status, tmp, statlen);
			if (strlcat(status, l_c->stream->ue->uri.uri, statlen) >= statlen)
				ERRX("string truncation"); /*XXX*/
		}

		if ((l_c->stream != NULL) &&
		    (l_c->stream->type == STREAM_FROM_MEM)) {
			snprintf(tmp, sizeof(tmp), " avail: %lld",
				 (long long)stream_availbytes(l_c->stream));
			if (strlcat(status, tmp, statlen) >= statlen)
				ERRX("string truncation"); /*XXX*/

			peer = activesource(l_c);
			if (peer != NULL) {
				snprintf(tmp, sizeof(tmp), " source: %s",
					 peer->name);
				if (strlcat(status, tmp, statlen) >= statlen)
					ERRX("string truncation"); /*XXX*/
			}
		}

		/*
		 * XXX have ulstate2string func which allows upper
		 *     layer to return a string describing current
		 *     state?
		 */

#if HAVE_OPENSSL
		if (l_c->ssl_active)
			strlcat(status, " SSL", statlen);
#endif /* HAVE_OPENSSL */

		dbg("%s", status);

		i++;
	}
}


/* activesource - returns peer currently sending first noncomplete block */
struct peer *
activesource(struct connection *c)
{
	struct block *block;
	int32_t bnum;

	bnum = c->stream->bnum;

	block = NULL;
	while (bnum < c->stream->ue->meta->blockcnt) {
		block = c->stream->ue->file->blocks[bnum];

		if ((urient_haveblock(c->stream->ue, bnum) == 0) &&
		    (block->actlen != block->blocklen))
			break;

		bnum++;
	}

	if (block != NULL && block->writercnt > 0)
		return block->source;
	else
		return NULL;
}
