/* digest.c - message digest handling, wrapper around openssl/native variants */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: digest.c,v 1.26 2010/10/11 09:07:47 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "log.h"
#include "digest.h"

char digest_default[DIGEST_NAMELEN];

#if !DIGEST_OPENSSL

/* digest_preinit - setup digest handling environment */
void
digest_preinit(char *mdname)
{
	const char *f = "digest_preinit";
	size_t maxlen;

	maxlen = sizeof(digest_default);
	if (mdname == NULL) {
		if (strlcpy(digest_default, DIGEST_DEFAULT, maxlen) >= maxlen)
			ERRX("DIGEST_DEFAULT too long");
	} else
		if (strlcpy(digest_default, mdname, maxlen) >= maxlen)
			serr("digest name too long: '%s'", mdname);

	if (!digest_validtype(digest_default))
		serr("unsupported digest: %s", digest_default);

	dbg2("%s: using md5/rmd160 digests", f);
	dbg("%s: default digest: %s", f, digest_default);
}


/* digest_init - init digest of type mdname */
struct digest *
digest_init(char *mdname)
{
	const char *f = "digest_init";
	struct digest *md;

	md = xmalloc(sizeof(struct digest));
	if (strcmp(mdname, DIGEST_MD5) == 0) {
		md->md5ctx = xmalloc(sizeof(MD5_CTX));
		MD5Init(md->md5ctx);
		md->type = md_md5;
	} else if (strcmp(mdname, DIGEST_RMD160) == 0) {
		md->rmd160ctx = xmalloc(sizeof(RMD160_CTX));
		RMD160Init(md->rmd160ctx);
		md->type = md_rmd160;
	} else {
		swarn("%s: unsupported digest: %s", f, mdname);
		free(md);
		return NULL;
	}

	return md;
}


/* digest_validtype - return true if mdname is a supported digest */
int
digest_validtype(char *mdname)
{
	if (strcmp(mdname, DIGEST_MD5) == 0 ||
	    strcmp(mdname, DIGEST_RMD160) == 0)
		return 1;
	else
		return 0;
}


/* digest_mdlen - return length (including NUL) of digest with given name */
int
digest_mdlen(char *mdname)
{
	const char *f = "digest_mdlen";

	if (strcmp(mdname, DIGEST_MD5) == 0) {
		return MD5_DIGEST_STRING_LENGTH;
	} else if (strcmp(mdname, DIGEST_RMD160) == 0) {
		return RMD160_DIGEST_STRING_LENGTH;
	} else {
		dbg("%s: unsupported digest: %s", f, mdname);
		return -1;
	}
}


/* digest_update - digest with data of specified length */
void
digest_update(struct digest *md, const char *data, size_t length)
{
	switch(md->type) {
	case md_md5:
		MD5Update(md->md5ctx, (const uint8_t *)data, length);
		break;
	case md_rmd160:
		RMD160Update(md->rmd160ctx, (const uint8_t *)data, length);
		break;
	default:
		ERRX("unexpected digest name type");
	}
}


/* digest_end - generate and return final digest */
char *
digest_end(struct digest *md)
{
	char *res;

	switch(md->type) {
	case md_md5:
		res = MD5End(md->md5ctx, NULL);
		break;
	case md_rmd160:
		res = RMD160End(md->rmd160ctx, NULL);
		break;
	default:
		ERRX("unexpected digest type");
	}

	return res;
}


/* digest_free - release digest resources */
void
digest_free(struct digest *md)
{
	/* free ctx here, errors might lead to digest_end not being called */
	switch(md->type) {
	case md_md5:
		free(md->md5ctx);
		break;
	case md_rmd160:
		free(md->rmd160ctx);
		break;
	default:
		ERRX("unexpected digest type");
	}
	free(md);
}

#else /* DIGEST_OPENSSL */

#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>

/* digest_preinit - setup digest handling environment */
void
digest_preinit(char *mdname)
{
	const char *f = "digest_preinit";
	size_t maxlen;

	/* use DIGEST_DEFAULT as default unless mdname is specified */
	maxlen = sizeof(digest_default);
	if (mdname == NULL) {
		if (strlcpy(digest_default, DIGEST_DEFAULT, maxlen) >= maxlen)
			ERRX("DIGEST_DEFAULT too long");
	} else
		if (strlcpy(digest_default, mdname, maxlen) >= maxlen)
			serr("digest name too long: '%s'", mdname);

	OpenSSL_add_all_digests();
	if (!digest_validtype(digest_default))
		serr("unsupported digest: %s", digest_default);

	dbg2("%s: using OpenSSL digests", f);
	dbg("%s: default digest: %s", f, digest_default);
}


/* digest_init - init digest of type mdname */
struct digest *
digest_init(char *mdname)
{
	const char *f = "digest_init";
	const EVP_MD *evpmd;
	struct digest *md;

	md = xmalloc(sizeof(struct digest));
	md->evpctx = EVP_MD_CTX_create();
	ASSERT(md->evpctx != NULL, "NULL expctx");
	if ((evpmd = EVP_get_digestbyname(mdname)) == NULL) {
		dbg2("%s: EVP_get_digestbyname failed (%s)", f, mdname);
		swarn("%s: unsupported digest: %s", f, mdname);
		free(md);
		return NULL;
	}
	if (EVP_DigestInit_ex(md->evpctx, evpmd, NULL) == 0)
		ERRX("EVP_DigestUpdate failed");
	md->len = digest_mdlen(mdname);

	return md;
}


/* digest_validtype - return true if mdname is a supported digest */
int
digest_validtype(char *mdname)
{
	if (EVP_get_digestbyname(mdname) != NULL)
		return 1;
	else
		return 0;
}


/* digest_mdlen - return length (including NUL) of digest with given name */
int
digest_mdlen(char *mdname)
{
	const char *f = "digest_preinit";
	const EVP_MD *evpmd;

	if ((evpmd = EVP_get_digestbyname(mdname)) == NULL) {
		dbg2("%s: EVP_get_digestbyname failed (%s)", f, mdname);
		swarn("%s: unsupported digest: %s", f, mdname);
		return -1;
	}

	return EVP_MD_size(evpmd) * 2 + 1;
}


/* digest_update - digest with data of specified length */
void
digest_update(struct digest *md, const char *data, size_t length)
{
	if (EVP_DigestUpdate(md->evpctx, data, length) == 0)
		ERRX("EVP_DigestUpdate failed");
}


/* digest_end - generate and return final digest */
char *
digest_end(struct digest *md)
{
	unsigned char md_value[EVP_MAX_MD_SIZE];
	unsigned int md_len;
	unsigned int i;
	char *sum;
	int off;

	if (EVP_DigestFinal_ex(md->evpctx, md_value, &md_len) == 0)
		ERRX("EVP_DigestFinal_ex failed");
	ASSERT(md_len <= sizeof(md_value), "md_value too short");

	sum = xmalloc(md->len);
	for (i = 0; i < md_len; i++) {
		off = i * 2;
		ASSERT(off + 2 < md->len, "offset too high");
		snprintf(sum + off, 3, "%02x", md_value[i]);
	}

	return sum;
}


/* digest_free - release digest resources */
void
digest_free(struct digest *md)
{
	/* free ctx here, errors might lead to digest_end not being called */
	EVP_MD_CTX_destroy(md->evpctx);

	free(md);
}

#endif /* DIGEST_OPENSSL */
