/* diskcache.c - best effort disk cache management */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: diskcache.c,v 1.57 2010/10/06 08:17:24 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/stat.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "log.h"
#include "alist.h"
#include "bitfield.h"
#include "block.h"
#include "blocksum.h"
#include "callback.h"
#include "case.h"
#include "file.h"
#include "filemeta.h"
#include "cachefile.h"
#include "diskcache.h"
#include "diskio.h"
#include "sumfile.h"
#include "uri.h"
#include "urient.h"

#define DISKCACHE_MAX 10 /* 1 GB */
#define DISKCACHE_PARAMUNIT (100 * 1024 * 1024) /* 100 MB */

#define DISKCACHE_MAX_ENTRIES 1024

#define NEWFILE_MAXRETRY 3

struct diskcache diskcache;

char *cachedir = NULL;

/*
 * XXX could use actual block size when updating used rather than file
 * blocksize (would only make a difference for last block in file).
 */


/* diskcache_preinit - initial diskcache setup */
void
diskcache_preinit(unsigned int cachemax, char *dir)
{
	const char *f = "diskcache_preinit";

	ASSERT(dir != NULL, "cachedir not specified");
	cachedir = dir;

	diskcache.used = 0;
	if (cachemax != 0)
		diskcache.max = cachemax * DISKCACHE_PARAMUNIT;
	else
		diskcache.max = DISKCACHE_MAX * DISKCACHE_PARAMUNIT;
	diskcache.commit = 0;

	sinfo("%s: diskcache in '%s', size: %lld bytes (%lld MB)", f,
	      cachedir, (long long)diskcache.max,
	      (long long)diskcache.max / (1024 * 1024));
}


/* diskcache_newfile - create new file on disk */
int
diskcache_newfile(struct urient *ue)
{
	const char *f = "diskcache_newfile";
	char path[PATH_MAX];
	size_t maxlen;
	int retries;
	char *gen;

/*	dbg("%s: %s %lld", f, ue->uri.path, ue->meta->size);*/

	retries = 0;
retry:
	/* generate possible filename based on urient */
	gen = cachefile_namegen(ue);

	/* build path relative to cachedir to test/create generated filename */
	maxlen = sizeof(path);
	strlcpy(path, cachedir, maxlen);
	strlcat(path, "/", maxlen);
	if (strlcat(path, gen, maxlen) >= maxlen) {
		dbg("%s: path too long", f);
		return -1;
	}

	/* attempt to create */
	if (cachefile_create(path, ue->meta->size) == -1) {
		if (retries++ < NEWFILE_MAXRETRY)
			goto retry;
		else {
			swarn("%s: unable to create cachefile", f);
			return -1; /* give up */
		}
	}

	if (urient_setfile(ue, cachedir, gen) == -1)
		return -1;

	/* XXX choose to deny based on size/cache full? */
	diskcache.commit += ue->meta->size;

	return 0;
}


/* diskcache_expire - cachefile expiry (LRU); delete files from disk cache */
void
diskcache_expire(void)
{
	const char *f = "diskcache_expire";
	struct urient *ue, *oldue;
	time_t accesstime;
	int i;

	/* all files should have an access time lower than this */
	accesstime = time(NULL) + 1;

	oldue = NULL;
	/* look for inactive urient with lowest access time */
	for (i = 0; (size_t)i < ALIST_SIZE(al_urient); i++) {
		ue = ALIST_PVAL(al_urient, i);
		/* nothing cached on disk, skip */
		if (ue->cf == NULL)
			continue;
		/* active file, skip */
		if (ue->file != NULL)
			continue;

		/* older? */
		if (ue->lastreq < accesstime) {
			accesstime = ue->lastreq;
			oldue = ue;
		}
		ue = NULL;
	}
	if (oldue == NULL) {
		dbg("%s: no expirable file found", f);
		return;
	}

	dbg("%s: expiring %s (%s)", f, oldue->filename, oldue->meta->uri->uri);
	for (i = 0; i < oldue->cf->bm.size; i++) {
		if (blockmap_test(&oldue->cf->bm, i))
			diskcache_lostblock(oldue, i);
	}
	ASSERT(diskcache.used >= 0, "diskcache.used negative");

	/* delete from disk */
	diskcache_delfile(oldue);

	/* free */
	cachefile_free(oldue->cf);
	oldue->cf = NULL;
}


/* diskcache_delfile - delete cachefile (and sumfile) */
void
diskcache_delfile(struct urient *ue)
{
	const char *f = "diskcache_delfile";
	char path[PATH_MAX];
	size_t maxlen;
	char *sumpath;

	ASSERT(ue->cf != NULL, "deleting NULL cf cachefile");

	/* build path relative to cachedir to delete file */
	maxlen = sizeof(path);
	strlcpy(path, ue->prefix, maxlen);
	strlcat(path, "/", maxlen);
	if (strlcat(path, ue->filename, maxlen) >= maxlen) {
		dbg("%s: path too long", f);
		return;
	}

	if (unlink(path) == -1)
		swarn("%s: unlink: %s: %m", f, path);
	if ((sumpath = sumfile_namegen(ue->meta, path)) != NULL)
		if (unlink(sumpath) == -1)
			swarn("%s: unlink: %s: %m", f, sumpath);

	diskcache.commit -= ue->meta->size;
}


/* diskcache_clean - delete all disk cache files (to be used before exit) */
void
diskcache_clean(void)
{
	const char *f = "diskcache_clean";
	struct urient *ue;
	int i, j;

	for (i = 0; (size_t)i < ALIST_SIZE(al_urient); i++) {
		ue = ALIST_PVAL(al_urient, i);
		if (ue->cf == NULL)
			continue;
		/* XXX skip active files? ignore if only used at exit */
		if (ue->file != NULL)
			swarn("%s: deleting active file", f);

		dbg("%s: deleting %s: %s", f, ue->meta->uri->uri, ue->filename);
		for (j = 0; j < ue->cf->bm.size; j++)
			if (blockmap_test(&ue->cf->bm, j))
				diskcache_lostblock(ue, j);
		ASSERT(diskcache.used >= 0, "diskcache.used negative");

		diskcache_delfile(ue);
		cachefile_free(ue->cf);
		ue->cf = NULL;
	}
}


/* diskcache_sizeleft - return size remaining in disk cache */
off_t
diskcache_sizeleft(void)
{
	return diskcache.max - diskcache.used;
}


/* diskcache_gotblock - notification of block download completion */
void
diskcache_gotblock(struct urient *ue, int32_t bnum)
{
	const char *f = "diskcache_gotblock";

	/* XXX file possibly deleted if file NULL? */
	ASSERT(ue->file != NULL, "NULL file");

#if 0 /* expiry now responsibility of caller */
	/* cache full? */
	if (diskcache.used + ue->meta->blocksize >= diskcache.max) {
		sarn("%s: disk cache full, initiating expiry", f);
		diskcache_expire();
	}
#endif

	/* do actual write to disk */
	if (disk_blockwrite(ue, ue->file->blocks[bnum], bnum) == -1)
		/*
		 * write failed, do not mark as cached.
		 * will only be kept in memory.
		 */
		swarn("%s: block not written to to cache", f);
	else {
		/*
		 * with wrapping download of partial blocks, checksum
		 * calculation is done simplest when the entire block is
		 * complete.
		 */
		digest_blocksum(ue, bnum);

		/* update cache info */
		blockmap_set(&ue->cf->bm, bnum);
		diskcache.used += ue->meta->blocksize;
	}

	ASSERT(ue->state == URIENT_STATE_META ||
	       ue->state == URIENT_STATE_INCOMPLETE,
	       "unexpected ue state");
	if (ue->state == URIENT_STATE_META)
		ue->state = URIENT_STATE_INCOMPLETE;
	/* has file been completed?*/
	if (urient_iscomp(ue)) {
		ASSERT(ue->state == URIENT_STATE_INCOMPLETE, "invalid state");
		ue->state = URIENT_STATE_COMPLETE;
	}
}


/* diskcache_lostblock - block no longer found on disk */
void
diskcache_lostblock(struct urient *ue, int32_t bnum)
{
	const char *f = "diskcache_lostblock";

	ASSERT(blockmap_test(&ue->cf->bm, bnum), "lost block not present");

	dbg("%s: %s %d", f, ue->uri.uri, bnum);
	blockmap_clear(&ue->cf->bm, bnum);
	diskcache.used -= ue->meta->blocksize;
}


/* diskcache_loadidx - load index file, update diskcache and cfdb */
void
diskcache_loadidx(void)
{
	const char *f = "diskcache_loadidx";
	char buf1[CACHEIDX_MAXLINE], buf2[CACHEIDX_MAXLINE],
	     buf3[CACHEIDX_MAXLINE], buf4[CACHEIDX_MAXLINE];
	char *l1, *l2, *l3, *l4;
	struct callback *cb;
	struct urient *ue;
	struct uri uri;
	char *filep;
	FILE *fp;
	char *p;
	int i;

	/*
	 * Information about all files in the disk cache is kept in
	 * memory (as opposed to only loading this information when
	 * needed).
	 */

	dbg("%s", f);

	if (access(CACHEIDX_NAME, R_OK) == -1 && errno == ENOENT) {
		dbg("%s: no diskcache index", f);
		return;
	}

	if ((fp = fopen(CACHEIDX_NAME, "r")) == NULL) {
		swarn("%s: %s: %m", f, CACHEIDX_NAME);
		return;
	}

	/* check header */
	if (fgets(buf1, sizeof(buf1), fp) == NULL) {
		swarn("%s: short file", f);
		fclose(fp);
		return;
	}
	if ((p = strchr(buf1, '\n')) == NULL) {
		swarn("%s: too short file", f);
		fclose(fp);
		return;
	}
	*p = '\0';
	if (strcmp(CACHEIDX_HDR, buf1) != 0) {
		swarn("%s: unexpected header", f);
		fclose(fp);
		return;
	}

	/*
	 * XXX two things can happen:
	 *     -cache contents can change; files might have been deleted
	 *      manually after the disk cache index was written
	 *     -the disk cache index might contain syntax errors,
	 *      either because it was modified/manually generated,
	 *      or because of version mismatch between index file/server
	 *      (or due to parsing bug)
	 */
	for (;;) {
		/* read first line */
		if (fgets(buf1, sizeof(buf1), fp) == NULL)
			break; /* EOF */
		/* read remaining lines */
		if (fgets(buf2, sizeof(buf2), fp) == NULL ||
		    fgets(buf3, sizeof(buf3), fp) == NULL ||
		    fgets(buf4, sizeof(buf4), fp) == NULL) {
			swarn("%s: short read", f);
			break;
		}

		/*
		 * replace trailing newlines with NUL
		 */

		if ((p = strchr(buf1, '\n')) == NULL) {
			swarn("%s: invalid format", f);
			break;
		}
		*p = '\0';

		if ((p = strchr(buf2, '\n')) == NULL) {
			swarn("%s: invalid format", f);
			break;
		}
		*p = '\0';

		if ((p = strchr(buf3, '\n')) == NULL) {
			swarn("%s: invalid format", f);
			break;
		}
		*p = '\0';

		if ((p = strchr(buf4, '\n')) == NULL) {
			swarn("%s: invalid format", f);
			break;
		}
		*p = '\0';

		/*
		 * locate beginning of data lines
		 * (all but first line should start with whitespace)
		 */

		l1 = buf1;

		l2 = buf2;
		if (*l2 != ' ') {
			swarn("%s: invalid format", f);
			break;
		}
		l2++;

		l3 = buf3;
		if (*l3 != ' ') {
			swarn("%s: invalid format", f);
			break;
		}
		l3++;

		l4 = buf4;
		if (*l4 != ' ') {
			swarn("%s: invalid format", f);
			break;
		}
		l4++;

		/*
		 * start decoding data
		 */

		if ((p = strchr(l1, '\t')) == NULL) {
			swarn("%s: invalid format", f);
			break;
		}
		*p = '\0';

		filep = p + 1;

		/* uri */
		if (uri_parse(&uri, l1) == -1) {
			swarn("%s: invalid URI", f);
			break;
		}
		/*
		 * verify that port number is set
		 * (no port number only accepted in user input)
		 */
		if (uri.port[0] == '\0') {
			swarn("%s: uri without port: %s", f, uri.uri);
			ERRX("invalid database");
			continue;
		}
		if ((ue = urient_add(&uri)) == NULL)
			break;

		/* check filename */
#if 1 /* likely expensive if cache contains many files */
		/*XXX add prefix */
		if (access(filep, R_OK|W_OK) != 0) {
			swarn("%s: cachefile access failed", f);
			urient_del(ue);
			continue;
		}
#endif

		if (urient_setfile(ue, cachedir, filep) == -1) {
			urient_del(ue);
			break;
		}

		/* second line, urient data */
		if (urient_decstr(ue, l2) == -1) {
			urient_del(ue);
			break;
		}

		/* third line, filemeta data (uri needs to be from ue) */
		if ((ue->meta = filemeta_decstr(&ue->uri, l3)) == NULL) {
			swarn("%s: invalid filemeta", f);
			urient_del(ue);
			break;
		}
		if ((ue->cf = cachefile_new(ue->meta)) == NULL) {
			urient_del(ue);
			break;
		}

		/* decode fourth line, block status */
		if (bitfield_decstr(&ue->cf->bm, l4) == -1) {
			urient_del(ue);
			break;
		}

		/* XXX verify cache file and contents if requested */

		/* update cache usage */
		/*XXX overflow */
		diskcache.commit += ue->meta->size;
		for (i = 0; i < ue->cf->bm.size; i++)
			if (blockmap_test(&ue->cf->bm, i))
				diskcache.used += ue->meta->blocksize;
		if (urient_iscomp(ue))
			ue->state = URIENT_STATE_COMPLETE;
		else
			ue->state = URIENT_STATE_INCOMPLETE;

		/*
		 * XXX how to handle situation where on disk data exceeds
		 *     current maximum cache setting?
		 *     add all entries found in cache, then call expiry
		 *     if needed
		 */

		if ((cb = callback_find(ucase(ue->uri.scheme))) == NULL)
			ERRX("no handler found for scheme");
		ue->uldat = cb->new_urient(ue);

		/* file contents ok, init urient state */
		if (ue->ueops != NULL)
			ue->ueops->cacheinit(ue->uldat);
	}
	fclose(fp);

	diskcache_status(0);
}


/* diskcache_saveidx - write list of all files currently cached */
void
diskcache_saveidx(void)
{
	const char *f = "diskcache_saveidx";
	char *ue_s, *fm_s, *bf_s;
	struct urient *ue;
	FILE *fp;
	int i;

	dbg("%s", f);

	if ((fp = fopen(CACHEIDX_NAME, "w")) == NULL) {
		swarn("%s: %s: %m", f, CACHEIDX_NAME);
		return;
	}

	/* write header */
	if (fprintf(fp, "%s\n", CACHEIDX_HDR) == -1) {
		swarn("%s: fprintf failure, aborting", f);
		fclose(fp);
		return;
	}

	for (i = 0; (size_t)i < ALIST_SIZE(al_urient); i++) {
		ue = ALIST_PVAL(al_urient, i);

		/* nothing written to disk? */
		if (ue->cf == NULL)
			continue;

		if ((ue_s = urient_encstr(ue)) == NULL) {
			swarn("%s: urient encoding failure for %s",
			     f, ue->meta->uri->uri);
			continue;
		}
		if ((fm_s = filemeta_encstr(ue->meta)) == NULL) {
			swarn("%s: filemeta encoding failure for %s",
			     f, ue->meta->uri->uri);
			continue;
		}
		if ((bf_s = bitfield_encstr(&ue->cf->bm)) == NULL) {
			/* can be NULL if no data is stored on disk */
			swarn("%s: NULL cachefile encoding for %s",
			     f, ue->meta->uri->uri);
			continue;
		}
		/* preceede second and third line with space for readability */
		if (fprintf(fp, "%s\t%s\n %s\n %s\n %s\n", ue->uri.uri,
			    ue->filename, ue_s, fm_s, bf_s) == -1) {
			swarn("%s: fprintf failure, aborting", f);
			break;
		}
	}

	fclose(fp);
}


/* diskcache_status - show diskcache status information */
void
diskcache_status(int verbose)
{
	const char *f = "diskcache_status";
	struct urient *ue;
	int i;

	sinfo("%s: used: %0.3f MB max: %0.3f MB", f,
	      (float)diskcache.used / (1024 * 1024),
	      (float)diskcache.max / (1024 * 1024));

	if (verbose)
		for (i = 0; (size_t)i < ALIST_SIZE(al_urient); i++) {
			ue = ALIST_PVAL(al_urient, i);
			dbg(" CF %s\t%ld", ue->filename, (long)ue->lastreq);
		}
}
