/* diskio.c - disk related input/output */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: diskio.c,v 1.28 2010/06/30 08:37:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>

#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "log.h"
#include "block.h"
#include "diskcache.h"
#include "file.h"
#include "diskio.h"
#include "iostat.h"
#include "urient.h"

/*
 * the lack of symmetry between the functions (read is partial block,
 * write is full block) is due to the desire to avoid writing data to
 * disk for blocks with incorrect checksum.
 */


/* disk_blockread - read rlen more of block data from disk */
void
disk_blockread(struct urient *ue, struct block *block, int32_t bnum,
	       size_t rlen)
{
/*	const char *f = "disk_blockread";*/
	off_t startoff;
	ssize_t rres;

	ASSERT(urient_haveblock(ue, bnum), "block not cached");
	ASSERT(block->blockdat != NULL, "no blockdat");

#if 0
	if (stream->blockoff == 0)
		dbg("%s: reading %d bytes of %d,%d from disk", f, rlen,
		    stream->bnum, block->actlen);
#endif

	startoff = block2off(ue, bnum) + block->actlen;
	if ((rres = pread(ue->file->fd, block->blockdat + block->actlen, rlen,
			  startoff)) != (ssize_t)rlen) {
		/* XXX error handling */
		if (rres == 0)
			serr("EOF during read (%ld)", (long)rlen);
		else if (rres == -1)
			serr("read: %m");
		else
			serr("short read: %ld != %ld", (long)rres, (long)rlen);
	}
	block->actlen += (uint32_t)rres;
	ios.disk_rdat += rres;

	ASSERT(block->actlen <= block->blocklen, "invalid block");
}


/* disk_blockwrite - save a complete block to disk */
int
disk_blockwrite(struct urient *ue, struct block *block, int32_t bnum)
{
	const char *f = "disk_blockwrite";
	ssize_t wres, wlen;
	off_t wtot, offset;

	ASSERT(ue->file != NULL, "NULL file");
	ASSERT(block->actlen == block->blocklen, "block incomplete");

	dbg("%s: %s block %d actlen: %d blocklen: %d", f, ue->uri.uri,
	    bnum, block->actlen, block->blocklen);

	wtot = 0;
	offset = block2off(ue, bnum);
	while (wtot < block->blocklen) {
		wlen = MIN(ue->file->opt_blksize, block->blocklen - wtot);
		if ((wres = pwrite(ue->file->fd, block->blockdat + wtot,
				   (size_t)wlen, offset + wtot)) != wlen) {
			swarn("%s: write failure block %d: %ld != %ld: %m", f,
			      bnum, (long)wres, (long)wlen);
			if (wres > 0)
				ios.disk_wdat += wres;
			return -1;
		}
		ios.disk_wdat += wres;
		wtot += wres;
	}

	dbg("%s: %s block %d complete, storing to disk complete", f,
	    ue->uri.uri, bnum);

	return 0;
}
