/* file.c - file structure management functions */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: file.c,v 1.30 2010/06/30 08:37:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include "log.h"
#include "alist.h"
#include "block.h"
#include "cachefile.h"
#include "file.h"
#include "filemeta.h"
#include "memcache.h"
#include "servfile.h"
#include "sumfile.h"
#include "urient.h"

#define FILE_MIN_IOSIZE 1024 /* minimum size used for IO operations */


/* file_open - create new file structure, or return existing one */
struct file *
file_open(struct urient *ue)
{
	const char *f = "file_open";
	char path[MAXPATHLEN];
	struct file *file;
	struct stat sb;
	size_t maxlen;

	ASSERT(strlen(ue->filename) != 0, "ue filename not set");

	dbg("%s: %s", f, ue->meta->uri->uri);

	file = xmalloc(sizeof(struct file));
	bzero(file, sizeof(struct file));

	file->fd = -1;

	file->blocks = xcalloc(ue->meta->blockcnt, sizeof(struct block *));

	/* build file path */
	maxlen = sizeof(path);
	strlcpy(path, ue->prefix, maxlen);
	strlcat(path, "/", maxlen);
	if (strlcat(path, ue->filename, maxlen) >= maxlen) {
		swarn("%s: path too long", f);
		free(file->blocks);
		free(file);
		return NULL;
	}

	/*XXX check for file existence first? */
	if (ue->master) {
		if ((file->fd = open(path, O_RDONLY, 0)) == -1) {
			swarn("%s: open: %s: %m", f, path);
			free(file->blocks);
			free(file);
			return NULL;
		}

		if ((file->suf = sumfile_open(ue->meta,
					      sumfile_namegen(ue->meta, path),
					      serversumfiles ?
					      o_rdwr : o_rdwr)) == NULL)
			swarn("%s: not using sum file", f);
	} else {
		/* require both read and write for cache files */
		if ((file->fd = open(path, O_RDWR, 0)) == -1) {
			swarn("%s: open: %s: %m", f, path);
			free(file->blocks);
			free(file);
			return NULL;
		}

		if ((file->suf = sumfile_open(ue->meta,
					      sumfile_namegen(ue->meta, path),
					      o_rdwr)) == NULL)
			swarn("%s: not using sum file", f);
	}

	/* get recommended io size */
	if (fstat(file->fd, &sb) == -1)
		serr("fstat: %m");
	ASSERT(ue->meta->size == sb.st_size, "invalid file size");
	if (sb.st_blksize < FILE_MIN_IOSIZE)
		file->opt_blksize = FILE_MIN_IOSIZE;
	else
		file->opt_blksize = sb.st_blksize;
/*	dbg("%s: IO buffer size for file: %d", f, file->opt_blksize);*/

	return file;
}


/* file_blockinit - initialize block data for urient file */
void
file_blockinit(struct urient *ue)
{
	const char *f = "file_blockinit";
	int i;

	for (i = 0; i < ue->meta->blockcnt; i++) {
		ue->file->blocks[i] = block_init(ue, i);
		if (ue->master)
			ue->file->blocks[i]->sum =
				sumfile_read(ue->file->suf, i);
		else if (blockmap_test(&ue->cf->bm, i)) {
			if ((ue->file->blocks[i]->sum =
			     sumfile_read(ue->file->suf, i)) == NULL) {
				dbg("%s: no digest for file in diskmap, "
				    "treated as not present", f);
				urient_lostblock(ue, i);
			}
		}
	}
}


/* file_close - close file and free all file resources */
void
file_close(struct urient *ue)
{
	const char *f = "file_close";
	int i;

	ASSERT(ue->file->refcnt == 0, "close on file with refcnt");

	dbg("%s: %s", f, ue->meta->uri->uri);

	xclose(ue->file->fd);
	sumfile_close(ue->file->suf);

	memcache_flush(ue);

	for (i = 0; i < ue->meta->blockcnt; i++)
		block_free(ue->file->blocks[i]);

	free(ue->file->blocks);
	free(ue->file);
	ue->file = NULL;
}


/* file_status - list all active files */
void
file_status(void)
{
	struct urient *ue;
	int i;

	dbg("active files:");
	/*XXX inefficient, no file-only list */
	  for (i = 0; (size_t)i < ALIST_SIZE(al_urient); i++) {
		  if ((ue = ALIST_PVAL(al_urient, i)) == NULL)
			  continue;
		  if (ue->file == NULL)
			  continue;

		  dbg(" FILE %s (%s)", ue->filename, ue->uri.uri);
	  }
}
