/* filemeta.c - in memory database of URI metadata, including peer block data */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: filemeta.c,v 1.46 2010/06/29 14:03:25 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>

#include "log.h"
#include "alist.h"
#include "block.h"
#include "filemeta.h"
#include "peerset.h"
#include "range.h"
#include "uri.h"

#define FILEMETA_ENCSTR_MAX (URI_MAXLEN + 100)

struct peerset *fm_mirrors = NULL;

/*
 * XXX should be possible to use pointer of struct peer as id for many
 * of alist users that index struct peers
 */

/*
 * the filemeta structure is referenced from urient structure, which
 * is assumed to exist before and after the filemeta structure (which
 * has a pointer to the struct uri in the urient structure).
 */


/* filemeta_addserv - for server mode, adds address of mirror/caches */
int
filemeta_addserv(char *serv, char *port)
{
	const char *f = "peer_addserv";
	struct peer *peer;

	if ((peer = peer_add(serv, port)) == NULL) {
		swarn("%s: unable to add mirror server", f);
		return -1;
	}

	dbg("%s: adding mirror for all served files: %s", f, peer->id);
	if (fm_mirrors == NULL)
		fm_mirrors = pset_create();

	pset_addpeer(fm_mirrors, peer);

	return 0;
}


/* filemeta_new - create new filemeta structure */
struct filemeta *
filemeta_new(struct uri *uri, off_t size, int blocksize, char *digest)
{
	const char *f = "filemeta_new";
	struct filemeta *fm;
	size_t maxlen;
	int i;

	ASSERT(uri != NULL, "NULL uri pointer");

/*	dbg("%s: %s %lld %d %s", f, uri->uri, (long long)size, blocksize,
	    digest);*/

	/*XXX any point in testing for zero sized files? */
	/* zero byte files are likely an indication of error, disallow */
	if (size == 0 || size > FILESIZE_MAX) {
		swarn("%s: invalid file size %lld", f, (long long)size);
		return NULL;
	}

	/* block size should be checked by caller */
	ASSERT(block_validsize(blocksize), "invalid blocksize");

	if (!digest_validtype(digest)) {
		swarn("%s: unsupported digest: %s", f, digest);
		return NULL;
	}

	fm = xmalloc(sizeof(struct filemeta));
	bzero(fm, sizeof(struct filemeta));

	/* filemeta is freed by urient_free, don't update uri refcnt */
	fm->uri = uri;

	/* set meta data */
	fm->size = size;
	fm->blocksize = blocksize;
	fm->blockcnt = filemeta_size2blockcnt(size, blocksize);
	ASSERT(size < fm->blocksize * (fm->blockcnt + 1), "blockcnt error");
	/* XXX fixed size rather than growing? */
	fm->al_peers = alist_init((void (*)(const void *))filemeta_printpeer,
				  filemeta_freepeer, 1,
				  ALIST_SIZE_GROW, ALIST_TYPE_STR);

	maxlen = sizeof(fm->digest);
	if (strlcpy(fm->digest, digest, maxlen) >= maxlen)
		ERRX("length of supported digest too long");

	dbg("%s: %s %lld %d (%d blocks) %s", f, fm->uri->uri,
	    (long long)fm->size, fm->blocksize, fm->blockcnt, fm->digest);

	/* add any mirrors */
	if (fm_mirrors)
		for (i = 0; i < fm_mirrors->peercnt; i++)
			filemeta_peerupd(fm, pset_peer(fm_mirrors, i), 0,
					 fm->blockcnt - 1, "MIRROR");

	return fm;
}


/* filemeta_free - free filemeta resources */
void
filemeta_free(struct filemeta *fm)
{
	alist_free(fm->al_peers);
	free(fm);
}


/* filemeta_newpeer - allocate new fm_peer element */
struct fm_peer *
filemeta_newpeer(struct peer *peer, int blockcnt)
{
	const char *f = "filemeta_newpeer";
	struct fm_peer *fmp;

	ASSERT(blockcnt > 0, "invalid blockcnt");

	dbg("%s: %s (%d blocks)", f, peer->id, blockcnt);

	fmp = xmalloc(sizeof(*fmp));
	bzero(fmp, sizeof(*fmp));
	fmp->peer = peer;
	blockmap_init(&fmp->bm, blockcnt);

	/* increase peer reference count */
	peer_addref(peer);

	return fmp;
}


/* filemeta_freepeer - release fm_peer resources */
void
filemeta_freepeer(void *data)
{
	struct fm_peer *fmp;

	fmp = data;
	peer_unref(fmp->peer);
	blockmap_free(&fmp->bm);
	free(fmp);
}


/* filemeta_peerupd - update credit information on file for given peer */
void
filemeta_peerupd(struct filemeta *fm, struct peer *peer, int32_t bnum,
		 int32_t seq, char *source)
{
	const char *f = "filemeta_peerupd";
	struct fm_peer *fmp;
	int32_t i;
	void *res;
	int new;

	dbg("%s: for %s, block %d + %d blocks", f, peer->id, bnum, seq);

	if ((fmp = alist_find(fm->al_peers, peer->id)) == NULL) {
		fmp = filemeta_newpeer(peer, fm->blockcnt);
		res = alist_add(fm->al_peers, peer->id, fmp);
		/* alist should grow, and never fail */
		ASSERT(res != NULL, "alist_add failed");
	}
	ASSERT(fmp != NULL, "NULL fm_peer");

	ASSERT(bnum + seq < fm->blockcnt, "bnum/seq combination too large");

	new = 0;
	for (i = bnum; i <= bnum + seq; i++)
		if (blockmap_test(&fmp->bm, i) == 0) {
			blockmap_set(&fmp->bm, i);
			new++;
		}

	dbg("blockinfo update for %s %d new blocks source: %s", peer->id, new,
	    source);

/*	filemeta_dump(fm);*/
}


/* filemeta_peerseq - count the number of subsequent blocks at peer */
int32_t
filemeta_peerseq(struct filemeta *fm, int32_t bnum, struct peer *peer)
{
	struct fm_peer *fmp;

	fmp = alist_find(fm->al_peers, peer->id);
	ASSERT(fmp != NULL, "unable to locate fm_peer");

	return blockmap_subseq(&fmp->bm, bnum);
}


/* filemeta_delpeer - delete peer from filemeta alist */
void
filemeta_delpeer(struct filemeta *fm, struct peer *peer)
{
	/* alist_del calls filemeta_freepeer */
	alist_del(fm->al_peers, peer->id);
}


/* filemeta_printpeer - fm_peer entry print function */
void
filemeta_printpeer(const struct fm_peer *fmp)
{
	dbg(" %s: %s", fmp->peer->id, range_strgen(&fmp->bm));
}


/* filemeta_dump - dump filemeta contents */
void
filemeta_dump(struct filemeta *fm)
{
	dbg("%s:", fm->uri->uri);
	dbg(" (%lld bytes, %d blocks of %d bytes)", (long long)fm->size,
	    fm->blockcnt, fm->blocksize);
	alist_dump(fm->al_peers);
}


/* filemeta_size2blockcnt - return number of blocks in file of given size */
int32_t
filemeta_size2blockcnt(off_t filesize, int blocksize)
{
	int32_t res;

	if (filesize == 0)
		return 0;

	res = filesize/blocksize;
	if (blocksize * res != filesize)
		res++;
	/*
	 * the size is off_t and count int32_t, so overflow is possible,
	 * but should not be a problem in practice.
	 * Maximum size for int32_t is 2147483647 (2 GB).
	 * A block size of 512 bytes would allow a file of
	 * 1TB to be used.
	 *
	 * The assert below should catch any cases of overflow.
	 */
	ASSERT(filesize <= blocksize * res, "incorrect fileblock calculation");

	return res;
}


/* filemeta_encstr - return pointer to (static) string with filemeta data */
char *
filemeta_encstr(struct filemeta *fm)
{
	static char str[FILEMETA_ENCSTR_MAX];
	int l;

	/*
	 * The peer list is not encoded. The encoding function is used
	 * by the diskcache, and the information on peers is transient
	 * data that is best obtained from the parent node when
	 * needed.
	 *
	 * the remaining fields are stored, separated by the tab character:
	 *
	 *  uri, file size, block size, digest
	 *
	 */

	l = snprintf(str, sizeof(str), "%lld\t%d\t%s", (long long)fm->size,
		     fm->blocksize, fm->digest);
	ASSERT((l != -1) && ((size_t)l < sizeof(str)), "buffer too short");

	return str;
}


/* filemeta_decstr - decode string with meta data and return filemeta struct */
struct filemeta *
filemeta_decstr(struct uri *uri, char *str)
{
	const char *f = "filemeta_decstr";
	char *sizep, *blocksizep, *digestp;
	const char *err;
	int blocksize;
	off_t size;
	char *p;

/*	off_t (size)\t int (blocksize)\t char * (digest) */

	/* identify elements */
	sizep = str;
	if ((p = strchr(str, '\t')) == NULL) {
		swarn("%s: invalid format", f);
		return NULL;
	}
	*p = '\0';
	p++;
	blocksizep = p;

	if ((p = strchr(p, '\t')) == NULL) {
		swarn("%s: invalid format", f);
		return NULL;
	}
	*p = '\0';
	p++;
	digestp = p;

	if (strchr(p, '\t') != NULL) {
		swarn("%s: too many fields", f);
		return NULL;
	}

	/* ensure that values are sensible */
	if (strlen(sizep) == 0 || strlen(blocksizep) == 0 ||
	    strlen(digestp) == 0) {
		swarn("%s: invalid format", f);
		return NULL;
	}

	/*
	 * decode strings
	 */

	/* don't bother with size checking here, is done in filemeta_new */
	size = strtonum(sizep, 0, SIZE_MAX, &err);
	if (err) {
		swarn("%s: invalid file size %s: %s", f, sizep, err);
		return NULL;
	}

	blocksize = strtonum(blocksizep, 0, INT_MAX, &err);
	if (err) {
		swarn("%s: invalid block size %s: %s", f, blocksizep, err);
		return NULL;
	}

	return filemeta_new(uri, size, blocksize, digestp);
}
