/* io.c - file and network input/output operations */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: io.c,v 1.82 2010/10/12 08:21:36 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>
#if HAVE_SYS_SENDFILE_H
#include <sys/sendfile.h>
#endif /* HAVE_SYS_SENDFILE_H */
#include <sys/uio.h>

#include <errno.h>
#include <fcntl.h> /* used by open() in debug code */
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "log.h"
#include "block.h"
#include "file.h"
#include "conn.h"
#include "diskio.h"
#include "io.h"
#include "stream.h"
#include "urient.h"

ssize_t block_send_sendfile(struct stream *, struct block *, size_t);
void verify_read(struct stream *, struct block *, size_t);


/* block_send - transmit block data over network */
ssize_t
block_send(struct stream *stream, struct block *block, size_t wlen)
{
	const char *f = "block_send";
	ssize_t wres;
	size_t rlen;

#if HAVE_SENDFILE
	/* sendfile only needed if file is stored on disk */
	if (stream->sendfile)
		return block_send_sendfile(stream, block, wlen);
#else
	ASSERT(stream->sendfile == 0, "sendfile wanted but not supported");
#endif /* HAVE_SENDFILE */

	if (urient_haveblock(stream->ue, stream->bnum)) {
		/* read from disk */
		/*XXX read more than requests? might be needed later */
		ASSERT(stream->blockoff <= block->blocklen, "negative rlen");
		ASSERT(block->actlen <= block->blocklen, "negative rlen");
		rlen = MIN(stream->readsize, block->blocklen - block->actlen);
		while ((rlen > 0) &&
		       (stream->blockoff + wlen > block->actlen)) {
			disk_blockread(stream->ue, block, stream->bnum, rlen);
			rlen = MIN(stream->readsize,
				   block->blocklen - block->actlen);
		}
	} else
		if (stream->blockoff == 0)
			dbg("%s: using in memory block data from %d,%d", f,
			    stream->bnum, stream->blockoff);

#if 0
	verify_read(stream, block, wlen);
#endif

	if ((wres = write(stream->dfd, block->blockdat + stream->blockoff,
		     (size_t)wlen)) == -1)
		swarn("%s: write: %m", f);

#if 0
	if (wres > 0)
		verify_read(stream, block, (size_t)wres);
#endif

	return wres;
}


#if HAVE_SENDFILE
/* block_send_sendfile - send using sendfile */
ssize_t
block_send_sendfile(struct stream *stream, struct block *block, size_t wlen)
{
	const char *f = "block_send_sendfile";
	ssize_t wres;
	off_t offset;

	ASSERT(block->blockdat == NULL, "sendfile but blockdat set");
	ASSERT(urient_haveblock(stream->ue, stream->bnum), "block not cached");

	if (stream->blockoff == 0)
		dbg("%s: sending %d bytes from block %d,%d via sendfile", f,
		    (int)wlen, stream->bnum, stream->blockoff);

	offset = stream_curroff(stream);
	wres = sendfile(stream->dfd, stream->ue->file->fd, &offset, wlen);
	/* sendfile, no ios.disk_rdat update */
	if (wres == -1)
		dbg("%s: write: %m", f);

	return wres;
}
#endif /* HAVE_SENDFILE */


#if 1
/* verify_read - debug function */
void
verify_read(struct stream *stream, struct block *block, size_t wlen)
{
	const char *f = "verify_read";
	char testbuf[NETWORK_IOSIZE];
	unsigned int i;
	off_t offset;
	ssize_t rres;
	int fd;

	offset = stream_curroff(stream);

	if ((fd = open("/home/media/DuckandC1951.mpg", O_RDONLY, 0)) == -1)
		ERRX("verification file");
	dbg("%s: from %lld, wlen %u", f, (long long)offset, (unsigned int)wlen);
	if (lseek(fd, offset, SEEK_SET) == -1)
		serr("lseek: %m");
	if ((rres = read(fd, testbuf, wlen)) != (ssize_t)wlen) {
		if (rres == 0)
			serr("EOF during read (%ld)", (long)wlen);
		else if (rres == -1)
			serr("read: %m");
		else
			serr("short read: %ld != %ld", (long)rres, (long)wlen);
	}
	xclose(fd);
	if (memcmp(testbuf, block->blockdat + stream->blockoff, wlen) != 0) {
		dbg("wlen %ld, blockoff %d actlen: %d blocklen: %d off: %lld",
		    (long)wlen, stream->blockoff, block->actlen,
		    block->blocklen, (long long)offset);

		for (i = 0; i < wlen; i++)
			dbg("%d 0x%X 0x%X", i, testbuf[i],
			    *(block->blockdat + stream->blockoff + i));
		ERRX("data mismatch");
	}
}
#endif
