/* ipc.c - inter process communication functions */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: ipc.c,v 1.22 2010/10/19 13:58:12 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>

#include "log.h"
#include "ipc.h"

struct addrinfo_hdr {
	uint8_t entries; /* number of struct addrinfo entries */
	int code; /* return value from getaddrinfo */
};


/* senddesc - send descriptor fd via s */
int
senddesc(int s, int fd)
{
	const char *f = "senddesc";
	/* based on OpenBSD CMSG_DATA manpage */
	struct msghdr msg;
	struct cmsghdr *cmsg;
	union {
		struct cmsghdr hdr;
		unsigned char buf[CMSG_SPACE(sizeof(int))];
	} cmsgbuf;
	struct iovec iov[1];
	unsigned char x;

	bzero(&msg, sizeof(msg));
	msg.msg_control = &cmsgbuf.buf;
	msg.msg_controllen = sizeof(cmsgbuf.buf);

	bzero(&cmsgbuf, sizeof(cmsgbuf));
	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_len = CMSG_LEN(sizeof(int));
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_RIGHTS;
	*(int *)CMSG_DATA(cmsg) = fd;

	/*
	 * solaris and linux does not appear to like passing
	 * descriptors without data; send dummy byte along with descriptor.
	 */
	x = 'X';
	iov[0].iov_base = &x;
	iov[0].iov_len = sizeof(x);
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;

	if (sendmsg(s, &msg, 0) == -1) {
		swarn("%s: sendmsg: %m", f);
		return -1;
	}

	return 0;
}


/* recvdesc - receive descriptor via s, returns desc or -1 on failure */
int
recvdesc(int s)
{
	const char *f = "recvdesc";
	/* based on OpenBSD CMSG_DATA manpage */
	struct msghdr msg;
	struct cmsghdr *cmsg;
	union {
		struct cmsghdr hdr;
		unsigned char buf[CMSG_SPACE(sizeof(int))];
	} cmsgbuf;
	int fd;
	struct iovec iov[1];
	unsigned char b;

	bzero(&msg, sizeof(msg));
	msg.msg_control = &cmsgbuf.buf;
	msg.msg_controllen = sizeof(cmsgbuf.buf);

	iov[0].iov_base = &b;
	iov[0].iov_len = sizeof(b);
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;

	if (recvmsg(s, &msg, MSG_WAITALL) == -1) {
		swarn("%s: recvmsg: %m", f);
		return -1;
	}
	ASSERT(b == 'X', "unexpected data followed descriptor");
	dbg("recvdesc returned");
	if ((msg.msg_flags & MSG_TRUNC) || (msg.msg_flags & MSG_CTRUNC))
		ERRX("truncated message");
	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg != NULL;
	     cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_len == CMSG_LEN(sizeof(int)) &&
		    cmsg->cmsg_level == SOL_SOCKET &&
		    cmsg->cmsg_type == SCM_RIGHTS) {
/*			dbg("got desc");*/
			fd = *(int *)CMSG_DATA(cmsg);
			return fd;
		}
	}

	return -1;
}


/* sendaddrinfo - send struct addrinfo over descriptor */
int
sendaddrinfo(int fd, struct addrinfo *addr, int code)
{
	struct addrinfo rescopy, *tres;
	struct addrinfo_hdr hdr;
	int i;

	if (code != 0) {
		/* lookup failed, return error value only */
		ASSERT(addr == NULL, "non-NULL addrinfo on error");
		hdr.entries = 0;
		hdr.code = code;
	} else {
		ASSERT(addr != NULL, "NULL addrinfo on non-error");
		/* count entires */
		i = 0;
		tres = addr;
		do {
			i++;
		} while ((tres = tres->ai_next) != NULL);
		hdr.entries = i;
		hdr.code = 0;
	}

	/* write reply header */
	if (writen(fd, &hdr, sizeof(hdr)) == -1)
		return -1;

	/* send addrinfo, if lookup was successful */
	if (hdr.entries > 0) {
		tres = addr;
		do {
			/*
			 * make some slight changes to addrinfo before sending:
			 * -send sockaddr after addrinfo
			 * -ignore ai_canoninfo
			 * -set ai_next to NULL (client must rebuild)
			 */
			bcopy(tres, &rescopy, sizeof(rescopy));
			rescopy.ai_addr = NULL;
			rescopy.ai_canonname = NULL;
			rescopy.ai_next = NULL;

			/* write rescopy, followed by sockaddr */
			if (writen(fd, &rescopy, sizeof(rescopy)) == -1)
				return -1;

			if (writen(fd, tres->ai_addr, tres->ai_addrlen) == -1)
				return -1;
		} while (tres != NULL && (tres = tres->ai_next) != NULL);
	}

	return 0;
}


/* readaddrinfo - read and reconstruct struct addrinfo from descriptor */
int
readaddrinfo(int fd, struct addrinfo **resp, int *code)
{
	struct addrinfo **rp, *res, *res0;
	struct addrinfo_hdr rep;
	int i;

	if (readn(fd, &rep, sizeof(rep)) == -1)
		return -1;

	if (rep.entries == 0) {
		/* resolve failed, finished */
		ASSERT(rep.code != 0, "no reply entries but no error");
		*code = rep.code;
		*resp = NULL;
		return 0;
	}
	ASSERT(rep.code == 0, "reply entries but error");
	*code = rep.code;

	res0 = NULL;
	rp = &res0;
	for (i = 0; i < rep.entries; i++) {
		/* allocate and reconstruct addrinfo structure */
		res = xmalloc(sizeof(struct addrinfo));
		res->ai_next = NULL;
		if (readn(fd, res, sizeof(struct addrinfo)) == -1) {
			freeaddrinfo(res0);
			free(res);
			return -1;
		}

#ifdef __APPLE__
		res->ai_addr = xmalloc(res->ai_addrlen);
#else
		/* use realloc to all data can be freed with one free() */
		/* XXX does ai_addrlen reflect the full structure size? */
		res = xrealloc(res, sizeof(struct addrinfo) + res->ai_addrlen);
		res->ai_addr = (struct sockaddr *)(res + 1);
#endif /* __APPLE__ */
		if (readn(fd, res->ai_addr, res->ai_addrlen) == -1) {
			freeaddrinfo(res0);
			free(res);
			return -1;
		}

		*rp = res;
		rp = &res->ai_next;
	}
	/* set ai_next of last entry to NULL */
	*rp = NULL;

	/* set addrinfo argument value */
	*resp = res0;

	return 0;
}


/*XXX readn/writen should return zero on eof */

/* writen - write until nbytes written, or failure */
ssize_t
writen(int fd, void *buf, size_t nbytes)
{
	const char *f = "writen";
	ssize_t res;
	size_t cnt;

	cnt = 0;
	res = 0;
	while (cnt < nbytes) {
		res = write(fd, (char *)buf + cnt, nbytes);
		if (res == 0) {
			dbg("%s: short write, eof", f);
			return -1;
		}
		if (res == -1 && errno == EAGAIN)
			continue;
		if (res == -1) {
			dbg("%s: write: %m", f);
			return -1;
		}
		cnt += res;
	}

	return res;
}


/* readn - read until nbytes read, or failure */
ssize_t
readn(int fd, void *buf, size_t nbytes)
{
	const char *f = "readn";
	ssize_t res;
	size_t cnt;

	cnt = 0;
	res = 0;
	while (cnt < nbytes) {
		res = read(fd, (char *)buf + cnt, nbytes);
		if (res == 0) {
			dbg("%s: short read, eof after %lld bytes read", f,
			    (long long)cnt);
			return -1;
		}
		if (res == -1 && errno == EAGAIN)
			continue;
		if (res == -1) {
			dbg("%s: read: %m", f);
			return -1;
		}
		cnt += res;
	}

	return res;
}
