/* log.c - system logging to stdout or syslog */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: log.c,v 1.33 2010/10/11 09:07:47 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <errno.h>
#include <stdarg.h>
#include <syslog.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "log.h"

int debug = 0; /* 0 - no debug, 1 - application debug, 2 - code debug */
int logsyslog = 0; /* log using syslog */
int logdatepid = 1; /* disable date in logmessages, used by testing framework */
int abortonerr = 1; /* call abort in ERR, zero value used testing framework */
int logerrline = 1; /* log line in addition to file where ERR occurs */

/* the name of the application is added to logoutput and set with log_preinit */
#define MAX_APPNAME 100
char appname[MAX_APPNAME] = "app_notset";

FILE *logfp = NULL; /* log to stderr by default */

void vslog(int, const char *, va_list);


/* log_preinit - set application name and choose log to stderr/syslog */
int
log_preinit(char *app, enum logtype logtype, char *file)
{
	if (strlcpy(appname, app, sizeof(appname)) >= sizeof(appname))
		abort(); /* application name too long */

	switch (logtype) {
		case log_syslog:
			openlog(app, LOG_PID, LOG_USER);
			logfp = NULL;
			logsyslog = 1;
			break;

		case log_stdout:
			logfp = stdout;
			break;

		case log_stderr:
			logfp = stderr;
			break;

		case log_file:
			if ((logfp = fopen(file, "a")) == NULL) {
				fprintf(stderr, "fopen: %s: %s", file,
					strerror(errno));
				return -1;
			}
			break;
	}

	return 0;
}


/* vslog - write logmessage, either via syslog or to stderr */
void
vslog(int pri, const char *fmt, va_list args)
{
	char fmtcpy[1024];
	char buf2[1024];
	char buf[1024];
	char str[10];
	size_t cpylen;
	struct tm *tm;
	time_t now;
	int i, j;

	if (!logsyslog && logfp == NULL)
		abort(); /* log_preinit not called */

	if (logsyslog)
		vsyslog(pri, fmt, args);
	else {
		ASSERT(strlen(fmt) + 1 <= sizeof(fmtcpy), "fmt too long");

		buf[0] = '\0';

		if (logdatepid) {
			now = time(NULL);
			tm = localtime(&now);
			if (strftime(buf, sizeof(buf), "%s %b %d %T %Y", tm) == 0)
				abort(); /* abort to avoid loop */
			(void)strlcat(buf, " ", sizeof(buf));
		}
		(void)strlcat(buf, appname, sizeof(buf));
		if (logdatepid) { /* pid */
			snprintf(str, sizeof(str), "[%d]", getpid());
			strlcat(buf, str, sizeof(buf));
		}
		(void)strlcat(buf, ": ", sizeof(buf));

		if (pri == LOG_ERR)
			(void)strlcat(buf, "error: ", sizeof(buf));
		else if (pri == LOG_WARNING)
			(void)strlcat(buf, "warning: ", sizeof(buf));

		/* replace any %m in fmt with strerror(errno) */
		cpylen = sizeof(fmtcpy);
		bzero(fmtcpy, cpylen);
		j = 0;
		for (i = 0; fmt[i] != '\0' && (size_t)j < cpylen; ) {
			if (fmt[i] == '%' && fmt[i + 1] == 'm') {
				/* ignore trunctaion in log message */
				strlcat(fmtcpy, strerror(errno), cpylen);
				fmtcpy[sizeof(fmtcpy) - 1] = '\0';

				j = strlen(fmtcpy);
				i += 2;
			} else {
				fmtcpy[j] = fmt[i];
				j++;
				i++;
			}
		}

		(void)vsnprintf(buf2, sizeof(buf2), fmtcpy, args);
		(void)strlcat(buf, buf2, sizeof(buf));
		/* ignore return code; message truncated if too long */
		(void)strlcat(buf, "\n", sizeof(buf));

		fprintf(logfp, "%s", buf);
	}
}


/* serr - vslog wrapper, exits after printing log message */
void
serr(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vslog(LOG_ERR, fmt, ap);
	va_end(ap);

	exit(1);
	/* NOTREACHED */
}


/* slog - vslog wrapper, simply prints log message */
void
slog(int pri, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vslog(pri, fmt, ap);
	va_end(ap);
}


/* dbg - debugging messages (application level debugging) */
void
dbg(const char *fmt, ...)
{
	va_list ap;

	if (!debug)
		return;

	va_start(ap, fmt);
	vslog(LOG_DEBUG, fmt, ap);
	va_end(ap);
}


/* dbg2 - verbose debug messages (code level debugging) */
void
dbg2(const char *fmt, ...)
{
	va_list ap;

	if (debug < 2)
		return;

	va_start(ap, fmt);
	vslog(LOG_DEBUG, fmt, ap);
	va_end(ap);
}


/* swarn - logging of messages relevant for user/system operator */
void
swarn(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vslog(LOG_WARNING, fmt, ap);
	va_end(ap);
}


/* sinfo - logging of informational messages */
void
sinfo(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vslog(LOG_INFO, fmt, ap);
	va_end(ap);
}


/* xmalloc - malloc or die */
void *
xmalloc(size_t size)
{
	void *res;

	if ((res = malloc(size)) == NULL)
		serr("malloc");

	return res;
}


/* xcalloc - calloc or die */
void *
xcalloc(size_t nmemb, size_t size)
{
	void *res;

	if ((res = calloc(nmemb, size)) == NULL)
		serr("calloc");

	return res;
}


/* xrealloc - realloc or die */
void *
xrealloc(void *ptr, size_t size)
{
	void *res;

	if ((res = realloc(ptr, size)) == NULL)
		serr("realloc");

	return res;
}


/* xstrdup - strdup or die */
char *
xstrdup(const char *s)
{
	char *res;
	if ((res = strdup(s)) == NULL)
		serr("strdup");

	return res;
}


/* xgettimeofday - get time or die */
int
xgettimeofday(struct timeval *tp, struct timezone *tzp)
{
	int res;

	if ((res = gettimeofday(tp, tzp)) == -1)
		serr("gettimeofday");

	return res;
}


/* xclose - close, should never fail */
void
xclose(int d)
{
	if (close(d) == -1)
		ERR("close");
}
