/* mconn.c - initiate and monitor multiple parallel connections */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: mconn.c,v 1.10 2010/06/25 22:02:27 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <poll.h>
#include <string.h>
#include <unistd.h>

#include "log.h"
#include "ipc.h"
#include "mconn.h"
#include "socket.h"
#include "tv.h"

void mconn_err(void);

/*
 * this file holds the code (to be used with the proc interface) for
 * connecting to a set of addresses in parallel.
 *
 * the implementation is fairly simple, allowing a struct addrinfo
 * (containing a list of addresses) to be submitted over a descriptor.
 * the server process reads the list, calls connect() for each
 * one. succeeding descriptors are sent back as soon as possible,
 * errors are sent in one batch for all remaining descriptors upon
 * connect timeout or when there are no descriptors that have not
 * failed or been returned. the connect time is measured and returned
 * if the connect operation is successful.
 *
 * the main limitation is that only one parallel connection operation
 * can be active at a time. multiple processes can however be used to
 * perform this task, which ideally should not be needed often enough
 * for this to be a problem. having multiple requests active at any
 * time would complicate code both here and in the callers.
 */

/*
 * XXX have timeout as a request parameter
 */
#define MCONN_TIMEOUT 2 /* max connect time in seconds */


/* mconn_serv - proc child process mconnect server */
void
mconn_serv(int fd)
{
	const char *f = "mconn_serv";
	struct mconn_req mreq;
	int n;

	for (;;) {
		dbg2("%s: starting doing mconn_loop", f);
		/* read query */
		if (mconn_readreq(fd, &mreq) == -1)
			break;
		dbg2("%s: got mconn request", f);

		/* attempt to connect */
		if ((n = mconn_setupfds(fd, &mreq)) == -1)
			break;
		else if (n > 0)
			/* poll remaining descriptors */
			if (mconn_pollfds(fd, &mreq) == -1)
				break;

		/* nope, free resources and send any errors */
		freeaddrinfo(mreq.addrs[0]);
		if (mconn_senderrs(fd, &mreq) == -1)
			break;
	}

	swarn("%s: request processing error, mconn server ending", f);

	/* caller is assumed to call _exit (so no cleanup done) */
}


/* mconn_sendreq - send addrinfo struct as request to mconn server */
int
mconn_sendreq(int fd, struct addrinfo *addr)
{
	return sendaddrinfo(fd, addr, 0);
}


/* mconn_readreq - read request (addrinfo struct) from fd */
int
mconn_readreq(int fd, struct mconn_req *mreq)
{
	const char *f = "mconn_readreq";
	struct addrinfo *res, *tres;
	int code;

	/* chain of addrinfo structures expected */
	if (readaddrinfo(fd, &res, &code) == -1)
		return -1;

	/* read as request, code should be zero, entries larger than zero */
	ASSERT(code == 0, "invalid request");

	bzero(mreq, sizeof(*mreq));
	/* place pointers to each entry in mreq */
	tres = res;
	do {
		mreq->addrs[mreq->cnt] = tres;
		mreq->cnt++;
	} while (tres != NULL && (tres = tres->ai_next) != NULL &&
		 mreq->cnt < MCONNMAXCONN);
	/* max limit reached? */
	if (tres != NULL) {
		/*XXX could alternatively return error */
		swarn("%s: too many entries in request, aborting", f);
		return -1;
	}

	if (res == NULL || mreq->cnt == 0) {
		swarn("%s: invalid request", f);
		return -1;
	}

	return 0;
}


/* mconn_setupfds - call connect, return number of in-progress descriptors */
int
mconn_setupfds(int fd, struct mconn_req *mreq)
{
	const char *f = "mconn_setupfds";
	int r;
	int s;
	int i;

	/* initiate connect(), return descriptor on immediate success */
	r = 0;
	for (i = 0; i < mreq->cnt; i++) {
		/* default to -1, for 'close not needed' */
		mreq->fds[i] = -1;
		if ((s = socket(mreq->addrs[i]->ai_family,
				mreq->addrs[i]->ai_socktype,
				mreq->addrs[i]->ai_protocol)) == -1) {
			swarn("%s: socket: %s", f, strerror(errno));
			mreq->errs[i] = errno;
			continue;
		}
		if (socketprep(s) == -1) {
			mreq->errs[i] = errno;
			xclose(s);
			continue;
		}

		xgettimeofday(&mreq->starttimes[i], NULL);
		errno = 0;
		if (connect(s, mreq->addrs[i]->ai_addr,
			    mreq->addrs[i]->ai_addrlen) == 0) {
			/* immediate success */
			dbg("%s: immediate connect success", f);
			mreq->conntimes[i] = tv_dur2ms(&mreq->starttimes[i]);
			mreq->fds[i] = s;
			if (mconn_sendrep(fd, mreq, i) == -1)
				mconn_err();
			continue;
		}
		if (errno != EINPROGRESS) {
			dbg("%s: immediate connect failure", f);
			mreq->errs[i] = errno;
			xclose(s);
			continue;
		}

		/* connect in progress, need to monitor fd */
		mreq->fds[i] = s;
		r++; /* count remaining descriptors */
	}

	return r;
}


/* mconn_pollfds - poll, waiting for connect response/timeout */
int
mconn_pollfds(int fd, struct mconn_req *mreq)
{
	const char *f = "mconn_pollfds";
	struct pollfd pfd[MCONNMAXCONN];
	struct timeval pollend;
	nfds_t pfdcnt;
	int timeout;
	int polln;
	int i;

	/* prepare for poll()ing of remaining connecting descriptors */
	bzero(&pfd, sizeof(pfd));
	for (i = 0; i < mreq->cnt; i++) {
		if (mreq->fds[i] == -1)
			continue;
		pfd[i].fd = mreq->fds[i];
		pfd[i].events = POLLIN|POLLOUT;
	}
	pfdcnt = mreq->cnt;

	/* connect end time */
	xgettimeofday(&pollend, NULL);
	pollend.tv_sec += MCONN_TIMEOUT;
	/* remaining time */
	timeout = 1 - tv_dur2ms(&pollend);
	dbg2("%s: timeout in %d ms", f, timeout);
	while ((polln = poll(pfd, mreq->cnt, timeout)) != -1) {
		if (polln == 0) {
			dbg("%s: connect timeout", f);
			/* timeout/ETIMEDOUT, senderr in parent */
			for (i = 0; i < mreq->cnt; i++) {
				if (mreq->fds[i] == -1)
					continue;
				/* don't overwrite already set errors */
				if (mreq->errs[i] == 0)
					mreq->errs[i] = ETIMEDOUT;
			}
			break;
		}

		for (i = 0; i < mreq->cnt; i++) {
			if (pfd[i].revents == 0)
				continue;

			/* socketerr sets errno in case of error */
			if (socketerr(pfd[i].fd) != -1) {
				ASSERT(pfd[i].revents & POLLOUT, "bad revent");
				/* connect succeeded */
/*				dbg2("%s: connect success", f);*/
				mreq->conntimes[i] =
					tv_dur2ms(&mreq->starttimes[i]);
				mreq->errs[i] = 0;
				if (mconn_sendrep(fd, mreq, i) == -1)
					mconn_err();
			} else {
				/* connect failed */
				dbg("%s: connect failure", f);
				mreq->errs[i] = errno;
			}
			pfd[i].fd = -1;
			pfdcnt--;
			continue;
		}

		if (pfdcnt == 0)
			break; /* no descriptors left */

		/* update timeout to remaining time */
		timeout = 1 - tv_dur2ms(&pollend);
		dbg2("next timeout in %d ms", timeout);
	}
	if (polln == -1) {
		/* do not expect any signals; poll err is not good */
		swarn("%s: poll: %m", f);
		return -1;
	}

	return 0;
}


/* mconn_senderrs - return errors for all failed descriptors */
int
mconn_senderrs(int fd, struct mconn_req *mreq)
{
	int i;

	for (i = 0; i < mreq->cnt; i++) {
		if (mreq->errs[i] == 0) {
			/* senderrs should be called after all descs sent */
			ASSERT(mreq->fds[i] == -1, "ok descriptor not sent");
			continue;
		}
		if (mreq->fds[i] != -1) {
			xclose(mreq->fds[i]);
			mreq->fds[i] = -1;
		}
/*		ASSERT(mreq->fds[i] == -1, "unexpected fd value, conn failed");*/
		if (mconn_sendrep(fd, mreq, i) == -1)
			return -1;
	}

	return 0;
}


/* mconn_sendrep - send back single connected file descriptor (or error) */
int
mconn_sendrep(int fd, struct mconn_req *mreq, int idx)
{
	struct mconn_rep rep;

	/* write reply header */
	bzero(&rep, sizeof(struct mconn_rep));
	rep.idx = idx;
	rep.err = mreq->errs[idx];
	if (rep.err == 0) {
		ASSERT(mreq->fds[idx] != -1, "no err and descriptor not set");
		rep.conntime = mreq->conntimes[idx];
		rep.descpass = 1;
	} else {
		rep.conntime = 0;
		rep.descpass = 0;
	}
	if (writen(fd, &rep, sizeof(rep)) == -1)
		return -1;

	/* sending descriptor? */
	if (rep.descpass) {
		if (senddesc(fd, mreq->fds[idx]) == -1)
			return -1;
	}
	/* descriptor no longer needed here, close */
	if (mreq->fds[idx] != -1) {
		xclose(mreq->fds[idx]);
		mreq->fds[idx] = -1;
	}

	return 0;
}


/* mconn_readrep - read (one) mconn reply, fd invalid if err set */
int
mconn_readrep(int fd, int *idx, int *err, float *conntime)
{
	struct mconn_rep rep;

	/* XXX usable in loop? for reading multiple replies */
	if (readn(fd, &rep, sizeof(rep)) == -1)
		return -1;

	/* connect error? */
	if (rep.err != 0) {
		*idx = rep.idx;
		*err = rep.err;
		/* conntime not set */
		/*XXX zero is a valid descriptor */
		return 0;
	}

	/* connect success */
	*idx = rep.idx;
	*err = 0;
	*conntime = rep.conntime;

	return recvdesc(fd);
}


/* mconn_err - exit after process communication error */
void
mconn_err(void)
{
	const char *f = "mconn_err";
	/*
	 * room for recovery is limited if an error occurs in
	 * communication with the parent process, exit and let process
	 * be restarted by proc if still wanted.
	 */
	swarn("%s: fatal error, mconn server exiting", f);
	_exit(1);
}
