/* memcache.c - memory cache for (currently) unused block data */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: memcache.c,v 1.32 2010/06/30 08:37:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "log.h"
#include "alist.h"
#include "alist_int.h"
#include "block.h"
#include "file.h"
#include "filemeta.h"
#include "memcache.h"
#include "urient.h"

/*
 * Overview
 *
 * Space is needed for block data when data is transmitted either to
 * or from the network (or both). The block data will be cached when
 * the reference count of the block reaches zero. This memory cache
 * stores the block data under the assumption that it might be needed
 * again a short time into the future. Data is removed from the cache
 * when the maximum cache limit is reached.
 *
 * The cache handles any block size (smaller than the maximum cache size).
 *
 *
 * Memory usage
 *
 * The total memory usage from block data is the sum of the data for
 * the blocks used in active streams and the data currently
 * cached. The first value is limited by the number of active streams
 * (which might use different block sizes), while the second is
 * limited by the maximum cache size.
 */

/*
 * TODO: stream aware memory caching
 *
 * general philosophy; worry about short term future:
 * -give highest priority to blocks closest to active clients
 * -lower priority towards max available range
 *
 * blockno  0123456789
 * cached    CCCCCC
 * clientA    X
 *  priA       1234
 * clientB      X
 *  priB         12
 * cum(min)  U-1-12
 *
 * blocks marked with U should already have been removed (not referenced)
 * low digits indicate high priority; both blocks with priority 1 have
 * highest priority. in case of tie, use client priority? not sure if
 * there is any point; blocks will have to be read in any case, giving
 * same overhead for server.
 *
 * use array for ordering, first sorting criteria is priority above,
 * if priority is the same, use current block offset, keeping the block
 * for the client that is closest to the end. if offset same, choose any.
 *
 * first pass through memcache, set sorting fields to zero for all
 * blocks. second pass, look at each active stream, adding priority
 * values as above, and overwriting current values for a block if new
 * values represent a higher priority (e.g. block 5 in example above
 * has priority 1 and not 3.). then sort array with e.g., qsort and
 * remove block with lowest priority.
 *
 * for internal requests the request pattern is likely sequential,
 * while external requests might not be. the approach outlined above is
 * likely to be most efficient in the first case; an scc node might
 * benefit from a different approach.
 */

/* the size of the memory cache is specified in mb */
#define MEMCACHE_PARAMUNIT  (1 * 1024 * 1024) /* 1 MB */
#define MEMCACHE_MAX 10

/*
 * XXX could have higher level function (specified as argument to
 *     preinit function) which does more fancy expiry, like interval
 *     caching or stream aware memory caching described above.
 */
#define memcache_expire(size) memcache_expire_lru(size)

struct alist *al_memcache = NULL;
struct memstat memstat;


/* memcache_preinit - setup memory cache, using default size if none given */
void
memcache_preinit(int memsize)
{
	const char *f = "memcache_preinit";

	bzero(&memstat, sizeof(memstat));
	if (memsize != 0)
		memstat.totsize = memsize * MEMCACHE_PARAMUNIT;
	else
		memstat.totsize = MEMCACHE_MAX * MEMCACHE_PARAMUNIT;
	memstat.freemem = memstat.totsize;

	al_memcache = alist_init((void (*)(const void *))memcache_print,
				 (void (*)(void *))memcache_free,
				 0, ALIST_SIZE_GROW, ALIST_TYPE_UNSORTED_PTR);

	dbg("%s: memcache size: %d MB", f, memstat.totsize/(1024*1024));
}


/* memcache_store - cache memory for given block */
void
memcache_store(struct urient *ue, int32_t bnum)
{
	const char *f = "memcache_store";
	struct mement *ment;
	struct block *block;

	ASSERT(ue->file != NULL, "NULL file");
	ASSERT((uint32_t)ue->meta->blocksize < memstat.totsize,
	       "block too large");

	block = ue->file->blocks[bnum];
	ASSERT(alist_find(al_memcache, block) == NULL, "already cached");

	/*
	 * XXX what is least preferable, wasting cache memory or
	 * potentially doing more malloc()/free() operations?
	 */
	if (block->actlen == 0)
		dbg("%s: caching empty block", f);

	/* enough room? */
	if (memstat.freemem < block->blocklen)
		memcache_expire(block->blocklen);
	ASSERT(memstat.freemem >= block->blocklen, "should be room for entry");

	ment = xmalloc(sizeof(struct mement));
	ment->ue = ue;
	ment->bnum = bnum;
	ment->blocklen = block->blocklen;
	/* data is 'moved' from block, need to keep copy of actlen value */
	ment->actlen = block->actlen;
	ment->blockdat = block->blockdat;
	block->blockdat = NULL; /* update block */

	if (alist_add(al_memcache, block, ment) == NULL)
		ERR("alist_add failed");
	memstat.freemem -= ment->blocklen;
	dbg("%s: caching %s, block %d", f, ue->uri.uri, bnum);

	memstat.stored++;
}


/* memcache_lookup - set data in block if cached, return -1 otherwise */
int
memcache_lookup(struct urient *ue, int32_t bnum)
{
	struct mement *ment;
	struct block *block;
	ssize_t idx;

	block = ue->file->blocks[bnum];
	if ((idx = alist_findidx(al_memcache, block,
				 ALIST_SIZE(al_memcache))) == -1)
		return -1;
	ment = ALIST_PVAL(al_memcache, idx);

	/* restore values */
	block->blockdat = ment->blockdat;

	/* delete from cache */
	alist_delidx(al_memcache, idx);

	memstat.hits++;

	return 0;
}


/* memcache_free - expire block data */
void
memcache_free(struct mement *ment)
{
	const char *f = "memcache_free";
	struct block *block;

	ASSERT(ment->blockdat != NULL, "NULL blockdat");

	block = ment->ue->file->blocks[ment->bnum];
	/* data expired, update block */
	if (block->actlen > 0 && block->actlen != block->blocklen)
		dbg("%s: incomplete block data lost", f);

	memstat.freemem += ment->blocklen;
	ASSERT(memstat.freemem <= memstat.totsize, "too much free mem");

	free(ment);
}


/* memcache_expire - LRU based memory cache replacement */
void
memcache_expire_lru(uint32_t size)
{
	const char *f = "memcache_expire_lru";
	struct mement *ment;
	int cnt;

	ASSERT(ALIST_SIZE(al_memcache) > 0, "expire on empty cache");

	/*
	 * the least-recently-used entry will automatically be the
	 * first entry in the list.
	 */
	cnt = 0;
	while (ALIST_SIZE(al_memcache) > 0 && size > memstat.freemem) {
		ment = ALIST_PVAL(al_memcache, 0);
		dbg("%s: expiring %s, block %d (%d kb)", f, ment->ue->uri.uri,
		    ment->bnum, ment->blocklen/1024);

		free(ment->blockdat);
		ment->ue->file->blocks[ment->bnum]->actlen = 0;
		alist_delidx(al_memcache, 0);
		memstat.expired++;
		cnt++;
	}
	dbg("%s: expired %d blocks to free %d kb", f, cnt, size/1024);
}


/* memcache_flush - delete any cache entries for ue, O(N) */
void
memcache_flush(struct urient *ue)
{
	const char *f = "memcache_flush";
	struct mement *ment;
	size_t i;
	int cnt;

	cnt = 0;
	/* alist is an array, traverse in reverse */
	for (i = ALIST_SIZE(al_memcache); i > 0; i--) {
		ment = ALIST_PVAL(al_memcache, i - 1);
		if (ment->ue == ue) {
			free(ment->blockdat);
			ment->ue->file->blocks[ment->bnum]->actlen = 0;
			alist_delidx(al_memcache, i - 1);
			cnt++;
		}
	}

	if (cnt > 0) {
		dbg("%s: %s %d blocks flushed", f, ue->uri.uri, cnt);
		memstat.flushed += cnt;
	}
}


/* memcache_status - dump current memory cache status */
void
memcache_status(void)
{
	const char *f = "memcache_status";

	dbg("%s: free: %0.2f MB stored: %d hits: %d expired: %d flushed: %d", f,
	    (float)memstat.freemem/(1024*1024), memstat.stored, memstat.hits,
	    memstat.expired, memstat.flushed);

	alist_dump(al_memcache);
}


/* memcache_print - alist print function for cache entry */
void
memcache_print(struct mement *ment)
{
	dbg(" %s, block %d (%0.2f/%0.2f KB)",  ment->ue->uri.uri, ment->bnum,
	    (float)ment->actlen/1024, (float)ment->blocklen/1024);
}
