/* peer.c - peer management functions (for all known peers) */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: peer.c,v 1.34 2010/10/11 09:07:47 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <string.h>

#include "log.h"
#include "alist.h"
#include "ip.h"
#include "peer.h"
#include "peer_int.h"

/*
 * the peers database is used for two types of information,
 * with the following keys:
 * -symbolic names for servers (servers, from uris)
 * -numeric addresses (for peers)
 *
 * the peer entries stored with symbolic names as keys (such as
 * spp.example.org) are primarily used for caching dns lookup results.
 * supporting non-blocking dns lookup makes it convenient to create
 * connection/peer structures early and use the peer structures to
 * store lookup results.
 *
 * the peer entries with numeric keys are used to store information on
 * the machine with that specific ip-address; latency, bandwidth etc.
 *
 * a server will exist with multiple entries in the database:
 * -one for the symbolic name
 * -one or more for each ip address
 *
 * no information is maintained on the relation between these entries
 * (though it can be determined from the dns information in the
 * symbolic name entry).
 * [rather, have peerset for each address in the symbolic name peer entry]
 *
 * it would have been possible to maintain a single peer entry in the
 * database for each server (with multiple keys pointing to the same
 * peer structure), but there are practical reasons for why this is
 * not done:
 *
 * -a server name can resolve to multiple addresses for machines that
 *  are located in different parts of the Internet; it is necessary to
 *  maintain separate information on network parameters for each ip
 *  address.
 *
 * -maintaining a relationship between the addresses creates several
 *  problems where correct behavior is unclear, such as how to handle
 *  changes in the list of addresses. the old address might still be
 *  referenced, should it now be treated as a normal peer or a server?
 *  the content stored on it would likely still correspond to that of
 *  a server.
 *
 * as is, it is simple to lookup the list of addresses, should this be
 * needed (XXX need to handle dns lookup cache expiry with this in mind,
 * if the information is important it would need to be refreshed regularly
 *  with a timer).
 */

/* XXX use peerset for maintaining peers? */

#define PEER_DEFAULT_SIZE 100

struct alist *al_peers;

void (*peer_initfunc)(struct peer *) = NULL;
void (*peer_freefunc)(struct peer *) = NULL;


/* peer_preinit - set init function to be called when a new peer is added */
void
peer_preinit(void (*initfunc)(struct peer *), void (*freefunc)(struct peer *))
{
	ASSERT(peer_initfunc == NULL, "peer_init() already set");

	peer_initfunc = initfunc;
	peer_freefunc = freefunc;

	al_peers = alist_init((void (*)(const void *))peer_print,
			      (void (*)(void *))peer_free,
			      PEER_DEFAULT_SIZE, ALIST_SIZE_GROW,
			      ALIST_TYPE_STR);
}


/* peer_find - return pointer to peer if it exists, NULL otherwise */
struct peer *
peer_find(char *name, char *port)
{
/*	const char *f = "peer_find";*/
	struct peer *peer;
	char *key;

	/*
	 * map ipv6 localhost to ipv4.
	 * XXX should have parameter for denying localhost added as peer
	 * unless doing testing.
	 */
	if (strcmp(name, "::1") == 0)
		return peer_find("127.0.0.1", port);

	key = peer_genid(name, port);
	if ((peer = alist_find(al_peers, key)) == NULL)
		return NULL;

	return peer;
}


/* peer_add - add new peer, simply return pointer if already exists */
struct peer *
peer_add(char *name, char *port)
{
	const char *f = "peer_add";
	struct addrinfo hints, *res;
	struct peer *peer;
	size_t maxlen;
	char *key;
	int n;

	key = peer_genid(name, port);

	/* already exists? */
	if ((peer = alist_find(al_peers, key)) != NULL)
		return peer;

	/*
	 * map ipv6 localhost to ipv4.
	 * XXX should have parameter for denying localhost added as peer
	 * unless doing testing. (XXX check in peers command rather than here?)
	 */
	if (strcmp(name, "::1") == 0)
		return peer_add("127.0.0.1", port);

/*	dbg("%s: %s %s", f, name, port);*/

	peer = xmalloc(sizeof(struct peer));
	bzero(peer, sizeof(struct peer));

	/* set address info unless symbolic host name */
	if (ip_isnumeric(name)) {
		bzero(&hints, sizeof(struct addrinfo));
		hints.ai_family = AF_UNSPEC;
		hints.ai_socktype = SOCK_STREAM;
		if ((n = getaddrinfo(name, port, &hints, &res)) != 0) {
			swarn("getaddrinfo failure for %s:%s: %s", name, port,
			      gai_strerror(n));
			peer_free(peer);
			return NULL;
		}
		peer->addr = res;
	} else
		peer->addr = NULL;

	maxlen = sizeof(peer->name);
	if (strlcpy(peer->name, name, maxlen) >= maxlen) {
		swarn("%s: name to long: %s", f, peer->name);
		peer_free(peer);
		return NULL;
	}
	maxlen = sizeof(peer->port);
	if (strlcpy(peer->port, port, maxlen) >= maxlen) {
		swarn("%s: port name to long: %s", f, peer->port);
		peer_free(peer);
		return NULL;
	}
	maxlen = sizeof(peer->id);
	if (strlcpy(peer->id, key, maxlen) >= maxlen)
		/* should be long enough to contain max name and port values */
		ERRX("node id buffer too short");

	if (alist_add(al_peers, peer->id, peer) == NULL)
		ERRX("unable to add peer");

	/* XXX call init callback for all registered upper layers */
	if (peer_initfunc != NULL)
		peer_initfunc(peer);

	return peer;
}


/* peer_del - delete peer, error if refcnt not zero */
void
peer_del(struct peer *peer)
{
	ASSERT(peer->refcnt == 0, "deleting referenced peer");

	alist_del(al_peers, peer->id);
}


/* peer_free - release peer resources */
void
peer_free(struct peer *peer)
{
	ASSERT(peer->refcnt == 0, "deleting referenced peer");

	if (peer->addr != NULL)
		freeaddrinfo(peer->addr);
	if (peer_freefunc != NULL)
		peer_freefunc(peer);
	free(peer);
}


/* peer_addref - increase peer reference count */
void
peer_addref(struct peer *peer)
{
	peer->refcnt++;
}


/* peer_unref - decrease peer reference count */
void
peer_unref(struct peer *peer)
{
	ASSERT(peer->refcnt > 0, "negative reference count");

	peer->refcnt--;

	/* XXX if (peer->refcnt == 0) create deletion timer/use expire queue? */
}


/* peer_genid - return pointer to string identifier for host/port */
char *
peer_genid(char *name, char *port)
{
	static char key[NI_MAXHOST + NI_MAXSERV]; /* first NUL replaced by : */

	strlcpy(key, name, sizeof(key));
	strlcat(key, ":", sizeof(key));
	if (strlcat(key, port, sizeof(key)) >= sizeof(key))
		ERRX("key too short");

	return key;
}


/* peer_print - show peer info */
void
peer_print(struct peer *peer)
{
	dbg(" %s (%d)", peer->id, peer->refcnt);
}


/* peer_dump - list all peers in database */
void
peer_dump(void)
{
	alist_dump(al_peers);
}
