/* peer.h - peer related structures, defines and function prototypes */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: peer.h,v 1.27 2010/06/30 08:37:59 karls Exp $ */

#ifndef _PEER_H_
#define _PEER_H_

#include <netdb.h>

#if HAVE_OPENSSL
#include <openssl/ssl.h>
#endif /* HAVE_OPENSSL */

#include "peerpar.h"

#define PEER_RTT_MAX 20000 /* assume values above this are invalid */
#define PEER_RTT_UNKN 999999

/*
 * node -> peerid (ssl cert), stored on disk/database
 *        support Berkeley db +
 *        forking with pipe to external tool
 *        -simple text protocol: GETID/STORECRED
 *        (external tool to avoid supporting many databased directly)
 *
 *  majority of operations on nodelist (list of ip/port touples, short term)
 *  rather than peerid/identities (ssl cert, long term)
 * -with the exception of credit handling on server, all protocol
 *  operations are only concerned with the nodes that can be used
 *  at the moment. peerid/longer term identities can be used for
 *  storing state (e.g., freeloader/malicious node), in order to
 *  bootstrap state of ip/port node based on knowledge from past
 *  behavior, assuming peerid is known.
 * -clients do not worry about peerid; do not present certificates
 *  to each other (currently), but doing this is a possibility.
 *
 */

enum protoffs {
	PROTO_SPP = 0,
/*	PROTO_FOO,*/
	PROTO_MAX,
};

#define PROTOS_MAX 5 /*XXX?*/

struct peer {
	char name[NI_MAXHOST];
	char port[NI_MAXSERV];
	char id[NI_MAXHOST + NI_MAXSERV]; /* unique key value ('name:port') */
	int refcnt;
	struct peerpar peerpar;

	void *uldat[PROTO_MAX + 1]; /* pointer to data for upper layers */

	/* XXX conn layer values */
	struct peerset *dns; /* list of peers with getaddrinfo data */
	time_t dnstime; /* time of dns entry being set (resolve time) */
	int reserr; /* return value from getaddrinfo */
	struct addrinfo *addr; /* getaddrinfo output (for numeric peers) */
	int resolveretries; /* number of EAI_AGAIN retries */

	/* connection cache array */
	struct connection *cached; /* cached connection */

	int sslactive; /*XXX igots are per file/peer */
	/* ssl connection cache array */
	struct connection *sslcached;
#if HAVE_OPENSSL
	SSL *ssl;
	SSL_SESSION *sess;
#endif /* HAVE_OPENSSL */
	float minmbps; /* lower limit for transfers from this peer */
};

void peer_preinit(void (*)(struct peer *), void (*)(struct peer *));
struct peer *peer_find(char *, char *);
struct peer *peer_add(char *, char *);
void peer_del(struct peer *);
void peer_free(struct peer *);
void peer_addref(struct peer *);
void peer_unref(struct peer *);
char *peer_genid(char *, char *);
void peer_print(struct peer *);
void peer_dump(void);

/* peer_ops */
void peer_opsinit(struct peer *);
void peer_opsfree(struct peer *);
void peer_setdns(struct peer *, struct peerset *, int);
int peer_validaddr(struct peer *);

#endif /* _PEER_H_ */
