/* peer_ops.c - various peer manipulation functions */

/*
 * Copyright (c) 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: peer_ops.c,v 1.7 2010/06/29 15:14:38 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <time.h>

#include "log.h"
#include "ip.h"
#include "peer.h"
#include "peerset.h"

/*XXX file should have name related to user, likely networking layer */


/* peer_opsinit - used to set default values when new peers are added */
void
peer_opsinit(struct peer *peer)
{
	peerpar_settime(peer, PEERPAR_PEER_CREATED);
	peerpar_settime(peer, PEERPAR_PEER_LASTACTIVE);
}


/* peer_opsfree - release conn layer peer resources */
void
peer_opsfree(struct peer *peer)
{
	if (peer->dns != NULL)
		pset_free(peer->dns);
}


/* peer_setdns - for symbolic names, add peerset pointing resolved peers */
void
peer_setdns(struct peer *peer, struct peerset *pset, int reserr)
{
	int i;

	if (reserr) {
		ASSERT(pset == NULL, "reserr but pset set");
		dbg("peer_setdns: %s (0 entries, err: %d)", peer->id, reserr);
	} else {
		dbg("peer_setdns: %s (%d entries)", peer->id,
		    pset->peercnt);
/*		pset_dump(pset, -1);*/
	}

	/* verify that all nodes are numeric (debug onlyXXX) */
	if (pset != NULL)
		for (i = 0; i < pset->peercnt; i++) {
			ASSERT(pset_peer(pset, i)->dns == NULL, "dns info set");
			ASSERT(peer_validaddr(pset_peer(pset, i)), "bad addr");
			ASSERT(ip_isnumeric(pset_peer(pset, i)->name), "symb");
		}

	if (peer->dns != NULL)
		pset_free(peer->dns);
	peer->dns = pset;
	peer->reserr = reserr;
	peer->dnstime = time(NULL);
}


/* peer_validaddr - true if peer has a valid and non-expired resolved address */
int
peer_validaddr(struct peer *peer)
{
	/* numeric address? */
	if (peer->addr != NULL) {
		ASSERT(ip_isnumeric(peer->name), "addr set, addr non-numeric");
		return 1;
	}

	/*
	 * XXX expiry time can probably be similar to typical name
	 *     cache expiry time, no need to refresh if retrieved from
	 *     named cache.
	 */
	if (peer->dnstime != 0 &&
	    peer->dnstime + RESOLVE_EXPIRYTIME > time(NULL))
		return 1;
	else
		return 0;
}
