/* peerpar.c - peer parameter manipulation functions */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: peerpar.c,v 1.10 2010/06/25 22:02:28 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "log.h"
#include "peer.h"
#include "peer_int.h"
#include "peerpar_int.h"


/* peerpar_setfloat - set single float value, replacing current values */
void
peerpar_setfloat(struct peer *peer, int par, float val)
{
	struct mfloat *mf;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");
	ASSERT(peerpars[par].type == PEERPAR_TYPE_FLOAT, "invalid type");

	mf = &peer->peerpar.mfloatv[peerpars[par].idx];
	/* setting a value is equivalent to SINGLE valop */
	peerpar_fadd(mf, PEERPAR_VALOP_SINGLE, val);
}


/* peerpar_addfloat - add new float parameter value to peer */
void
peerpar_addfloat(struct peer *peer, int par, float val)
{
	struct mfloat *mf;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");
	ASSERT(peerpars[par].type == PEERPAR_TYPE_FLOAT, "invalid type");

	mf = &peer->peerpar.mfloatv[peerpars[par].idx];
	peerpar_fadd(mf, peerpars[par].valop, val);
}


/* peerpar_getfloat - get float parameter value, or 0 if not set */
float
peerpar_getfloat(const struct peer *peer, int par)
{
	float f;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");
	ASSERT(peerpars[par].type == PEERPAR_TYPE_FLOAT, "invalid type");

	if (peerpar_getfloat2(peer, par, &f) == -1)
		return 0;
	else
		return f;
}


/* peerpar_getfloat2 - return current float value, or -1 it not set */
int
peerpar_getfloat2(const struct peer *peer, int par, float *val)
{
	const struct mfloat *mf;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");
	ASSERT(peerpars[par].type == PEERPAR_TYPE_FLOAT, "invalid type");

	mf = &peer->peerpar.mfloatv[peerpars[par].idx];
	return peerpar_fget(mf, val);
}


/* peerpar_fcmp - compare float values */
int
peerpar_fcmp(const void *a, const void *b)
{
	const float *fa = (const float *)a;
	const float *fb = (const float *)b;

	if (*fa < *fb)
		return -1;
	else if (*fa > *fb)
		return 1;
	else
		return 0;
}


/* peerpar_fadd - add float value */
void
peerpar_fadd(struct mfloat *mf, int valop, float val)
{
	int i;

	/* add */
	switch (valop) {
	case PEERPAR_VALOP_SINGLE:
		mf->vals[0] = val;
		mf->cnt = 1;
		break;
	case PEERPAR_VALOP_MED: /* FALLTHROUGH */
	case PEERPAR_VALOP_AVG: /* FALLTHROUGH */
	case PEERPAR_VALOP_MAX: /* FALLTHROUGH */
	case PEERPAR_VALOP_MIN:
		for (i = PEERPAR_HISTCNT - 2; i + 1 > 0; i--)
			mf->vals[i + 1] = mf->vals[i];
		mf->vals[0] = val;
		if (mf->cnt < PEERPAR_HISTCNT)
			mf->cnt++;
		break;
	default:
		ERRX("invalid partype");
	}

	/* recalculate value */
	switch (valop) {
	case PEERPAR_VALOP_SINGLE:
		mf->currval = mf->vals[0];
		break;
	case PEERPAR_VALOP_MED:
		mf->currval = peerpar_fmed(mf);
		break;
	case PEERPAR_VALOP_AVG:
		mf->currval = peerpar_favg(mf);
		break;
	case PEERPAR_VALOP_MAX:
		mf->currval = peerpar_fmax(mf);
		break;
	case PEERPAR_VALOP_MIN:
		mf->currval = peerpar_fmin(mf);
		break;
	default:
		ERRX("invalid partype");
	}
}


/* peerpar_fget - return current float value, or -1 if not set */
int
peerpar_fget(const struct mfloat *mf, float *val)
{
	/* no value set? */
	if (mf->cnt == 0)
		return -1;
	else {
		if (val != NULL)
			*val = mf->currval;
		return 0;
	}
}


/* peerpar_fmed - calculate median value */
float
peerpar_fmed(struct mfloat *mf)
{
	float med[PEERPAR_HISTCNT];
	float val;
	int idx;
	int i;

	if (mf->cnt == 0)
		return 0;

	/* if even number of values, use average of middle values */
	if (mf->cnt == 1)
		val = mf->vals[0];
	else if (mf->cnt == 2) {
		val = mf->vals[0] + mf->vals[1];
		val /= 2;
	} else {
		for (i = 0; i < mf->cnt; i++)
			med[i] = mf->vals[i];
		qsort(med, mf->cnt, sizeof(float), peerpar_fcmp);
		/* even/odd? */
		if (mf->cnt % 2 == 1) {
			idx = mf->cnt/2;
			val = med[idx];
		} else {
			idx = mf->cnt/2;
			ASSERT(idx - 1 >= 0, "invalid index");
			val = med[idx - 1] + med[idx];
			val /= 2;
		}
	}

	return val;
}


/* peerpar_favg - calculate average value */
float
peerpar_favg(struct mfloat *mf)
{
	float val;
	int i;

	if (mf->cnt == 0)
		return 0;

	val = 0;
	for (i = 0; i < mf->cnt; i++)
		val += mf->vals[i];
	val /= mf->cnt;

	return val;
}


/* peerpar_fmin - return the lowest value */
float
peerpar_fmin(struct mfloat *mf)
{
	float val;
	int i;

	if (mf->cnt == 0)
		return 0;

	val = mf->vals[0];
	for (i = 0; i < mf->cnt; i++)
		if (mf->vals[i] < val)
			val = mf->vals[i];

	return val;
}


/* peerpar_fmax - return the highest value */
float
peerpar_fmax(struct mfloat *mf)
{
	float val;
	int i;

	if (mf->cnt == 0)
		return 0;

	val = mf->vals[0];
	for (i = 0; i < mf->cnt; i++)
		if (mf->vals[i] > val)
			val = mf->vals[i];

	return val;
}


/* peerpar_setint - set peerparam integer value */
void
peerpar_setint(struct peer *peer, int par, int val)
{
	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");
	ASSERT(peerpars[par].type == PEERPAR_TYPE_INT, "invalid type");

	peer->peerpar.intv[peerpars[par].idx] = val;
}


/* peerpar_getint - get integer parameter value */
int
peerpar_getint(const struct peer *peer, int par)
{
	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");
	ASSERT(peerpars[par].type == PEERPAR_TYPE_INT, "invalid type");

	return peer->peerpar.intv[peerpars[par].idx];
}


/* peerpar_adjint - adjust integer value, returning new value */
int
peerpar_adjint(struct peer *peer, int par, int adj)
{
	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");
	ASSERT(peerpars[par].type == PEERPAR_TYPE_INT, "invalid type");

	peer->peerpar.intv[peerpars[par].idx] += adj;

	return peer->peerpar.intv[peerpars[par].idx];
}


/* peerpar_settime - set current time in given field */
void
peerpar_settime(struct peer *peer, int par)
{
	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");
	ASSERT(peerpars[par].type == PEERPAR_TYPE_TIME, "invalid type");

	peer->peerpar.timev[peerpars[par].idx] = time(NULL);
}


/* peerpar_gettime - return current time in given field */
time_t
peerpar_gettime(const struct peer *peer, int par)
{
	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");
	ASSERT(peerpars[par].type == PEERPAR_TYPE_TIME, "invalid type");

	return peer->peerpar.timev[peerpars[par].idx];
}


/* peerpar_cmpfield - peer comparison function, for given field */
int
peerpar_cmpfield(const struct peer *na, const struct peer *nb, int par)
{
	float floata, floatb;
	time_t timea, timeb;
	int undefa, undefb;
	int inta, intb;
	int res;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	undefa = 0;
	undefb = 0;
	switch (peerpars[par].type) {
	case PEERPAR_TYPE_FLOAT:
		/* first do comparison based on whether values are defined */
		if (peerpar_getfloat2(na, par, &floata) == -1)
			undefa = 1;
		if (peerpar_getfloat2(nb, par, &floatb) == -1)
			undefb = 1;
		if (undefa || undefb) {
			slog(LOG_DEBUG, "sort with undefined values");

			if (undefa && undefb)
				res = 0;
			else if (undefa && !undefb)
				res = -1;
			else
				res = 1;
		} else {
			if (floata > floatb)
				res = 1;
			else if (floata < floatb)
				res = -1;
			else
				res = 0;
		}
		break;
	case PEERPAR_TYPE_INT:
		inta = peerpar_getint(na, par);
		intb = peerpar_getint(nb, par);

		if (inta < intb)
			res = -1;
		else if (inta == intb)
			res = 0;
		else
			res = 1;
		break;
	case PEERPAR_TYPE_TIME:
		timea = peerpar_gettime(na, par);
		timeb = peerpar_gettime(nb, par);

		if (timea < timeb)
			res = -1;
		else if (timea > timeb)
			res = 1;
		else
			res = 0;
		break;
	default:
		ERRX("invalid valtype");
	}

	return res;
}


/* eval_floatop - return true false depending on (vala op valb), O(1) */
int
eval_floatop(float vala, int op, float valb)
{
	int res;

	switch (op) {
	case PEERPAR_CMPOP_EQ:
		res = (vala == valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_NE:
		res = (vala != valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_LT:
		res = (vala < valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_GT:
		res = (vala > valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_GE:
		res = (vala >= valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_LE:
		res = (vala <= valb ? 1 : 0);
		break;
	default:
		ERRX("invalid operator");
	}

	return res;
}


/* eval_intop - return true false depending on (vala op valb), O(1) */
int
eval_intop(int vala, int op, int valb)
{
	int res;

	switch (op) {
	case PEERPAR_CMPOP_EQ:
		res = (vala == valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_NE:
		res = (vala != valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_LT:
		res = (vala < valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_GT:
		res = (vala > valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_GE:
		res = (vala >= valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_LE:
		res = (vala <= valb ? 1 : 0);
		break;
	default:
		ERRX("invalid operator");
	}

	return res;
}


/* eval_timeop - return true false depending on (vala op valb), O(1) */
int
eval_timeop(time_t vala, int op, time_t valb)
{
	int res;

	switch (op) {
	case PEERPAR_CMPOP_EQ:
		res = (vala == valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_NE:
		res = (vala != valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_LT:
		res = (vala < valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_GT:
		res = (vala > valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_GE:
		res = (vala >= valb ? 1 : 0);
		break;
	case PEERPAR_CMPOP_LE:
		res = (vala <= valb ? 1 : 0);
		break;
	default:
		ERRX("invalid operator");
	}

	return res;
}


/* peerpar_eval - evaluate if params are true, return true/false, O(1) */
int
peerpar_eval(struct peer *peer, int par, int op, void *val)
{
	time_t tval;
	float fval;
	int ival;
	int res;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	switch (peerpars[par].type) {
	case PEERPAR_TYPE_FLOAT:
		/* XXX what if undefined? */
		fval = peerpar_getfloat(peer, par);
		res = eval_floatop(fval, op, *(float *)val);
		break;
	case PEERPAR_TYPE_INT:
		ival = peerpar_getint(peer, par);
		res = eval_intop(ival, op, *(int *)val);
		break;
	case PEERPAR_TYPE_TIME:
		tval = peerpar_gettime(peer, par);
		res = eval_timeop(tval, op, *(time_t *)val);
		break;
	default:
		ERRX("invalid valtype");
	}

	return res;
}


/* peerpar_fieldop - perform 'respar = par1 op par2', O(1) */
int
peerpar_fieldop(struct peer *peer, int par1, int op, int par2, int respar)
{
	float fval1, fval2, fres;
	int ival1, ival2, ires;
	int res;

	ASSERT(par1 >= 0 && par1 < PEERPAR_MAX, "invalid parameter");
	ASSERT(par2 >= 0 && par2 < PEERPAR_MAX, "invalid parameter");
	ASSERT(respar >= 0 && respar < PEERPAR_MAX, "invalid parameter");
	ASSERT(peerpars[par1].type == peerpars[par2].type, "type mismatch");

	switch (peerpars[par1].type) {
	case PEERPAR_TYPE_FLOAT:
		/* use getfloat -> zero if undefined XXX */
		fval1 = peerpar_getfloat(peer, par1);
		fval2 = peerpar_getfloat(peer, par2);
		if ((res = peerpar_floatcalc(fval1, op, fval2, &fres)) == 0)
			peerpar_setfloat(peer, respar, fres);
		break;
	case PEERPAR_TYPE_INT:
		ival1 = peerpar_getint(peer, par1);
		ival2 = peerpar_getint(peer, par2);
		if ((res = peerpar_intcalc(ival1, op, ival2, &ires)) == 0)
			peerpar_setint(peer, respar, ires);
		break;
	case PEERPAR_TYPE_TIME:
		/* currently deemed not useful */
		ERRX("fieldop not supported on time types");
		break;
	default:
		ERRX("invalid valtype");
	}

	return res;
}


/* peerpar_floatcalc - do operation 'op' on val1 and val2, results in res */
int
peerpar_floatcalc(float val1, int op, float val2, float *res)
{
	double tres;

	switch (op) {
	case PEERPAR_FOP_ADD:
		tres = (double)val1 + val2;
		break;
	case PEERPAR_FOP_SUB:
		tres = (double)val1 - val2;
		break;
	case PEERPAR_FOP_MULT:
		tres = (double)val1 * val2;
		break;
	case PEERPAR_FOP_DIV:
		/* divide by zero? */
		if (val2 == 0) {
			slog(LOG_DEBUG, "peerpar_floatop: div by zero");
			return -1;
		}
		tres = (double)val1 / val2;
		break;
	default:
		ERRX("invalid operator");
	}

	/* XXX set to maximum/min value in case of overflow */
	if (tres > MAXFLOAT) {
		slog(LOG_WARNING, "peerpar_floatcalc: positive overflow");
		*res = MAXFLOAT;
	} else
		*res = (float)tres;

	return 0;
}


/* peerpar_intcalc - perform operation 'op' on val1 and val2, results in res */
int
peerpar_intcalc(int val1, int op, int val2, int *res)
{
	long long tres;

	switch (op) {
	case PEERPAR_FOP_ADD:
		tres = (long long)val1 + val2;
		break;
	case PEERPAR_FOP_SUB:
		tres = (long long)val1 - val2;
		break;
	case PEERPAR_FOP_MULT:
		tres = (long long)val1 * val2;
		break;
	case PEERPAR_FOP_DIV:
		/* divide by zero? */
		if (val2 == 0) {
			slog(LOG_DEBUG, "peerpar_intop: div by zero");
			return -1;
		}
		tres = (long long)val1 / val2;
		break;
	default:
		ERRX("invalid operator");
	}

	/* set to maximum/min value in case of overflow */
	if (tres > INT_MAX) {
		slog(LOG_WARNING, "peerpar_intcalc: positive overflow");
		*res = INT_MAX;
	} else if (tres < INT_MIN) {
		slog(LOG_WARNING, "peerpar_intcalc: negative overflow");
		*res = INT_MIN;
	} else
		*res = (int)tres;

	return 0;
}


#if 0
/* peerpar_applyall - apply function to all peers, O(N) */
void
peerpar_applyall(int par, int op, void *val,
		 void (*fn)(struct peer *, int, int, void *))
{
	struct peer *l_n;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	SLIST_FOREACH(l_n, &peers, n) {
		fn(l_n, par, op, val);
	}
}
#endif


/* peerpar_dump - show contents of peerpar */
void
peerpar_dump(struct peer *peer, int par)
{
	char buf[1024];
	char val[20];
	size_t r;
	int idx;
	int i;


	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	snprintf(buf, sizeof(buf), "%s %s:", peer->id, peerpars[par].name);
	idx = peerpars[par].idx;
	switch (peerpars[par].type) {
	case PEERPAR_TYPE_INT:
		snprintf(val, sizeof(val), " %d", peer->peerpar.intv[idx]);
		r = strlcat(buf, val, sizeof(buf));
		break;
	case PEERPAR_TYPE_FLOAT:
		r = 0;
		if (peer->peerpar.mfloatv[idx].cnt == 0) {
			r = strlcat(buf, " N/A", sizeof(buf));
			break;
		}
		for (i = 0; i < peer->peerpar.mfloatv[idx].cnt; i++) {
			snprintf(val, sizeof(val), " %f",
				 peer->peerpar.mfloatv[idx].vals[i]);
			r = strlcat(buf, val, sizeof(buf));
		}
		break;
	case PEERPAR_TYPE_TIME:
		snprintf(val, sizeof(val), " %ld",
			 (long)peer->peerpar.timev[idx]);
		r = strlcat(buf, val, sizeof(buf));
		break;
	default:
		ERRX("invalid valtype");
	}

	ASSERT(r < sizeof(buf), "peerpar_dump buffer too short");
	slog(LOG_INFO, "%s", buf);
}
