/* peerset.c - functions for operating on sets of peers */

/*
 * Copyright (c) 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: peerset.c,v 1.11 2010/06/25 22:02:28 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include "log.h"
#include "alist.h"
#include "peer.h"
#include "peer_int.h"
#include "peerset.h"

int sortkey_par = -1; /* XXX use global to determine sorting parameter */

void pset_freepeer(struct peer *);
void pset_dumpfunc(struct peer *, int, void *);
int pset_peercmp(const void *, const void *);
int pset_peercmprev(const void *, const void *);


/* pset_create - create and return pointer to a nset structure, O(1) */
struct peerset *
pset_create(void)
{
	struct peerset *pset;

	pset = xmalloc(sizeof(struct peerset));
	bzero(pset, sizeof(struct peerset));

	pset->al_peers = alist_init(NULL, (void (*)(void *))pset_freepeer,
				    PEERSET_MAX_PEERS, ALIST_SIZE_GROW,
				    ALIST_TYPE_UNSORTED_STR);
	return pset;
}


/* pset_free - release peerset resources, O(N) */
void
pset_free(struct peerset *pset)
{
	alist_free(pset->al_peers);
	free(pset);
}


/* pset_cpy - return duplicate of pset (working set only) */
struct peerset *
pset_cpy(struct peerset *pset)
{
	struct peerset *new;
	int i;

	new = pset_create();
	/* XXX probably faster way to do this */
	for (i = 0; i < pset->peercnt; i++)
		pset_addpeer(new, pset_peer(pset, i));

	return new;
}


/* pset_freepeer - unreference peer, O(1) */
void
pset_freepeer(struct peer *peer)
{
	peer_unref(peer);
}


/* pset_addpeer - add specified peer to the end of the peerset, O(N) */
void
pset_addpeer(struct peerset *pset, struct peer *new)
{
	const char *f = "pset_addpeer";
	struct peer *peer;

	if (alist_find(pset->al_peers, new->id) != NULL) {
		swarn("%s: host already exists in peerset: %s", f, new->id);
		return;
	}

	peer = alist_add(pset->al_peers, new->id, new);
	/* growing alist, should never fail */
	ASSERT(peer != NULL, "alist_add failed");

	/* successfully added, increase refcount */
	peer_addref(new);

	/* do not increment peercnt if it is different from fullpeercnt */
	if (pset->fullpeercnt == pset->peercnt)
		pset->peercnt++;
	pset->fullpeercnt++;
}


/* pset_delpeer - remove specified peer from peerset, O(2N) */
int
pset_delpeer(struct peerset *pset, struct peer *pr)
{
	int inrange;

	/* peercnt should only be decremented for peers in the working set */
	inrange = 0;
	if (pset->fullpeercnt != pset->peercnt)
		if (alist_rfind(pset->al_peers, pr->id, pset->peercnt) != NULL)
			inrange = 1;

	/* alist_del fails if peer does not exist. peer unref done by alist */
	alist_del(pset->al_peers, pr->id);

	/* decrement if peer was in working set */
	if ((pset->fullpeercnt == pset->peercnt) || inrange)
		pset->peercnt--;
	pset->fullpeercnt--;

	return 0;
}


/* pset_import - import peers based on parameters, O(N) */
int
pset_import(struct peerset *pset, int par, int op, void *val)
{
	struct peer *l_n;
	int cnt;
	int i;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	cnt = 0;
	for (i = 0; (size_t)i < ALIST_SIZE(al_peers); i++) {
		l_n = ALIST_PVAL(al_peers, i);
		if (peerpar_eval(l_n, par, op, val)) {
			pset_addpeer(pset, l_n);
			cnt++;
		}
	}

	return cnt;
}


/* pset_softtrim - reduce working size of set, without removing any peers */
void
pset_softtrim(struct peerset *pset, int newsize)
{
	ASSERT(newsize >= 0, "negative newsize");
	ASSERT(newsize <= pset->fullpeercnt, "peer offset out of bounds");

	pset->peercnt = newsize;
}


/* pset_evalsofttrim - remove matching nodes from working set */
void
pset_evalsofttrim(struct peerset *pset, int par, int op, void *val)
{
	struct peer *peer;
	int res;
	int ok;
	int i;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	if (pset->peercnt == 0)
		return;

	/* eval each node, counting true values */
	ok = 0;
	for (i = 0; i < pset->peercnt; i++) {
		peer = pset_peer(pset, i);
		res = peerpar_eval(peer, par, op, val);
		if (res)
			/* trim */
			peerpar_setint(peer, PEERPAR_INT_INT1, 1);
		else {
			/* leave */
			peerpar_setint(peer, PEERPAR_INT_INT1, 0);
			ok++;
		}
	}
	/* sort to get true values at beginning */
	pset_sort(pset, PEERPAR_INT_INT1, 0);

	/* reduce to only true nodes */
	pset_softtrim(pset, ok);
}


/* pset_trim - reduce set size of, removing any peers outsize new range */
void
pset_trim(struct peerset *pset, int newsize)
{
	ASSERT(newsize >= 0, "negative newsize");
	ASSERT(newsize <= pset->fullpeercnt, "peer offset out of bounds");

	alist_trim(pset->al_peers, newsize);

	/* only adjust peercnt if it is affected by trim */
	if ((pset->fullpeercnt == pset->peercnt) || newsize < pset->peercnt)
		pset->peercnt = newsize;
	pset->fullpeercnt = newsize;
}


/* pset_shift - shift working set of values in set either left or right, O(N) */
void
pset_shift(struct peerset *pset, int dir)
{
	int aldir;

	if ((pset->peercnt == 0) || (pset->peercnt == 1))
		return;

	if (dir == PEERSET_SHIFT_LEFT)
		aldir = ALIST_SHIFT_LEFT;
	else if (dir == PEERSET_SHIFT_RIGHT)
		aldir = ALIST_SHIFT_RIGHT;
	else
		ERRX("invalid direction specifier");

	alist_shift(pset->al_peers, 0, pset->peercnt - 1, aldir);
}


/* pset_peercmp - compare given field values of given field */
int
pset_peercmp(const void *a, const void *b)
{
	const struct peer *na = (const struct peer *)a;
	const struct peer *nb = (const struct peer *)b;

	return peerpar_cmpfield(na, nb, sortkey_par);
}


/* pset_peercmprev - compare given field values of given field, reverse sort */
int
pset_peercmprev(const void *a, const void *b)
{
	const struct peer *na = (const struct peer *)a;
	const struct peer *nb = (const struct peer *)b;

	return -peerpar_cmpfield(na, nb, sortkey_par);
}


/* pset_sort - sort working set with supplied function */
void
pset_sort(struct peerset *pset, int par, int rev)
{
	int (*compar)(const void *, const void *);

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	if (rev)
		compar = pset_peercmprev;
	else
		compar = pset_peercmp;
	pset_mysort(pset, par, compar);
}


/* pset_mysort - sort working set with supplied function */
void
pset_mysort(struct peerset *pset, int par,
	    int (*comp)(const void *a, const void *b))
{
	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	/* the field value would ideally be a function parameter */
	ASSERT(sortkey_par == -1, "sortkey param already set");
	sortkey_par = par;

	alist_mysort(pset->al_peers, pset->peercnt, comp);

	sortkey_par = -1;
}


/*XXX more useful if output from function is placed in par */
/* pset_apply - apply function to working set */
void
pset_apply(struct peerset *pset, int par, void *val,
	   void (*func)(struct peer *, int, void *))
{
	int i;

	/* allow -1 for no specific parameter */
	ASSERT(par >= -1 && par < PEERPAR_MAX, "invalid parameter");

	for (i = 0; i < pset->peercnt; i++)
		func(pset_peer(pset, i), par, val);
}


/* pset_applyfloat - set value of float param to output from function */
void
pset_applyfloat(struct peerset *pset, int par, void *val,
	   float (*func)(struct peer *, int, void *))
{
	float fres;
	int i;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	for (i = 0; i < pset->peercnt; i++) {
		fres = func(pset_peer(pset, i), par, val);
		peerpar_setfloat(pset_peer(pset, i), par, fres);
	}
}


/* pset_applyint - set value of int param to output from function */
void
pset_applyint(struct peerset *pset, int par, void *val,
	   int (*func)(struct peer *, int, void *))
{
	int fres;
	int i;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	for (i = 0; i < pset->peercnt; i++) {
		fres = func(pset_peer(pset, i), par, val);
		peerpar_setint(pset_peer(pset, i), par, fres);
	}
}


#if 0
/* pset_setrank - place current offset of peer in peerset in given field */
void
pset_setrank(struct peerset *pset, int par)
{
	int i;

	ASSERT(par >= 0 && par < PEERPAR_MAX, "invalid parameter");

	for (i = 0; i < pset->peercnt; i++)
		peerpar_setint(ns_peer(pset, i), par, &i); /*setparXXX*/
}
#endif


/* pset_setintrank - place rank of integer field in output field */
void
pset_setintrank(struct peerset *pset, int ipar, int opar)
{
	int val, val2;
	int rank;
	int i;

	ASSERT(ipar >= 0 && ipar < PEERPAR_MAX, "invalid parameter");
	ASSERT(opar >= 0 && opar < PEERPAR_MAX, "invalid parameter");

	/* how about undefined values? */
	rank = 0;
	val = peerpar_getint(pset_peer(pset, 0), ipar);
	peerpar_setint(pset_peer(pset, 0), opar, rank);
	for (i = 1; i < pset->peercnt; i++) {
		val2 = peerpar_getint(pset_peer(pset, i), ipar);
		if (val != val2)
			rank++;
		peerpar_setint(pset_peer(pset, i), opar, rank);

		val = val2;
	}
}


/* pset_setfloatrank - place (integer) rank of float field in output field */
void
pset_setfloatrank(struct peerset *pset, int ipar, int opar)
{
	float val, val2;
	int rank;
	int i;

	ASSERT(ipar >= 0 && ipar < PEERPAR_MAX, "invalid parameter");
	ASSERT(opar >= 0 && opar < PEERPAR_MAX, "invalid parameter");

	/* how about undefined values? */
	rank = 0;
	val = peerpar_getfloat(pset_peer(pset, 0), ipar);
	peerpar_setint(pset_peer(pset, 0), opar, rank);
	for (i = 1; i < pset->peercnt; i++) {
		val2 = peerpar_getfloat(pset_peer(pset, i), ipar);
		if (val != val2)
			rank++;
		peerpar_setint(pset_peer(pset, i), opar, rank);

		val = val2;
	}
}


/* pset_fieldop - apply operation on all peers in set */
void
pset_fieldop(struct peerset *pset, int f1, int op, int f2, int fres)
{
	int i;

	ASSERT(f1 >= 0 && f1 < PEERPAR_MAX, "invalid parameter");
	ASSERT(f2 >= 0 && f2 < PEERPAR_MAX, "invalid parameter");

	for (i = 0; i < pset->peercnt; i++)
		peerpar_fieldop(pset_peer(pset, i), f1, op, f2, fres);
}


/* pset_dumpfunc - print peer info */
/* ARGSUSED */
void
pset_dumpfunc(struct peer *p, int par, void *val)
{
	(void)val; /* unused */

	/* par == -1 -> name only */
	ASSERT(par >= -1 && par < PEERPAR_MAX, "invalid parameter");

	/* only peer name? */
	if (par == -1)
		dbg(" %s", p->id);
	else
		peerpar_dump(p, par);
}


/* pset_dump - list all peers in peerset */
void
pset_dump(struct peerset *pset, int par)
{
	const char *f = "pset_dump";

	/* par == -1 -> name only */
	ASSERT(par >= -1 && par < PEERPAR_MAX, "invalid parameter");

	dbg("%s: peerset: ", f);
	pset_apply(pset, par, NULL, pset_dumpfunc);
}
