/* poll.c - main event loop */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: poll.c,v 1.117 2010/11/04 21:53:34 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include <fcntl.h>

#include <errno.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "log.h"
#include "alist.h"
#include "conn.h"
#include "peer.h"
#include "poll.h"
#include "proc.h"
#include "protio.h"
#include "servop.h"
#include "socket.h"
#include "stream.h"
#include "timers.h"
#include "queue.h"

#define POLLUPDATE_NOACCEPT 1

int pollchange;

volatile int pollquit;

short connevents(struct connection *);
void ioevent(struct connection *);
void empty_queues(void);


/* serverop - main poll function */
void
serverop(void)
{
	const char *f = "serverop";
	struct pollfd pfds[POLL_MAX]; /* XXX dynamic? */
	struct connection *pollcon[POLL_MAX];
	struct connection *l_c;
	int oldpfdcnt;
	int noaccept;
	long timeout;
	int pfdcnt; /* number of entries in active array */
	int polln;
	int i;

	/* add accept() sockets */
	pfdcnt = 0;
	SLIST_FOREACH(l_c, &connections, n) {
		pfds[pfdcnt].fd = l_c->fd;
		pfds[pfdcnt].events = (short)connevents(l_c);
		pollcon[pfdcnt] = l_c;
		l_c->pollidx = pfdcnt;
		pfdcnt++;
	}
	oldpfdcnt = pfdcnt;
poll:
	timeout = timeout_next();
/*	dbg("%s: next timeout in %d ms", f, timeout);*/
	while ((polln = poll(pfds, (nfds_t)pfdcnt, timeout)) != -1) {
/*		ASSERT(pollquit == 0, "signal did not interrupt poll");*/

		pollchange = 0;

#ifdef __APPLE__
		if (polln > pfdcnt)
			swarn("%s: polln unexpectedly large: %d > %d", f,
			      polln, pfdcnt);
#else
		ASSERT(polln <= pfdcnt, "polln inconsistency");
#endif /* __APPLE__ */

		for (i = 0; i < pfdcnt; i++) {
			if (polln == 0)
				break;

			if (pfds[i].revents == 0)
				continue;
			else
				polln--;

			if (socketerr(pfds[i].fd) == -1 ||
			    pfds[i].revents & (POLLERR|POLLHUP|POLLNVAL)) {
				if (pollcon[i]->peer != NULL)
					swarn("%s: socket err: %s", f,
					      pollcon[i]->peer->id);
				else
					swarn("%s: socket err", f);
				if (pollcon[i]->proc != NULL) {
					swarn("%s: failure in proc conn", f);
					proc_returnproc(pollcon[i]->proc);
					pollcon[i]->proc = NULL;
					pollcon[i]->fd = -1;
				}
				if (pollcon[i]->dnsproc != NULL) {
					swarn("%s: failure in dnsproc conn", f);
					proc_returnproc(pollcon[i]->dnsproc);
					pollcon[i]->dnsproc = NULL;
					pollcon[i]->fd = -1;
				}
				if (pollcon[i]->protops == NULL) {
					ASSERT(pollcon[i]->type == CONN_TYPE_INCOMING, "unexpected type");
					ASSERT(pollcon[i]->status == CONNECTION_STATUS_READREQ, "unexpected status");
					dbg("%s: conn failure, no protops", f);
				} else if (pollcon[i]->status == CONNECTION_STATUS_CONNECTING) {
					pollcon[i]->protops->connevent(pollcon[i],
							      CONNEVENT_CONNECT_FAIL,
							      errno);

/*					pollcon[i]->protops->connerr(pollcon[i]->uldat, errno);*/
					failedconn(pollcon[i]->peer);
				} else
					pollcon[i]->protops->connevent(pollcon[i],
							      CONNEVENT_REQUEST_FAIL,
							      errno);

/*					pollcon[i]->protops->failedreq(pollcon[i]->uldat);*/
				changecstatus(pollcon[i], CONNECTION_STATUS_ENDED);
				continue;
			}

			if (pfds[i].revents & (POLLIN|POLLOUT)) {
				pollcon[i]->pollevent = pfds[i].revents & (POLLIN|POLLOUT);
				ioevent(pollcon[i]);
			} else
				ERRX("unexpected event");
		}

#ifdef __APPLE__
		if (polln != 0)
			swarn("%s: %d poll events remaining", f, polln);
#else
		ASSERT(polln == 0, "poll events remaining");
#endif /* __APPLE__ */

		/* (XXX always) check timeouts */
		timeout_handle();

		empty_queues();

		/* setup active streams */
		if (pollchange) {
			pollchange = 0;

			noaccept = 0;
			pfdcnt = 0;
			SLIST_FOREACH(l_c, &connections, n) {
				if (pfdcnt >= POLL_MAX) {
					dbg("%s: too many connection, "
					    "disabling accept", f);
					noaccept = 1;
					break;
				}

				/* awaiting any events on connection? */
				pfds[pfdcnt].events = (short)connevents(l_c);
				if (pfds[pfdcnt].events == 0) {
					l_c->pollidx = -1;
					continue;
				} else
					pfds[pfdcnt].fd = l_c->fd;

				pollcon[pfdcnt] = l_c;
				l_c->pollidx = pfdcnt;

				pfdcnt++;
			}

			if (pfdcnt != oldpfdcnt) {
/*			dbg("%s: newdescs: %d -> %d", f, oldpfdcnt, pfdcnt);*/
				oldpfdcnt = pfdcnt;
			}

			/* disable poll on accept socket? */
			if (noaccept)
				for (i = 0; i < pfdcnt; i++) {
					if (pollcon[i]->status ==
					    CONNECTION_STATUS_ACCEPTING)
						pfds[i].fd = -1;
				}
		}

		/* XXX poll NOT interrupted */
		if (pollquit == 1) {
			dbg("%s: got termination signal (poll)", f);
			return;
		}

		timeout = timeout_next();
/*		dbg("%s: next timeout in %d ms", f, timeout);*/
	}

	dbg("%s: poll: %m", f);

	/* poll interrupted */
	if (pollquit == 1) {
		dbg("%s: got termination signal", f);
		return;
	} else {
		proc_checksig(); /* SIGCHLD? */
		dbg("%s: poll interrupted, restarting", f);
		goto poll;
	}
}


/* pollend - end server operation (typically called from signal handler) */
void
pollend(void)
{
	pollquit = 1;
}


/* connevents - return appropriate poll event values for connection */
short
connevents(struct connection *c)
{
/*	const char *f = "connevents";*/
	short events;

	events = 0;

#if HAVE_OPENSSL
	if (c->ssl_cmd != SSLCMD_NONE) {
		ASSERT(c->status != CONNECTION_STATUS_ENDED, "ended ssl");

		switch (c->ssl_cmdstate) {
		case SSLCMD_WANT_READ:
			events |= POLLIN;
			break;
		case SSLCMD_WANT_WRITE:
			events |= POLLOUT;
			events |= POLLIN; /* for error */
			break;
		default:
			ERRX("unexpected ssl_cmdstate");
		}

		return events;
	}
#endif /* HAVE_OPENSSL */

	/* XXX can be simplified by having a READ/WRITE flags in status value */
	switch (c->status) {
	case CONNECTION_STATUS_ACCEPTING:
		events |= POLLIN;
		break;

	case CONNECTION_STATUS_DEADLINESTREAM: /* FALLTHROUGH */
	case CONNECTION_STATUS_FOREVERSTREAM:
		if (c->stream->state == STREAM_ACTIVE) {
			if (c->stream->type == STREAM_FROM_MEM) {
				events |= POLLOUT; /* write to net */
				events |= POLLIN; /* for error */
			} else
				events |= POLLIN;
		} else /* non-active */
			if (c->stream->type == STREAM_FROM_MEM)
				events |= POLLIN; /* for error */
		break;

	case CONNECTION_STATUS_IDLE:
		events |= POLLIN;
		break;

	case CONNECTION_STATUS_SSLCACHED: /* FALLTHROUGH */
	case CONNECTION_STATUS_CACHED:
		/* input would indicate error*/
		events |= POLLIN;
		break;

	case CONNECTION_STATUS_READREQ:
		events |= POLLIN;
		break;

	case CONNECTION_STATUS_READREQDAT:
		events |= POLLIN;
		break;

	case CONNECTION_STATUS_READREP:
		events |= POLLIN;
		break;

	case CONNECTION_STATUS_READREPDAT:
		events |= POLLIN;
		break;

	case CONNECTION_STATUS_WRITEREP:
		events |= POLLOUT; /* write to net */
		events |= POLLIN; /* for error */
		break;

	case CONNECTION_STATUS_WRITEREQ:
		events |= POLLOUT; /* write to net */
		events |= POLLIN; /* for error */
		break;

	case CONNECTION_STATUS_CONNECTING:
		events |= POLLOUT;
		events |= POLLIN; /* for error */
		break;

	case CONNECTION_STATUS_RESOLVING:
		events |= POLLIN;
		break;

	case CONNECTION_STATUS_MCONN: /* FALLTHROUGH */
		events |= POLLIN;
		break;

	case CONNECTION_STATUS_ENDED:       /* FALLTHROUGH */
	case CONNECTION_STATUS_MCONNWAIT:   /* FALLTHROUGH */
	case CONNECTION_STATUS_RESOLVEWAIT: /* FALLTHROUGH */
	case CONNECTION_STATUS_UNCONNECTED:
		events = 0;
		break;
	}

#if 0
	if (events & POLLIN)
		dbg("%s: POLLIN: %s", f, conn_state2str(c->status));
	if (events & POLLOUT)
		dbg("%s: POLLOUT: %s", f, conn_state2str(c->status));
	if ((events & POLLOUT) == 0 && (events & POLLIN) == 0)
		dbg("%s: noevents: %s", f, conn_state2str(c->status));
#endif

/*	dbg("%s: %d", f, events);*/

	return events;
}


/* ioevent - input/output event on connection, place in appropriate queue */
void
ioevent(struct connection *c)
{
	/*
	 * active connections imply data io, protocol io otherwise.
	 * protocol io can be either spp commands, or ssl protocol
	 * handling (need_read/need_write).
	 */

	if (c->status == CONNECTION_STATUS_DEADLINESTREAM ||
	    c->status == CONNECTION_STATUS_FOREVERSTREAM)
		queue_add(QUEUE_DATAIO, c);
	else {
#if HAVE_OPENSSL
		/* ssl protocol management expected? */
		if (c->ssl_cmd)
			queue_add(QUEUE_SSLPROT, c);
		else
#endif /* HAVE_OPENSSL */
			queue_add(QUEUE_PROTIO, c);
	}
}


/* empty_queues - process all event queues */
void
empty_queues(void)
{
	struct connection *c;

	/*
	 * io queues: protocol io, bulk data io (for resource sharing)
	 *            ^^ always read
	 *                         ^^^ read based on priority/usage
	 */

	/* empty protio queue */
	while ((c = queue_pop(QUEUE_PROTIO)) != NULL)
		protio(c);

	/* empty iodone queue */
	while ((c = queue_pop(QUEUE_IODONE)) != NULL)
		iodone(c);

	/* empty (XXXpriority?) dataio queue */
	while ((c = queue_pop(QUEUE_DATAIO)) != NULL)
		dataio(c);

#if HAVE_OPENSSL
	/* XXX ssl_protio can result in addition to PROTIO queue */
	/* empty sslprot queue */
	while ((c = queue_pop(QUEUE_SSLPROT)) != NULL)
		ssl_protio(c);

	/* empty sslcomp queue */
	while ((c = queue_pop(QUEUE_SSLCOMP)) != NULL)
		sslcomp(c);
#endif /* HAVE_OPENSSL */

	/* empty connend queue */
	while ((c = queue_pop(QUEUE_CONNEND)) != NULL)
		end_connection(c);
}
