/* proc.c - generic child process management interface */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: proc.c,v 1.14 2010/06/25 22:02:28 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

/*
 * the interface in this file wraps all operations related to forking
 * and handling child death related signals.
 * motivation for this approach is as follows:
 * -some parts (signal blocking etc.) slightly tricky, less scope for
 *  errors if done in one place
 * -use of fork() not trivial to test, especially due to limitations
 *  in gcov support for forked processes. keeping this code independent
 *  of the code for communication with child processes makes this code
 *  more general and simplifies testing, which only needs to be done once.
 */
/*
 * no code for releasing resources after shutdown, seems overkill for now
 */

/*
 * XXX could have timeout for removing unused processes.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "log.h"
#include "aqueue.h"
#include "proc.h"

#define PROC_STATESTR_MAX 100

#ifndef WAIT_ANY
#define WAIT_ANY -1
#endif

int sigchld_received = 0; /* set if at least one signal has been received */
int proc_didpreinit = 0; /* set if preinit done */

int proc_enabled[PROC_MAX];

/* NOTE: needs to be synced with proc_systnames below */

struct proc_conf {
	int initial; /* number of processes to fork at startup */
	int max;     /* max number of processes to fork */
	int curr;    /* current number of running processes */
	int acceptfail; /* whether to accept subsystem failing */
	int deaths; /* number of child deaths */
	enum proc_type type; /* load handling method type */
	void (*childfunc)(int); /* child process function */
	aqueue_t *aq_waiting; /* queue of waiting requests */
};

/* queue element callback data */
struct proc_callback {
	void *data; /* caller data pointer */
	void (*gotdesc)(void *, struct proc *); /* child process available */
	void (*noserv)(void *); /* unable to process request callback */
};

struct proc *procs[PROC_MAX];
struct proc_conf proc_conf[PROC_MAX];
char *proc_systnames[] = {
	"resolver",
	"mconn",
};

char *proc_state2str(int);


/* proc_preinit - setup proc handler */
void
proc_preinit(void)
{
	struct sigaction act;
	int i;

	for (i = 0; i < PROC_MAX; i++)
		proc_enabled[i] = 0; /* disabled until enabled via proc_init */

	bzero(&act, sizeof(act));
	act.sa_handler = proc_sigchld;
	sigemptyset(&act.sa_mask);
	/*
	 * restart function calls, detecting child deaths currently not
	 * considered time critical (especially if there is any function
	 * calls worth interrupting; if application hangs in poll(), there
	 * will be little happening and restarting any dead children can
	 * wait.
	 */
	act.sa_flags = SA_RESTART;

	/*
	 * not always available; not relied on but setting it reduces
	 * scope for unneeded operations (only care about child
	 * deaths).
	 */
	act.sa_flags = SA_NOCLDSTOP;
	if (sigaction(SIGCHLD, &act, NULL) == -1)
		serr("sigaction");

	proc_didpreinit = 1;
}


/* proc_init - setup specified subsystem */
void
proc_init(enum proc_subsyst subsyst, int initial, int max, int acceptfail,
	  enum proc_type type, void (*childfunc)(int))
{
	const char *f = "proc_init";
	int i;

	ASSERT(initial <= max, "invalid initial value");
	ASSERT(max > 0, "zero max value, useless init");
	ASSERT(childfunc != NULL, "no childfunc");
	ASSERT(initial > 0 || type == PROC_TYPE_GROW, "useless init");
	/* for fixed, max should be either same as initial or zero */
	ASSERT(type == PROC_TYPE_GROW || (max == initial || max == 0),
	       "invalid max value for PROC_TYPE_FIXED");
	ASSERT(proc_didpreinit != 0, "proc_preinit not called");

	procs[subsyst] = xcalloc(max, sizeof(struct proc));
	proc_conf[subsyst].initial = initial;
	proc_conf[subsyst].max = max;
	proc_conf[subsyst].curr = 0;
	proc_conf[subsyst].acceptfail = acceptfail;
	proc_conf[subsyst].type = type;
	proc_conf[subsyst].childfunc = childfunc;
	proc_conf[subsyst].aq_waiting = aqueue_init(NULL);

	for (i = 0; i < max; i++) {
		procs[subsyst][i].pid = -1;
		procs[subsyst][i].fd = -1;
		procs[subsyst][i].idx = i;
		procs[subsyst][i].subsyst = subsyst;
		procs[subsyst][i].status = PROC_STATUS_NOPROC;
	}

	for (i = 0; i < initial; i++) {
		if (proc_fork(subsyst, i) == -1)
			continue; /* no process added */
		proc_conf[subsyst].curr++;
	}

	if (initial > 0 && proc_conf[subsyst].curr == 0) {
		swarn("%s: unable to start any '%s' processes", f,
		      proc_systnames[subsyst]);
		return;
	}

	proc_enabled[subsyst] = 1;

	sinfo("%s: starting '%s' subsystem, %d initial process(es)", f,
	      proc_systnames[subsyst], proc_conf[subsyst].curr);
}


/* proc_enqueue - submit request for given subsystem */
void
proc_enqueue(enum proc_subsyst subsyst, void *data,
	     void (*gotdesc)(void *, struct proc *),
	     void (*noserv)(void *))
{
	const char *f = "proc_enqueue";
	struct proc_callback *pcb;
	struct proc *proc;
	int idx;
	int i;

	/* call for for convenience (death can happen after call) */
	proc_checksig();

	if (proc_enabled[subsyst] == 0) {
		swarn("%s: request to disabled '%s' subsystem", f,
		      proc_systnames[subsyst]);
		/* call noserv at once if system disabled */
		noserv(data);
		return;
	}

	/* submit at once if any processes are available */
	if ((proc = proc_getproc(subsyst)) != NULL) {
		dbg("%s: sending job to process %d", f, proc->idx);
		gotdesc(data, proc);
		return;
	}

	/* fork new process for request? */
	if (proc_conf[subsyst].type == PROC_TYPE_GROW &&
	    proc_conf[subsyst].curr < proc_conf[subsyst].max) {
		dbg("%s: starting new '%s' subsystem process", f,
		    proc_systnames[subsyst]);
		idx = -1;
		for (i = 0; i < proc_conf[subsyst].max; i++)
			if (procs[subsyst][i].status == PROC_STATUS_NOPROC) {
				idx = i;
				break;
			}
		ASSERT(idx != -1, "should be free proc but none found");
		if (proc_fork(subsyst, idx) == 0) {
			proc_conf[subsyst].curr++;
			/* a process should be available now */
			proc = proc_getproc(subsyst);
			ASSERT(proc != NULL, "still no available proc");
			dbg("%s: sending job to process %d", f, proc->idx);
			gotdesc(data, proc);
			return;
		}
	}

	/* nothing free, put on wait queue XXX could have max queue length */
	dbg("%s: putting '%s' request on queue (%d previous entries)", f,
	    proc_systnames[subsyst], proc_queuesize(subsyst));
	pcb = xmalloc(sizeof(struct proc_callback));
	pcb->data = data;
	pcb->gotdesc = gotdesc;
	pcb->noserv = noserv;
	aqueue_enqueue(proc_conf[subsyst].aq_waiting, pcb);
}


/* proc_queuesize - return size of wait for for given subsystem */
int
proc_queuesize(enum proc_subsyst subsyst)
{
	return aqueue_size(proc_conf[subsyst].aq_waiting);
}


/* proc_returnproc - caller finished with process */
void
proc_returnproc(struct proc *proc)
{
	const char *f = "proc_returnproc";
	struct proc_callback *pcb;
	aqueue_t *aq;

	/* call for for convenience (death can happen after call) */
	proc_checksig();

	dbg("%s: '%s' process returned", f, proc_systnames[proc->subsyst]);

	if (proc->status == PROC_STATUS_ACTIVE)
		proc->status = PROC_STATUS_FREE;
	else if (proc->status == PROC_STATUS_DEAD) {
		xclose(proc->fd);
		/* child died, attempt to create new process */
		if (proc_enabled[proc->subsyst])
			proc_refork(proc);
	}

	/* process any outstanding requests */
	aq = proc_conf[proc->subsyst].aq_waiting;
	if (aqueue_size(aq) > 0) {
		if (proc_enabled[proc->subsyst])
			proc_submitjob(proc->subsyst);
		else
			while ((pcb = aqueue_dequeue(aq)) != NULL) {
				pcb->noserv(pcb->data);
				free(pcb);
			}
	}
}


/* proc_kill - problem with process, kill if not already dead */
void
proc_kill(struct proc *proc)
{
	const char *f = "proc_kill";

	ASSERT(proc->status == PROC_STATUS_FREE ||
	       proc->status == PROC_STATUS_DEAD ||
	       proc->status == PROC_STATUS_ACTIVE,
	       "invalid state for proc_kill");

	/* process might already be dead, check for SIGCHLD */
	proc_checksig();

	if (proc->status == PROC_STATUS_DEAD)
		return; /* already dead */

	/*
	 * XXX is there a race here? can child die and a new process
	 * get same pid before we do kill(), resulting in the wrong
	 * process being killed?
	 */
	dbg("%s: killing '%s' process %d", f, proc_systnames[proc->subsyst],
	    proc->idx);
	if (kill(proc->pid, SIGINT) == -1) {
		if (errno == ESRCH)
			swarn("%s: pid: (%ld): %m ", f, (long)proc->pid);
		else
			ERR("kill");
	}
	if (proc->status != PROC_STATUS_ACTIVE) {
		xclose(proc->fd);
		proc->killed = 1; /* mark as killed and closed */
	}
}


/* proc_checksig - update process status in case of child death */
void
proc_checksig(void)
{
	const char *f = "proc_checksig";
	sigset_t chldmask, savemask;
	enum proc_status oldstate;
	struct proc *proc;
	int status;
	pid_t pid;

	/*
	 * this function calls waitpid() and updates the child status
	 * after one or more SIGCHLD signals have been received.
	 *
	 * calling this function regularly will cause free/inactive
	 * processes that die to be restarted at the time of calling.
	 *
	 * not calling this function regularly will not have any
	 * significant effect for the caller because it is called
	 * anyway at the beginning of the caller accessible functions.
	 * some more work will be done when it is eventually called.
	 * subsystem shutdown will also take longer to occur if a
	 * child function that dies quickly and repeatedly.
	 * (which would leave zombie processes hanging around).
	 *
	 * in summary, this function should be used regularly by the
	 * caller. not because correct the behavior of these function
	 * depends upon it, but in order to cleanup any zombie
	 * processes and quickly detect and shutdown failing
	 * subsystems.
	 */

	if (!sigchld_received)
		return;

	while ((pid = waitpid(WAIT_ANY, &status, WNOHANG)) > 0) {
		if (!WIFEXITED(status) && !WIFSIGNALED(status)) {
#if 0 /* XXX should not happen with SA_NOCLDSTOP */
			/* not dead, ignore */
			continue;
#else
			dbg("%s: pid: %ld", f, (long)pid);
			ERRX("child not dead");
#endif
		}

		if ((proc = proc_pid2proc(pid)) == NULL)
			/*
			 * likely indication of illegitimate child
			 * (i.e., not created by proc interface)
			 */
			ERRX("unknown child died");

		if (proc->killed)
			/* fd already closed */
			proc->killed = 0;
		/* active descriptors will be closed upon being returned */
		else if (proc->status != PROC_STATUS_ACTIVE)
			/* close now */
			xclose(proc->fd);

		/* if disabled, only call waitpid and continue */
		if (proc_enabled[proc->subsyst] == 0) {
			proc->status = PROC_STATUS_DEAD;
			continue;
		}

		swarn("%s: '%s' child %d %s", f, proc_systnames[proc->subsyst],
		      proc->idx, proc_state2str(status));
		/* store old state and set new state to be DEAD */
		oldstate = proc->status;
		proc->status = PROC_STATUS_DEAD;
		/* try to restart now if process was free */
		if (oldstate == PROC_STATUS_FREE) {
			dbg("%s: attempting to restart '%s' child", f,
			      proc_systnames[proc->subsyst]);
			proc_refork(proc);
		} else
			dbg("%s: postponing restart for '%s' child", f,
			    proc_systnames[proc->subsyst]);
	}
	/* XXX when is return value zero? */
	if (pid == -1 && errno != ECHILD)
		ERR("waitpid");

	/* block SIGCHLD while resetting signal notifier */
	sigemptyset(&chldmask);
	sigaddset(&chldmask, SIGCHLD);
	if (sigprocmask(SIG_BLOCK, &chldmask, &savemask) == -1)
		ERR("sigprocmask");

	/* critical region */
	sigchld_received = 0;

	/* restore signal mask */
	sigprocmask(SIG_SETMASK, &savemask, NULL);
}


/* proc_shutdown - shutdown subsystem, killing all active child processes */
void
proc_shutdown(enum proc_subsyst subsyst)
{
	const char *f = "proc_shutdown";
	int i;

	/* call for for convenience (death can happen after call) */
	proc_checksig();

	if (proc_enabled[subsyst] == 0) {
		dbg("%s: '%s' already inactive, doing nothing", f,
		    proc_systnames[subsyst]);
		return;
	}
	ASSERT(proc_enabled[subsyst] != 0, "shutdown of inactive subsystem");

	/* disable subsystem */
	sinfo("%s: shutting down '%s' subsystem", f, proc_systnames[subsyst]);
	proc_enabled[subsyst] = 0;

	/*
	 * kill any running processes.
	 * need to walk through entire array, in case of child deaths,
	 * there might be holes with dead processes in the array.
	 */
	for (i = 0; i < proc_conf[subsyst].max; i++) {
		if (procs[subsyst][i].status == PROC_STATUS_ACTIVE ||
		    procs[subsyst][i].status == PROC_STATUS_FREE)
			proc_kill(&procs[subsyst][i]);
	}
}


/* proc_sigchld - handler for SIGCHLD signal */
void
proc_sigchld(int sig)
{
	ASSERT(sig == SIGCHLD, "unexpected signal received");

	sigchld_received = 1;
}


/* proc_pid2proc - obtain process pointer for given pid */
struct proc *
proc_pid2proc(pid_t pid)
{
	enum proc_subsyst ss;
	int i;

	/* walk through subsystems/processes looking for pid */
	for (ss = 0; ss < PROC_MAX; ss++) {
		for (i = 0; i < proc_conf[ss].max; i++)
			if (procs[ss][i].pid == pid)
				return &procs[ss][i];
	}

	/* pid not found, usually bad news */
	return NULL;
}


/* proc_getproc - obtain unused resolver socket, or NULL if all busy */
struct proc *
proc_getproc(enum proc_subsyst subsyst)
{
	const char *f = "proc_getproc";
	int i;

	/* try to find free resolver process */
	/*
	 * need to walk through entire array, in case of child deaths,
	 * there might be holes with dead processes in the array
	 */
	for (i = 0; i < proc_conf[subsyst].max; i++) {
		/*
		 * if the child died but proc_checksig has not yet
		 * been called, operating on the file descriptor
		 * should show as poll error for the caller
		 */
		if (procs[subsyst][i].status == PROC_STATUS_FREE) {
			procs[subsyst][i].status = PROC_STATUS_ACTIVE;
			dbg("%s: activating '%s' process", f,
			    proc_systnames[subsyst]);
			return &procs[subsyst][i];
		}
	}

	return NULL;
}


/* proc_submitjob - send next request in queue to specified process */
void
proc_submitjob(enum proc_subsyst subsyst)
{
	const char *f = "proc_submitjob";
	struct proc_callback *pcb;
	struct proc *proc;

	/* expectations towards function behavior expressed in asserts */

	ASSERT(proc_enabled[subsyst] != 0, "submit to disabled subsystem");

	proc = proc_getproc(subsyst);
	ASSERT(proc != NULL, "expected process to be available");
	ASSERT(proc->status == PROC_STATUS_ACTIVE, "bad status");

	pcb = aqueue_dequeue(proc_conf[subsyst].aq_waiting);
	/* would need to hand back process if nothing on queue */
	ASSERT(pcb != NULL, "submitjob called with empty queue");
	dbg("%s: sending job to process %d", f, proc->idx);
	pcb->gotdesc(pcb->data, proc);
	free(pcb);
}


/* proc_fork - fork new process for given slot of specified subsystem */
int
proc_fork(enum proc_subsyst subsyst, int idx)
{
	const char *f = "proc_fork";
	sigset_t chldmask, savemask;
	struct proc *proc;
	int sv[2];
	int i;

	ASSERT(procs[subsyst][idx].status == PROC_STATUS_DEAD ||
	       procs[subsyst][idx].status == PROC_STATUS_NOPROC,
	       "invalid process entry status");

	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, sv) == -1) {
		swarn("%s: socketpair: %m", f);
		return -1;
	}
	/* get proc pointer, changed values will be replaced below */
	proc = &procs[subsyst][idx];

	/* block SIGCHLD until handler for new pid has been set up */
	sigemptyset(&chldmask);
	sigaddset(&chldmask, SIGCHLD);
	if (sigprocmask(SIG_BLOCK, &chldmask, &savemask) == -1)
		ERR("sigprocmask failure");
	if ((proc->pid = fork()) == -1) {
		swarn("%s: fork: %m", f);
		/* close should not fail if code is correct */
		xclose(sv[0]);
		xclose(sv[1]);
		proc->status = PROC_STATUS_NOPROC;
		return -1;
	}
	if (proc->pid == 0) {
		/*
		 * child
		 */

		/* restore signal mask */
		sigprocmask(SIG_SETMASK, &savemask, NULL);

		if (close(sv[1]) == -1) {
			swarn("%s: close: %m", f);
			_exit(1);
		}
		setproctitle("%s", proc_systnames[subsyst]);
/*		dbg("%s: resolver child, starting servop", f);*/

		/* close other descs */
		for (i = 0; i < getdtablesize(); i++) {
			/*XXX syslog, call log_preinit? */
			if (i == STDIN_FILENO ||
			    i == STDOUT_FILENO ||
			    i == STDERR_FILENO)
				continue;

			if (i != sv[0])
				(void)close(i);
		}

		/* start server */
		proc_conf[subsyst].childfunc(sv[0]);
		swarn("%s: '%s' process exited", f, proc_systnames[subsyst]);
		_exit(1);
	}

	/*
	 * parent
	 */

	xclose(sv[0]);

	proc->fd = sv[1];
	proc->status = PROC_STATUS_FREE;

	/* restore signal mask */
	sigprocmask(SIG_SETMASK, &savemask, NULL);

	return 0;
}


/* proc_refork - attempt to restart dead child process */
void
proc_refork(struct proc *proc)
{
	const char *f = "proc_refork";
	struct proc_callback *pcb;
	enum proc_subsyst subsyst;

	subsyst = proc->subsyst;

	/* check subsystem child mortality rate */
	proc_conf[subsyst].deaths++;
	if (proc_conf[subsyst].deaths > PROC_FATALITY_MAX) {
		swarn("%s: too many '%s' child deaths, giving up subsystem", f,
		      proc_systnames[subsyst]);
	} else {
		/* attempt to create new child */
		dbg("%s: attempting to restart '%s' child", f,
		      proc_systnames[subsyst]);
		if (proc_fork(subsyst, proc->idx) == 0)
			return; /* ok, reforked */
	}

	/* unable to keep proc active, reduce subsystem child count */
	ASSERT(proc_conf[subsyst].curr > 0, "invalid number of procs");
	proc_conf[subsyst].curr--;

	/* still any processes active? */
	if (proc_conf[subsyst].curr > 0)
		return; /* ok, let remaining processes do their work */

	/*
	 * no subsystem processes remaining, subsystem has failed.
	 */

	/* failure accepted? */
	if (proc_conf[subsyst].acceptfail == 0)
		serr("%s: '%s' child subsystem failed, exiting", f,
		     proc_systnames[subsyst]);

	/* disable subsystem and report failure to queued entries. */
	proc_enabled[subsyst] = 0;
	if (proc_queuesize(subsyst) > 0)
		dbg("%s: sending error to queued entries (%d)", f,
		    proc_queuesize(subsyst));
	while ((pcb = aqueue_dequeue(proc_conf[subsyst].aq_waiting)) != NULL) {
		pcb->noserv(pcb->data);
		free(pcb);
	}
}


/* proc_dump - show process status */
void
proc_dump(void)
{
	const char *f = "proc_dump";
	enum proc_subsyst ss;
	int i;

	sinfo("%s: child subsystem process status:", f);

	/* walk through subsystems/processes */
	for (ss = 0; ss < PROC_MAX; ss++) {
		if (proc_enabled[ss])
			sinfo(" -%s (enabled), %d enqueued", proc_systnames[ss],
			      proc_queuesize(ss));
		else
			sinfo(" -%s (disabled)", proc_systnames[ss]);
		for (i = 0; i < proc_conf[ss].max; i++)
			if (procs[ss][i].status == PROC_STATUS_NOPROC)
				sinfo("  %d NOPROC", procs[ss][i].idx);
			else if (procs[ss][i].status == PROC_STATUS_FREE)
				sinfo("  %d FREE", procs[ss][i].idx);
			else if (procs[ss][i].status == PROC_STATUS_ACTIVE)
				sinfo("  %d ACTIVE", procs[ss][i].idx);
			else if (procs[ss][i].status == PROC_STATUS_DEAD)
				sinfo("  %d DEAD", procs[ss][i].idx);
	}
}


/* proc_state2str - return pointer to string representation of wait status */
char *
proc_state2str(int status)
{
	static char str[PROC_STATESTR_MAX];

	if (WIFEXITED(status)) {
		snprintf(str, sizeof(str), "exited with return value %d",
			WEXITSTATUS(status));
		str[sizeof(str) - 1] = '\0';
	} else if (WIFSIGNALED(status)) {
		snprintf(str, sizeof(str),
			 "died after receiving signal %d%s",
			WTERMSIG(status),
			 WCOREDUMP(status) ? " (core generated)" : "");
		str[sizeof(str) - 1] = '\0';
	} else
		ERRX("unexpected status value");

	return str;
}
