/* protio.c - functions for protocol data io */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: protio.c,v 1.100 2010/08/02 22:24:56 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <errno.h>
#include <poll.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "log.h"
#include "block.h"
#include "callback.h"
#include "conn.h"
#include "io.h"
#include "peer.h"
#include "protio.h"
#include "queue.h"
#include "socket.h"
#include "stream.h"
#include "tcpmodel.h"
#include "tv.h"

#define READ_TIMEOUT_SEC 2
#define READ_TIMEOUT_MSEC 900000

#define MAX_LINE 1024 /* maximum length for line in request/response */

char *proto_idreq(char *);


/* protio - perform state dependent io-operation on connection */
void
protio(struct connection *c)
{
	const char *f = "protio";
	float td;

	/* system io possible? (input) */
	if ((c->pollevent & POLLOUT) == 0) {
		switch (c->status) {
		case CONNECTION_STATUS_RESOLVEWAIT: /* FALLTHROUGH */
		case CONNECTION_STATUS_RESOLVING:   /* FALLTHROUGH */
		case CONNECTION_STATUS_MCONN:       /* FALLTHROUGH */
		case CONNECTION_STATUS_MCONNWAIT:   /* FALLTHROUGH */
		case CONNECTION_STATUS_ACCEPTING:   /* FALLTHROUGH */
		case CONNECTION_STATUS_ENDED:
			/* ignore */
			break;

		case CONNECTION_STATUS_DEADLINESTREAM: /* FALLTHROUGH */
		case CONNECTION_STATUS_FOREVERSTREAM:  /* FALLTHROUGH */
		case CONNECTION_STATUS_UNCONNECTED:    /* FALLTHROUGH */
		case CONNECTION_STATUS_CONNECTING:     /* FALLTHROUGH */
			ERRX("unexpected pollevent");
			break;

		/* protocol input data in these states */
		case CONNECTION_STATUS_IDLE:       /* FALLTHROUGH */
		case CONNECTION_STATUS_READREP:    /* FALLTHROUGH */
		case CONNECTION_STATUS_READREPDAT: /* FALLTHROUGH */
		case CONNECTION_STATUS_READREQ:    /* FALLTHROUGH */
		case CONNECTION_STATUS_READREQDAT: /* FALLTHROUGH */
		/* a POLLIN event indicates error for cached/write states */
		case CONNECTION_STATUS_CACHED:     /* FALLTHROUGH */
		case CONNECTION_STATUS_SSLCACHED:  /* FALLTHROUGH */
		case CONNECTION_STATUS_WRITEREP:   /* FALLTHROUGH */
		case CONNECTION_STATUS_WRITEREQ:
			dbg("%s: calling read_conn: ibuflen: %d obuflen: %d",
			    f, c->ibuflen, c->obuflen);
			if (read_conn(c) == -1) {
				if (c->type == CONN_TYPE_OUTGOING &&
				    c->protops != NULL) {
					/*XXX return errno? */
					c->protops->connevent(c, CONNEVENT_REQUEST_FAIL, 0);
/*					c->protops->failedreq(c->uldat);*/
				}
				return;
			}
			dbg("%s: called read_conn: ibuflen: %d obuflen: %d", f,
			    c->ibuflen, c->obuflen);
			break;
		}
	}

	switch (c->status) {
	case CONNECTION_STATUS_UNCONNECTED:/* FALLTHROUGH */
	case CONNECTION_STATUS_RESOLVEWAIT:/* FALLTHROUGH */
	case CONNECTION_STATUS_MCONNWAIT:  /* FALLTHROUGH */
		ERRX("unexpected pollevent");
		break;

	case CONNECTION_STATUS_ACCEPTING:
		ASSERT((c->pollevent & POLLOUT) == 0, "input socket POLLOUT");

		/* new connection */
		(void)accept_connection(c->fd);
		break;

	case CONNECTION_STATUS_CACHED:
		ASSERT((c->pollevent & POLLOUT) == 0, "input socket POLLOUT");

		/* XXX close or set readreq */
		dbg("%s: cached peer: %s", f, c->peer->id);
		ERRX("new data on cached connection");
		/* NOTREACHED */

	case CONNECTION_STATUS_IDLE:
		ASSERT((c->pollevent == -1) ||
		       ((c->pollevent & POLLOUT) == 0), "input socket POLLOUT");

		/* connection reuse */
/*		dbg("%s: new data on idle connection", f);*/

		changecstatus(c, CONNECTION_STATUS_READREQ);

		/* FALLTHROUGH */
	case CONNECTION_STATUS_READREP: /* FALLTHROUGH */
	case CONNECTION_STATUS_READREQ:
		if (ibuf_nlnlterm(c) == 0)
			queue_add(QUEUE_IODONE, c);
		break;

	case CONNECTION_STATUS_READREPDAT: /* FALLTHROUGH */
	case CONNECTION_STATUS_READREQDAT:
		if (ibuf_explen(c) == 0)
			queue_add(QUEUE_IODONE, c);
		break;

	case CONNECTION_STATUS_RESOLVING:
		/* resolve response available */
		conn_getdnsrep(c);
		break;

	case CONNECTION_STATUS_MCONN:
		/* resolve response available */
		conn_getmconnrep(c);
		break;

	case CONNECTION_STATUS_CONNECTING:
		if (socketerr(c->fd)) {
			c->protops->connevent(c, CONNEVENT_CONNECT_FAIL, errno);
/*			c->protops->connerr(c->uldat, errno);*/
			changecstatus(c, CONNECTION_STATUS_ENDED);
		} else {
			td = tv_dur2s(&c->connstart);
			peer_addrtt(c->peer, td);
			dbg("%s: connect time to %s %0.2f s", f, c->peer->id,
			    td);
			if (conncomp(c) == 0) {
				c->protops->connevent(c, CONNEVENT_CONNECT_OK, 0);
/*				c->protops->conndone(c->uldat);*/
			}
		}
		break;

	case CONNECTION_STATUS_WRITEREP: /* FALLTHROUGH */
	case CONNECTION_STATUS_WRITEREQ:
		if (write_obuf(c) == 0)
			queue_add(QUEUE_IODONE, c);
		break;

	case CONNECTION_STATUS_ENDED:
		/* connection ended, nothing to do */
		break;

	default:
		dbg("%s: status: %s", f, conn_state2str(c->status));
		ERRX("unexpected status");
	}
}


/* dataio - perform data io read/write operation on ready connection */
void
dataio(struct connection *c)
{
	const char *f = "dataio";
	ssize_t res;
	char byte;

	/* already ended? */
	if (c->status == CONNECTION_STATUS_ENDED)
		return;

	ASSERT(c->status == CONNECTION_STATUS_DEADLINESTREAM ||
	       c->status == CONNECTION_STATUS_FOREVERSTREAM,
	       "conn not active");
	ASSERT(c->stream != NULL, "NULL stream");

	/* attempt to detect error on outgoing stream */
	if ((c->stream->type == STREAM_FROM_MEM) && (c->pollevent & POLLIN)) {
		/* try to read XXX other way to detect error? */
		if ((res = read(c->fd, &byte, 1)) == 0) {
			dbg("%s: end of connection detected", f);
			changecstatus(c, CONNECTION_STATUS_ENDED);
		} else if (res == -1) {
			dbg("%s: error on stream connection: %m", f);
			changecstatus(c, CONNECTION_STATUS_ENDED);
		} else
			ERRX("non-error pollevent on inactive stream");
		return;
	}

	if (c->stream->type == STREAM_TO_FILE) {
		ASSERT(c->pollevent & POLLIN, "STREAM_TO_FILE, but not POLLIN");

		stream_recv(c);
	} else if (c->stream->type == STREAM_FROM_MEM) {
		/* look at active stream */
		ASSERT(c->pollevent & POLLOUT, "expected POLLOUT");

		stream_send(c);
	} else
		ERRX("unexpected stream state");
}


/* iodone - protocol i/o done, request processing starting/finished */
void
iodone(struct connection *c)
{
	const char *f = "iodone";
	struct callback *cb;
	char *protname;

	/* XXX have protocol id in conn struct, reference callback array */
	switch (c->status) {
	case CONNECTION_STATUS_READREP:
		c->protops->connevent(c, CONNEVENT_GOT_REPLY, 0);
/*		c->protops->got_rep(c->uldat);*/
		break;

	case CONNECTION_STATUS_READREPDAT:
		c->protops->connevent(c, CONNEVENT_GOT_REPLYDAT, 0);
/*		c->protops->got_repdat(c->uldat);*/
		break;

	case CONNECTION_STATUS_WRITEREQ:
		/* read reply after having sent request */
		changecstatus(c, CONNECTION_STATUS_READREP);
		break;

	case CONNECTION_STATUS_READREQ:
		if ((protname = proto_idreq(c->ibuf)) == NULL ||
		    (cb = callback_find(protname)) == NULL) {
			swarn("%s: unsupported protocol '%s', "
			      "ending connection", f, protname);
			changecstatus(c, CONNECTION_STATUS_ENDED);
		} else {
/*			c->protops->connevent(c, CONNEVENT_GOT_REQUEST, 0);*/
			cb->got_req(c);
		}
		break;

	case CONNECTION_STATUS_READREQDAT:
		c->protops->connevent(c, CONNEVENT_GOT_REQUESTDAT, 0);
/*		c->protops->got_reqdat(c->uldat);*/
		break;

	case CONNECTION_STATUS_WRITEREP:
/*		ASSERT(c->ioendstat != 0, "invalid status");*/

		c->protops->connevent(c, CONNEVENT_WROTE_REPLY, 0);
/*		c->protops->sent_repdat(c->uldat);*/
/*		connstatus();*/
		if (c->ioendstat != 0)
			ioendcstatus(c);
		break;

	case CONNECTION_STATUS_ENDED:
		/* connection ended, nothing to do */
		break;

	default:
		dbg("%s: status: %s", f, conn_state2str(c->status));
		ERRX("unexpected state for completed io");
	}
}


/* proto_idreq - attempt to identify protocol in http-like request */
char *
proto_idreq(char *reqstr)
{
	const char *f = "proto_idreq";
	static char protname[CALLBACK_MAXNAME];
	size_t len;
	char *e;
	char *p;

	/* look for first whitespace */
	if ((p = strchr(reqstr, ' ')) == NULL) {
		dbg("%s: no whitespace", f);
		return NULL;
	}
	p++;
	/* look for second whitespace (should be before potential protname) */
	if ((p = strchr(p, ' ')) == NULL) {
		dbg("%s: no second whitespace", f);
		return NULL;
	}
	p++;
	/* look for / (char after potential protname) */
	if ((e = strchr(p, '/')) == NULL) {
		dbg("%s: no slash", f);
		return NULL;
	}
	len = e - p;
	if (len >= sizeof(protname)) {
		dbg("%s: too long (%ld)", f, (long)len);
		return NULL; /* too long */
	}
	strncpy(protname, p, len);
	protname[len] = '\0';

	dbg("%s: got request in protocol '%s'", f, protname);

	return protname;
}
