/* queues.c - simple connection queue handling functions */

/*
 * Copyright (c) 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: queue.c,v 1.15 2010/06/25 22:02:28 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include "log.h"
#include "aqueue.h"
#include "conn.h"
#include "urient.h"
#include "queue.h"

void queue_connprint(const struct connection *);
void queue_sslconnprint(const struct connection *);

/*
 * TODO: A way of handling different client priorities
 */

aqueue_t *aq_protio;
aqueue_t *aq_dataio;
aqueue_t *aq_connend;
aqueue_t *aq_reqs;
aqueue_t *aq_iodone;
aqueue_t *aq_resdone;
#if HAVE_OPENSSL
aqueue_t *aq_sslcomp;
aqueue_t *aq_sslprot;
#endif /* HAVE_OPENSSL */


/* queue_preinit - setup queues, should be used before usage */
void
queue_preinit(void)
{
	aq_protio = aqueue_init((void (*)(const void *))queue_connprint);
	aq_dataio = aqueue_init((void (*)(const void *))queue_connprint);
	aq_connend = aqueue_init((void (*)(const void *))queue_connprint);
	aq_reqs = aqueue_init((void (*)(const void *))queue_connprint);
	aq_iodone = aqueue_init((void (*)(const void *))queue_connprint);
	aq_resdone = aqueue_init((void (*)(const void *))queue_connprint);
#if HAVE_OPENSSL
	aq_sslcomp = aqueue_init((void (*)(const void *))queue_sslconnprint);
	aq_sslprot = aqueue_init((void (*)(const void *))queue_sslconnprint);
#endif /* HAVE_OPENSSL */
}


/* queue_add - add connection to specified queue */
void
queue_add(enum queue_type type, struct connection *c)
{
	switch (type) {
	case QUEUE_PROTIO:
		aqueue_enqueue(aq_protio, c);
		break;
	case QUEUE_DATAIO:
		aqueue_enqueue(aq_dataio, c);
		break;
	case QUEUE_CONNEND:
		aqueue_enqueue(aq_connend, c);
		break;
	case QUEUE_REQUESTS:
		aqueue_enqueue(aq_reqs, c);
		break;
	case QUEUE_IODONE:
		aqueue_enqueue(aq_iodone, c);
		break;
	case QUEUE_RESDONE:
		aqueue_enqueue(aq_resdone, c);
		break;
#if HAVE_OPENSSL
	case QUEUE_SSLCOMP:
		aqueue_enqueue(aq_sslcomp, c);
		break;
	case QUEUE_SSLPROT:
		aqueue_enqueue(aq_sslprot, c);
		break;
#else
	case QUEUE_SSLCOMP:
	case QUEUE_SSLPROT:
		ERRX("no ssl support");
#endif /* HAVE_OPENSSL */
	}
}


/* queue_empty - returns 1 if specified queue is empty, 0 otherwise */
int
queue_empty(enum queue_type type)
{
	int empty;

	empty = -1;
	switch (type) {
	case QUEUE_PROTIO:
		empty = aqueue_isempty(aq_protio);
		break;
	case QUEUE_DATAIO:
		empty = aqueue_isempty(aq_dataio);
		break;
	case QUEUE_CONNEND:
		empty = aqueue_isempty(aq_connend);
		break;
	case QUEUE_REQUESTS:
		empty = aqueue_isempty(aq_reqs);
		break;
	case QUEUE_IODONE:
		empty = aqueue_isempty(aq_iodone);
		break;
	case QUEUE_RESDONE:
		empty = aqueue_isempty(aq_resdone);
		break;
#if HAVE_OPENSSL
	case QUEUE_SSLCOMP:
		empty = aqueue_isempty(aq_sslcomp);
		break;
	case QUEUE_SSLPROT:
		empty = aqueue_isempty(aq_sslprot);
		break;
#else
	case QUEUE_SSLCOMP:
	case QUEUE_SSLPROT:
		ERRX("no ssl support");
#endif /* HAVE_OPENSSL */
	}
	ASSERT(empty != -1, "empty not set");

	if (empty)
		return 1;
	else
		return 0;
}


/* queue_pop - get first entry from specified queue, NULL if queue is empty */
struct connection *
queue_pop(enum queue_type type)
{
	struct connection *c;

	if (queue_empty(type))
		return NULL;

	c = NULL;
	switch (type) {
	case QUEUE_PROTIO:
		c = aqueue_dequeue(aq_protio);
		break;
	case QUEUE_DATAIO:
		c = aqueue_dequeue(aq_dataio);
		break;
	case QUEUE_CONNEND:
		c = aqueue_dequeue(aq_connend);
		break;
	case QUEUE_REQUESTS:
		c = aqueue_dequeue(aq_reqs);
		break;
	case QUEUE_IODONE:
		c = aqueue_dequeue(aq_iodone);
		break;
	case QUEUE_RESDONE:
		c = aqueue_dequeue(aq_resdone);
		break;
#if HAVE_OPENSSL
	case QUEUE_SSLCOMP:
		c = aqueue_dequeue(aq_sslcomp);
		break;
	case QUEUE_SSLPROT:
		c = aqueue_dequeue(aq_sslprot);
		break;
#else
	case QUEUE_SSLCOMP:
	case QUEUE_SSLPROT:
		ERR("no ssl support");
#endif /* HAVE_OPENSSL */
	}

	ASSERT(c != NULL, "NULL connection");

	return c;
}


/* queue_status - dump contents of all queues */
void
queue_status(void)
{
	dbg("queue_status: ");

	if (!aqueue_isempty(aq_protio)) {
		dbg(" protio: ");
		aqueue_dump(aq_protio);
	}

	if (!aqueue_isempty(aq_dataio)) {
		dbg(" dataio: ");
		aqueue_dump(aq_dataio);
	}

	if (!aqueue_isempty(aq_connend)) {
		dbg(" connend: ");
		aqueue_dump(aq_connend);
	}

	if (!aqueue_isempty(aq_reqs)) {
		dbg(" reqs: ");
		aqueue_dump(aq_reqs);
	}

	if (!aqueue_isempty(aq_iodone)) {
		dbg(" iodone: ");
		aqueue_dump(aq_iodone);
	}

	if (!aqueue_isempty(aq_resdone)) {
		dbg(" resdone: ");
		aqueue_dump(aq_resdone);
	}

#if HAVE_OPENSSL
	if (!aqueue_isempty(aq_sslcomp)) {
		dbg(" sslcomp: ");
		aqueue_dump(aq_sslcomp);
	}

	if (!aqueue_isempty(aq_sslprot)) {
		dbg(" sslprot: ");
		aqueue_dump(aq_sslprot);
	}
#endif /* HAVE_OPENSSL */
}


/* queue_connprint - print information about connection */
void
queue_connprint(const struct connection *c)
{
	dbg(" %s:%s - status: %d (fd: %d)", c->peername, c->peerport,
	    c->status, c->fd);
}


#if HAVE_OPENSSL
/* queue_sslconnprint - print information about ssl connection */
void
queue_sslconnprint(const struct connection *c)
{
	dbg(" %s:%s - status: %d ssl_cmd: %d (fd: %d)", c->peername,
	    c->peerport, c->status, c->ssl_cmd, c->fd);
}
#endif /* HAVE_OPENSSL */
