/* range.c - range string parsing functions */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: range.c,v 1.14 2010/06/25 22:02:29 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "log.h"
#include "blockmap.h"
#include "range.h"

static char *nextrange(char *);
static int isrange(char *);

/*
 * The functions in this file manipulate strings with number ranges
 * such as 0-19,25-199.
 */


/* range_strgen - generate string with block ranges for a file */
char *
range_strgen(const struct blockmap *bm)
{
	const char *f = "range_strgen";
	static char str[BLOCKRANGE_MAXLEN];
	int rangestart;
	int needcomma;
	size_t strlen;
	char num[10];
	int inrange;
	int32_t i;

	rangestart = 0; /* set to a value to please gcc */
	needcomma = 0; /* only first digit does not need comma */
	inrange = 0;

	strlen = sizeof(str);
	bzero(str, strlen);
	for (i = 0; i < bm->size; i++) {
		if (blockmap_test(bm, i) && !inrange) {
			/* start of new range */
			if (needcomma)
				strlcat(str, ",", strlen);

			snprintf(num, sizeof(num), "%d", i);
			if (strlcat(str, num, strlen) >= strlen) {
				dbg("%s: range to long", f);
				return NULL;
			}

			rangestart = i;
			needcomma = 1;
			inrange = 1;
		} else if (!blockmap_test(bm, i) && inrange) {
			/* end of range, add previous digit */
			if (rangestart != i - 1) {
				strlcat(str, "-", strlen);

				snprintf(num, sizeof(num), "%d", i - 1);
				if (strlcat(str, num, strlen) >= strlen) {
					dbg("%s: range too long", f);
					return NULL;
				}
			}
			inrange = 0;
		}
	}

	if (inrange && rangestart != i - 1) {
		/* end of range at end of file, add last block number */
		strlcat(str, "-", strlen);
		snprintf(num, sizeof(num), "%d", bm->size - 1);
		if (strlcat(str, num, strlen) >= strlen) {
			dbg("%s: range too long", f);
			return NULL;
		}
	}

	return str;
}


/* range_parse - parse range string, setting corresponding map values */
int
range_parse(struct blockmap *bm, char *range)
{
	const char *f = "range_parse";
	const char *errstr;
	int start, end;
	char *p, *q;
	char *next;
	int32_t i;

	/* do nothing in case of empty string */
	if (strlen(range) == 0)
		return 0;

	next = range;

	do  {
		p = next;

		/* start-end? */
		if (isrange(p)) {
			q = strchr(p, '-');
			ASSERT(q != NULL, "isrange without range char");
			*q = '\0';

			start = strtonum(p, 0, bm->size - 1, &errstr);
			if (errstr) {
				dbg("%s: invalid number %s: %s", f, p, errstr);
				return -1;
			}

			q++;
			p = q;

			/* one more range? */
			if ((next = nextrange(p)) != NULL) {
				*next = '\0';
				next++;
			}

			end = strtonum(p, 0, bm->size - 1, &errstr);
			if (errstr) {
				dbg("%s: invalid number %s: %s", f, p, errstr);
				return -1;
			}

			if (start >= end) {
				dbg("%s: end lower than start", f);
				return -1;
			}

			for (i = start; i <= end; i++)
				blockmap_set(bm, i);
		} else {
			/* only a single digit range */

			/* one more range? */
			if ((next = nextrange(p)) != NULL) {
				*next = '\0';
				next++;
			}

			start = strtonum(p, 0, bm->size - 1, &errstr);
			if (errstr) {
				dbg("%s: invalid number %s: %s", f, p, errstr);
				return -1;
			}

			blockmap_set(bm, start);
		}
	} while (next != NULL);

	return 0;
}


/* nextrange - return pointer to first comma, or NULL if no comma before NUL */
static char *
nextrange(char *str)
{
	int32_t i;

	for (i = 0; (size_t)i < strlen(str); i++) {
		if (str[i] == ',')
			return str + i;
	}

	return NULL;
}


/* isrange - is the first non-digit character a '-'? */
static int
isrange(char *str)
{
	int32_t i;

	for (i = 0; (size_t)i < strlen(str); i++) {
		if (!isdigit(str[i])) {
			if (str[i] == '-')
				return 1;
			else
				return 0;
		}
	}

	return 0;
}
