/* resolve.c - functions for non-blocking dns resolution via proc interface */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: resolve.c,v 1.7 2010/06/25 22:02:29 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/socket.h>

#include <netdb.h>
#include <string.h>

#include "log.h"
#include "ipc.h"
#include "proc.h"
#include "resolve.h"
#include "tv.h"

/*
 * storage for requests, field lengths should be equal to, or larger than,
 * the maximum length for the host/service names that can be submitted
 * typically the struct peer name/port fields).
 */
struct resolve_req {
	char name[NI_MAXHOST]; /* name to resolve */
	char port[NI_MAXSERV]; /* service to resolve */
};

struct resolve_rep {
	uint8_t entries; /* number of struct addrinfo entries */
	int code; /* return value from getaddrinfo */
};

int resolve_readreq(int, struct resolve_req *);
int resolve_handlereq(int, struct resolve_req *);


/* resolve_serv - resolver process, loop forever doing dns lookups */
void
resolve_serv(int fd)
{
	const char *f = "resolve_serv";
	struct resolve_req req;

	for (;;) {
		/* read query */
		if (resolve_readreq(fd, &req) == -1)
			break;

		/* process and send reply */
		if (resolve_handlereq(fd, &req) == -1)
			break;
	}
	swarn("%s: request processing error, resolve server ending", f);
}


/*
 * very little scope for error recovery if client and server gets out
 * of sync. little choice but to kill process/exit and hope problem
 * does not appear again. too many failures will result in shutdown of
 * the resolver subsystem.
 *
 * how many failures should be accepted?
 * -none, kill entire application (seems too strict, being able to handle
 *  some failures (such as user inadvertently killing resolver, seems sensible)
 * -small number (ok?)
 * -many/unlimited (seems likely to only hide serious problems)
 */

/* resolve_readreq - read request (at server) */
int
resolve_readreq(int fd, struct resolve_req *req)
{
	const char *f = "resolve_readreq";
	uint16_t reqlen;

	/* read length, then host name */
	if (readn(fd, &reqlen, sizeof(reqlen)) == -1)
		return -1;
	if (reqlen > sizeof(req->name)) {
		/*
		 * XXX could read request, ignore it and return
		 * EAI_OVERFLOW, which would remove the need to
		 * restart the child process, but probably overkill;
		 * same max lengths should be used in all parts of
		 * code.
		 */
		swarn("%s: host name too long", f);
		return -1;
	}
	if (readn(fd, req->name, reqlen) == -1)
		return -1;
	if (req->name[reqlen - 1] != '\0') {
		swarn("%s: host name improperly terminated", f);
		return -1;
	}

	/* read length, then port */
	if (readn(fd, &reqlen, sizeof(reqlen)) == -1)
		return -1;
	if (reqlen > sizeof(req->port)) {
		swarn("%s: service name too long", f);
		return -1;
	}
	if (readn(fd, req->port, reqlen) == -1)
		return -1;
	if (req->port[reqlen - 1] != '\0') {
		swarn("%s: service name improperly terminated", f);
		return -1;
	}

	return 0;
}


/* resolve_handlereq - process dns query and send reply */
int
resolve_handlereq(int fd, struct resolve_req *req)
{
	const char *f = "resolve_handlereq";
	struct addrinfo hints, *res;
	struct timeval start;
	int error;
	int ret;

	/* resolve query */
	bzero(&hints, sizeof(struct addrinfo));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	xgettimeofday(&start, NULL);
	error = getaddrinfo(req->name, req->port, &hints, &res);
	dbg("%s: getaddrinfo() time for %s:%s %0.2fs", f, req->name, req->port,
	    tv_dur2s(&start));

	ret = sendaddrinfo(fd, (error ? NULL : res), error);
	if (!error)
		/* result sent, free in child */
		freeaddrinfo(res);

	return ret;
}


/* resolve_sendreq - send host/service pair as text to server */
int
resolve_sendreq(int fd, char *host, char *port)
{
	uint16_t reqlen;

	/* send host name length (including NUL), followed by host name */
	reqlen = strlen(host) + 1;
	if (writen(fd, &reqlen, sizeof(reqlen)) == -1)
		return -1;
	if (writen(fd, host, reqlen) == -1)
		return -1;

	/* send service name length (including NUL), followed by service name */
	reqlen = strlen(port) + 1;
	if (writen(fd, &reqlen, sizeof(reqlen)) == -1)
		return -1;
	if (writen(fd, port, reqlen) == -1)
		return -1;

	return 0;
}


/* resolve_readrep - read reply from resolver process */
int
resolve_readrep(int fd, struct addrinfo **resp, int *code)
{
	return readaddrinfo(fd, resp, code);
}
