/* servfile.c - file related functions relevant only for server */

/*
 * Copyright (c) 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: servfile.c,v 1.23 2010/06/30 08:37:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include "log.h"
#include "block.h"
#include "filemeta.h"
#include "uri.h"
#include "urient.h"
#include "servfile.h"

int serversumfiles = 0;

char *pubdir = NULL; /* server media file directory */
uint32_t blocksize_default = 0;
extern char digest_default[];


/* servfile_preinit - specify default blocklen and content directory */
int
servfile_preinit(uint32_t blocksize, char *dir)
{
	const char *f = "servfile_preinit";

	/* only assert here, tests should be done earlier */
	ASSERT(block_validsize(blocksize), "invalid blocksize");

	if (access(dir, R_OK|X_OK) == -1) {
		swarn("%s: %s: %m", f, dir);
		return -1;
	}
	pubdir = dir;

	blocksize_default = blocksize;

	return 0;
}


/* servfile_metainit - initialize filemeta for given file on server */
int
servfile_metainit(struct urient *ue)
{
	const char *f = "servfile_metainit";
	char path[MAXPATHLEN];
	struct stat sb;
	size_t maxlen;

	dbg("%s", f);

	maxlen = sizeof(path);
	strlcpy(path, ue->prefix, maxlen);
	strlcat(path, "/", maxlen);
	if (strlcat(path, ue->filename, maxlen) >= maxlen) {
		dbg("%s: path too long", f);
		return -1;
	}

	if (stat(path, &sb) == -1) {
		dbg("%s: stat: %m", f);
		return -1;
	}

	/* XXX read metadata from file/command line */
	if ((ue->meta = filemeta_new(&ue->uri, sb.st_size,
				     blocksize_default,
				     digest_default)) == NULL) {
		dbg("%s: unable to add filemeta", f);
		return -1;
	}

	ue->master = 1;

	/* file on server, set as complete */
	ue->state = URIENT_STATE_COMPLETE;

	return 0;
}


/* published file - returns 0 if the file exists, -1 otherwise */
int
published_file(struct uri *uri)
{
	const char *f = "published_file";
	char path[MAXPATHLEN];
	size_t maxlen;

	/*
	 * Note, this function works independently of the
	 * rest of the functions that use urient structures.
	 * This function can in other words be called before
	 * an urient structure is created, but needs to be kept
	 * in sync with urient path changes.
	 */
	maxlen = sizeof(path);
	strlcpy(path, pubdir, maxlen);
	strlcat(path, "/", maxlen);
	if (strlcat(path, uri->path, maxlen) >= maxlen) {
		swarn("%s: path too long", f);
		return -1;
	}

	if (access(path, R_OK) == 0)
		return 0;
	else {
		swarn("%s: %s: %m", f, path);
		return -1;
	}
}


/* valid_path - check (decoded) path for undesired elements */
int
valid_path(char *filename)
{
	/* XXX incomplete */
	if (filename[0] == '.')
		return -1;

	/* XXX expanding path alternative */
	if (strstr(filename, "/.") != NULL)
		return -1;

	return 0;
}
