/* ssl.c - functions for ssl establishment and management */

/*
 * Copyright (c) 2006, 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: ssl.c,v 1.99 2010/10/11 08:58:00 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

/*
 * SSL code using OpenSSL.
 * Based on example code from 'SSL and TLS' by Eric Rescorla
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>

#include <string.h>
#include <unistd.h>

#include "log.h"
#include "conn.h"
#include "file.h"
#include "peer.h"
#include "peerset.h"
#include "queue.h"
#include "protio.h"
#include "socket.h"
#include "ssl.h"
#include "tv.h"

#if HAVE_OPENSSL

#define RANDFILE "/dev/urandom" /* XXX alt /dev/random */
#define PRIVATE_KEYFILE "privkey.pem"
#define USER_CERTFILE "hostcert.pem"
#define CA_CERTFILE "ca-cert.pem"

int s_server_session_id_context = 1; /*XXX*/
int use_ssl = 0;
SSL_CTX *ssl_ctx;

int password_cb(char *, int, int, void *);
void sslerr(void);
void ssl_endstate(struct connection *);


/* password_cb - dummy password callback */
int
password_cb(char *buf, int num, int rwflag, void *userdata)
{
	(void)buf;
	(void)num;
	(void)rwflag;
	(void)userdata;

	/* should be no passwords */
	ERRX("passwd");

	return 0;
}


/* sslerr - log ssl errors */
void
sslerr(void)
{
	const char *f = "sslerr";
	unsigned long error;

	while ((error = ERR_get_error()) != 0)
		swarn("%s: %s", f, ERR_error_string(error, NULL));
}


/* ssl_preinit - ssl initialization, must be called after digest_preinit */
int
ssl_preinit(void)
{
	const char *f = "ssl_preinit";
	SSL_METHOD *meth;

	SSL_library_init();
	SSL_load_error_strings();

	if ((access(USER_CERTFILE, R_OK) == -1) ||
	    (access(PRIVATE_KEYFILE, R_OK) == -1) ||
	    (access(CA_CERTFILE, R_OK) == -1)) {
		swarn("%s: no SSL keys/certificates", f);
		return -1;
	}

	dbg("%s: SSL initialization starting", f);

	meth = SSLv3_method();
	if ((ssl_ctx = SSL_CTX_new(meth)) == NULL) {
		swarn("%s: SSL_CTX_new", f);
		sslerr();
		return -1;
	}

	if (SSL_CTX_use_certificate_file(ssl_ctx, USER_CERTFILE,
					 SSL_FILETYPE_PEM) != 1) {
		swarn("%s: SSL_CTX_use_certificate_file: %s", f, USER_CERTFILE);
		sslerr();
		return -1;
	}

	SSL_CTX_set_default_passwd_cb(ssl_ctx, password_cb);

	if (SSL_CTX_use_PrivateKey_file(ssl_ctx, PRIVATE_KEYFILE,
					SSL_FILETYPE_PEM) != 1) {
		swarn("%s: SSL_CTX_use_PrivateKey_file: %s", f,
		      PRIVATE_KEYFILE);
		sslerr();
		return -1;
	}

	if (SSL_CTX_load_verify_locations(ssl_ctx, CA_CERTFILE, 0) != 1) {
		swarn("%s: SSL_CTX_load_verify_locations: %s", f, CA_CERTFILE);
		sslerr();
		return -1;
	}
	SSL_CTX_set_verify(ssl_ctx, SSL_VERIFY_PEER|SSL_VERIFY_CLIENT_ONCE|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);
	SSL_CTX_set_verify_depth(ssl_ctx, 1);

	if (RAND_load_file(RANDFILE, 1024) != 1024) {
		swarn("%s: RAND_load_file", f);
		sslerr();
		return -1;
	}

	if (SSL_CTX_set_cipher_list(ssl_ctx, "RC4-SHA") == 0) {
		swarn("%s: SSL_CTX_set_cipher_list", f);
		sslerr();
		return -1;
	}

	SSL_CTX_set_session_id_context(ssl_ctx,
				       (void *)&s_server_session_id_context,
				       sizeof(s_server_session_id_context));

	dbg("%s: SSL initialization finished (%s)", f, OPENSSL_VERSION_TEXT);
	use_ssl = 1;

	return 0;
}


/* ssl_protio - perform ssl related io operations */
void
ssl_protio(struct connection *c)
{
	const char *f = "ssl_protio";

	dbg("%s: %s %s", f, ssl_cmd2str(c->ssl_cmd),
	    ssl_state2str(c->ssl_cmdstate));
	switch (c->ssl_cmd) {
	case SSLCMD_CONNECT:
		ssl_tryconn(c);
		break;

	case SSLCMD_ACCEPT:
		ssl_tryaccept(c);
		break;

	/*XXX?*/
	case SSLCMD_READ:
		read_conn(c);
		break;

	case SSLCMD_WRITE:
		ERRX("SSLCMD_WRITE");
		break;

	case SSLCMD_CLOSE:
		ssl_close(c);
		break;

	default:
		ERRX("unexpected SSL state");
	}
}


/* ssl_close - end ssl protected connection */
void
ssl_close(struct connection *c)
{
	const char *f = "ssl_close";
	int ssl_serv;
	int ssl_err;
	int res;

	dbg("%s", f);

	if (c->peer->sslcached == c) {
		dbg("%s: removing cached ssl connection", f);
		c->peer->sslcached = NULL;
	}
	c->peer->sslactive = 0; /*XXX?*/

	ASSERT(c->ssl != NULL, "no ssl state");

	if (c->ssl_cmd != SSLCMD_CLOSE) {
		if (c->ssl_cmd != SSLCMD_NONE)
			dbg("%s: already executing cmd: %s", f,
			    ssl_cmd2str(c->ssl_cmd));

		/* XXX sess is per peer, sslconn per file */
		if (c->peer->sess != NULL) {
			dbg("%s: have session for %s, killing ssl connection",
			    f, c->peer->id);
			ssl_kill(c);
			return;
		}

		if (c->ssl_serv == 0)
			if ((c->peer->sess = SSL_get1_session(c->ssl)) == NULL)
				ERRX("SSL_get1_session");

		c->ssl_active = 0;
	}

	res = SSL_shutdown(c->ssl);
	if (res == 1) {
		dbg("%s: successful shutdown", f);
		SSL_free(c->ssl);
		c->ssl = NULL;
		/* store server state, will be cleared by ssl_endstate() */
		ssl_serv = c->ssl_serv;
		ssl_endstate(c);
		/* idle/cache depends on client/server/scc */
		if (ssl_serv == 1)
			idle_connection(c);
		else
			cache_connection(c);
	} else if (res == 0) {
		/*XXX ssl_close might be called multiple times */
		if (c->ssl_cmd == SSLCMD_CLOSE) {
			dbg("%s: multiple shutdown, kill", f);
			ssl_kill(c);
			return;
		}

		c->ssl_cmd = SSLCMD_CLOSE;
		dbg("%s: starting shutdown", f);
		c->ssl_cmdstate = SSLCMD_WANT_READ;
		/* XXX might be called on ended protocol */
/*		changecstatus(c, CONNECTION_STATUS_READREP);*/
	} else if (res == -1) {
		c->ssl_cmd = SSLCMD_CLOSE;

		ssl_err = SSL_get_error(c->ssl, res);
		switch(ssl_err) {
		case SSL_ERROR_WANT_READ:
			dbg("%s: SSL_ERROR_WANT_READ", f);
			c->ssl_cmdstate = SSLCMD_WANT_READ;
			break;
		case SSL_ERROR_WANT_WRITE:
			dbg("%s: SSL_ERROR_WANT_WRITE", f);
			c->ssl_cmdstate = SSLCMD_WANT_WRITE;
			break;
		default:
			sslerr();
			ERRX("ssl_close");
		}
	} else {
		sslerr();
		ERRX("SSL_shutdown");
	}
}


/* ssl_connect - start ssl connection negotiation */
void
ssl_connect(struct connection *c)
{
	const char *f = "ssl_connect";

	dbg("%s", f);

	ASSERT(c->uldat != NULL, "NULL uldat");
	ASSERT(c->obuflen == 0, "buffered data");

	if ((c->ssl = SSL_new(ssl_ctx)) == NULL) {
		sslerr();
		ERRX("SSL_new");
	}
	c->sbio = BIO_new_socket(c->fd, BIO_NOCLOSE);
	SSL_set_bio(c->ssl, c->sbio, c->sbio);
	if (c->peer->sess != NULL) {
		if (SSL_set_session(c->ssl, c->peer->sess) != 1) {
			sslerr();
			ERRX("SSL_set_session");
		}
		c->peer->sess = NULL;
	}

	c->ssl_cmd = SSLCMD_CONNECT;

	xgettimeofday(&c->sslstart, NULL);

	ssl_tryconn(c);
}


/* ssl_tryconn - check if ssl connection establishment is complete */
void
ssl_tryconn(struct connection *c)
{
	const char *f = "ssl_tryconn";
	int ssl_err;
	int r;

	ASSERT(c->ssl_cmd == SSLCMD_CONNECT, "not connect ssl cmd");

	dbg("%s", f);

	if ((r = SSL_connect(c->ssl)) <= 0) {
		ssl_err = SSL_get_error(c->ssl, r);
		switch(ssl_err) {
		case SSL_ERROR_WANT_READ:
			dbg("%s: SSL_ERROR_WANT_READ", f);
			c->ssl_cmdstate = SSLCMD_WANT_READ;
			return;
		case SSL_ERROR_WANT_WRITE:
			dbg("%s: SSL_ERROR_WANT_WRITE", f);
			c->ssl_cmdstate = SSLCMD_WANT_WRITE;
			return;
		default:
			sslerr();
			dbg("%s: %d errno: %d", f, ssl_err, errno);
			c->protops->connevent(c, CONNEVENT_SSLCONNECT_FAIL,
					      ssl_err);
			changecstatus(c, CONNECTION_STATUS_ENDED);
			c->ssl_cmd = SSLCMD_NONE;
			c->ssl_cmdstate = 0;
			return;
		}
	}
	dbg("%s: ssl connect time %0.2f s", f, tv_dur2s(&c->sslstart));

	if (SSL_session_reused(c->ssl) == 1)
		dbg("%s: previous ssl session reused", f);
	else
		dbg("%s: new ssl session (tryconn)", f);

	/* check_cert_chain()*//*XXX check that valid cache/server? */

	c->ssl_cmd = SSLCMD_NONE;
	c->ssl_cmdstate = 0;
	c->ssl_active = 1;

#if 0 /* don't want to lose c->uldat if cached */
	/* ready for use, cache connection */
	sslconn_cache(c);
#endif

	/* add to ssl establishment queue */
	queue_add(QUEUE_SSLCOMP, c);
}


/* ssl_tryaccept - server part of ssl establishment */
void
ssl_tryaccept(struct connection *c)
{
	const char *f = "ssl_tryaccept";
	char certname[256];
	int ssl_err;
	X509 *cert;
	int r;

	ASSERT(c->ssl_cmd == SSLCMD_ACCEPT, "not accept ssl cmd");

	dbg("%s", f);

	if ((r = SSL_accept(c->ssl)) <= 0) {
		ssl_err = SSL_get_error(c->ssl, r);
		switch(ssl_err) {
		case SSL_ERROR_WANT_READ:
			dbg("%s: SSL_ERROR_WANT_READ", f);
			c->ssl_cmd = SSLCMD_ACCEPT;
			c->ssl_cmdstate = SSLCMD_WANT_READ;
			return;
			/* NOTREACHED */
		case SSL_ERROR_WANT_WRITE:
			dbg("%s: SSL_ERROR_WANT_WRITE", f);
			c->ssl_cmd = SSLCMD_ACCEPT;
			c->ssl_cmdstate = SSLCMD_WANT_WRITE;
			return;
			/* NOTREACHED */
		default:
			sslerr();
			dbg("%s: %d errno: %d", f, ssl_err, errno);
			changecstatus(c, CONNECTION_STATUS_ENDED);
			SSL_free(c->ssl);
			c->ssl = NULL;
			return;
		}
	}
	dbg("%s: ssl accepted succeeded", f);

	c->ssl_active = 1;
	c->ssl_cmd = SSLCMD_NONE;
	c->ssl_cmdstate = 0;

	if (SSL_session_reused(c->ssl) == 1)
		dbg("%s: previous ssl session reused", f);
	else
		dbg("%s: new ssl session (tryaccept)", f);

	if (SSL_get_verify_result(c->ssl) != X509_V_OK) {
		sslerr();
		swarn("%s: invalid certificate", f);
		changecstatus(c, CONNECTION_STATUS_ENDED);
		return;
	}

	if ((cert = SSL_get_peer_certificate(c->ssl)) == NULL) {
		sslerr();
		dbg("%s: SSL_get_peer_certificate failed", f);
		ERR_print_errors(c->sbio);
		changecstatus(c, CONNECTION_STATUS_ENDED);
		return;
	}
	X509_NAME_get_text_by_NID(X509_get_subject_name(cert), NID_commonName,
				  certname, sizeof(certname));
	dbg("%s: got certificate for: %s", f, certname);
	X509_free(cert);
	/* XXX could retrieve info about user from db */

	/*
	 * new command will be received, set connection idle to clear
	 * state from starttls
	 */
	idle_connection(c);
}


/* ssl_starttls_serv - server part of start of ssl connection establishment */
void
ssl_starttls_serv(struct connection *c)
{
	const char *f = "ssl_starttls_serv";

	dbg("%s", f);

	ASSERT(c->ssl == NULL, "ongoing ssl session");
	ASSERT(c->obuflen == 0, "obuf nonzero");

	c->sbio = BIO_new_socket(c->fd, BIO_NOCLOSE);
	if ((c->ssl = SSL_new(ssl_ctx)) == NULL) {
		sslerr();
		ERRX("SSL_new");
	}
	SSL_set_bio(c->ssl, c->sbio, c->sbio);

	c->ssl_serv = 1;
	c->ssl_cmd = SSLCMD_ACCEPT;

	ssl_tryaccept(c);
}


/* ssl_read - perform read io on ssl connection */
int
ssl_read(struct connection *c, void *buf, size_t nbytes)
{
	const char *f = "ssl_read";
	unsigned int tot;
	int ssl_err;
	int r;

/*	dbg("%s", f);*/

	ASSERT(c->ssl != NULL, "no ssl data");

	tot = 0;
	do {
		r = SSL_read(c->ssl, (char *)buf + tot, nbytes - tot);
		if (r <= 0) {
			ssl_err = SSL_get_error(c->ssl, r);
			switch(ssl_err) {
			case SSL_ERROR_ZERO_RETURN:
				dbg("%s: zero return", f);
				c->ssl_cmd = SSLCMD_NONE;
				c->ssl_cmdstate = 0;
				ssl_close(c);
				break;
			case SSL_ERROR_WANT_READ:
				dbg("%s: SSL_ERROR_WANT_READ", f);
				dbg("%s: state: %s tot: %d", f,
				    conn_state2str(c->status), tot);

				c->ssl_cmd = SSLCMD_READ;
				c->ssl_cmdstate = SSLCMD_WANT_READ;
				if (tot == 0) {
					errno = EAGAIN;
					return -1;
				}
				break;
			case SSL_ERROR_WANT_WRITE:
				dbg("%s: SSL_ERROR_WANT_WRITE", f);
				c->ssl_cmd = SSLCMD_READ;
				c->ssl_cmdstate = SSLCMD_WANT_WRITE;
				if (tot == 0) {
					errno = EAGAIN;
					return -1;
				}
				break;
			case SSL_ERROR_SYSCALL:
				sslerr();
				if (r == 0)
					dbg("%s: unexpected EOF", f);
				else if (r == -1)
					dbg("%s: err: %m", f);
				else
					ERRX("ssl_read: SSL_ERROR_SYSCALL");
				ssl_kill(c);
				changecstatus(c, CONNECTION_STATUS_ENDED);
				break;
			default:
				sslerr();
				dbg("%s: %d errno: %d len: %d", f, ssl_err,
				    errno, r);
				ssl_kill(c);
				changecstatus(c, CONNECTION_STATUS_ENDED);
				c->ssl_cmd = SSLCMD_NONE;
				c->ssl_cmdstate = 0;
			}
		} else {
			tot += r;
			c->ssl_cmd = SSLCMD_NONE;
			c->ssl_cmdstate = 0;
		}
	} while (c->ssl != NULL && SSL_pending(c->ssl) && tot < nbytes);

/*	dbg("%s: (%d) %s", f, r, c->ibuf);*/

	return tot;
}


/* ssl_write - transmit data in buf via ssl protected connection */
int
ssl_write(struct connection *c, void *buf, size_t nbytes)
{
	const char *f = "ssl_write";
	int ssl_err;
	int r;

/*	dbg("%s: %s", f, c->obuf);*/

	ASSERT(c->ssl != NULL, "no ssl data");

	r = SSL_write(c->ssl, (char *)buf, nbytes);
	if (r <= 0) {
		ssl_err = SSL_get_error(c->ssl, r);
		switch(ssl_err) {
		case SSL_ERROR_ZERO_RETURN:
			dbg("%s: connection closed", f);
			c->ssl_cmd = SSLCMD_NONE;
			c->ssl_cmdstate = 0;
			ssl_close(c);
			r = -1;
			break;

		case SSL_ERROR_WANT_READ:
			dbg("%s: SSL_ERROR_WANT_READ", f);
			c->ssl_cmd = SSLCMD_WRITE;
			c->ssl_cmdstate = SSLCMD_WANT_READ;
			errno = EAGAIN;
			r = -1;
			break;

		case SSL_ERROR_WANT_WRITE:
			dbg("%s: SSL_ERROR_WANT_WRITE", f);
			c->ssl_cmd = SSLCMD_WRITE;
			c->ssl_cmdstate = SSLCMD_WANT_WRITE;
			errno = EAGAIN;
			r = -1;
			break;

		case SSL_ERROR_SYSCALL:
			sslerr();
			/*
			 * XXX can r==0 happen? will not SSL_ERROR_ZERO_RETURN
			 *     result instead?
			 */
			if (r == 0)
				dbg("%s: unexpected EOF", f);
			else if (r == -1)
				dbg("%s: %m", f);
			else
				ERRX("ssl_write: SSL_ERROR_SYSCALL");
			ssl_kill(c);
			changecstatus(c, CONNECTION_STATUS_ENDED);
			r = -1;
			break;

		default:
			sslerr();
			dbg("%s: %d errno: %d len: %d", f, ssl_err, errno, r);
			ssl_kill(c);
			changecstatus(c, CONNECTION_STATUS_ENDED);
			r = -1;
		}
	} else {
		c->ssl_cmd = SSLCMD_NONE;
		c->ssl_cmdstate = 0;
	}

	return r;
}


/* ssl_kill - forcefully end ssl state on connection */
void
ssl_kill(struct connection *c)
{
	const char *f = "ssl_kill";

	dbg("%s", f);

	ssl_endstate(c);
	changecstatus(c, CONNECTION_STATUS_ENDED);
}


/* ssl_endstate - reset all state on connection */
void
ssl_endstate(struct connection *c)
{
	if (c->ssl != NULL) {
		SSL_free(c->ssl);
		c->ssl = NULL;
	}
	c->peer->sess = NULL;
	if (c->peer->sslcached == c) {
		c->peer->sslcached = NULL;
		c->peer->sslactive = 0;
	}
	c->ssl_cmd = SSLCMD_NONE;
	c->ssl_cmdstate = 0;
	c->ssl_serv = 0;
	c->ssl_active = 0;

	c->protops->connevent(c, CONNEVENT_SSLCONNECTION_END, 0);
}


/* sslcomp - ssl operation completed, initiate ssl protected operation */
void
sslcomp(struct connection *c)
{
	ASSERT(c->ssl_active, "sslcomp called on inactive connection");

	changecstatus(c, CONNECTION_STATUS_WRITEREQ);
	xgettimeofday(&c->reqstart, NULL);

	c->peer->sslactive = 1;
	c->protops->connevent(c, CONNEVENT_SSLCONNECT_OK, 0);
/*	c->protops->sslcomp(c->uldat);*/
}


/* sslconn_cache - cache ssl connection to peer */
void
sslconn_cache(struct connection *c)
{
	const char *f = "sslconn_cache";

	ASSERT(c->peer->sslcached == NULL, "preexisting cached ssl conn");
	ASSERT(c->peer->sslactive == 1, "peer not ssl active");

	ASSERT(c->peer->addr != NULL, "caching to symbolic peer");

	dbg("%s: caching ssl connection to %s", f, c->peer->id);

	conn_endstate(c);

	c->peer->sslcached = c;
	changecstatus(c, CONNECTION_STATUS_SSLCACHED);
}

#else

int use_ssl = 0;

#endif /* HAVE_OPENSSL */


/* ssl_cmd2str - return string representation of ssl command */
char *
ssl_cmd2str(enum sslcmd cmd)
{
	switch(cmd) {
	case SSLCMD_NONE:
		return "SSLCMD_NONE";
	case SSLCMD_CONNECT:
		return "SSLCMD_CONNECT";
	case SSLCMD_ACCEPT:
		return "SSLCMD_ACCEPT";
	case SSLCMD_READ:
		return "SSLCMD_READ";
	case SSLCMD_WRITE:
		return "SSLCMD_WRITE";
	case SSLCMD_CLOSE:
		return "SSLCMD_CLOSE";
	}
	ERRX("no command matched");
	/*NOTREACHED*/
}


/* ssl_state2str - return string representation of ssl state */
char *
ssl_state2str(enum sslcmdstate state)
{
	switch(state) {
	case SSLCMD_WANT_READ:
		return "SSLCMD_WANT_READ";
	case SSLCMD_WANT_WRITE:
		return "SSLCMD_WANT_WRITE";
	}
	ERRX("no command matched");
	/*NOTREACHED*/
}
