/* stream.c - function for management of data streams (non-protocol data) */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: stream.c,v 1.79 2010/10/14 15:18:01 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/time.h>

#include <string.h>
#include <unistd.h>

#include "log.h"
#include "block.h"
#include "blocksum.h"
#include "conn.h"
#include "file.h"
#include "filemeta.h"
#include "io.h"
#include "iostat.h"
#include "peer.h"
#include "protio.h"
#include "servop.h"
#include "stream.h"
#include "tcpmodel.h"
#include "urient.h"
#include "tv.h"

int active_streams = 0;

/*
 * XXX
 * move from stream_foo to connstream_foo/conn_streamfoo?
 * change all functions taking 'stream *' arguments to 'connection *'
 */

/*
 * XXX TODO: make streams stop caring about block boundaries?  for
 *     stream data, outgoing packet sizes would ideally not be limited
 *     by block boundaries (or ignore this, letting kernel handle
 *     this?)
 */


/* stream_new - create a new stream for the given range of the given file */
struct stream *
stream_new(enum streamtype type, int fd, int fd2, struct urient *ue,
	   off_t startoff, off_t endoff)
{
	const char *f = "stream_new";
	struct stream *stream;
	struct block *block;
	uint32_t offset;

/*	dbg("%s: %d-%d", f, startoff, endoff);*/

	urient_fileref(ue);
	ASSERT(ue->file != NULL, "NULL file");

	stream = xmalloc(sizeof(struct stream));
	bzero(stream, sizeof(struct stream));

	stream->type = type;
	stream->sfd = fd;
	stream->dfd = fd2;
	stream->ue = ue;
	stream->bnum = off2block(ue, startoff);
	stream->state = STREAM_STOPPED;

	block = ue->file->blocks[stream->bnum];
	/* offset relative to start of block */
	offset = (uint32_t)startoff - stream->bnum * ue->meta->blocksize;
	if (type == STREAM_FROM_MEM)
		stream->blockoff = offset;
	else if (type == STREAM_TO_FILE) {
		/* continue incomplete downloads, if any */
		if (block->actlen > 0)
			stream->blockoff = block->actlen;
	} else
		ERRX("unexpected stream type")
	ASSERT(stream->blockoff < ue->meta->blocksize, "offset too large");

	stream->startoff = startoff;
	stream->endoff = endoff;

	dbg("%s: from %lld, blockoff: %d", f, (long long)stream_curroff(stream),
	    stream->blockoff);

	if (type == STREAM_FROM_MEM) {
		stream->readsize = ue->file->opt_blksize;
		stream->writesize = NETWORK_IOSIZE;
	} else if (type == STREAM_TO_FILE) {
		stream->readsize = NETWORK_IOSIZE;
		stream->writesize = ue->file->opt_blksize;
	} else
		ERRX("unexpected stream type")
	ASSERT(stream->readsize != 0, "zero readsize");

	dbg("%s: %s block %d,%d of type %s", f, ue->uri.uri, stream->bnum,
	    stream->blockoff, stream_type2str(type));

	active_streams++;

	xgettimeofday(&stream->ct, NULL);

	return stream;
}


/* stream_sendstart - start outgoing data stream */
void
stream_sendstart(enum streamdeadline dltype, struct connection *c,
		 struct urient *ue, off_t start, off_t end)
{
	c->stream = stream_new(STREAM_FROM_MEM, -1, c->fd, ue, start, end);
	c->stream->conn = c;

	c->stream->state = STREAM_ACTIVE;
	/* XXX could return e.g., start bnum + length */
	c->protops->streamevent(c->uldat, STREAMEVENT_STREAM_START, 0, 0);
/*	c->protops->streamstart(c->uldat);*/

	if (dltype == STREAM_NODEADLINE)
		changecstatus(c, CONNECTION_STATUS_FOREVERSTREAM);
	else if (dltype == STREAM_DEADLINE)
		changecstatus(c, CONNECTION_STATUS_DEADLINESTREAM);
	else
		ERRX("unexpected stream deadline type");

	stream_blockinit(c, c->stream->bnum);
}


/* stream_recvstart - start incoming data stream */
void
stream_recvstart(enum streamdeadline dltype, struct connection *c,
		 struct urient *ue, off_t start, off_t end)
{
	ASSERT(ue->file != NULL, "NULL file");

	c->stream =
		stream_new(STREAM_TO_FILE, c->fd, ue->file->fd, ue, start, end);
	c->stream->conn = c;

	c->stream->state = STREAM_ACTIVE;
	c->protops->streamevent(c->uldat, STREAMEVENT_STREAM_START, 0, 0);
/*	c->protops->streamstart(c->uldat);*/

	if (dltype == STREAM_NODEADLINE)
		changecstatus(c, CONNECTION_STATUS_FOREVERSTREAM);
	else if (dltype == STREAM_DEADLINE)
		changecstatus(c, CONNECTION_STATUS_DEADLINESTREAM);
	else
		ERRX("unexpected stream deadline type");

	stream_blockinit(c, c->stream->bnum);
}


/* stream_end - free resources associated with stream */
void
stream_end(struct stream *stream)
{
	const char *f = "stream_end";
	struct block *block;

	ASSERT(stream->ue->file != NULL, "NULL stream file");

	block = stream->ue->file->blocks[stream->bnum];

/*	dbg("%s", f);*/

	dbg("%s: %s: ending %s, block %d - %s after %0.2f iosum: %lld", f,
	    stream->ue->uri.uri, stream_type2str(stream->type), stream->bnum,
	    stream->conn->peer->id, tv_dur2s(&stream->ct),
	    (long long)stream->nbytes);

	block->source = NULL; /*did not get last byte from peer, clear */

	/* close descriptors XXX? */
	if (stream->type == STREAM_TO_FILE)
		/* no longer downloading block */
		block_unref(stream->ue, stream->bnum, BLOCKREF_WRITER);
	else if (stream->type == STREAM_FROM_MEM) {
		/* no longer sending block */
		if (!stream->sendfile)
			block_unref(stream->ue, stream->bnum, BLOCKREF_READER);
	} else
		ERRX("unexpected stream type")

	stream_free(stream);
}


/* stream_free - release stream resources (and notify stream end) */
void
stream_free(struct stream *stream)
{
	ASSERT(active_streams > 0, "negative active_streams value");
	active_streams--;

	stream->conn->protops->streamevent(stream->conn->uldat, STREAMEVENT_STREAM_END, 0, 0);
/*	stream->conn->protops->streamend(stream->conn->uldat);*/
	urient_fileunref(stream->ue);

	free(stream);
}


/* stream_wake - wake stream, called upon data becoming available */
void
stream_wake(struct stream *stream)
{
	const char *f = "stream_wake";

	/* should primarily be relevant for outgoing streams */
	ASSERT(stream->type == STREAM_FROM_MEM, "unexpected type");

	dbg("%s", f);

	stream_restart(stream);
	/* data should be available, start send */
	stream_send(stream->conn);
}


/* stream_blockinit - prepare stream for transmission of bnum */
void
stream_blockinit(struct connection *c, int32_t bnum)
{
	struct stream *stream;
	struct block *block;

	stream = c->stream;
	block = stream->ue->file->blocks[bnum];
	stream->bnum = bnum;

	if (stream->type == STREAM_TO_FILE) {
		block_ref(stream->ue, bnum, BLOCKREF_WRITER);
		/* set most recent content source for block */
		block->source = c->peer;
	} else if (stream->type == STREAM_FROM_MEM) {
#if HAVE_SENDFILE
		/*XXX send from memory if cached? */
		if (urient_haveblock(stream->ue, stream->bnum))
			stream->sendfile = 1;
#endif /* HAVE_SENDFILE */

		if (!stream->sendfile)
			block_ref(stream->ue, bnum, BLOCKREF_READER);
	} else
		ERRX("unexpected stream type");

	if (!stream->sendfile)
		ASSERT(block->blockdat != NULL, "no block data");

	xgettimeofday(&c->stream->blockstart, NULL);
}


/* stream_blockend - stream transfer of block bnum ended */
void
stream_blockend(struct connection *c, int32_t bnum)
{
	const char *f = "stream_blockend";
	struct stream *stream;
	float dur;

	stream = c->stream;

	/* only base reported speeds on whole block downloads */
	dur = tv_dur2ms(&stream->blockstart);
	/* only add rate for whole downloaded blocks */
	if (stream->blockstart.tv_sec != 0 &&
	    stream->type == STREAM_TO_FILE &&
	    stream->nbytes == stream->ue->meta->blocksize)
		peer_addrate(c->peer, dur, stream->nbytes);
	bzero(&stream->blockstart, sizeof(stream->blockstart));

	/* note: byte count might include multiple blocks for stream clients */
	dbg("%s: %s %s %s block %d complete after %0.2fs (%lld bytes, %0.2f mbps)",
	    f, stream->ue->uri.uri, c->peer->id, stream_type2str(stream->type),
	    bnum, dur/1000, (long long)stream->nbytes,
	    rate_len2mbps(dur, stream->nbytes));

	/* XXX for debugging */
	if (stream->type == STREAM_TO_FILE)
		digest_blocksum(stream->ue, stream->bnum);

	if (stream->type == STREAM_TO_FILE)
		block_unref(stream->ue, stream->bnum, BLOCKREF_WRITER);
	else if (stream->type == STREAM_FROM_MEM) {
		if (!stream->sendfile)
			block_unref(stream->ue, stream->bnum, BLOCKREF_READER);
	} else
		ERRX("unexpected stream type");

	c->age++;
}


/* stream_send - transmit data over stream (called when descriptor writable) */
void
stream_send(struct connection *c)
{
	const char *f = "stream_send";
	struct stream *stream;
	struct block *block;
	ssize_t wres;
	size_t wlen;
	off_t avail;

/*	dbg("%s", f);*/

	stream = c->stream;
	block = stream->ue->file->blocks[stream->bnum];

	ASSERT(stream != NULL, "stream NULL");
	ASSERT(stream->ue->file != NULL, "stream file NULL");
	ASSERT(block->actlen <= block->blocklen, "invalid block");

	if (stream->blockoff == 0)
		dbg("%s: sending block from %d,%d (%lld), %s to %s", f,
		    stream->bnum, stream->blockoff,
		    (long long)stream_curroff(stream),
		    stream->ue->uri.uri, c->peer->id);

	/* data available? */
	if ((avail = stream_availblock(stream)) == 0) {
		stream_stop(stream);
		c->protops->streamevent(c->uldat, STREAMEVENT_STREAM_STOPPED, 0, 0);
/*		c->protops->stream_stopped(c->uldat);*/
		return;
	}
	wlen = MIN(avail, stream->endoff + 1 - stream_curroff(stream));
	wlen = MIN(wlen, stream->writesize);
	if ((wres = block_send(stream, block, wlen)) == -1) {
		dbg("%s: send error, ending connection to %s", f, c->peer->id);
		changecstatus(c, CONNECTION_STATUS_ENDED);
		return;
	} else if (wres == 0)
		ERRX("zero length write");
	stream->blockoff += (uint32_t)wres;

	ios.net_wdat += wres;
	stream->nbytes += wres;
	c->protops->streamevent(c->uldat, STREAMEVENT_STREAM_WROTEDATA, 0, wres);
/*	c->protops->stream_wrotedata(c, wres);*/

	/*
	 * block complete?
	 *
	 * for a full block the blockoff value will be identical to
	 * the block length, while the block end offset (as returned
	 * by stream_blockendoff()) will be the block length - 1;
	 * need to compare the blockoff to the block end offset + 1
	 * to correctly detect the end.
	 */
	if (stream->blockoff < stream_blockendoff(stream, block) + 1)
		return;

	/*
	 * block complete, notify upper layer and optionally do next block
	 */

	/* range request? */
	if (stream->endoff + 1 < stream_curroff(stream))
		ERRX("stream passed endpoint");

	/* notification of block completion */
	stream_blockend(c, stream->bnum);
	c->protops->streamevent(c->uldat, STREAMEVENT_STREAM_SENTBLOCK,
				stream->bnum, 0);
/*	c->protops->stream_sentblock(c->uldat, stream->bnum);*/

	if (stream_complete(stream)) {
		stream_free(stream);
		c->stream = NULL;
/*		changecstatus(c, CONNECTION_STATUS_IDLE);*/
		idle_connection(c);
	} else {
		/* proceed with next block */
		stream->bnum++;
		stream->blockoff = 0;
		stream_blockinit(c, stream->bnum);
	}
}


/* stream_recv - receive data for stream (called when data readable) */
void
stream_recv(struct connection *c)
{
	const char *f = "stream_recv";
	struct stream *stream;
	struct block *block;
	ssize_t rres;
	size_t rlen;

/*	dbg("%s", f);*/

	stream = c->stream;
	block = stream->ue->file->blocks[stream->bnum];

	ASSERT(stream != NULL, "stream NULL");
	ASSERT(stream->ue->file != NULL, "stream file NULL");
	ASSERT(stream->endoff + 1 >= block2off(stream->ue, stream->bnum),
	       "stream ends before block");
	ASSERT(block != NULL, "block NULL");
	ASSERT(block->actlen <= block->blocklen, "invalid block");

	/*
	 * XXX ibufread used due to asymmetry between protocol and data i/o.
	 *     the length of data io is known beforehand, but this is not
	 *     the case for protocol data, where the endpoint is not known
	 *     before parsing. for e.g., the get command, this can result
	 *     in data being read into the ibuf buffer. calling ibufread
	 *     here results in additional copy operations (i.e., potentially
	 *     lower performance), but simplifies the code by not having
	 *     to first look for data in the ibuf buffer.
	 */
	rlen = MIN((off_t)stream->readsize,
		   stream_blockendoff(stream, block) + 1 - stream->blockoff);
	ASSERT(rlen != 0, "zero rlen");
	rres = ibufread(c, block->blockdat + stream->blockoff, rlen);
	if (rres == -1) {
		dbg("%s: ibufread failure", f);
		/* ibufread/read_conn marks connection as ended */
		c->protops->streamevent(c->uldat, STREAMEVENT_STREAM_FAIL, 0, 0);
/*		c->protops->failedreq(c->uldat);*/
		return;
	}
	ASSERT(rres != 0, "ibufread returned zero rres");

#if 0
	dbg("%s: data verification", f);
	verify_read(stream, block, rres);
#endif

	stream->nbytes += rres;
	stream->blockoff += (uint32_t)rres;
	block->actlen += (uint32_t)rres;
	ASSERT(block->actlen <= block->blocklen, "invalid block");
	ASSERT(stream->blockoff <= block->blocklen, "stream passed block end");

	c->protops->streamevent(c->uldat, STREAMEVENT_STREAM_READDATA, 0, rres);
/*	c->protops->stream_readdata(c, rres);*/

	/* not yet complete? */
	if (stream->blockoff != block->blocklen) {
		ASSERT(block->actlen != block->blocklen, "seemingly complete");
		return; /* not yet finished */
	}

	/* block download complete */
	ASSERT(stream->endoff + 1 >= stream_curroff(stream), "endpoint passed");

	urient_gotblock(stream->ue, stream->bnum);
	c->age++;
	stream_blockend(c, stream->bnum);
	stream_free(stream);
	c->stream = NULL;

	/* completion notification */
	c->protops->streamevent(c->uldat, STREAMEVENT_STREAM_COMPLETE, 0, 0);
/*	c->protops->stream_comp(c->uldat);*/
}


/* stream_stop - stop stream due (due to data being unavailable) */
void
stream_stop(struct stream *stream)
{
	const char *f = "stream_stop";

	ASSERT(stream->state == STREAM_ACTIVE, "stopping non-active stream");
	stream->state = STREAM_STOPPED;
	pollchange++; /* needed due to stream state change */

	dbg("%s: %s: stopping %s, block %d - %s after %0.2fs nbytes: %lld", f,
	    stream->ue->uri.uri, stream_type2str(stream->type), stream->bnum,
	    stream->conn->peer->id, tv_dur2s(&stream->ct),
	    (long long)stream->nbytes);
	xgettimeofday(&stream->stoptime, NULL);
}


/* stream_restart - restart stopped stream (due to data being available) */
void
stream_restart(struct stream *stream)
{
	const char *f = "stream_restart";

	ASSERT(stream->state == STREAM_STOPPED, "stream not waiting");
	stream->state = STREAM_ACTIVE;
	pollchange++; /* needed due to stream state change */

	ASSERT(stream_availblock(stream) > 0, "restart but has no data");

	/*XXX log duration stream stopped */
	dbg("%s: %s: restarting %s, block %d - %s after %0.2fs", f,
	    stream->ue->uri.uri, stream_type2str(stream->type), stream->bnum,
	    stream->conn->peer->id, tv_dur2s(&stream->stoptime));
}


/* stream_tooslow - abort too slow stream */
void
stream_tooslow(struct connection *c)
{
	const char *f = "stream_tooslow";
	struct stream *s;
	float dur;

	s = c->stream;

	sinfo("%s: aborting %s, block %d", f, s->ue->uri.uri, s->bnum);
	dur = tv_dur2ms(&s->blockstart);
	peer_tooslow(c->peer, s->nbytes, dur);

/*	dbg("%s: ending too slow connection: %p", f, c);*/
	changecstatus(c, CONNECTION_STATUS_ENDED);

	c->protops->streamevent(c->uldat, STREAMEVENT_STREAM_TOOSLOW, 0, 0);
/*	c->protops->stream_tooslow(c->uldat);*/

	/*
	 * end stream now to avoid race where new stream for same block
	 * started before this one properly ended
	 */
	stream_end(c->stream);
	c->stream = NULL;
}


/* stream_curroff - return current offset for stream */
off_t
stream_curroff(struct stream *stream)
{
	return block2off(stream->ue, stream->bnum) + stream->blockoff;
}


/* stream_blockendoff - end of block for current stream, from start of block */
uint32_t
stream_blockendoff(struct stream *stream, struct block *block)
{
	int32_t blockend;

	/*
	 * the block end offset can have values in the range (0, BLOCKLEN - 1).
	 * a block will typically be incomplete while it has a blockoff
	 * values in the same range, meaning that to detect block completion
	 * it will be necessary to check against 'stream_blockendoff() + 1'.
	 * this is due to the blockoff value starting at zero before anything
	 * is read/written. reading a single byte will give an blockoff of 1,
	 * and assuming a blocklength of 1, this will mean that block end offset
	 * is 1. the block will be complete when blockoff is 1 + 1.
	 */

	/* nearing end of stream request? */
	blockend = MIN(block->blocklen - 1,
		       stream->endoff - block2off(stream->ue, stream->bnum));
	ASSERT(blockend >= 0, "non-positive blockend");

	return blockend;
}


/* stream_complete - returns 1 if stream finished */
int
stream_complete(struct stream *stream)
{
/*	const char *f = "stream_complete";*/

	/*
	 * the stream offset is incremented upon i/o; the end is reached
	 * when the stream offset has passed the stream endoff
	 */
	if (stream->endoff + 1 == stream_curroff(stream)) {
		/* end of ranged stream request */
/*		dbg("%s: request complete at %d,%d", f, stream->bnum,
		    stream->blockoff);*/
		return 1;
	}

	return 0;
}


/* stream_type2str - return string representation of stream type */
char *
stream_type2str(enum streamtype type)
{
	switch (type) {
	case STREAM_FROM_MEM:
		return "STREAM_FROM_MEM";
	case STREAM_TO_FILE:
		return "STREAM_TO_FILE";
	}
	ERRX("no type matched");
	/*NOTREACHED*/
}


/* stream_state2str - return string representation of stream state */
char *
stream_state2str(enum streamstate state)
{
	switch (state) {
	case STREAM_ACTIVE:
		return "STREAM_ACTIVE";
	case STREAM_STOPPED:
		return "STREAM_STOPPED";
	}
	ERRX("no state matched");
	/*NOTREACHED*/
}


/* conn_streamev2str - return text representation of stream event */
char *
conn_streamev2str(enum streamevent event)
{
	switch (event) {
	case STREAMEVENT_STREAM_START:
		return "STREAMEVENT_STREAM_START";
	case STREAMEVENT_STREAM_TOOSLOW:
		return "STREAMEVENT_STREAM_TOOSLOW";
	case STREAMEVENT_STREAM_STOPPED:
		return "STREAMEVENT_STREAM_STOPPED";
	case STREAMEVENT_STREAM_FAIL:
		return "STREAMEVENT_STREAM_FAIL";
	case STREAMEVENT_STREAM_COMPLETE:
		return "STREAMEVENT_STREAM_COMPLETE";
	case STREAMEVENT_STREAM_END:
		return "STREAMEVENT_STREAM_END";
	case STREAMEVENT_STREAM_WROTEDATA:
		return "STREAMEVENT_STREAM_WROTEDATA";
	case STREAMEVENT_STREAM_READDATA:
		return "STREAMEVENT_STREAM_READDATA";
	case STREAMEVENT_STREAM_SENTBLOCK:
		return "STREAMEVENT_STREAM_SENTBLOCK";
	}
	ERRX("no state matched");
	/*NOTREACHED*/
}
