/* sumfile.c - file block digest io functions */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: sumfile.c,v 1.27 2010/06/25 22:02:29 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "log.h"
#include "digest.h"
#include "filemeta.h"
#include "iostat.h"
#include "sumfile.h"

/*
 * sumfile usage possibilities:
 *  -server
 *   A: do not want to generate sumfiles, no sumfiles pregenerated
 *   B: do not want to generate sumfiles, already pregenerated
 *      (apart from read/write open, no difference to below entry)
 *   C: want to generate sumfiles
 *    C1: pregenerated sumfiles exist (can be readonly->try open rdonly)
 *    C2: pregenerated sumfiles do not exist
 *  -cache
 *   D: want to generate sumfiles
 *    D1: pregenerated sumfiles exist
 *    D2: pregenerated sumfiles do not exist
 *
 *  A: no read, no write
 *  B: read, no write [sumfiles might be read-only, should open as RDONLY]
 *  C: read/write
 *  --
 *  D: read/write
 *
 * Gives following options:
 *
 *  A: open(O_RDONLY) -> fail, return NULL
 *  B: open(O_RDONLY) -> ok
 *  C: open(O_RDWR,O_CREAT) || open(O_RDONLY)
 *  --
 *  D: open(O_RDWR,O_CREAT)
 *
 *  O_CREAT redundant in this case
 *  O_RDONLY - server
 * (O_WRONLY - not applicable)
 *  O_RDWR - server/cache
 */


/* sumfile_open - open sumfile with given name, returning descriptor */
struct sumfile *
sumfile_open(struct filemeta *fm, char *path, enum sumfile_flags_t flags)
{
	const char *f = "sumfile_open";
	struct sumfile *suf;
	int fd;

	if (path == NULL) /* sumfile_namegen returns NULL on error */
		return NULL;

	switch (flags) {
	case o_rdwr:
		if ((fd = open(path, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR)) != -1)
			break;
		if (errno != EACCES) {
			swarn("%s: open: %s: %m", f, path);
			return NULL;
		}
		/* EACCES, try to open O_RDONLY in case of read-only file */
		/* FALLTHROUGH */
	case o_rdonly:
		if ((fd = open(path, O_RDONLY, S_IRUSR|S_IWUSR)) == -1) {
			swarn("%s: open: %s: %m", f, path);
			return NULL;
		}
		break;

	default:
		ERRX("unexpected flag");
	}

	suf = xmalloc(sizeof(struct sumfile));
	suf->fd = fd;
	suf->mdlen = digest_mdlen(fm->digest);

	return suf;
}


/* sumfile_write - store entry to disk */
void
sumfile_write(struct sumfile *suf, int32_t bnum, const char *entry)
{
	const char *f = "sumfile_write";
	ssize_t wres;
	char *asum;

	if (suf == NULL)
		return; /* ignore if no sumfile */

	ASSERT(strlen(entry) + 1 == (size_t)suf->mdlen, "invalid entry len");

	/* replace trailing zero with \n; */
	asum = xstrdup(entry);
	asum[strlen(asum)] = '\n';

	/* write at appropriate location */
	if (lseek(suf->fd, (off_t)bnum * suf->mdlen, SEEK_SET) == -1)
		serr("%s: lseek: %m", f);
	if ((wres = write(suf->fd, asum, suf->mdlen)) != (ssize_t)suf->mdlen)
		/*XXX what if disk full, only log failures for now */
		/*
		 * XXX sumentry will be invalid if the write was impartial,
		 *      will need to handle in some way
		 */
		swarn("%s: write: %m", f);
	if (wres > 0)
		ios.disk_woh += wres;

	free(asum);
}


/* sumfile_read - read sumfile entry from disk, must be freed by caller */
char *
sumfile_read(struct sumfile *suf, int32_t bnum)
{
	const char *f = "sumfile_read";
	ssize_t rres;
	char *sum;

/*	dbg("%s: %d", f, bnum);*/

	if (suf == NULL)
		return 0; /* ignore if no sumfile */

	/* seek to offset for block */
	if (lseek(suf->fd, (off_t)bnum * suf->mdlen, SEEK_SET) == -1)
		serr("%s: lseek: %m", f);

	sum = xmalloc(suf->mdlen);
	if ((rres = read(suf->fd, sum, suf->mdlen)) != (ssize_t)suf->mdlen) {
		if (rres == -1)
			swarn("%s: read: %m", f);
		else
			dbg2("%s: short read (read %lld)", f, (long long)rres);
		if (rres > 0)
			ios.disk_roh += rres;
		free(sum);
		return NULL;
	}
	ios.disk_roh += rres;

	/* valid format? */
	if (sum[0] == '\0') {
		/* assume not present, return NULL */
		free(sum);
		return NULL;
	} else if (sum[suf->mdlen - 1] != '\n') {
		swarn("%s: invalid sumfile format", f);
		free(sum);
		return NULL;
	}
	sum[suf->mdlen - 1] = '\0'; /* replace trailing '\n' with '\0' */

	return sum;
}


/* sumfile_close - close sumfile and free sumfile struct */
void
sumfile_close(struct sumfile *suf)
{
	if (suf == NULL)
		return;

	xclose(suf->fd);
	free(suf);
}


/* sumfile_namegen - generate name that encodes digest name and block size */
char *
sumfile_namegen(struct filemeta *fm, char *filepath)
{
	const char *f = "sumfile_namegen";
	static char path[MAXPATHLEN];
	size_t maxlen;
	int l;

	/*
	 * having digest name and blocksize in filename avoids
	 * checksum failures due to metadata changes.
	 * (XXX filesize could also be included)
	 */
	maxlen = sizeof(path);
	l = snprintf(path, maxlen, "%s-%s-%d" SUMFILE_POSTFIX,
		     filepath, fm->digest, fm->blocksize);
	if ((l == -1) || ((size_t)l >= maxlen)) {
		swarn("%s: path too long: %s", f, path);
		return NULL;
	}

	return path;
}
