/* sumfile.h - block digest functions, defines, and documentation */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: sumfile.h,v 1.9 2010/06/30 08:37:59 karls Exp $ */

#define SUMFILE_POSTFIX ".sum"

/* named flags based on open(2) param */
enum sumfile_flags_t { o_rdonly, o_rdwr };

struct sumfile {
	int fd;
	int mdlen;
};

/**
 * sumfile - disk based block digest caching
 *
 * The so-called sumfiles are used to store the file block message
 * digests to disk. When used, the sumfiles remove the need for
 * calculating the digest of a block more than once (ideally). As long
 * as the sumfile is maintained, it is possible to read the block
 * checksum from disk rather than having to recalculate it.
 *
 * The contents of a sumfile can be additionally be used to
 * reconstruct an overview of the contents of the cache in case of an
 * improper server shutdown, because the contents are updated each
 * time a block is written to disk (as opposed to on server shutdown).
 *
 * The on-disk format is as follows:
 *
 *  <block digest for block 0> '\n'
 *  <block digest for block 1> '\n'
 *  ...
 *  <block digest for block N> '\n'
 *
 *  Each line is as long as the block digest for the file, followed by
 *  a newline. The offset of a the digest for a given block is well
 *  defined as long as the digest used by the file is known.
 *
 *  The file name for the sumfiles is chosen to include the digest
 *  name and block size, to avoid incorrect checksums being used if
 *  these meta data values are changed (a file size change might still
 *  change the checksums, but this is likely not common enough to
 *  warrant the file size to be included).
 */

/**
 * sumfile_open - open block digest file
 *
 * The sumfile_open() function attempts to open a sumfile for the
 * specified filename, based on the given path name for a cache
 * file. The flags parameter is used to specify whether the file
 * should be attempted opened as read/write (o_rdwr) or read-only
 * (o_rdonly).
 *
 * If o_rdwr fails, the sumfile will be attempted opened as o_rdonly.
 *
 * The function returns a pointer to a new sumfile struct, or NULL in
 * case of failure.
 */
struct sumfile *
sumfile_open(struct filemeta *fm, char *, enum sumfile_flags_t);

/**
 * sumfile_write - add block digest to sumfile
 *
 * The supplied block digest for the given block is stored to disk
 * with this function. No operation is performed if the sumfile
 * pointer is NULL.
 */
void
sumfile_write(struct sumfile *, int32_t, const char *);

/**
 * sumfile_read - read block digest from sumfile
 *
 * The block digest for the given block is read from disk, if it
 * exists.
 *
 * Memory for the block is allocated and a pointer to it is returned
 * in case of success. The memory must be freed by the caller. NULL is
 * returned in case of failure.
 *
 * No operation is performed, and NULL is returned if the sumfile
 * pointer is NULL.
 */
char *
sumfile_read(struct sumfile *, int32_t);

/**
 * sumfile_close - close sumfile
 *
 * Close specified sumfile and free sumfile resources.
 */
void
sumfile_close(struct sumfile *);

/**
 * sumfile_namegen - generate sumfile name
 *
 * Returns pointer to a (static) string buffer with the sumfile
 * pathname for the specified file. Uses metadata to construct a
 * filename that includes the digest and blocksize.
 */
char *
sumfile_namegen(struct filemeta *, char *);
