/* tcpmodel.c - transfer rate estimation functions */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: tcpmodel.c,v 1.8 2010/10/11 09:07:47 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include "log.h"
#include "peer.h"
#include "peerpar.h"
#include "tcpmodel.h"

/*
 * TCP estimation functions. Based on the article
 * "The Macroscopic Behavior of the TCP Congestion Avoidance Algorithm"
 * by Matthew Mathis, Je rey Semke, Jamshid Mahdavi and Teunis Ott.
 *
 * Original function slightly modified in order to handle LAN speeds
 * (correlation between rtt/bw not as relevant as on a WAN).
 */


/* tcpmodel - return estimated (maximum) rate in mbps, from rtt (in seconds) */
float
tcpmodel(float rtt, int mss)
{
	float mbps;
	float k;

	ASSERT(rtt != 0 && mss != 0, "zero rtt/mss");

	mss *= 8; /* bytes to bits */
	k = mss / RTT_LOSSP_SQRT;
	mbps = k/rtt;
	mbps /= 1024 * 1024; /* bps to mbps */

	/* minor modification to model, attempt to handle lan speeds */
	return MIN(RTT_MINLAN, mbps);
}


/* rttmbps - return estimate of transfer rate from peer, based on rtt */
float
rttmbps(struct peer *peer)
{
	float rtt;
	int mss;

	if (peerpar_getfloat2(peer, PEERPAR_LINK_RTT, &rtt) == -1)
		return 0;
	if (rtt == 0)
		return 0;
	rtt /= 1000; /* ms to sec */
	if ((mss = peerpar_getint(peer, PEERPAR_LINK_MSS)) == 0)
		return 0;

	return tcpmodel(rtt, mss);
}


/* peer_rate - return best available estimate of peer speed */
float
peer_rate(struct peer *peer)
{
	float val;

	/* use throughput estimate if it is available */
	if (peerpar_getfloat2(peer, PEERPAR_LINK_RATE, &val) == 0)
		return val;

	/* if not, use latency estimate if it is available */
	if (peerpar_getfloat2(peer, PEERPAR_LINK_RTT, &val) == 0)
		return rttmbps(peer); /* rtt based tcp throughput estimate */

#if 0
	/*XXX needs to be returned to client */
	/* if not, use server rtt estimate, if available */
	if (peerpar_getfloat2(peer, PEERPAR_SPP_SERVEST, &val) == 0)
		return rttmbps(peer); /* rtt based tcp throughput estimate */
#endif

	/* if no data is available for throughput estimate, set to zero */
	return 0.0;
}


/* peer_xfrdeadline - return deadline for transfer from peer (in seconds) */
float
peer_xfrdeadline(struct peer *peer, off_t length)
{
	const char *f = "peer_xfrdeadline";
	float deadline;
	float range_mb;
	int have_rate, have_rtt;
	float rate, rtt;

	ASSERT(length != 0, "zero length");

	have_rate = peerpar_getfloat2(peer, PEERPAR_LINK_RATE, &rate);
	have_rtt = peerpar_getfloat2(peer, PEERPAR_LINK_RTT, &rtt);
	if (have_rate == -1 && have_rtt == -1) {
		peer->minmbps = PEER_MINSPEED;
		dbg("%s: minbps for %s %0.2f Mbit/s, MINSPEED, no data known",
		    f, peer->id, peer->minmbps);
	} else if (have_rate == -1) {
		peer->minmbps = rttminmbps(peer);
		dbg("%s: minbps for %s %0.2f Mbit/s, based on RTT (%0.2f ms)",
		    f, peer->id, peer->minmbps, rtt);
	} else {
		peer->minmbps = rate / peerdiv(peer);
		dbg("%s: minbps for %s %0.2f Mbit/s, based on rate (%0.2f MBit/s)",
		    f, peer->id, peer->minmbps, rate);
	}
	ASSERT(peer->minmbps != 0, "zero minmbps");

	/*XXX could consider urgency */

	/* calculate byte range in mbit */
	range_mb = length; /*byte length*/
	range_mb *= 8; /* ->bits */
	range_mb /= 1024 * 1024; /* ->mbit */

/*	dbg("%s: %lld (%f mbit)", f, length, range_mb);*/

	/* get max duration of block transfer for peer */
	deadline = range_mb / peer->minmbps;
	/* too much variability at low values */
	if (deadline < 0.5)
		deadline = 0.5;
	dbg("%s: deadline for %lld bytes %0.2f s (%0.2f mbps)", f,
	    (long long)length, deadline, peer->minmbps);

	return deadline;
}


/* peer_tooslow - notification of too slow transfer, dur in ms */
void
peer_tooslow(struct peer *peer, off_t transmitted, float dur)
{
	const char *f = "peer_tooslow";
	int have_rate, have_rtt;
	float oldrate, oldrtt;
	int tooslow;
	float mbps;

	ASSERT(peer->minmbps != 0, "no rate limit set");

	mbps = rate_len2mbps(dur, transmitted);

	/* use 0/PEER_RTT_UNKN in logmessage if no value set */
	oldrate = 0;
	oldrtt = PEER_RTT_UNKN;
	have_rate = peerpar_getfloat2(peer, PEERPAR_LINK_RATE, &oldrate);
	have_rtt = peerpar_getfloat2(peer, PEERPAR_LINK_RTT, &oldrtt);

	tooslow = peerpar_getint(peer, PEERPAR_PEER_TOOSLOW);

	dbg("%s: %s (%f < %f), len: %lld, rtt: %0.2f mbps: %0.2f (#%d)", f,
	    peer->id, mbps, peer->minmbps, (long long)transmitted, oldrtt,
	    oldrate, tooslow);

	/* adjust estimate if nothing transmitted */
	if (transmitted == 0) {
		if (have_rate != -1 && oldrate > 0) {
			dbg("%s: nothing transmitted, halving rate", f);
			oldrate /= 2;
			peerpar_setfloat(peer, PEERPAR_LINK_RATE, oldrate);
		} else if (have_rtt != -1 && oldrtt > 0) {
			dbg("%s: nothing transmitted, doubling rtt", f);
			oldrtt *= 2;
			peerpar_setfloat(peer, PEERPAR_LINK_RTT, oldrtt);
		}
	} else {
		/* only add rate if no value known, halve otherwise */
		if (have_rate == -1)
			peer_addrate(peer, dur, transmitted);
		else
			peerpar_setfloat(peer, PEERPAR_LINK_RATE, oldrate / 2);
	}

#if 0
	/* XXX have 'too slow' code? */
	if (tooslow > 3)
		peer->lastres = PEER_PREVFAIL;
#endif

	peerpar_adjint(peer, PEERPAR_PEER_TOOSLOW, 1);
}


/* peer_addrtt - sanity check rtt (given in ms) value and add to node */
void
peer_addrtt(struct peer *peer, float frtt)
{
	const char *f = "peer_addrtt";

	if (frtt > PEER_RTT_MAX || frtt == 0) {
		swarn("%s: ignoring invalid RTT %0.2f", f, frtt);
		return;
	}

	peerpar_addfloat(peer, PEERPAR_LINK_RTT, frtt);

	dbg("%s: new rtt for %s, %0.2f ms (added %0.2f)", f, peer->id,
	    peerpar_getfloat(peer, PEERPAR_LINK_RTT), frtt);
}


/* peer_addrate - add rate value for peer, dur in ms */
void
peer_addrate(struct peer *peer, float dur, off_t len)
{
	const char *f = "peer_addrate";
	float mbps;

	dbg("%s: %lld bytes transmitted in %f ms", f, (long long)len, dur);
	if (dur == 0 || len == 0)
		ERRX("peer_addrate zero param");

	mbps = rate_len2mbps(dur, len);
/*	dbg("%s: %f", f, mbps);*/

	if (mbps == 0) {
		dbg("%s: ignoring zero value", f);
		return;
	}

	peerpar_addfloat(peer, PEERPAR_LINK_RATE, mbps);

	dbg("%s: new rate for %s, %0.2f Mbit/s (added %0.2f)", f, peer->id,
	    peerpar_getfloat(peer, PEERPAR_LINK_RATE), mbps);
}


/* peerdiv - factor to set minimum accepted transfer rate, based on rtt */
int
peerdiv(struct peer *peer)
{
	if (peerpar_getfloat(peer, PEERPAR_LINK_RTT) < PEER_LAN_RTTLIM)
		return PEER_MINBPS_LANRATIO;
	else
		return PEER_MINBPS_WANRATIO;
}


/* rate_len2mbps - return rate based on length and duration (in ms) */
float
rate_len2mbps(float dur, off_t len)
{
	float mbps;

	mbps = len * 8;
	mbps /= dur/1000;
	mbps /= 1024 * 1024;

	return mbps;
}
