/* timers.c - timer queue management */

/*
 * Copyright (c) 2007, 2008, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: timers.c,v 1.23 2010/06/29 14:35:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <sys/time.h>

#include <string.h>

#include "log.h"
#include "timers.h"
#include "timers_int.h"

#define TIMEOUTS_MAX 1000 /*XXX*/

struct timeout *timeouts[TIMEOUTS_MAX];
int timeouts_size = 0;


/* tocmp - time based timeout sorting function for qsort */
int
tocmp(const void *a, const void *b)
{
	struct timeout *ta = *(struct timeout * const*)a;
	struct timeout *tb = *(struct timeout * const*)b;
	struct timeval *tva = &ta->when;
	struct timeval *tvb = &tb->when;

	/* active timeout? */
	if (!ta->active || !tb->active) {
		/* invert to get inactive timeouts at the end */
		if (ta->active > tb->active)
			return -1;
		else if (ta->active == tb->active)
			return 0;
		else
			return 1;
	}

	if (timercmp(tva, tvb, <))
		return -1;
	else if (timercmp(tva, tvb, ==))
		return 0;
	else
		return 1;
}


/* timeouts_sort - sort timeout values in timeout order */
void
timeouts_sort(void)
{
	qsort(timeouts, (size_t)timeouts_size, sizeof(struct timeout *), tocmp);
}


/* timeout_add - add new timeout */
struct timeout *
timeout_add(void (*callback)(void *), struct timeval *tv, void *dat)
{
	struct timeout *to;

	ASSERT(callback != NULL, "NULL callback");

	if (timeouts_size >= TIMEOUTS_MAX)
		ERRX("timeouts array full"); /* XXX grow */

	to = xmalloc(sizeof(struct timeout));
	bzero(to, sizeof(struct timeout));
	to->callback = callback;
	/* set to active if a timeout value was supplied */
	if (tv == NULL)
		to->active = 0;
	else {
		to->active = 1;
		to->when = *tv;
	}
	to->dat = dat;

	timeouts[timeouts_size] = to;
	timeouts_size++;

	timeouts_sort();

	return to;
}


/* timeout_on - enable previously disabled timeout */
void
timeout_on(struct timeout *to)
{
	const char *f = "timeout_on";

	ASSERT(to != NULL, "NULL timeout");

	if ((to->when.tv_sec == 0) && (to->when.tv_usec == 0))
		ERRX("enabling timer with unset time value");

	if (to->active == 1)
		swarn("%s: enabling already enabled timer", f);

	to->active = 1;
	timeouts_sort();
}


/* timeout_off - disable timeout */
void
timeout_off(struct timeout *to)
{
	const char *f = "timeout_off";

	ASSERT(to != NULL, "NULL timeout");

	if (to->active == 0)
		dbg("%s: deactivating deactivated timeout", f);

	to->active = 0;
	timeouts_sort();
}


/* timeout_upd - reset timeout value with the relative time specified by tv */
void
timeout_upd(struct timeout *to, struct timeval *tv)
{
	struct timeval now, new_tv;

	ASSERT((to != NULL) && (tv != NULL), "NULL input value");

	/* timeval specifies timeout time relative to current time */
	xgettimeofday(&now, NULL);
	timeradd(&now, tv, &new_tv);

	to->active = 1;
	to->when = new_tv;

	timeouts_sort();
}


/* timeout_upd - reset timeout value with the relative sec value */
void
timeout_secupd(struct timeout *to, int sec)
{
	struct timeval now;

	/* timeval specifies timeout time relative to current time */
	xgettimeofday(&now, NULL);
	now.tv_sec += sec;

	to->active = 1;
	to->when = now;

	timeouts_sort();
}


/* timeout_isactive - returns true if timer is active */
int
timeout_isactive(struct timeout *to)
{
	if (to->active)
		return 1;
	else
		return 0;
}


/* timeout_del - remove specified timeout */
void
timeout_del(struct timeout *to)
{
	int i;

	for (i = 0; i < timeouts_size; i++)
		if (timeouts[i] == to)
			break;
	ASSERT(i < timeouts_size, "timeout_del: timeout not found");

	free(timeouts[i]);
	if (i < timeouts_size - 1)
		timeouts[i] = timeouts[timeouts_size - 1];
	timeouts_size--;

	timeouts_sort();
}


/* timeouts_next - returns time to next timeout in milliseconds, -1 if none */
long
timeout_next(void)
{
	const char *f = "timeout_next";
	struct timeval now, tdiff;
	long dur;

	dur = -1;
	if ((timeouts_size > 0) && timeouts[0]->active) {
		xgettimeofday(&now, NULL);
		timersub(&timeouts[0]->when, &now, &tdiff);
		dur = tdiff.tv_sec * 1000;
		dur += (tdiff.tv_usec/1000);
		/* set minimal value for already expired timers */
		if (dur < 0) {
			dbg("%s: timeout already expired", f);
			dur = 0;
		}
	}

	return dur;
}


/* timeout_expired - return pointer to first expired timeout, NULL if none */
struct timeout *
timeout_expired(void)
{
	struct timeval now;
	struct timeout *to;

	to = timeouts[0];
	if ((timeouts_size > 0) && to->active) {
		xgettimeofday(&now, NULL);

		if (timercmp(&to->when, &now, <=)) {
			to->active = 0; /* set inactive upon expiry */
			timeouts_sort();

			return to;
		}
	}

	return NULL;
}


/* timeout_handle - call callbacks for pending timeouts */
void
timeout_handle(void)
{
	struct timeout *to;

	while ((to = timeout_expired()) != NULL)
		to->callback(to->dat);
}


/* timeout_status - list all currently active timeouts */
void
timeout_status(void)
{
	const char *f = "timeout_status";
	struct timeval now, td;
	int i;

	if (timeouts_size == 0)
		return;

	dbg("%s: timeouts:", f);
	xgettimeofday(&now, NULL);
	for (i = 0; i < timeouts_size; i++) {
		timersub(&timeouts[i]->when, &now, &td);
		if (timeouts[i]->active)
			dbg(" %ld.%06ld remaining", (long)td.tv_sec,
			    (long)td.tv_usec);
		else
			dbg(" inactive");
	}
}
