/* uri.c - URI decoding */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: uri.c,v 1.36 2010/10/06 08:17:24 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "log.h"
#include "uri.h"

/*
 * URIs are treated the following way:
 * -decoding of encoded characters
 * -normalization (spp:// prefix, lowercase host, always port number)
 * -division into separate parts (scheme, host, port, path)
 *
 * This still leaves room for some aliasing, e.g., no normalization of
 * path names is done; these paths might be the same depending on the
 * platform:
 *
 *  -foo/bar and foo//bar
 *  -foo/bar foo/../foo/bar
 *
 * The path is not checked for dangerous elements (such as files with
 * leading '.' characters in their filename). Any code using the path
 * value must handle this in an appropriate way.
 *
 * A more general solution to this problem would be to identify
 * files/blocks by their checksum value rather than their url.
 */

/*
 * Encoding needs to be done by the lhc/scc before sending requests.
 * Either receive the uri already encoded and it would be possible to
 * use this already encoded uri when sending it to the next step (this
 * might also be correct approach, XXX check rfc).
 *
 * Here the normalized uri is encoded before transmission. The
 * argument for this is that the received encoding might not be
 * valid. Reencoding attempts to ensure that it will be correct
 * (barring implementation bugs).
 */


/* uri_encode - encode uri, return pointer to static buffer with result */
char *
uri_encode(char *uri)
{
	const char *f = "uri_encode";
	static char enc[URI_MAXLEN * 3]; /* encoding requires three chars */
	char str[4];
	int i, j;
	int l;

	if (strlen(uri) >= sizeof(enc)) {
		dbg("%s: uri too long", f);
		return NULL;
	}

	bzero(enc, sizeof(enc));
	j = 0;
	for (i = 0; (size_t)i < strlen(uri); i++) {
		if (isalnum(uri[i]) ||
		    uri[i] == '-' || uri[i] == '/' || uri[i] == '_' ||
		    uri[i] == ':' || uri[i] == '.' || uri[i] == '#') {
			enc[j] = uri[i];
			j++;
		} else {
			l = snprintf(str, sizeof(str), "%%%02x",
				     uri[i]);
			if ((l == -1) || ((size_t)l >= sizeof(str)))
				ERRX("str buffer too short");
			if (strlcat(enc, str, sizeof(enc)) >= sizeof(enc))
				ERRX("encoded uri too long");
			j += strlen(str);
		}
	}

	return enc;
}


/* uri_unescape - ASCII charset URI decoding, dstlen includes trailing '\0' */
int
uri_unescape(char *dst, const char *src, size_t dstlen)
{
	const char *f = "uri_unescape";
	unsigned int val;
	const char *p;
	size_t i;
	char *q;
	int n;

	/* cannot add trailing '\0', return error */
	if (dstlen == 0)
		return -1;

	p = src;
	q = dst;
	i = 0;
	while ((i < dstlen - 1) && (*p != '\0')) {
		if (((*p != '%') || (strlen(p) < 3)) ||
		    !(isxdigit(*(p + 1)) && isxdigit(*(p + 2)))) {
			*(q + i) = *p;
		} else {
			n = sscanf(p, "%%%2X", &val);
			ASSERT(n == 1, "single value should be parsed");

			/* only ASCII for now */
			ASSERT(val <= 255, "non-ASCII value in URI");

			/* control characters are invalid XXXiso8859? */
			if ((val <= 0x1F) || (val >= 0x7F)) {
				dbg("%s: disallowed character encoding", f);
				return -1;
			}

			*(q + i) = (char)val;
			p += 2;
		}
		i++;
		p++;
	}
	*(q + i) = '\0';

	/*XXX have semantics more like strlcpy? */
	if (*p != '\0') {
		dbg("%s: encoded string too long", f);
		return -1;
	}

	return 0;
}


/* uri_parse - populate uri with fields from uritxt (modifies uritxt) */
int
uri_parse(struct uri *uri, const char *uritxt)
{
	const char *f = "uri_parse";
	char *scheme, *hier, *host, *port, *path;
	char uricpy[URI_MAXLEN];
	const char *errstr;
	unsigned int i;
	size_t maxlen;

	/*
	 * RFC2396 says decoding must be done on each element
	 * separately: first identify elements, then call unescape on
	 * each one, normalize, and finally, build a full normalized
	 * uri from elements.
	 */

	/* look for invalid characters in encoded uri */
	for (i = 0; i < strlen(uritxt); i++)
		if (!isgraph(uritxt[i])) {
			dbg("%s: invalid char in URI", f);
			return -1;
		}

	/* copy uri, to be able to modify */
	maxlen = sizeof(uricpy);
	if (strlcpy(uricpy, uritxt, maxlen) >= maxlen) {
		dbg("%s: URI too long", f);
		return -1;
	}

	/*
	 * identify uri components
	 */
	scheme = uricpy;

	if ((hier = strstr(uricpy, "://")) == NULL) {
		dbg("%s: not proper URI", f);
		return -1;
	}
	*hier = '\0';

	host = hier + (long)strlen("://");

	/* path, locate if exists */
	if ((path = strchr(host, '/')) != NULL) {
		*path = '\0';
		path++; /* path */
	}

	/* port, locate if exists */
	if ((port = strchr(host, ':')) != NULL) {
		*port = '\0';
		port++; /* port */
	}

	/*
	 * verify presence of mandatory fields
	 */
	if (strlen(scheme) == 0 || strlen(host) == 0) {
		dbg("%s: invalid URI", f);
		return -1;
	}

	/*
	 * start building (and normalizing) uri elements
	 */

	bzero(uri, sizeof(struct uri));

	/* scheme */
	if (uri_unescape(uri->scheme, scheme, sizeof(uri->scheme)) == -1) {
		dbg("%s: scheme decoding failed", f);
		return -1;
	}
	/* lowercase scheme */
	for (i = 0; (size_t)i < strlen(uri->scheme); i++)
		uri->scheme[i] = (char)tolower(uri->scheme[i]);

	/* host */
	if (uri_unescape(uri->host, host, sizeof(uri->host)) == -1) {
		dbg("%s: URI host decoding failed", f);
		return -1;
	}
	/* lowercase host */
	for (i = 0; (size_t)i < strlen(uri->host); i++)
		uri->host[i] = (char)tolower(uri->host[i]);

	/* port number, verify if set (keep as NUL otherwise) */
	if (port != NULL) {
		if (uri_unescape(uri->port, port, sizeof(uri->port)) == -1) {
			dbg("%s: URI port decoding failed", f);
			return -1;
		}
		(void)strtonum(uri->port, 1, 65535, &errstr);
		if (errstr) {
			dbg("%s: URI port: '%s': %s", f, port, errstr);
			return -1;
		}
	}

	/* path, simply copy (users of field need to treat it with caution) */
	if (path != NULL) {
		if (uri_unescape(uri->path, path, sizeof(uri->path)) == -1) {
			dbg("%s: URI path decoding failed", f);
			return -1;
		}
	} else /* path can remain NUL string */
		dbg("%s: notice: no path in URI", f);

	/*
	 * finally, assemble uri parts in full uri (normalized)
	 */
	if (uri_joinparts(uri) == -1)
		return -1;
	else
		return 0;
}


/* uri_joinparts - join each element into normalized uri */
int
uri_joinparts(struct uri *uri)
{
	const char *f = "uri_joinpart";
	size_t maxlen;

	maxlen = sizeof(uri->uri);
	/* scheme always spp, original value kept in uri->scheme */
	(void)strlcpy(uri->uri, uri->scheme, sizeof(uri->scheme));
	(void)strlcat(uri->uri, "://", maxlen);
	(void)strlcat(uri->uri, uri->host, maxlen);
	if (uri->port[0] != '\0') {
		(void)strlcat(uri->uri, ":", maxlen);
		if (strlcat(uri->uri, uri->port, maxlen) >= maxlen) {
			dbg("%s: URI too long", f);
			return -1;
		}
	}
	(void)strlcat(uri->uri, "/", maxlen);
	if (strlcat(uri->uri, uri->path, maxlen) >= maxlen) {
		dbg("%s: URI path too long", f);
		return -1;
	}

/*	dbg("%s: uri: %s %s %s %s -> %s", f, uri->scheme, uri->host, uri->port,
	    uri->path, uri->uri);*/

	return 0;
}
