/* urient.c - in memory database of URIs */

/*
 * Copyright (c) 2007, 2008, 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: urient.c,v 1.44 2010/07/21 13:44:46 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include <time.h>

#include "log.h"
#include "alist.h"
#include "file.h"
#include "cachefile.h"
#include "filemeta.h"
#include "urient.h"

struct alist *al_urient = NULL;

void urient_print(const struct urient *);
void urient_free(struct urient *);
char *urient_state2str(enum uestate);


/* urient_preinit - setup alist, must be called before other functions */
void
urient_preinit(void)
{
	al_urient = alist_init((void (*)(const void *))urient_print,
			       (void (*)(void *))urient_free,
			       URIENT_MAX_ENTRIES, ALIST_SIZE_FIXED,
			       ALIST_TYPE_STR);
}


/* urient_find - check if given uri exists in uri database, with expiry */
struct urient *
urient_find(struct uri *uri)
{
	const char *f = "urient_find";
	struct urient *ue;

	dbg("%s: %s", f, uri->uri);

	if ((ue = alist_find(al_urient, uri->uri)) == NULL)
		return NULL;

	return ue;
}


/* urient_add - add new uri to uri database (uri is copied) */
struct urient *
urient_add(struct uri *uri)
{
	const char *f = "urient_add";
	struct urient *ue;

	dbg("%s: %s", f, uri->uri);

	if (alist_find(al_urient, uri->uri) != NULL) {
		dbg("%s: uri already exists", f);
		return NULL;
	}

	ue = xmalloc(sizeof(struct urient));
	bzero(ue, sizeof(*ue));
	ue->uri = *uri;
	if (alist_add(al_urient, uri->uri, ue) == NULL) {
		free(ue);
		return NULL;
	}

	ue->state = URIENT_STATE_URI;

	return ue;
}


/* urient_del - remove urient from database */
void
urient_del(struct urient *ue)
{
	const char *f = "urient_del";

	dbg("%s: %s", f, ue->uri.uri);

	alist_del(al_urient, ue->uri.uri);
}


/* urient_setfile - specify filename/prefix (only name will be copied) */
int
urient_setfile(struct urient *ue, char *prefix, char *name)
{
	const char *f = "urient_setfile";
	size_t maxlen;

/*	ASSERT(ue->state == URIENT_STATE_xxx?, "invalid state");*/
	ASSERT(ue->prefix == NULL, "prefix already set");

	ue->prefix = prefix;
	maxlen = sizeof(ue->filename);
	if (strlcpy(ue->filename, name, maxlen) >= maxlen) {
		dbg("%s: path too long", f);
		return -1;
	}

	return 0;
}


/* urient_gotreq - got request, update lastreq time */
void
urient_gotreq(struct urient *ue)
{
	ue->lastreq = time(NULL);
}


/* urient_print - print urient state to log */
void
urient_print(struct urient const *ue)
{
	dbg(" URIENT: %s %s %s:%s", urient_state2str(ue->state),
	    ue->uri.uri,
	    ue->prefix == NULL ? "-" : ue->prefix,
	    ue->prefix == NULL ? "-" : ue->filename);
}


/* urient_free - release urient resources */
void
urient_free(struct urient *ue)
{
	/* urient_free is responsible for freeing filemeta if it is set */
	if (ue->meta != NULL)
		filemeta_free(ue->meta);
	if (ue->cf != NULL)
		cachefile_free(ue->cf);
	if (ue->uldat != NULL)
		ue->ueops->uldat_free(ue->uldat);

	free(ue);
}


/* urient_status - dump all urient entries */
void
urient_status(void)
{
	const char *f = "urient_status";

	dbg("%s: urient entries:", f);
	alist_dump(al_urient);
}


/* urient_decstr - parse string and set values in urient structure */
int
urient_decstr(struct urient *ue, char *str)
{
	const char *f = "urient_decstr";
	const char *err;
	char *timep;

	/*
	 * identify string elements
	 */

	timep = str;

	/*
	 * decode string elements
	 */

	/* last access time, should not be larger than current time */
	ue->lastreq = (time_t)strtonum(timep, 0, time(NULL), &err);
	if (err) {
		swarn("%s: invalid time %s: %s", f, timep, err);
		return -1;
	}

	return 0;
}


/* urient_encstr - return pointer to (static) string with urient data */
char *
urient_encstr(struct urient *ue)
{
	static char str[URIENT_ENCSTR_MAX];
	int l;

	/* build string */
	l = snprintf(str, sizeof(str), "%ld", (long)ue->lastreq);
	ASSERT((l != -1) && ((size_t)l < sizeof(str)), "data too long");

	return str;
}


/* urient_state2str - return string representation of urient state */
char *
urient_state2str(enum uestate state)
{
	switch (state) {
	case URIENT_STATE_NONE:
		return "NONE";
	case URIENT_STATE_URI:
		return "URI";
	case URIENT_STATE_META:
		return "META";
	case URIENT_STATE_INCOMPLETE:
		return "INCOMPLETE";
	case URIENT_STATE_COMPLETE:
		return "COMPLETE";
	}
	ERRX("no state matched");
	/*NOTREACHED*/
}
