/* urient_block.c - urient block related utility functions */

/*
 * Copyright (c) 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: urient_block.c,v 1.10 2010/06/25 22:02:29 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include <time.h>

#include "log.h"
#include "alist.h"
#include "block.h"
#include "blockmap.h"
#include "file.h"
#include "filemeta.h"
#include "cachefile.h"
#include "urient.h"


/* urient_subseq - return the number of blocks after bnum stored locally */
int32_t
urient_subseq(struct urient *ue, int32_t bnum)
{
	ASSERT(bnum < ue->meta->blockcnt, "invalid bnum");

	if (ue->state == URIENT_STATE_COMPLETE)
		return ue->meta->blockcnt - (bnum + 1);
	else {
		ASSERT(ue->cf != NULL, "cf not set");
		return blockmap_subseq(&ue->cf->bm, bnum);
	}
}


/* urient_blockcnount - return number of continous blocks from bnum */
int32_t
urient_blockcount(struct urient *ue, int32_t bnum)
{
	ASSERT(bnum < ue->meta->blockcnt, "invalid bnum");

	if (ue->state == URIENT_STATE_COMPLETE)
		return ue->meta->blockcnt - bnum;
	else {
		ASSERT(ue->cf != NULL, "cf not set");
		return blockmap_blockcount(&ue->cf->bm, bnum);
	}
}


/* urient_haveblock - return true if bnum is stored locally */
int
urient_haveblock(struct urient *ue, int32_t bnum)
{
	int res;

	ASSERT(bnum < ue->meta->blockcnt, "invalid bnum");

	if (ue->state == URIENT_STATE_COMPLETE)
		res = 1;
	else if (ue->cf != NULL && blockmap_test(&ue->cf->bm, bnum))
		res = 1;
	else
		res = 0;

	return res;
}


/* urient_availbytes - return number of bytes available from offset */
off_t
urient_availbytes(struct urient *ue, off_t offset)
{
	struct block *block;
	int32_t bnum;
	off_t bstart;
	off_t bend;
	off_t avail;

	ASSERT(offset < ue->meta->size, "invalid offset");

	if (ue->state == URIENT_STATE_COMPLETE)
		return ue->meta->size - offset;
	else {
		bnum = off2block(ue, offset);
		bstart = block2off(ue, bnum);
		block = ue->file->blocks[bnum];
		if (blockmap_test(&ue->cf->bm, bnum))
			bend = bstart + block->blocklen;
		else
			bend = bstart + block->actlen;
		if (offset >= bend)
			return 0; /* nothing available */
		avail = bend - offset;
		if (bnum + 1 == ue->meta->blockcnt ||
		    (blockmap_test(&ue->cf->bm, bnum) == 0 &&
		     block->actlen < block->blocklen))
			return avail;
		else
			return avail +
				urient_availbytes(ue, bend);
	}
}


/* urient_availblock - return number of bytes in block after offset */
uint32_t
urient_availblock(struct urient *ue, int32_t bnum, uint32_t blockoff)
{
	struct block *block;
	int32_t validlen;

	block = ue->file->blocks[bnum];

	/* whole block available? */
	if (urient_haveblock(ue, bnum))
		validlen = block->blocklen - blockoff;
	else {
		/* no, calculate how much available data there is */
		if (block->actlen > blockoff)
			validlen = block->actlen - blockoff;
		else
			validlen = 0;
	}

	return validlen;
}


/* off2block - return block number for given offset */
int32_t
off2block(struct urient *ue, off_t offset)
{
	ASSERT(offset < ue->meta->size, "invalid offset");

	return (int32_t)offset/ue->meta->blocksize;
}


/* use size - 1, when total filesize given as parameter, rather than offset */
/* block2off - return block number given offset */
off_t
block2off(struct urient *ue, int32_t bnum)
{
	return bnum * ue->meta->blocksize;
}


/* urient_iscomp - return true if all data has been downloaded */
int
urient_iscomp(struct urient *ue)
{
	if (ue->meta->blockcnt == urient_blockcount(ue, 0))
		return 1;
	else
		return 0;
}
