/* urient_state.c - urient state handling */

/*
 * Copyright (c) 2009, 2010
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  info@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadalleen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

/* $Id: urient_state.c,v 1.15 2010/06/30 08:37:59 karls Exp $ */

#ifdef HAVE_CONFIG_H
#include "autoconf.h"
#endif /* HAVE_CONFIG_H */

#include "log.h"
#include "block.h"
#include "conn.h"
#include "cachefile.h"
#include "diskcache.h"
#include "file.h"
#include "filemeta.h"
#include "servfile.h"
#include "socket.h"
#include "timers.h"
#include "uri.h"
#include "urient.h"

extern char *pubdir; /*XXX*/


/* urient_servinit - add urient for file, if it exists */
struct urient *
urient_servinit(struct uri *uri)
{
	const char *f = "urient_servinit";
	struct urient *ue;

	dbg("%s", f);

	/* request for valid file? (checks path validity) */
	if (valid_path(uri->path) == -1) {
		swarn("%s: path not permitted: %s", f, uri->path);
		return NULL;
	}

	/* do this check to avoid adding urient if file does not exist */
	if (published_file(uri) == -1) {
		swarn("%s: unable to serve requested uri %s", f, uri->uri);
		return NULL;
	}

	/* create urient for new request */
	if ((ue = urient_add(uri)) == NULL)
		return NULL;


	if (urient_setfile(ue, pubdir, uri->path) == -1) {
		urient_del(ue);
		return NULL;
	}

	if (servfile_metainit(ue) == -1)
		return NULL;

	return ue;
}


/* urient_cacheinit - initialize cache file (set metadata and create on disk) */
int
urient_cacheinit(struct urient *ue, off_t size, int32_t blocksize,
		 char *digest)
{
	const char *f = "urient_cacheinit";

	if ((ue->meta = filemeta_new(&ue->uri, size, blocksize, digest)) == NULL)
		return -1;
	ue->state = URIENT_STATE_META;

	/*XXX retry for each downloaded block? */
	if (diskcache_newfile(ue) == -1) {
		dbg("%s: unable to create diskcache file", f);
		/* writing to disk not possible, simply return */
		return -1;
	}
	ASSERT(ue->cf == NULL, "cf already set");
	if ((ue->cf = cachefile_new(ue->meta)) == NULL) {
		dbg("%s: cachefile_new failed", f);
		return -1;
	}

	if (ue->ueops != NULL && ue->ueops->cacheinit != NULL)
		ue->ueops->cacheinit(ue->uldat);
	ue->state = URIENT_STATE_INCOMPLETE;

	return 0;
}


/* urient_fileref - prepare file, returns -1 on failure, 0 otherwise */
int
urient_fileref(struct urient *ue)
{
	const char *f = "urient_fileref";

#if 0
	ASSERT(UTOSPP(ue)->refcnt > 0, "fileref with no sppfile refs"); /*XXX*/
#endif

	if (ue->file != NULL) {
/*		dbg("%s: reusing file struct for %s", f, ue->uri.uri);*/
		ue->file->refcnt++;

		return 0;
	}

	if ((ue->file = file_open(ue)) == NULL)
		return -1;
	file_blockinit(ue);
	ue->file->refcnt++;

	dbg2("%s: refcnt: %d", f, ue->file->refcnt);

	return 0;
}


/* urient_fileunref - reduce file reference count */
void
urient_fileunref(struct urient *ue)
{
	const char *f = "urient_fileunref";

	ue->file->refcnt--;
	ASSERT(ue->file->refcnt >= 0, "negative refcnt");

	dbg2("%s: refcnt: %d", f, ue->file->refcnt);

#if 0 /* stream references file but not sppfile */
	ASSERT(UTOSPP(ue)->refcnt > 0, "fileref with no sppfile refs"); /*XXX*/
#endif
}


/* urient_gotblock - notification of block download completion */
void
urient_gotblock(struct urient *ue, int bnum)
{
	const char *f = "urient_gotblock";

	ASSERT(ue->master == 0, "gotblock on server");

	/* save entire block to file */
	dbg("%s: block %d retrieved, storing to disk", f, bnum);

	/* XXX exact size for last block? */
	if (diskcache_sizeleft() < ue->meta->blocksize) {
		swarn("%s: disk cache full, initiating expiry", f);
		/*XXX call upper layer callback? */
		diskcache_expire();
	}

	diskcache_gotblock(ue, bnum);

	/* no free of block data, leave in memory for stream_send() */
}


/* urient_lostblock - block no longer found on disk */
void
urient_lostblock(struct urient *ue, int bnum)
{
	const char *f = "urient_lostblock";

	ASSERT(ue->master == 0, "lostblock on server");

	dbg("%s: %s block %d", f, ue->uri.uri, bnum);

	diskcache_lostblock(ue, bnum);
}


/* urient_purgeblock - remove block from memory, halt downloads */
void
urient_purgeblock(struct urient *ue, int bnum)
{
	const char *f = "urient_purgeblock";

	/*
	 * will typically be called due to igot digest being wrong.
	 * can occur in two cases:
	 *  -reporting of downloaded block with incorrect checksum
	 *   ->downloading can only occur to lhc stream client
	 *     (if this is allowed) client should be able to
	 *     handle data streams that are corrupted (ala mpeg)
	 *  -something changed on server (digest/blocksize/file contents)
	 *   -XXX do not care about this for now, but will require
	 *    deletion of all data (might be efficient way for server
	 *    to remove copies at clients).
	 */
	if (ue->file == NULL)
		return;
	if (ue->file->blocks[bnum]->actlen > 0) {
		dbg("%s: purging block %s:%d", f, ue->uri.uri, bnum);
		ue->file->blocks[bnum]->actlen = 0;
		/* XXX hopefully not necessary to do much more */
	}
	if (ue->file->blocks[bnum]->sum != NULL) {
		free(ue->file->blocks[bnum]->sum);
		ue->file->blocks[bnum]->sum = NULL;
	}
	if (ue->state == URIENT_STATE_COMPLETE)
		ue->state = URIENT_STATE_INCOMPLETE;

	/* XXX might have active dlagents, notify, might have to split new */
}

/*XXX purgefile function for when server starts returning errors? (file deleted) */
